//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_AXIS_H
#define GF_AXIS_H

#include "gftop.h"

extern bool gfThreeDimensions; 

extern const GString gfLAxisID;

//@ManMemo: Left-side modeling axis.
class GFLAxis: public GFTop {
  public:
    GFLAxis(const char*, GFPrimitive* = (GFPrimitive*)gfNoFather);
    
    virtual bool isA(const GString& pString) const {
	return(pString == gfLAxisID || GFTop::isA(pString));
    }

    virtual const char* getDescription(void) const { 
	return("An invisible node for trees of primitives to be used as a \
left corner. In dependence of the global variable gfThreeDimensions it will \
be mapped to the x-y plane or will be rotated by 90 degrees in counter-\
clockwise direction."); 
    }

    virtual const char* getClass(void) const { return(gfLAxisID); }

    virtual void rotate(const GVector3&) {}
    // No rotation allowed.

    virtual void translate(const GVector3& pVector) {
	GVector3 tVector(pVector);
	if (!gfThreeDimensions) {
	    // forbid y and z translations:
	    tVector.setY(0.0);
	    tVector.setZ(0.0);
	}
	GFPrimitive::translate(tVector);
    }
  public:
    virtual void shade(void);
    // * GF SHADE LIBRARY *
    static GFObject* classCMD(GFArgc, GFArgs);
};

extern const GString gfRAxisID;

//@ManMemo: Right-side modeling axis.
class GFRAxis: public GFTop {
  public:
    GFRAxis(const char*, GFPrimitive* = (GFPrimitive*)gfNoFather);
    
    virtual bool isA(const GString& pString) const {
	return(pString == gfRAxisID || GFTop::isA(pString));
    }

    virtual const char* getDescription(void) const { 
	return("An invisible node for trees of primitives to be used as a \
right corner. In dependence of the global variable gfThreeDimensions it will \
be mapped to the x-y plane or will be rotated by 90 degrees in clockwise \
direction."); 
    }

    virtual const char* getClass(void) const { return(gfRAxisID); }

    virtual void rotate(const GVector3&) {}
    // No rotation allowed.

    virtual void translate(const GVector3& pVector) {
	GVector3 tVector(pVector);
	if (!gfThreeDimensions) {
	    // forbid y and z translations:
	    tVector.setY(0.0);
	    tVector.setZ(0.0);
	}
	GFPrimitive::translate(tVector);
    }
  public:
    virtual void shade(void);
    // * GF SHADE LIBRARY *
    static GFObject* classCMD(GFArgc, GFArgs);
};

#endif
