//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_BLOCK_H
#define GF_BLOCK_H

#include "gfprimtv.h"
#include "gfimage.h"

extern const GString gfBlockID;

//@ManMemo: An orthogonal solid block.
class GFBlock: public GFPrimitive {
  public:
    //@ManMemo: Construct block from values.
    GFBlock(const char*, const GVector3&, GFPrimitive* = (GFPrimitive*)gfNoFather);

    //@ManMemo: Construct block by copying an existing one.
    GFBlock(const char*, const GFBlock*, GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual bool isA(const GString& pString) const {
	return(pString == gfBlockID);
    }

    virtual const char* getDescription(void) const { return("An orthogonal \
solid block. The block may be assigned an image. The display of the image, \
however, depends on the kind of rendering. The dimensions of the block may be \
changed after creation. For image mapping in shading mode the dimensions of \
the image must be powers of two. Otherwise, the image mapping will be \
ignored."); }
    virtual const char* getClass(void) const { return(gfBlockID); }

    virtual void saveConstr(ostream&);

    //@ManMemo: Set new dimensions.
    void setDimensions(const GVector3&);

    //@ManMemo: Inquire topical dimensions.
    const GVector3 getDimensions(void) const;

    //@ManMemo: Set image mapping.
    void setImage(const char*);

    //@ManMemo: Set globally width for image mapping.
    void setIWidth(GFloat);

    //@ManMemo: Set globally height for image mapping.
    void setIHeight(GFloat);

    // overloaded ray-tracing interface to do image mapping:
    virtual const GSurface& getSurface(const GPoint3&);

    // overload shading interface to do image mapping:
    virtual void shade(void); 
    // * GF SHADE LIBRARY *

  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    static class GFBlockP* sBlock;
    GVector3 mDimensions;

    GFImage* mImage;
};

#endif
