//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_COORDINATES_H
#define GF_COORDINATES_H

#include "gfprimtv.h"

//@ManMemo: A shadable axis of coordinates.
class GFCoordinates: public GFPrimitive {
/*@DOC:
A shadable axis of coordinates, only visible in shading mode.
This primitive is not designed for usage via external API.
Therefore some methods are empty.
*/
  public:
    GFCoordinates(GFPrimitive* pFather = (GFPrimitive*)gfNoFather): 
    GFPrimitive(0, pFather), mStart(0.0), mMajorDist(1.0), mMinorDist(0.1), 
    mLength(1), mChanged(true) {}
    
    void setStart(GFloat pStart) {
	mStart = pStart;
	mChanged = true;
    }

    void setDirection(GFDirection pDir) {
	mDir = pDir; 
	mChanged = true;
    }

    void setMajorDistance(GFloat pMajorDist) {
	mMajorDist = pMajorDist;
	mChanged = true;
    }

    void setMinorDistance(GFloat pMinorDist) {
	mMinorDist = pMinorDist;
	mChanged = true;
    }

    void setLength(GFloat pLength) {
	assert(pLength > 0);
	mLength = pLength;
	mChanged = true;
    }

    virtual bool isA(const GString &) const { return(false); }
    virtual const char *getDescription(void) const { return(0); }
    virtual const char *getClass(void) const { return(0); }

    virtual void saveConstr(ostream&) {}

    virtual void shade(void);
    // * GF SHADE LIBRARY *
  private:
    GFDirection mDir;
    GFloat mStart, mMajorDist, mMinorDist, mLength;
    bool mChanged;
};

#endif
