//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_CYLINDER_H
#define GF_CYLINDER_H

#include "gfprimtv.h"

extern const GString gfCylinderID;

//@ManMemo: A cylindrical primitive.
class GFCylinder: public GFPrimitive {
/*@DOC:
A cylinder along y axis. Radius and length may be specified.
The discretiziation in object space may be specified arbitraryly
and will multiple be used.
*/
  public:
    //@ManMemo: Construction from parameters.
    GFCylinder(const char*, GFloat LENGTH, GFloat RADIUS, 
	       GFPrimitive* = (GFPrimitive*)gfNoFather);
    //@ManMemo: Construction by copying.
    GFCylinder(const char*, const GFCylinder*, 
	       GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFCylinder(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfCylinderID);
    }
    virtual const char* getClass(void) const { return(gfCylinderID); }

    virtual const char* getDescription(void) const { return("Solid cylinder. \
The cylinder is centered around y axis and starts at x-z plane. Radius and \
length may be changed after creation."); }

    //@ManMemo: Set new radius > 0.
    void setRadius(GFloat);
    //@ManMemo: Return topical radius.
    GFloat getRadius(void) const;

    //@ManMemo: Set new length > 0.
    void setLength(GFloat);
    //@ManMemo: Return topical length.
    GFloat getLength(void) const;

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    
    virtual void resolutionChanged(void);
    GFloat mRadius;
    GFloat mLength;
    class GFCylinderP* mCylinder;
};

#endif
