//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_DATA_BASE_H
#define GF_DATA_BASE_H

#include <stdio.h>
#include <util/gstring.h>
#include "gfprogr.h"

extern const GString gfDbScene;
extern const GString gfDbSceneExt;
extern const GString gfDbClass;
extern const GString gfDbClassExt;
extern const GString gfDbOff;
extern const GString gfDbOffGeoExt;
extern const GString gfDbOffVnmExt;
extern const GString gfDbOffIpcExt;
extern const GString gfDbMaterial;
extern const GString gfDbMaterialExt;
extern const GString gfDbImage;
extern const GString gfDbImageExt;
extern const GString gfDbFont;
extern const GString gfDbFontExt;

extern const GString gfDbDefaultExt;

typedef enum {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Enumeration of 
    // database file formats.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    GF_SCENE,
    GF_CLASS,
    GF_OFF,
    GF_MATERIAL,
    GF_IMAGE,
    GF_FONT
} GFDataCategorie;

class GFDataBase {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Abstract definition
    // of Data base interface.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual const GString& getFile(GFDataCategorie, const GString&, const GString& = gfDbDefaultExt) = 0;
    // Return resolved filename if file found.
    // Else return empty string.
    virtual FILE* openFile(GFDataCategorie, const GString&, const GString& = gfDbDefaultExt) = 0;
    // Return file pointer or zero.
    virtual void closeFile(FILE*) = 0;
    // Close an already opened file.
    // If pointer is zero nothing will happen.
};

class GFLocalFileSystem: public GFDataBase {
    // CONCEPT: Furniture graphics kernel
    //
    // ABSTRACT PURPOSE: Implementation of the above class
    // basing on a local file system.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: getString() - return value must be
    // used immediately.
    // CONSTANTS: 
  public:
    virtual const GString& getFile(GFDataCategorie, const GString&, const GString& = gfDbDefaultExt);
    virtual FILE* openFile(GFDataCategorie, const GString&, const GString& = gfDbDefaultExt);
    virtual void closeFile(FILE*);
  private:
    static GString sString;
};

extern GFDataBase* gfDataBase;

#endif
