//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_ELLIPSOID_H
#define GF_ELLIPSOID_H

#include "gfprimtv.h"

extern const GString gfEllipsoidID;

//@ManMemo: A non-uniform scaled sphere.
class GFEllipsoid: public GFPrimitive {
  public:
    //@ManMemo: Construction from parameters.
    GFEllipsoid(const char*, const GVector3&, 
		GFPrimitive* = (GFPrimitive*)gfNoFather);
    //@ManMemo: Construction by copying. 
    GFEllipsoid(const char*, const GFEllipsoid*, 
		GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFEllipsoid(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfEllipsoidID);
    }

    virtual const char* getDescription(void) const { 
	return("AN ellipsoid. The ellipsoid is created centrically around \
the local origin. The dimensions may be changed after creation."); 
    }
    virtual const char* getClass(void) const { 
	return(gfEllipsoidID); 
    }

    //@ManMemo: Set new dimensions, each > 0.
    void setDimensions(const GVector3&);
    //@ManMemo: Return topical dimensions.
    const GVector3 getDimensions(void) const;

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    class GFSphereP* mSphere;
    GVector3 mDimensions;
    virtual void resolutionChanged(void);
};

#endif
