//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_FILTER_H
#define GF_FILTER_H

#include "gfmanip.h"

extern const GString gfFilterID;

//@ManMemo: An input class for filter purposes.
class GFFilter: public GFInputDevice {
/*@Doc:
An input class defined to filter events for an associated manipulator.
*/
  public:
    //@ManMemo: Create a filter with an associated manipulator.
    GFFilter(const char*, GFManipulator*);

    virtual ~GFFilter(void);

    virtual const char* getDescription(void) const {
	return("An input object that wraps a manipulator object. \
It may be used to do specific operations required be a graphical \
user interface, for instance.");
    }

    //@ManMemo: Activate encapsulated manipulator.
    void activate(void);

    //@ManMemo: Deactivate encapsulated manipulator.
    void deactivate(void);

    //@ManMemo: Inquire topical position.
    const GPoint3& getPosition(void) const {
	return(mPos);
    }

    //@ManMemo: Inquire topical mouse button state.
    int getButton(void) const {
	return(mButton);
    }
  public:
    virtual bool event(GEvent*);
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif 
  private:
    GFManipulator* mDevice;
    bool mAct;
    GPoint3 mPos;
    int mButton;
};    

#endif
