//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_LIGHT_H
#define GF_LIGHT_H

#include <render/grayligh.h>
#include "gftop.h"

extern const GString gfLightID;

//@ManMemo: Integration of ray-tracing and shading compatible light source.
class GFPointLight: public GRayPointLight, public GShadePointLight {
};

//@ManMemo: A primitive containing a light source.
class GFLight: public GFTop {
  public:
    //@ManMemo: Light source constructor.
    GFLight(const char*, GFPrimitive* = (GFPrimitive*)gfNoFather, 
	    GFTop* = (GFTop*)gfNoFather);

    virtual ~GFLight(void);
    
    virtual bool isA(const GString& pString) const {
	return(pString == gfLightID || GFTop::isA(pString));
    }

    virtual const char* getDescription(void) const { return("Invisible \
encapsulation of a point light source. The point light is located at the \
local origin. Light sources are acting globally, i.e., they do not need to be \
inserted into a scene. To pick and manipulate a light source just insert the \
light source into a scene and append a transparent primitive to it as a \
child."); }
    virtual const char* getClass(void) const { return(gfLightID); }

    //@ManMemo: Switch light on. 
    virtual void on(void) {
	mLight.on();
    }

    //@ManMemo: Switch light off. 
    virtual void off(void) {
	mLight.off();
    }
  public:
    virtual void saveConstr(ostream&); 

    // assign to the global light source pointer
    // used by renderers:
    static void initialize(void) {
	gShadeLights = &sShadeList;
	gRayLights = &sRayList;
    }
    GFPointLight& getLight(void) {
	return(mLight);
    }
    virtual void shade(void);
    // * GF SHADE LIBRARY *
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif  
  protected:
    virtual void modelingChanged(void);
    void setupLight(void);
  private:
    GFPointLight mLight;
    static GDLList<GRayLight> sRayList;
    static GDLList<GShadeLight> sShadeList;
    GFTop* mLocalRoot;
};

#endif
