//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_LOOKAT_H
#define GF_LOOKAT_H

#include <input/gdevice.h>
#include "gftypes.h"
#include "gfscene.h"
#include "gfpixmap.h"

extern const GString gfLookatCameraID;

//@ManMemo: GF renderer object. 
class GFLookatCamera: public GInputDevice, public GFObject {
/*@Doc:
A class hiding the lookat camera and containing both
shading and ray-tracing camera.
*/
  public:
    //@ManMemo: Create camera. Initial reference point is the WC origin.
    GFLookatCamera(const char*); 

    virtual const char* getDescription(void) const;

    //@ManMemo: Shading without swapping.
    void shading(void);

    //@ManMemo: On-line ray-tracing.
    void rayTracing(GFloat);

    //@ManMemo: Off-line ray-tracing.
    void trace2image(void);

    //@ManMemo: Set scene to render.
    void setScene(GFScene& pScene) {
	mScene = &pScene;
    }

    //@ManMemo: Inquire topical scene.
    GFScene& getScene(void) const {
	assert(mScene);
	return(*mScene);
    }

    //@ManMemo: Set pixmap to render in.
    void setPixmap(GFPixmap& pPixmap) {
	mPixmap = &pPixmap;
    }

    //@ManMemo: Inquire topical pixmap.
    GFPixmap& getPixmap(void) {
	assert(mPixmap);
	return(*mPixmap);
    }

    //@ManMemo: Set new zoom value in degrees.
    void setZoom(GFloat);

    //@ManMemo: Inquire topical zoom value.
    GFloat getZoom(void) const;

    //@ManMemo: Set new camera viewpoint.
    void setViewpoint(const GPoint3&);

    //@ManMemo: Inquire camera viewpoint.
    const GVector3& getViewpoint(void) const;

    //@ManMemo: Set new camera refpoint.
    void setRefpoint(const GPoint3&);

    //@ManMemo: Inquire camera refpoint.
    const GVector3& getRefpoint(void) const;

    //@ManMemo: Switch camera into perspectivic projection.
    void perspectivic(void) {
	mPersp = true;
	mShader.perspectivic();
	mTracer.perspectivic();
    }

    //@ManMemo: Switch camera into orthographic projection.
    void orthographic(void) {
	mPersp = false;
	mShader.orthographic();
	mTracer.orthographic();
    }
  public:
    GFLookatShader& getShader(void) {
	return(mShader);
    }
    const GString& getName(void) const {
	return(mName);
    }

    virtual bool event(GEvent*);

    static GFLookatCamera* getByName(const char*);
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    // draw w/o buffer swapping:
    void draw(void);
    // * GF SHADE LIBRARY *
    // swap buffers:
    void swap(void);
    // * GF SHADE LIBRARY *
  private:
    GFLookatShader mShader;
    GFLookatRayTracer mTracer; 
    bool mPersp, mDraw, mSwap;

    GFScene* mScene;
    GFPixmap* mPixmap;
    const GString mName;

    static GDLList<GFLookatCamera> sList;
};

#endif
