//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_M_LINE_H
#define GF_M_LINE_H

#include "gftext.h"

extern const GFloat gfMLineDefaultOffset;
extern const GString gfMLineID;

//@ManMemo: A measurement line.
class GFMLine: public GFMeasurement {
/*@Doc:
A measurement line to be added to primitives or top objects
for automatic measurement.
*/
  public:
    GFMLine(const char*, GFDirection, GFPrimitive*);
    ~GFMLine(void) {
	delete mText;
    }
    
    virtual bool isA(const GString& pString) const {
	return(pString == gfMLineID || GFMeasurement::isA(pString));
    }

    virtual const char* getDescription(void) const { return("An orthogonal \
measurement line that has to be bound to one side of a primitive. It then will \
automatically display the length of the bounding box on this side. \
You may also explicitely specify a text string to be displayed.\n\n\
The global variable gfMeasuring must be set to true to see the measuring."); }

    virtual const char* getClass(void) const { return(gfMLineID); }

    virtual void saveConstr(ostream&);

    void setOffset(GFloat pOff) {
	mOff = pOff;
    }

    GFloat getOffset(void) const {
	return(mOff);
    }

    void setText(const GString& pText) {
	mText->setText(pText);
	mMsr = false;
    }

    const GString& getText(void) const {
	return(mText->getText());
    }
  public:
    virtual void shade(void); 
    // * GF SHADE LIBRAY *

    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    GFDirection mDir;
    GFloat mOff;
    GFMText* mText;
    GFloat mDistance;
    bool mMsr;
};

#endif
