//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_M_SYMBOL_H
#define GF_M_SYMBOL_H

#include "gftypes.h"
#include "gfmeasr.h"

extern const GString gfMSymbolID;

typedef enum {
    GF_M_CIRCLE = 1,
    GF_M_RECTANGLE,
    GF_M_POLYLINE
} GFMSymbolType;

//@ManMemo: A 2D measurement symbol.
class GFMSymbol: public GFMeasurement {
  public:
    GFMSymbol(const char*, GFMSymbolType, const GFPolygon2D&, GFPrimitive*);

    virtual bool isA(const GString& pString) const {
	return(pString == gfMSymbolID || GFMeasurement::isA(pString));
    }

    virtual const char* getDescription(void) const { return("Instances of MSymbol are used to add two-dimensional graphics elements to 3D scenes. Concrete types of elements are: Circle, Rectangle and Polyline."); }

    virtual const char* getClass(void) const { return(gfMSymbolID); }

     //@ManMemo: Overwrite point at specified position.
    void setPoint(int, const GPoint2&);

    //@ManMemo: Get point from specified position. 
    const GPoint2& getPoint(int) const;

    //@ManMemo: Get type of 2D element.
    GFMSymbolType getType(void) const {
	return(mType);
    }

    //@ManMemo: Get size of point array.
    int getSize(void) const {
	return(mValues.getNumber());
    }
  public:
    virtual void saveConstr(ostream&);

    virtual void shade(void); 
    // * GF SHADE LIBRAY *
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    bool mOK;
    GFMSymbolType mType;
    GConstantVector<GPoint2> mValues;
    static int sMinSize[4];
};

#endif
