//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_OFF_POLYGON_H
#define GF_OFF_POLYGON_H

#include "gfprimtv.h"

extern const GString gfOffPolygonID;

//@ManMemo: A primitive encapsulating OFF polygon data sets.
class GFOffPolygon: public GFPrimitive {
  public:
    GFOffPolygon(const char*, const GString&, 
		 GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFOffPolygon(const char*, const GFOffPolygon*, 
		 GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFOffPolygon(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfOffPolygonID);
    }

    virtual const char* getDescription(void) const { return("A polygon set \
representing Object File Format (OFF) polygon sets. Such a polygon set must \
consist of simple, convex and planar polygons, defined in a clockwise \
direction. The set must specify a solid body, either alone or in cooperation \
with further primitives. If the geometry file cannot be found in the \
appropriate search path, the program terminates with an appropriate \
message."); }
    virtual const char* getClass(void) const {
	return(gfOffPolygonID); 
    }

    virtual void saveConstr(ostream&);

    // set/get a scaling factor:
    void setScale(GFloat);
    GFloat getScale(void) const { return(mScale); }

    const GString& getName(void);

    // overloaded ray-tracing interface:
    virtual const GSurface& getSurface(const GPoint3&);

    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    class GFOffPolygonP* mOffPolygon;
    GFloat mScale;
};

#endif
