//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_POLYGON_H
#define GF_POLYGON_H

#include "gfprimtv.h"
#include "gftypes.h"
#include "gfimage.h"

extern const GString gfPolygonID;

//@ManMemo: A polygon primitive.
class GFPolygon: public GFPrimitive {
/*@Doc:
A simple, planary and convexe polygon that may arbitrarily lying in the 3D room.
*/
  public:
    GFPolygon(const char*, const GFPolygon3D&, GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFPolygon(const char*, const GFPolygon*, GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual bool isA(const GString& pString) const {
	return(pString == gfPolygonID);
    }

    virtual const char* getDescription(void) const { return("A polygonal \
plane to be used to build solids or environments. The polygon may be assigned an image. \
The display of the image, however, depends on the kind of rendering. \
For image mapping in shading mode the dimensions of the image must be powers of two. \
Otherwise, the image mapping will be ignored. The polygon must be specified in clock-wise \
direction.\nNOTE: The image mapping is not implemented until now."); }
    virtual const char* getClass(void) const { return(gfPolygonID); }

    virtual void saveConstr(ostream&);

    void setImage(const char*);
    void setIWidth(GFloat);
    void setIHeight(GFloat);

    // overloaded ray-tracing interface to do image mapping:
    virtual const GSurface& getSurface(const GPoint3&);

    // overload shading interface to do image mapping:
    virtual void shade(void); 
    // * GF SHADE LIBRARY *

    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    class GFPolygonP* mPolygon;
    GFImage* mImage;
};

#endif
