//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_RENDER_H
#define GF_RENDER_H

#include <gdata.h>
#include <render/grender.h>
#include <render/gshade.h>
#include <render/graytrac.h>

#include "gfbasic.h"

enum GFRenderEnum {
    GF_NO_RENDERABLE,

    GF_CYLINDER,
    GF_FRAME,
    GF_POLYGON,
    GF_OFF_POLYGON,
    GF_SPHERE,
    GF_SWEEP_X_POLYGON,
    GF_SWEEP_Y_POLYGON,
    GF_SWEEP_Z_POLYGON,
    GF_FITTING_SURFACE,
    GF_ROTATION_SURFACE,
    GF_SWEEP_SURFACE
}; 

class GFPrimitive;

//@ManMemo: Abstract renderable primitive.
class GFRenderable: public GShadable, public GRayTraceable, 
public GGraphicalData {
/*@DOC:
Root class of the basic primitives of the furniture kernel.
Multiple references are supported using hash indexes to get a
shared represen tation from data base.
*/
  public:
    void setModelingPrimitive(GFPrimitive&);
    GFPrimitive& getModelPrimitive(void) const {
	assert(mModelPrim);
	return(*mModelPrim);
    }
    // Unreference a multiple referenced renderable:
    static void unreference(GFRenderable*);

    // Inquire run-time identification for storable renderables:
    virtual GFRenderEnum getID(void) const { return(GF_NO_RENDERABLE); }

    // Inquire cache value for storable renderables:
    virtual GCacheValue getCacheValue(void) const { return(0); }
  public:
    // rendering interface(s):
    virtual void shade(void);
    // * GF SHADE LIBRARY *

    virtual GFloat getNearest(const GRay3& pRay, GShadable*& rShad) {
	GFRenderable* tRend = (GFRenderable*)rShad;
	GFloat tDist = getNearest(pRay, tRend);
	rShad = (GShadable*)tRend;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3& pRay, GRayTraceable*& rRay) {
	GFRenderable* tRend = (GFRenderable*)rRay;
	GFloat tDist = getNearest(pRay, tRend);
	rRay = (GRayTraceable*)tRend;
	return(tDist);
    }
    virtual GFloat getNearest(const GRay3&, GFRenderable*&);
  protected:
    // utilitys for WC/MC and vice versa conversion:
    GRay3 ray2mc(const GRay3& WC_RAY);
    GFloat distance2wc(const GRay3& MC_RAY, const GRay3& WC_RAY, 
		       GFloat MC_DISTANCE);
    GVector3 normal2wc(const GPoint3& WC_POINT, const GVector3& MC_NORMAL);
  protected:
    static void insertInCache(GFRenderable*);
  private:
    GFPrimitive* mModelPrim;
};

typedef GDLList<GFRenderable>& GFRenderables;
GFRenderables gfGetRenderables(GCacheValue);
void gfPrintCacheStatistics(ostream&);

extern GFRenderable* gfNoRenderable;

#endif
