//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_ROTATION_SURFACE_H
#define GF_ROTATION_SURFACE_H

#include "gfprimtv.h"
#include "gftypes.h"

extern const GString gfRotationSurfaceID;

//@ManMemo: A rotational primitive.
class GFRotationSurface: public GFPrimitive {
  public:
    GFRotationSurface(const char*, GSplineMode, const GVector3&, 
		      const GFCurve3D&, GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFRotationSurface(const char*, const GFRotationSurface*, 
		      GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual ~GFRotationSurface(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfRotationSurfaceID);
    }

    virtual void setAxis(const GVector3&);
    const GVector3& getAxis(void) const {
	return(mAxis);
    }

    void setMode(GSplineMode);
    const GSplineMode& getMode(void) const;

    virtual const char* getDescription(void) const { return("Rotation surface. Watch out of points lying on the rotation axis!"); }
    virtual const char* getClass(void) const { return(gfRotationSurfaceID); }

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  protected:
    class GFRotationSurfaceP* mRot;
  private:
    virtual GFRenderable* getRenderable(void);
    void resolutionChanged(void);
    GVector3 mAxis;
    GSplineMode mMode;
};

extern const GString gfRotationXSurfaceID;

//@ManMemo: Rotation surface along x axis.
class GFRotationXSurface: public GFRotationSurface {
  public:
    GFRotationXSurface(const char*, GSplineMode, const GFCurve3D&, 
		       GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFRotationXSurface(const char*, const GFRotationXSurface*, 
		       GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual bool isA(const GString& pString) const {
	return(pString == gfRotationXSurfaceID);
    }
    
    virtual const char* getDescription(void) const { return("Rotation surface, defining curve is rotated around the X-axis. Watch out of points lying on the rotation axis."); }
    virtual const char* getClass(void) const { return(gfRotationXSurfaceID); }

    virtual void setAxis(const GVector3&) {
	gOutput->error("Cannot change axis of rotation.");
    }

  public:
    virtual void saveConstr(ostream&);
    static GFObject* classCMD(GFArgc, GFArgs);
};

extern const GString gfRotationYSurfaceID;

//@ManMemo: Rotation surface along x axis.
class GFRotationYSurface: public GFRotationSurface {
  public:
    GFRotationYSurface(const char*, GSplineMode, const GFCurve3D&,
		       GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFRotationYSurface(const char*, const GFRotationYSurface*,
		       GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual bool isA(const GString& pString) const {
	return(pString == gfRotationYSurfaceID);
    }

    virtual const char* getDescription(void) const { return("Rotation surface, defining curve is rotated around the Y-axis. Watch out of points lying on the rotation axis."); }
    virtual const char* getClass(void) const { return(gfRotationYSurfaceID); }

    virtual void setAxis(const GVector3&) {
	gOutput->error("Cannot change axis of rotation.");
    }
  public:
    virtual void saveConstr(ostream&);
    static GFObject* classCMD(GFArgc, GFArgs);
};

extern const GString gfRotationZSurfaceID;

//@ManMemo: Rotation surface along x axis.
class GFRotationZSurface: public GFRotationSurface {
  public:
    GFRotationZSurface(const char*, GSplineMode, const GFCurve3D&,
		       GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFRotationZSurface(const char*, const GFRotationZSurface*, 
		       GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual bool isA(const GString& pString) const {
	return(pString == gfRotationZSurfaceID);
    }

    virtual const char* getDescription(void) const { return("Rotation surface, defining curve is rotated around the Z-axis. Watch out of points lying on the rotation axis."); }
    virtual const char* getClass(void) const { return(gfRotationZSurfaceID); }

    virtual void setAxis(const GVector3&) {
	gOutput->error("Cannot change axis of rotation.");
    }

  public:
    virtual void saveConstr(ostream&);
    static GFObject* classCMD(GFArgc, GFArgs);
};

#endif


