//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_SCENE_H
#define GF_SCENE_H

#include "gfprimtv.h"

extern const GString gfSceneID;

//@ManMemo: The scene aggregate that will be rendered by cameras.
class GFScene: public GFObject {
  public:
    //@ManMemo: Create a scene with specified name.
    GFScene(const char*);
    virtual ~GFScene(void);

    //@ManMemo: Insert a primitive into scene.
    void insert(GFPrimitive& pPrim) {
	mScene.append(&pPrim);
    }

    //@ManMemo: Remove a primitive from scene.
    void remove(GFPrimitive& pPrim) {
	mScene.remove(&pPrim);
    }

    const char* getDescription(void) const {
	return("A dynamic aggregate of primitives which may be rendered by a \
camera. Only primitives without a father may be inserted into a scene. Do not \
insert a primitive more than once into one scene.");
    }

    //@ManMemo: Set global background color. 
    void setBackground(const GColorRGB& pColor) {
	mBackgr = pColor;
    }

    //@ManMemo: Inquire list of scene objects.
    GAggregate<GFPrimitive>& getObjects(void) { 
	return(mScene); 
    } 

    //@ManMemo: Compute recursively bounding boxes.
    void computeBounds(void);
    // This might be necessary after loading a scene dump.

    //@ManMemo: Evaluate recursively logical checks.
    void evalLogics(void);
  public:
    void insertFdb(GFPrimitive& pPrim) {
	mFeedback.append(&pPrim);
    }
    void removeFdb(GFPrimitive& pPrim) {
	mFeedback.remove(&pPrim);
    }
    
    const GColorRGB& getBackground(void) const {
	return(mBackgr);
    }
    
    static GFScene* getByName(const char*);

    GAggregate<GFPrimitive>& getFeedback(void) {
	return(mFeedback);
    }

    const GString& getName(void) const {
	return(mName);
    }
    
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    GString mName;
    GDLList<GFPrimitive> mScene;
    GDLList<GFPrimitive> mFeedback;
    
    static GDLList<GFScene> sList;
    GColorRGB mBackgr;
};

#endif
