//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_SPHERE_H
#define GF_SPHERE_H

#include "gfprimtv.h"

extern const GString gfSphereID;

//@ManMemo: An ideal sphere.
class GFSphere: public GFPrimitive {
/*@DOC:
An ideal sphere specified by a radius. The discretiziation in object space
may be specified arbitraryly and will multiple be used.
*/
  public:
    //@ManMemo: Construction from parameters.
    GFSphere(const char*, GFloat, 
	     GFPrimitive* = (GFPrimitive*)gfNoFather);
    //@ManMemo: Construction by copying.
    GFSphere(const char*, const GFSphere*, 
	     GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFSphere(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfSphereID);
    }

    virtual const char* getDescription(void) const { return("A homogeneous \
sphere. The sphere is centered around the local origin. The radius may be \
changed after creation."); }
    virtual const char* getClass(void) const { return(gfSphereID); }

    virtual void saveConstr(ostream&);

    //@ManMemo: Set new radius > 0.
    void setRadius(GFloat);

    //@ManMemo: Return topical radius.
    GFloat getRadius(void) const;
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    
    virtual void resolutionChanged(void);
    GFloat mRadius;
    class GFSphereP* mSphere;
};

#endif
