//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_SPLINE_SURF_H
#define GF_SPLINE_SURF_H

#include "gfprimtv.h"
#include "gftypes.h"

extern const GString gfFittingSurfaceID;

//@ManMemo: A fitting surface for GF.
class GFFittingSurface: public GFPrimitive {
  public:
    //@ManMemo: Construct surface from vertices.
    GFFittingSurface(const char*, int, int, const GFSurface3D&, GSplineMode = G_OPEN, GSplineMode = G_OPEN, GFPrimitive* = (GFPrimitive*)gfNoFather);

    //@ManMemo: Construct surface by copying an existing one.
    GFFittingSurface(const char*, const GFFittingSurface*, GFPrimitive* = (GFPrimitive*)gfNoFather);
    
    virtual ~GFFittingSurface(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfFittingSurfaceID);
    }

    virtual const char* getDescription(void) const { return("Fitting spline surface."); }
    virtual const char* getClass(void) const { return(gfFittingSurfaceID); }

    virtual void saveConstr(ostream&);

    //@ManMemo: Set spline mode for both directions.
    void setMode(GSplineMode, GSplineMode);

    //@ManMemo: Inquire spline mode for U direction.
    const GSplineMode& getModeU(void) const;

    //@ManMemo: Inquire spline mode for W direction.
    const GSplineMode& getModeW(void) const;

  public:
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    virtual GFRenderable* getRenderable(void);
    void resolutionChanged(void);
    class GFFittingSurfaceP* mFittingSurface;

    GSplineMode mModeU, mModeW;
};
#endif



