//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_SWEEP_POLYGON_H
#define GF_SWEEP_POLYGON_H

#include "gfprimtv.h"
#include "gftypes.h"

//@ManMemo: Common base class for sweep polygons.
class GFSweepPolygon: public GFPrimitive {
/*@Doc:
The polygon must be planar and convex. The polygon vertices have
to be specified in clock-wise direction. The polygon is closed automatically.
*/
  protected:
    GFSweepPolygon(const char* pName, GFPrimitive* pFather): 
    GFPrimitive(pName, pFather) {}
  public:
    virtual void setLength(GFloat) = 0;
    GFloat getLength(void) const {
	return(mLength);
    }
    virtual void saveConstr(ostream&);
  public:
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  protected:
    GFloat mLength;
};

extern const GString gfSweepXPolygonID;

//@ManMemo: A sweep along x axis.
class GFSweepXPolygon: public GFSweepPolygon {
/*@Doc: 
A solid created by sweeping a planar polygon in x direction. 
The polygon must be planar and convex. The polygon vertices have
to be specified in clock-wise direction. The polygon is closed automatically.
*/
  public:
    GFSweepXPolygon(const char*, const GFPolygon2D&, GFloat, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepXPolygon(const char*, const GFSweepXPolygon*, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFSweepXPolygon(void);

    virtual void setLength(GFloat);

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepXPolygonID);
    }

    virtual const char* getDescription(void) const { return("An X-axis \
sweeped planar and convexe polygon lying in y-z plane. The smallest point of the sweep \
must be (0, 0, 0). The length may be changed after creation."); }
    virtual const char* getClass(void) const { return(gfSweepXPolygonID); }
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
  private:
    virtual GFRenderable* getRenderable(void);
    class GFSweepXPolygonP* mSweepPolygon;
};

extern const GString gfSweepYPolygonID;

//@ManMemo: A sweep along y axis.
class GFSweepYPolygon: public GFSweepPolygon {
/*@Doc: 
A solid created by sweeping a planar polygon in y direction. 
The polygon must be planar and convex. The polygon vertices have
to be specified in clock-wise direction. The polygon is closed automatically.
*/
  public:
    GFSweepYPolygon(const char*, const GFPolygon2D&, GFloat, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepYPolygon(const char*, const GFSweepYPolygon*, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFSweepYPolygon(void);

    virtual void setLength(GFloat);

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepYPolygonID);
    }

    virtual const char* getDescription(void) const { return("A Y-axis sweeped \
planar and convexe polygon lying in x-z plane. The smallest point of the sweep must be \
(0, 0, 0). The length may be changed after creation."); }
    virtual const char* getClass(void) const { return(gfSweepYPolygonID); }
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
  private:
    virtual GFRenderable* getRenderable(void);
    class GFSweepYPolygonP* mSweepPolygon;
};

extern const GString gfSweepZPolygonID;

//@ManMemo: A sweep along z axis.
class GFSweepZPolygon: public GFSweepPolygon {
/*@Doc: 
A solid created by sweeping a planar polygon in z direction. 
The polygon must be planar and convex. The polygon vertices have
to be specified in clock-wise direction. The polygon is closed automatically.
*/
  public:
    GFSweepZPolygon(const char*, const GFPolygon2D&, GFloat, 
                    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepZPolygon(const char*, const GFSweepZPolygon*, 
                    GFPrimitive* = (GFPrimitive*)gfNoFather);
    virtual ~GFSweepZPolygon(void);

    virtual void setLength(GFloat);

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepZPolygonID);
    }

    virtual const char* getDescription(void) const { return("A Z-axis sweeped \
planar and convexe polygon lying in x-y plane. The smallest point of the sweep must be \
(0, 0, 0). The length may be changed after creation."); }
    virtual const char* getClass(void) const { return(gfSweepZPolygonID); }
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
  private:
    virtual GFRenderable* getRenderable(void);
    class GFSweepZPolygonP* mSweepPolygon;
};

#endif
