//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_SWEPT_SURFACE_H
#define GF_SWEPT_SURFACE_H

#include "gfprimtv.h"
#include "gftypes.h"

extern const GString gfSweepSurfaceID;

//@ManMemo: Basic sweeping surface.
class GFSweepSurface: public GFPrimitive {
  public:
    GFSweepSurface(const char*, GFloat, const GVector3&, const GFCurve3D&, 
		   GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepSurface(const char*, const GFSweepSurface*, 
		   GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual ~GFSweepSurface(void);

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepSurfaceID);
    }

    virtual void setSweepVector(const GVector3&);
    const GVector3& getSweepVector(void) const;

    virtual void setLength(GFloat);
    const GFloat& getLength(void) const;
    
    virtual const char* getDescription(void) const { 
	return("Surface swept in the direction given by a vector.");
    }

    virtual const char* getClass(void) const { return(gfSweepSurfaceID); }

    virtual void saveConstr(ostream&);

  public:
    static GFObject* classCMD(GFArgc, GFArgs);   
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  protected:
    class GFSweepSurfaceP* mSweep;
  private:
    virtual GFRenderable* getRenderable(void);
    void resolutionChanged(void);
    GVector3 mSweepVector;
    GFloat mLength;
};

extern const GString gfSweepXSurfaceID;

//@ManMemo: A sweep along x axis.
class GFSweepXSurface: public GFSweepSurface {
  public:
    GFSweepXSurface(const char*, GFloat, const GFCurve3D&,
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepXSurface(const char*, const GFSweepXSurface*, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual void setSweepVector(const GVector3&) {
	gOutput->warning("Will not change the sweep vector.");
    }

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepXSurfaceID);
    }
    virtual const char* getDescription(void) const { 
	return("Surface swept by the x axis.");
    }
    virtual const char* getClass(void) const { return(gfSweepXSurfaceID); }

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
};

extern const GString gfSweepYSurfaceID;

//@ManMemo: A sweep along y axis.
class GFSweepYSurface: public GFSweepSurface {
  public:
    GFSweepYSurface(const char*, GFloat, const GFCurve3D&,
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepYSurface(const char*, const GFSweepYSurface*, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual void setSweepVector(const GVector3&) {
	gOutput->warning("Will not change the sweep vector.");
    }

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepYSurfaceID);
    }
    virtual const char* getDescription(void) const { 
	return("Surface swept by the y axis.");
    }
    virtual const char* getClass(void) const { return(gfSweepYSurfaceID); }

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
};

extern const GString gfSweepZSurfaceID;

//@ManMemo: A sweep along z axis.
class GFSweepZSurface: public GFSweepSurface {
  public:
    GFSweepZSurface(const char*, GFloat, const GFCurve3D&, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);
    GFSweepZSurface(const char*, const GFSweepZSurface*, 
		    GFPrimitive* = (GFPrimitive*)gfNoFather);

    virtual void setSweepVector(const GVector3&) {
	gOutput->warning("Will not change the sweep vector.");
    }

    virtual bool isA(const GString& pString) const {
	return(pString == gfSweepZSurfaceID);
    }
    virtual const char* getDescription(void) const { 
	return("Surface swept by the z axis.");
    }
    virtual const char* getClass(void) const { return(gfSweepZSurfaceID); }

    virtual void saveConstr(ostream&);
  public:
    static GFObject* classCMD(GFArgc, GFArgs);
};

#endif
