//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_TCL_H
#define GF_TCL_H

#ifdef GWITH_TCL

#include "gfdevice.h"
#include "gfprimtv.h"
#include "gftop.h"

//@ManMemo: Generic Tcl callback procedure for input devices.
class GFTclCallback: public GCallback {
/*@Doc:
Generic Tcl callback procedure for input devices.
This callback may only be used for GF input device because it performs a type cast
to GF input device without any possibility to verify it.
*/
  public:
    //@ManMemo: Initialize callback with Tcl procedure.
    GFTclCallback(const char* pProc): mProc(pProc) {}
    
    virtual void action(GInputDevice&);
  private:
    GString mProc;
};

// Autoload a class description:
extern int gfInclude(const char*);

extern const GString gfGfmID;
extern const GString gfGfpID;

//@ManMemo: Base class for Tcl extensible primitive classes.
class GFTclPrimitive: public GFPrimitive, public GTclClass {
  protected:
    GFTclPrimitive(const char *, GFPrimitive*, GTclClassDesc*, char*[]);
  public:
    virtual bool isA(const GString& pString) const {
	return(GTclClass::isA(pString));
    }

    virtual const char* getDescription(void) const { 
	return(GTclClass::getDescription()); 
    }
    virtual const char* getClass(void) const { 
	return(GTclClass::getClass()); }

    virtual void saveConstr(ostream&);
    virtual void saveChildren(ostream& o) {
	if (wasChanged()) GFPrimitive::saveChildren(o);
    }
  public:
    static int addClassCMD(ClientData, Tcl_Interp*, int, char*[]);
    virtual bool objectCMD(GFArgs);
  private:
    GString mConstrArgs;
    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
};

extern const GString gfGftID;

//@ManMemo: Base class for Tcl extensible top classes.
class GFTclTop: public GFTop, public GTclClass {
  protected:
    GFTclTop(const char *, GFPrimitive*, GTclClassDesc*, char*[]);
  public:
    virtual bool isA(const GString& pString) const {
	return(GFTop::isA(pString) || GTclClass::isA(pString));
    }

    virtual const char* getDescription(void) const { 
	return(GTclClass::getDescription()); 
    }
    virtual const char* getClass(void) const { 
	return(GTclClass::getClass()); 
    }

    virtual void saveConstr(ostream&);
    virtual void saveChildren(ostream& o) {
	if (wasChanged()) GFPrimitive::saveChildren(o);
    }
  public:
    static int addClassCMD(ClientData, Tcl_Interp*, int, char*[]);
    virtual bool objectCMD(GFArgs);
  private:
    GString mConstrArgs;
    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
};

extern const GString gfGfoID;

//@ManMemo: Base class for Tcl extensible classes.
class GFTclObject: public GFObject, public GTclClass {
  protected:
    GFTclObject(const char *, GTclClassDesc*, char*[]);

    virtual bool isA(const GString& pString) const {
	return(GTclClass::isA(pString));
    }

    virtual const char* getDescription(void) const { 
	return(GTclClass::getDescription()); 
    }
    virtual const char* getClass(void) const { 
	return(GTclClass::getClass());
    }
  public:
    static int addClassCMD(ClientData, Tcl_Interp*, int, char*[]);

    virtual bool objectCMD(char*[]); 
  private:
    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
};

// initialize Tcl and return true if successful:
bool gfInitTcl(Tcl_Interp*);

#endif

#endif
