//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GF_TEXT_H
#define GF_TEXT_H

#include <format/ghfont.h>
#include <util/gtempl.h>
#include "gfdatabs.h"
#include "gfmeasr.h"

//@ManMemo: Hershey font encapsulation including a data base.
class GFFont: public GHersheyFont {
  public:
    GFFont(const GString&);
  public:
    static const GFFont* getFont(const GString&);
  private:
    static GDLList<GFFont> sFonts;
};

extern const GString gfMTextID;
extern const GFloat gfMTextDefaultScale;

typedef enum GFAlignment {
    GF_LEFT = -1,
    GF_CENTER,
    GF_RIGHT
};

typedef enum GFMTextMode {
    GF_NORMAL = 0,
    GF_UNDERLINE,
    GF_BOX
};

//@ManMemo: A text mainly for usage in measurements.
class GFMText: public GFMeasurement {
  public:
    //@ManMemo: Create text object with specified initial label.
    GFMText(const char*, const char*, GFPrimitive*);

    virtual bool isA(const GString& pString) const {
	return(pString == gfMTextID || GFMeasurement::isA(pString));
    }

    virtual const char* getDescription(void) const { return("A text \
primitive that may be used in measurements, for instance. \n\n\
The global variable gfMeasuring must be set to true to see the measuring."); }

    virtual const char* getClass(void) const { return(gfMTextID); }
    
    virtual void saveConstr(ostream&);
    
    //@ManMemo: Change text font.
    void setFont(const GString&);

    //@ManMemo: Inquire topical font.
    const GString& getFont(void) const;

    //@ManMemo: Change text string.
    void setText(const GString&);

    //@ManMemo: Inquire topical text.
    const GString& getText(void) const {
	return(mText);
    }

    //@ManMemo: Change text scaling.
    void setScale(GFloat);

    //@ManMemo: Inquire topical scaling
    GFloat getScale(void) const {
	return(mScale);
    }

    //@ManMemo: Change text alignment.
    void setAlignment(GFAlignment);

    //@ManMemo: Inquire topical alignment.
    GFAlignment getAlignment(void) const {
	return(mAlign);
    }

    //@ManMemo: Change text marking mode.
    void setMode(GFMTextMode pMode) {
	mMode = pMode;
    }

    //@ManMemo: Inquire topical mode.
    GFMTextMode getMode(void) const {
	return(mMode);
    }
  public:
    virtual void shade(void); 
    // * GF SHADE LIBRARY *
    static GFObject* classCMD(GFArgc, GFArgs);
#ifdef GWITH_TCL
    virtual bool objectCMD(GFArgs);
#endif
  private:
    void createLines(void);
    void updateBounds(void);
  private:
    GString mText;
    const GFFont* mFont;
    GDLList<GShadablePolyline> mLines;
    GFloat mScale;
    GFloat mWidth;
    GFAlignment mAlign;
    GFMTextMode mMode;
    GPoint3 mMax, mMin;
    static bool sAutoCreate;
};

#endif
