//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan 30 20:41:53 1996 - juhu@rz.tu-ilmenau.de

#include "3dschunk.h"
#include "3ds.h"

//////////////////////////////////////////////////////////////////////////////
// @static@ variables
//////////////////////////////////////////////////////////////////////////////


char  Scene3DS::mMsg[255] = "";
char  Scene3DS::mObjName[80] = "";
Mesh* Scene3DS::mMesh = 0;

// max. smooth-angle = 60 degre
GFloat Scene3DS::mMinDot = 0.5;




//// Ctor
Scene3DS::Scene3DS(void) 
:mWorldBound(GPoint3(MAXFLT, MAXFLT,  MAXFLT),
	     GPoint3(-MAXFLT, -MAXFLT,  -MAXFLT))
{
    // @ADD@ the default material to the list.
    MatProp *tMatProp = new MatProp;
    mMPropList.append(tMatProp);
}

Scene3DS::~Scene3DS(void) {
    // free up all allocated elments
    for(GDynaListIterator<OmniLight *> tOmniItr(mOmniList); 
	tOmniItr.more(); tOmniItr.next())
	delete tOmniItr.cur();

    for(GDynaListIterator<SpotLight *> tSpotItr(mSpotList); 
	tSpotItr.more(); tSpotItr.next())
	delete tSpotItr.cur();

    for(GDynaListIterator<Camera *> tCamItr(mCamList); 
	tCamItr.more(); tCamItr.next())
	delete tCamItr.cur();

    for(GDynaListIterator<Mesh *> tMeshItr(mMeshList); 
	tMeshItr.more(); tMeshItr.next())
	delete tMeshItr.cur();

    for(GDynaListIterator<Material *> tMatItr(mMtlList); 
	tMatItr.more(); tMatItr.next())
	delete tMatItr.cur();

    for(GDynaListIterator<MatProp *> tMatPItr(mMPropList); 
	tMatPItr.more(); tMatPItr.next())
	delete tMatPItr.cur();
    
    
}



////
//// read Color
////
DOS_ifstream& operator >> (DOS_ifstream& rIN, Color& rColor) {
    Chunk tChunk(rIN);
    
    switch (tChunk.tag()) {
      case COLOR_F: 
	rIN >> rColor.mR >> rColor.mG >> rColor.mB;
	break;

      case COLOR_24: 
	byte tR, tG, tB;
	
	rIN >> tR >> tG >> tB;
	rColor.mR = (float)tR/255.0;
	rColor.mG = (float)tG/255.0;
	rColor.mB = (float)tB/255.0;

	break;

      default:     
	error ("Error parsing color!");
    }

    tChunk.end();

    return(rIN);
}

////
//// read Percentage
////
DOS_ifstream& operator >> (DOS_ifstream& rIN, Percentage& rP) {
    rP.mP = 0.0;

    Chunk tChunk(rIN);
    switch (tChunk.tag()) {
      case INT_PERCENTAGE: 
	word tP;
	rIN >> tP;
	rP.mP = tP / 100.0;
	break;

      case FLOAT_PERCENTAGE: 
	rIN >> rP.mP;
	break;

      default:     
	error("WARNING: Error parsing percentage");
    }

    tChunk.end();
    return(rIN);
}



////
//// read from 3DS-file
////
void Scene3DS::read(const char *pInName) {
    // open the input-stream
    mIN.open(pInName, ios::in | ios::bin);
    if(!mIN.good()) {
	sprintf(mMsg, "Could not open %s %s", pInName, "to read !");
	error(mMsg);
    }
    
    Chunk tChunk(mIN);
        
    // check the .3DS identification
    if (tChunk.tag() == M3DMAGIC )
	parse3DS (tChunk);
    else
	error("Error: Input file is not .3DS format !");

    tChunk.end();

    mIN.close();
}


void Scene3DS::parse3DS (const Chunk &pChunk) {
    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case MDATA: 
	    parseMData (tChunk);
	    break;
	  default:
	    break;
	}
	
	if(!(tChunk < pChunk))
	    break;
    }
}

//// parsing the mesh-block
void Scene3DS::parseMData(const Chunk &pChunk) {
    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case AMBIENT_LIGHT: 
	    // @RESERVED@
	    break;
	      
	  case SOLID_BGND: 
	    mIN >> mBGColor;
	    break;
	      
	  case USE_SOLID_BGND: 
	    mUseBGColor = true;
	    break;
	      
	  case FOG: 
	    // @RESERVED@
	    break;
	    
	  case FOG_BGND: 
	    // @RESERVED@
	    break;
	    
	  case USE_FOG: 
	    // @RESERVED@
	    break;
		
	  case MAT_ENTRY: 
	    parseMatEntry(tChunk);
	    break;
	
	  case NAMED_OBJECT: 
	    parseNamedObject(tChunk);
	    break;
	}
    }
}




//// parsing named object
void Scene3DS::parseNamedObject (const Chunk &pChunk) {
    
    mIN >> mObjName;
    cleanupName (mObjName);

    cerr << "Working on: " <<  mObjName << "\n";

    mMesh = 0;

    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case N_TRI_OBJECT: 
	    parseNTriObject (tChunk);
	    break;
	      
	  case N_DIRECT_LIGHT: 
	    parseNDirectLight (tChunk);
	    break;
	      
	  case N_CAMERA: 
	    parseNCamera();
	    break;
		
	  case OBJ_HIDDEN: 
	    if (mMesh) mMesh->mHidden = true;
	    break;
	      
	  case OBJ_DOESNT_CAST: 
	    if (mMesh) mMesh->mShadow = false;
	    break;
	}
    }
    

    if (mMesh) 	{
	mMesh->computeNormals();
	
	// @ADD@ to the mesh-list
	mMeshList.append(mMesh);
    }
    
}


//// parsing named triangles object
void Scene3DS::parseNTriObject (const Chunk &pChunk) {
    if(!(mMesh = new Mesh (mObjName, 0, 0)) )
	error("Out of memory allocating mesh");

    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case POINT_ARRAY: 
	    parsePointArray();
	    break;
		
	  case FACE_ARRAY: 
	    parseFaceArray (tChunk);
	    break;
		
	  case MESH_MATRIX: 
	    if(mMesh) 
		mIN >> mMesh->mMatrix;
	    break;
	}
    }
}

//// parsing points array
void Scene3DS::parsePointArray(void) {
    mIN >> mMesh->mNumVerts;
    if(!(mMesh->mVertex = new GPoint3[mMesh->mNumVerts]))
	error("Out of memory allocating mesh");

    for (register int i = 0; i < mMesh->mNumVerts; i++) {
	
	mIN >> mMesh->mVertex[i];
	
	// update the world bound
	mWorldBound += mMesh->mVertex[i];
    }
}

//// parsing faces array
void Scene3DS::parseFaceArray (const Chunk &pChunk) {
    
    mIN >> mMesh->mNumFaces;
    if(!(mMesh->mFaces = new Face[mMesh->mNumFaces]))
	error("Out of memory allocating mesh");

    if(!(mMesh->mMtl = new (Material *)[mMesh->mNumFaces]))
	error("Out of memory allocating mesh");

    // read face-indices
    for (register int i = 0; i < mMesh->mNumFaces; i++) {
	mIN >> mMesh->mFaces[i];
	mMesh->mMtl[i] = 0;
    }

    // read face-material
    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case MSH_MAT_GROUP: 
	    parseMshMatGroup();
	    break;
		
	  case SMOOTH_GROUP: 
	    parseSmoothGroup();
	    mMesh->mSmooth = true;
	    break;
	}
    }

    // no material was found. Set it to default!
    for (i = 0; i < mMesh->mNumFaces; i++) {
	if (!mMesh->mMtl[i])
	    mMesh->mMtl[i] = updateMaterials ("Default", 0);
    }
}

////
//// Add a new material to the material list 
////
Material* Scene3DS::updateMaterials (char *pNewMatName, int pExt) {
    Material *tMat = 0;

    // searching named material in the list...
    for(GDynaListIterator<Material *> _Itr(mMtlList); _Itr.more(); _Itr.next()) {
	if(strcmp(_Itr.cur()->getName(), pNewMatName) == 0) {
	    tMat = _Itr.cur();
	    break;
	}
    }
    
    // adding the new mat. to the material list
    if (!tMat) {
	if(!(tMat = new Material(pNewMatName, pExt)))
	    error("Out of memory adding material");
	
	// @ADD@ to materials list
	mMtlList.append(tMat);
    }

    return (tMat);
}



////
//// read materials for faces of a mesh
////
void Scene3DS::parseMshMatGroup(void) {
    Material *tMat;
    char tName[80];
    word  tNum;
    word  tFaceNr;

    mIN >> tName;
    cleanupName (tName);

    tMat = updateMaterials (tName, 0);

    mIN >> tNum;

    for (register int i = 0; i < tNum; i++) {
	mIN >> tFaceNr;
	mMesh->mMtl[tFaceNr] = tMat;
    }
}


void Scene3DS::parseSmoothGroup(void) {
    // @RESERVED@
    cerr << "smooth!";
}



////
//// Parsing an material-entry
////
void Scene3DS::parseMatEntry (const Chunk &pChunk) {
   
    MatProp *tMatProp = new MatProp;
    
    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    {
	switch (tChunk.tag()) {
	  case MAT_NAME: 
	    mIN >> tMatProp->mName;
	    cleanupName (tMatProp->mName);
	    break;

	  case MAT_AMBIENT: 
	    mIN >> tMatProp->mAmbient;
	    break;

	  case MAT_DIFFUSE: 
	    mIN >> tMatProp->mDiffuse;
	    break;

	  case MAT_SPECULAR: 
	    mIN >> tMatProp->mSpecular;
	    break;

	  case MAT_SHININESS: {
	      
	      Percentage tP;
	      mIN >> tP;
	      tMatProp->mShininess = tP * 100.0;
	      break;
	  }
	    
	  case MAT_TRANSPARENCY: {
	      
	      Percentage tP;
	      mIN >> tP;
	      tMatProp->mTransparency = tP();
	      break;
	  }
	    
	  case MAT_SELF_ILLUM: 
	    tMatProp->mSelfIllum = true;
	    break;

	  case MAT_REFLMAP: 
	    // @RESERVED@
	    //tMatProp->mReflection = parsePercentage();
	    //(void)parse_mapname (tChunk);
	    break;

	  case MAT_ACUBIC: 
	    // @RESERVED@
	    //if (tMatProp->mReflection == 0.0)
	    //tMatProp->mReflection = 1.0;
	    break;

	  case MAT_TEXMAP: 
	    // @RESERVED@
	    //tMatProp->mTexStrength = parsePercentage();
	    //strcpy (tMatProp->mTexMap, parseMapName (tChunk));
	    break;

	  case MAT_BUMPMAP: 
	    // @RESERVED@
	    //tMatProp->mBumpStrength = parsePercentage();
	    //strcpy (tMatProp->mBumpMap, parseMapName (tChunk));
	    break;
	}
    }


    mMPropList.append(tMatProp);
}




////
//// read Cameras
////
void Scene3DS::parseNCamera(void) {
    Camera *tCam = new Camera;
    if(!tCam) error("Out of memory adding camera"); 
    
    mIN >> *tCam;
    
    mCamList.append(tCam);
}




//////////////////////////////////////////////////////////////////////////////
// Lights
//////////////////////////////////////////////////////////////////////////////
void Scene3DS::parseNDirectLight (const Chunk &pChunk) {
    bool light_off = false;
    bool spot_flag = false;

    GPoint3 tPos;
    Color tColor;

    // @WARNING@
    mIN >> tPos;
    mIN >> tColor;

     
    for(Chunk tChunk(mIN); tChunk <= pChunk; tChunk.next())	
    { 
	switch (tChunk.tag()) {
	      
	  case DL_OFF: 
	    light_off = true;
	    break;
	      
	  case DL_SPOTLIGHT: 
	    parseDLSpotlight();
	    spot_flag = true;
	    break;
	}
    }

    if (light_off)
	return;
    
    
    mOmniList.append(new OmniLight(mObjName, tPos, tColor));
}

////
//// spotlight unsupported in NFF 
////
void Scene3DS::parseDLSpotlight(void) {
    // @RESEVED@
    //mIN >> target >> hotspot >> falloff;
}


Mesh::Mesh (const char *pName, int pNumVerts, int pNumFaces) 
:NObject(pName)
{
    mNumVerts = pNumVerts;

    if (pNumVerts <= 0)
	mVertex = 0;
    else {
	mVertex = new GPoint3[pNumVerts];
	if (!mVertex) error("Out of memory allocating mesh");
    }

    mNumFaces = pNumFaces;

    if (pNumFaces <= 0) {
	mFaces = 0;
	mMtl = 0;
    }
    else {
	mFaces = new Face[pNumFaces];
	if (!mFaces) error("Out of memory allocating mesh");

	mMtl = new (Material*)[pNumFaces];
	if (!mMtl) error("Out of memory allocating mesh");
    }

    mNormals = 0;
    mHidden = false;
    mShadow = true;
    mSmooth = false;
}




Mesh::~Mesh (void) {
    gkillArray(mVertex);
    gkillArray(mFaces);
    gkillArray(mNormals);
    
    // @WARING@
    gkillArray(mMtl);

}


void Mesh::getSmoothData(PolySmoother *pSmoother) const {
    for (register int i = 0; i < mNumFaces; i++) {
	int a = mFaces[i].a;
	int b = mFaces[i].b;
	int c = mFaces[i].c;
	GPoint3 tVerts[3];
	tVerts[0] = mVertex[a]; 
	tVerts[1] = mVertex[b]; 
	tVerts[2] = mVertex[c];
	
	Polygon *tPolygon = new Polygon(3, tVerts, (void*)(getMaterials()[i])); 
	
	pSmoother->addPolygon(tPolygon);
    }

    pSmoother->computeNormals();
    
}

    
bool Mesh::computeNormals(void) {
    
    //// generate vertex-normals 
    if(mSmooth) {
	if(!(mNormals = new GVector3[mNumVerts]))
	    error("Out of memory allocating mesh");
		
	for (register int i = 0; i < mNumFaces; i++) {
	    int a = mFaces[i].a;
	    int b = mFaces[i].b;
	    int c = mFaces[i].c;
	    
	    // compute the face\'s normal
	    //GVector3 tNormal = (mVertex[c] - mVertex[b]).cross(mVertex[a] - mVertex[b]);
	    GVector3 tT1 = mVertex[a] - mVertex[b];
	    GVector3 tT2 = mVertex[a] + mVertex[b];
	    GVector3 tNormal = GVector3(tT1.getY() * tT2.getZ(),
					tT1.getZ() * tT2.getX(),
					tT1.getX() * tT2.getY());
	    
	    tT1 = mVertex[b] - mVertex[c];
	    tT2 = mVertex[b] + mVertex[c];
	    tNormal += GVector3(tT1.getY() * tT2.getZ(),
			       tT1.getZ() * tT2.getX(),
			       tT1.getX() * tT2.getY());
	    
	    tT1 = mVertex[c] - mVertex[a];
	    tT2 = mVertex[c] + mVertex[a];
	    tNormal += GVector3(tT1.getY() * tT2.getZ(),
				tT1.getZ() * tT2.getX(),
				tT1.getX() * tT2.getY());


	    if(tNormal.sqdist() < gEpsilon2) {
		// @WARNING@
		cerr << "WARNING: degenerated polygon removed!\n";
		mFaces[i].flags = 0;
	    }

	    else {
		mFaces[i].flags = 1;
		tNormal = tNormal.unitize();
		if(mNormals[a].sqdist() < gEpsilon2)
		    mNormals[a] += tNormal;
		else if(mNormals[a].unitize().dot(tNormal) > Scene3DS::mMinDot)
		    mNormals[a] += tNormal;
		
		if(mNormals[b].sqdist() < gEpsilon2)
		    mNormals[b] += tNormal;
		else if(mNormals[b].unitize().dot(tNormal) > Scene3DS::mMinDot)
		    mNormals[b] += tNormal;
		
		if(mNormals[c].sqdist() < gEpsilon2)
		    mNormals[c] += tNormal;
		else if(mNormals[c].unitize().dot(tNormal) > Scene3DS::mMinDot)
		    mNormals[c] += tNormal;
	    }
	}
    }
}

    




void error(const char *err) {
    cerr << err << '\n';
    exit(3);
}

void cleanupName (char *name) {
    char *tmp = new char[strlen(name)+1];
    int  i;

    // Remove any leading blanks or quotes 
    i = 0;
    while ((name[i] == ' ' || name[i] == '"') && name[i] != '\0')
	i++;
    
    strcpy (tmp, &name[i]);

    // Remove any trailing blanks or quotes 
    for (i = strlen(tmp)-1; i >= 0; i--) {
	if (isprint(tmp[i]) && !isspace(tmp[i]) && tmp[i] != '"')
	    break;
	else
	    tmp[i] = '\0';
    }

    strcpy (name, tmp);

    // Prefix the letter 'N' to materials that begin with a digit 
    if (!isdigit (name[0]))
	strcpy (tmp, name);
    else {
	tmp[0] = 'N';
	strcpy (&tmp[1], name);
    }

    // Replace all illegal charaters in name with underscores 
    for (i = 0; tmp[i] != '\0'; i++) {
	if (!isalnum(tmp[i]))
	    tmp[i] = '_';
    }

    strcpy (name, tmp);

    delete (tmp);
}



float Camera::getFOV (void) {
    static float sLensTable[13] = { 
	15.0, 17.0, 24.0, 35.0, 50.0, 
	85.0, 100.0, 135.0, 200.0,
	500.0, 625.0, 800.0, 1000.0 
    };
    
    static float sFovTable[13] = { 
	115.0, 102.0, 84.0, 63.0, 
	46.0, 28.0, 24.0, 18.0,
	12.0, 5.0, 4.0, 3.125, 2.5 
    };

    float fov, f1, f2, l1, l2;
    int   i;

    if (mLens < 15.0)
	mLens = 15.0;
    else if (mLens > 1000.0)
	mLens = 1000.0;

    for (i = 0; i < 13; i++)
	if (mLens < sLensTable[i])
	    break;

    if (i == 13)
	i = 12;
    else if (i == 0)
	i = 1;

    f1 = sFovTable[i-1];
    f2 = sFovTable[i];
    l1 = sLensTable[i-1];
    l2 = sLensTable[i];

    fov = f1 + (mLens - l1) * (f2 - f1) / (l2 - l1);

    return fov;
}

