//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Fri Jan 26 21:41:48 1996 - juhu@rz.tu-ilmenau.de

#include "3ds.h"

class SceneGeo: public Scene3DS {
  public:
    SceneGeo(void) : Scene3DS() { }
    ~SceneGeo(void) {}
    
    // write to a Geo-file after reading the 3DS-scene
    void write(const char *);

  private:
    ofstream mOUT;

    void writeMesh (Mesh *pMesh);
  
};

void SceneGeo::write(const char *pOutput) {
    // output
    mOUT.open(pOutput);
    if(!mOUT.good()) {
	char tMsg[255];
	sprintf(tMsg, "Could not open %s %s", pOutput, "to write.");
	error(tMsg);
    }

    char *tPtr = strchr(pOutput, '.');
    char tStr[255];
    strncpy(tStr, pOutput, tPtr - &(pOutput[0]));
    tStr[tPtr - &(pOutput[0])] = 0;
	 

    mOUT << "GFT " << tStr << " {} {} {n f} {String Object} {Creates a " <<  tStr
	 << "with specified {ARG 1 name} and {ARG 2 father}.} \n {\n";
    
    //// Writing meshes
    for(GDynaListIterator<Mesh *> tMeshItr(mMeshList); 
	tMeshItr.more(); tMeshItr.next()) 
    {
	strcpy(tStr, tMeshItr.cur()->getName());
	
	for(register int i=0; i < strlen(tStr); i++) {
	    tStr[i] = tolower(tStr[i]);
	}
    
	mOUT << "OffPolygon $this." << tStr << " " << tStr << " $this\n"
	     << "$this." << tStr << " -setScale 0.04\n";
	
	writeMesh(tMeshItr.cur());
    }
    
    mOUT << "\n}\n";
    

    // close the output stream
    mOUT.close();

    cerr << "Done.\n";
    
}



////
//// Write Mesh
////
void SceneGeo::writeMesh (Mesh *pMesh) {
    static Material *sMat = 0;

    
    register int i;
    
    if (pMesh->isHidden() )
	return;
 
    ofstream tOUT;
    
    char tName[255];
    sprintf(tName, "%s%s", pMesh->getName(), ".geo");
    for(i=0; i < strlen(tName); i++) {
	tName[i] = tolower(tName[i]);
    }
    

    // output
    tOUT.open(tName);
    if(!tOUT.good()) {
	char tMsg[255];
	sprintf(tMsg, "Could not open %s %s", tName, "to write.");
	error(tMsg);
    }

    cerr << "Writing " << tName << " ...\n";

    tOUT <<  pMesh->getNumVerts() << " " << pMesh->getNumFaces() << " 0\n";
    
    for (i = 0; i < pMesh->getNumVerts(); i++) {
	tOUT << pMesh->getVertex()[i].getX() << " "
	    << pMesh->getVertex()[i].getY() << " "
	    << pMesh->getVertex()[i].getZ() << "\n";
    }
    
    
    for (i = 0; i < pMesh->getNumFaces(); i++) {
// 	
// 	// bad face removed!
// 	if(!pMesh->getFaces()[i].flags)
// 	    continue;
	tOUT << "3 " << pMesh->getFaces()[i].a + 1 << " "
	     << pMesh->getFaces()[i].b + 1 << " " 
	     << pMesh->getFaces()[i].c + 1 << "\n";
	
    }

    tOUT.close();
    
}



static void printHelp(void) {
    cerr << "3DS2Geo 0.2: Convert .3DS to .Geo\n\n"; 
    cerr << "Usage: 3ds2Geo <filename> [options] \n";
    cerr << "  -oFile   set output-filename\n";
}









int main(int argc, char *argv[]) {
    char tInFileName[256]="";
    char tOutFileName[256]="";
    
    if(argc<2) {
	printHelp();
	exit(3);
    }
    

    argc--;
    argv++;
    while (argc--) {
	if (argv[0][0] != '-') {
	    strcpy(tInFileName, argv[0]);
	    argv++;
	    continue;
	}
	
	// Output
	else if (strncasecmp(argv[0]+1, "o", 1) == 0) {
	    strcpy(tOutFileName, argv[0] + 2);
	}
	
	else
	    cerr << "Ignoring unknown option " << argv[0] << '\n';
	argv++;
    }

    if(tInFileName[0]==0x0) {
	printHelp();
	exit(3);
    }
    
    
    if(tOutFileName[0]==0x0) {
	char *tPtr = strchr(tInFileName, '.');
	if(tPtr) {
	    strncpy(tOutFileName, tInFileName, tPtr - &(tInFileName[0]));
	    strcat(tOutFileName, ".cls");
	}
	else {
	    strcpy(tOutFileName, tInFileName);
	    strcat(tOutFileName, ".cls");
	}
    }
      
    SceneGeo aSc;

    cerr << "Parsing " << tInFileName << " ...\n";
    aSc.read(tInFileName);
   
    cerr << "Writing " << tOutFileName << " ...\n";
    aSc.write(tOutFileName);
    
 
    return (0);
}


