//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan 30 09:00:47 1996 - juhu@rz.tu-ilmenau.de

#include "3ds.h"

class SceneNFF: public Scene3DS {
  public:
    SceneNFF(void) : Scene3DS() { }
    ~SceneNFF(void) {}
    
    // write to a NFF-file after reading the 3DS-scene
    void write(const char *pOuput);
    
    void setReflectionFlag(bool pFlag) {
	mEnableSpeculare = pFlag;
    }
    
    
  private:
    ofstream mOUT;
    static bool mEnableSpeculare;
    
    void writeCamera (const char *pName, 
		      const GPoint3 &pPos, 
		      const GPoint3 &pTarget, 
		      float pFOV,
		      bool pCommentOut = false);

    void writeLight (const char *pName, const GPoint3 &pPos, const Color &pColor);
    
    void writeMesh (Mesh *pMesh);
    
    void writeMatProp(MatProp *pMP);
};

bool  SceneNFF::mEnableSpeculare = false;


void SceneNFF::write(const char *pOutput) {
    // output
    mOUT.open(pOutput);
    if(!mOUT.good()) {
	char tMsg[255];
	sprintf(tMsg, "Could not open %s %s", pOutput, "to write.");
	error(tMsg);
    }

    
    //// World\'s bounding box
    mOUT << "#####################################################################\n";
    mOUT << "# World's bounding box:\n";
    mOUT << "# Min :" << mWorldBound.getMin() << "\n";
    mOUT << "# Max :" << mWorldBound.getMax() << "\n";
    mOUT << "#####################################################################\n";

    //// solid background\'s color
    if(mUseBGColor)
	mOUT << "b " << mBGColor << "\n";
    

    //// write cameras
   
    if(!mCamList.count()) {
	cerr << "Warning: no camera was found! Standard camera inserted.\n";
	GPoint3 tTarget = (mWorldBound.getMin() +  mWorldBound.getMax())/2.0;
// 	GFloat tDist = gmax(mWorldBound.getMax().getY() - mWorldBound.getMin().getY(),
// 			    mWorldBound.getMax().getZ() - mWorldBound.getMin().getZ());
// 	// sqrt(3)/2
// 	tDist = tDist * 0.866  +  mWorldBound.getMax().getX();
// 	
// 	GPoint3 tPos(tDist, tTarget.getY(), tTarget.getZ());
//      writeCamera("Standard Camera", tPos, tTarget, 35.0, 0.0 );

	GPoint3 tPos(2.0 * mWorldBound.getMax().getX() - mWorldBound.getMin().getX(),
		     2.0 * mWorldBound.getMax().getY() - mWorldBound.getMin().getY(),
		     tTarget.getZ());
	
	writeCamera("Standard Camera", tPos, tTarget, 60.0 );
    }
    else {
	GDynaListIterator<Camera *> tCamItr(mCamList);
	Camera *tCam = tCamItr.cur();
	
	cerr << "Writing camera: " << tCam->getName() << "...\n";
	

	// write the first camera w/o comments
	writeCamera(tCam->getName(), 
		    tCam->getPos(), 
		    tCam->getTarget(), 
		    tCam->getFOV(), 
		    false);
	
	for(tCamItr.next(); tCamItr.more(); tCamItr.next()) {
	    tCam = tCamItr.cur();
	    // comment out all the others
	    writeCamera(tCam->getName(), 
			tCam->getPos(), 
			tCam->getTarget(), 
			tCam->getFOV(), 
			true);
	}
	
    }

    
    //// Writing lights

        

    // no lightsource was found!
    if(!mOmniList.count()) {
	cerr << "Warning: no lightsource was found! Standard point-light inserted.\n";
	
	GPoint3 tPos(2.0 * mWorldBound.getMax().getX() - mWorldBound.getMin().getX(),
		     2.0 * mWorldBound.getMax().getY() - mWorldBound.getMin().getY(),
		     (mWorldBound.getMin().getZ() +  mWorldBound.getMax().getZ())/2.0);
	writeLight ("Standard Pointlight", tPos, Color(1.,1.,1.));
    }
    else {
	
	for(GDynaListIterator<OmniLight *> tOmniItr(mOmniList); 
	    tOmniItr.more(); tOmniItr.next()) 
	{
	    OmniLight *tLight = tOmniItr.cur();
	    
	    cerr << "Writing light : " << tLight->getName() << "...\n";

	    writeLight(tLight->getName(), tLight->getPos(), tLight->getColor());
	}
    }
    

    //// Writing meshes
    for(GDynaListIterator<Mesh *> tMeshItr(mMeshList); 
	tMeshItr.more(); tMeshItr.next()) 
    {
	cerr << "Writing mesh: " << tMeshItr.cur()->getName() << "...\n";
	
	writeMesh(tMeshItr.cur());
    }
    


    // close the output stream
    mOUT.close();

    cerr << "Done.\n";
    
}



////
//// Write camera
////
#define CCOUT {if(pCommentOut) mOUT << "#";}
void SceneNFF::writeCamera (const char *pName, 
			    const GPoint3 &pPos, 
			    const GPoint3 &pTarget, 
			    float pFOV,
			    bool pCommentOut)
{
    mOUT << "##############################################################################\n";
    mOUT << "# Camera: " << pName << "\n";
    mOUT << "##############################################################################\n";
    
    CCOUT; mOUT << "v\n";
    CCOUT; mOUT << "from " << pPos.getX() << " " << pPos.getY() << " " << pPos.getZ() << "\n";
    CCOUT; mOUT << "at " << pTarget.getX() << " " << pTarget.getY() << " " << pTarget.getZ() << "\n";
    CCOUT; mOUT << "up 0 0 1\n";
    CCOUT; mOUT << "angle " << pFOV << "\n";
    CCOUT; mOUT << "hither 100\n";
    CCOUT; mOUT << "resolution 512 512\n";
    
}

////
//// Write Light
////
void SceneNFF::writeLight (const char *pName, const GPoint3 &pPos, const Color &pColor) {
    mOUT << "##############################################################################\n";
    mOUT << "# Light: " << pName << "\n";
    mOUT << "##############################################################################\n";
    
    mOUT << "l ";
    mOUT << pPos.getX() << " " << pPos.getY() << " " << pPos.getZ() << " " 
	 << pColor << "\n";
}




////
//// Write Mesh
////
void SceneNFF::writeMesh (Mesh *pMesh) {
    static Material *sMat = 0;

    
    register int i;
    
    if (pMesh->isHidden() )
	return;
 
    mOUT << "##############################################################################\n";
    mOUT << "# Object: " << pMesh->getName() << "\n";
    mOUT << "# Num. of Triangles: " <<  pMesh->getNumFaces() << "\n";
    mOUT << "# Num. of Vertices: " <<  pMesh->getNumVerts() << "\n";
    mOUT << "# Smoothing is:"; 
    if (pMesh->isSmooth()) mOUT << "ON\n"; else mOUT << "OFF\n";
    mOUT << "##############################################################################\n";
    
    if (!pMesh->isSmooth()) {
	for (i = 0; i < pMesh->getNumFaces(); i++) {
	
	    // remove bad face 
	    if(!pMesh->getFaces()[i].flags)
		continue;
	
	
	    // check for previos material
	    if(pMesh->getMaterials()[i] != sMat) {
		sMat = pMesh->getMaterials()[i];
	    
		MatProp *tProp = 0;  
		
		// searching for named material in the list...
		for(GDynaListIterator<MatProp *> _Itr(mMPropList); _Itr.more(); _Itr.next()) {
		    if(strcmp(_Itr.cur()->getName(), sMat->getName()) == 0) {
			tProp = _Itr.cur();
			break;
		    }
		}
		if(tProp) 
		    writeMatProp(tProp);
		else
		    cerr << "WARNING: material " << sMat->getName() << " not found !\n";
	    }

	
	    int a = pMesh->getFaces()[i].a;
	    int b = pMesh->getFaces()[i].b;
	    int c = pMesh->getFaces()[i].c;
	
	
	    GPoint3 tA = pMesh->getVertex()[a];
	    GPoint3 tB = pMesh->getVertex()[b];
	    GPoint3 tC = pMesh->getVertex()[c];

	
	    mOUT << "p 3\n" ;
	    mOUT << tA.getX() << " " << tA.getY() << " " << tA.getZ() << " " <<
		tB.getX() << " " << tB.getY() << " " << tB.getZ() << " " <<
		tC.getX() << " " << tC.getY() << " " << tC.getZ() << "\n";
	    
	}
    }
    // smooth
    else {
	
	

	PolySmoother tSmoother;
	tSmoother.enableEdgePreservation(mMinDot);
    
	pMesh->getSmoothData(&tSmoother);
	GDynaList<Polygon*> *tPolygons = tSmoother.getPolygons();

	for(GDynaListIterator<Polygon*> tItr(*tPolygons); tItr.more(); tItr.next()) {
	    Polygon *tPoly = tItr.cur();
	    
	    // check for previos material
	    if((Material*)(tPoly->mUser) != sMat) {
		sMat = (Material*)(tPoly->mUser);
	    
		MatProp *tProp = 0;  
		
		// searching for named material in the list...
		for(GDynaListIterator<MatProp *> _Itr(mMPropList); _Itr.more(); _Itr.next()) {
		    if(strcmp(_Itr.cur()->getName(), sMat->getName()) == 0) {
			tProp = _Itr.cur();
			break;
		    }
		}
		if(tProp) 
		    writeMatProp(tProp);
		else
		    cerr << "WARNING: material " << sMat->getName() << " not found !\n";
	    }

	    mOUT << "pp " << tPoly->mNumVerts << "\n" ;
	    for(register int i=0; i <  tPoly->mNumVerts; i++) {
		mOUT << tPoly->mVertices[i].getX() << " " 
		    << tPoly->mVertices[i].getY() << " " 
		    << tPoly->mVertices[i].getZ() << " "
		    << tPoly->mNormals[i].getX() << " "
		    << tPoly->mNormals[i].getY() << " "
		    << tPoly->mNormals[i].getZ() << "\n";
	    }
	    
	    

	}

	


// 
// 	
// 	if(pMesh->getNormals()[a].sqdist() < gEpsilon2 ||
// 	       pMesh->getNormals()[b].sqdist() < gEpsilon2 ||
// 	       pMesh->getNormals()[c].sqdist() < gEpsilon2) 
// 	    {
// 		mOUT << "p 3\n" ;
// 		mOUT << tA.getX() << " " << tA.getY() << " " << tA.getZ() << " " <<
// 		    tB.getX() << " " << tB.getY() << " " << tB.getZ() << " " <<
// 		    tC.getX() << " " << tC.getY() << " " << tC.getZ() << "\n";
// 	    }
// 
// 	    mOUT << "pp 3\n" ;
// 	    
// 	    GVector3 tNormal = pMesh->getNormals()[a].unitize();;
// 	    mOUT << tA.getX() << " " 
// 		 << tA.getY() << " " 
// 		 << tA.getZ() << " "
// 		 << tNormal.getX() << " " 
// 		 << tNormal.getY() << " " 
// 		 << tNormal.getZ() << "\n";
// 	      
// 	    tNormal = pMesh->getNormals()[b].unitize();
// 	    mOUT << tB.getX() << " " 
// 		 << tB.getY() << " " 
// 		 << tB.getZ() << " " 
// 		 << tNormal.getX() << " " 
// 		 << tNormal.getY() << " " 
// 		 << tNormal.getZ() << "\n";
// 	    
// 	    tNormal = pMesh->getNormals()[c].unitize();
// 	    mOUT << tC.getX() << " " 
// 		 << tC.getY() << " " 
// 		 << tC.getZ() << " " 
// 		 << tNormal.getX() << " " 
// 		 << tNormal.getY() << " " 
// 		 << tNormal.getZ() << "\n";
// 	}
	

    }
   
}


////
//// Write Material Properties
////
void SceneNFF::writeMatProp(MatProp *pMP) {
    float tR, tG, tB;

    
    mOUT << "# texture: " << pMP->getName() << "\n";
    mOUT << "f ";
   
    // max. specular color
    pMP->mSpecular.getRGB(tR, tG, tB);
    GFloat maxSpec = gmax(tR, tG); maxSpec = gmax(maxSpec, tB);
       
    // find max. diffuse color
    pMP->mDiffuse.getRGB(tR, tG, tB);
    GFloat maxDiff = gmax3(tR, tG, tB);
    
    GFloat tSpec = (maxDiff > gEpsilon) ? maxSpec / maxDiff : 0.0; 


    
    // color
    mOUT << tR << " " << tG << " " << tB << " ";
    
    // diffuse
    mOUT <<  "1.0 ";
    
    // specular
    if(mEnableSpeculare)
	mOUT << tSpec << " ";
    else
	mOUT << "0 ";
    
    // shininess
    mOUT << pMP->mShininess << " ";
    
    // transparency, IOR
    if (pMP->mTransparency > 0.0) {
	mOUT << pMP->mTransparency << " 1.3\n";
    }
    else {
	mOUT << "0 0\n";
    }
}




static void printHelp(void) {
    cerr << "3DS2NFF 0.2: Convert .3DS to .NFF\n\n"; 
    cerr << "Usage: 3ds2nff <filename> [options] \n";
    cerr << "  -oFile   set output-filename\n";
    cerr << "  -sAngle  set the max. angle for smoothing\n";
    cerr << "  -r-      set reflection OFF (default ON)\n";
}









int main(int argc, char *argv[]) {
    char tInFileName[256]="";
    char tOutFileName[256]="";
    GFloat tSmoothAngle = 60.0;
    bool tReflFlag = true;
    

    if(argc<2) {
	printHelp();
	exit(3);
    }
    

    argc--;
    argv++;
    while (argc--) {
	if (argv[0][0] != '-') {
	    strcpy(tInFileName, argv[0]);
	    argv++;
	    continue;
	}
		
	// Output
	else if (strncasecmp(argv[0]+1, "o", 1) == 0) {
	    strcpy(tOutFileName, argv[0] + 2);
	}
	// SmoothAngle
	else if (strncasecmp(argv[0]+1, "s", 1) == 0) {
	    tSmoothAngle = atof(argv[0] + 2);
	}
	// reflection ON/OFF
	else if (strncasecmp(argv[0]+1, "r", 1) == 0) {
	    if(strncmp(argv[0]+2, "-", 1) == 0)
		tReflFlag = false;
	}
	
	else
	    cerr << "Ignoring unknown option " << argv[0] << '\n';
	argv++;
    }

    if(tInFileName[0]==0x0) {
	printHelp();
	exit(3);
    }
    
    if(tOutFileName[0]==0x0) {
	char *tPtr = strchr(tInFileName, '.');
	if(tPtr) {
	    strncpy(tOutFileName, tInFileName, tPtr - &(tInFileName[0]));
	    strcat(tOutFileName, ".nff");
	}
	else {
	    strcpy(tOutFileName, tInFileName);
	    strcat(tOutFileName, ".nff");
	}
    }
    
    
  
    SceneNFF aSc;
    aSc.setSmoothAngle(tSmoothAngle * M_D2R);
    aSc.setReflectionFlag(tReflFlag);
    
    cerr << "Parsing " << tInFileName << " ...\n";
    aSc.read(tInFileName);
   
    cerr << "Writing " << tOutFileName << " ...\n";
    aSc.write(tOutFileName);
    
 
    return (0);
}


