//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Wed Dec  6 06:06:08 1995 - juhu@rz.tu-ilmenau.de
#include <fstream.h>

#include <time.h>
#include "render/graycam.h"
#include "render/gpixmap.h"

#include "gx/gx_obj.h"
#include "gx/gx_csg.h"
#include "gx/gx_light.h"
#include "gx/gx_cam.h"

#include "gt_obj.h"

int main(int argc, char *argv[]) {

    char tInFileName[256]="";
    char tBuf[256];	 
    bool tTargaFormat = TRUE;
    GXOptions::PixHeight = 300;
    GXOptions::PixWidth = 300;


    argc--;
    argv++;
    while (argc--) {
	if (argv[0][0] != '-') {
	    strcpy(tInFileName, argv[0]);
	    argv++;
	    continue;
	}
		
	// antialising ON/OFF
	else if (strncasecmp(argv[0]+1, "a", 1) == 0) {
	    GXOptions::Antialiasing = TRUE;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::Antialiasing = FALSE;
	}

	// direct display ON/OFF
	else if (strncasecmp(argv[0]+1, "d", 1) == 0) {
	    GXOptions::DisplayFlag = TRUE;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::DisplayFlag = FALSE;
	}
	// Verbose message ON/OFF
	else if (strncasecmp(argv[0]+1, "v", 1) == 0) {
	    GXOptions::VerboseFlag = TRUE;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::VerboseFlag = FALSE;
	}
	// pause before exit ON/OFF
	else if (strncasecmp(argv[0]+1, "p", 1) == 0) {
	    GXOptions::PauseFlag = TRUE;
	    if(strncmp(argv[0]+2, "-", 1) == 0)
	      GXOptions::PauseFlag = FALSE;
	}
	
	// max. number of objects in a BSP-node
	else if (strncasecmp(argv[0]+1, "n", 1) == 0) {
	    GXOptions::MaxObjInNode = atoi(*argv + 2);
	}
	// Output
	else if (strncasecmp(argv[0]+1, "o", 1) == 0) {
	    GXOptions::OutputFlag = TRUE;
	    if(strncmp(argv[0]+2, "-", 1) == 0) {
		GXOptions::OutputFlag = FALSE;
		argv++;
		continue;
	    }
	    strcpy(tBuf, argv[0] + 2);
	}
	// pause before exit ON/OFF
	else if (strncasecmp(argv[0]+1, "f", 1) == 0) {
	    
	    if(strncasecmp(argv[0]+2, "s", 1) == 0)
	      tTargaFormat = FALSE;
	}
	
	// Height
	else if (strncasecmp(argv[0]+1, "h", 1) == 0) {
	    GXOptions::PixHeight = atoi(*argv + 2);
	}
	// Width
	else if (strncasecmp(argv[0]+1, "w", 1) == 0) {
	    GXOptions::PixWidth = atoi(*argv + 2);
	}
	else
	    cout << "Ignoring unknown option " << argv[0] << '\n';
	argv++;
    }

    
    
    
    
    //// intialize GENERIC-kernel..
    gInitialize();
   
    //// raytracer 
    GXRayTracer  aRT;
    GXPrimitiveList aSc(true);
    
    //// Lights
    gGXLights  = new GDLList<GXLight>(true);
    GXLight *_pL;
    
    _pL = new GXRectangularLight(GPoint3(0, 0, 2),
			  GPoint3(0, 0, 0),
			  1, 1, 2, 2,
			  9, 16, 10);
    _pL->setColor(GColorRGB(0.6, 0.6, 0.6));
    //gGXLights->append(_pL);
    
    _pL = new GXPointLight();
    _pL->setPosition(GPoint3(-10, 10, 10 ));
    _pL->setColor(GColorRGB(1, 0, 0));
    gGXLights->append(_pL);
    _pL = new GXPointLight();
    _pL->setPosition(GPoint3(10, 10, 10 ));
    _pL->setColor(GColorRGB(0, 1, 0));
    gGXLights->append(_pL);
    _pL = new GXPointLight();
    _pL->setPosition(GPoint3(0, -10, 10 ));
    _pL->setColor(GColorRGB(0, 0, 1));
    gGXLights->append(_pL);
    
    
    
    GSurface *_pSurf;
    GPoint3 *tVerts = new GPoint3[4];
    GTCone *_pCone;
    GTDisc *tDisc;
    GTSphere *_pSph;
    GTPolygon *tPoly;
    GXObjectList tCSGList;
    
    
    _pSurf = new GSurface();
    _pSurf->setAmbient(GColorRGB(0.5, 0.5, 0.5));
    _pSurf->setDiffuse(GColorRGB(0.1, 0.1, 0.1));
    _pSurf->setSpecular(GColorRGB(0.6, 0.6, 0.6));
    _pSurf->setShininess(100);

    
    // background
    tVerts[0] = GPoint3(-20, -20, -3);
    tVerts[1] = GPoint3(20, -20, -3);
    tVerts[2] = GPoint3(20, 20, -3);
    tVerts[3] = GPoint3(-20, 20, -3);
    tPoly = new GTPolygon(4, tVerts);
    tPoly->setSurface(_pSurf);
    aSc.append(tPoly);
    
/****/
    
    // C
    _pCone = new GTCone(GPoint3(0, 0, -1), 3, 
			GPoint3(0, 0, 1),  3);
    _pCone->setSurface(_pSurf);    
    tCSGList.append(_pCone);
    tDisc = new GTDisc(GPoint3(0, 0, -1), GVector3(0, 0, -1), 3);
    tDisc->setSurface(_pSurf);    
    tCSGList.append(tDisc);
    tDisc = new GTDisc(GPoint3(0, 0, 1), GVector3(0, 0, 1), 3);
    tDisc->setSurface(_pSurf);    
    tCSGList.append(tDisc);
    GXPrimitive *tCSG1 = new GX_CSG_Union(tCSGList);
    
    tCSGList.reset();
    _pCone = new GTCone(GPoint3(0, 0, -1.2), 2.55, 
			GPoint3(0, 0, 1.2),  2.55);
    _pCone->setSurface(_pSurf);    
    tCSGList.append(_pCone);
    tDisc = new GTDisc(GPoint3(0, 0, -1.2), GVector3(0, 0, -1), 2.55);
    tDisc->setSurface(_pSurf);    
    tCSGList.append(tDisc);
    tDisc = new GTDisc(GPoint3(0, 0, 1.2), GVector3(0, 0, 1), 2.55);
    tDisc->setSurface(_pSurf);    
    tCSGList.append(tDisc);
    GXPrimitive *tCSG2 = new GX_CSG_Union(tCSGList);
    
    tCSGList.reset();
    tCSGList.append(tCSG1);
    tCSGList.append(tCSG2);
    
    _pSph = new GTSphere(GPoint3(2.55, 0, 0), 0.6);
    _pSph->setSurface(_pSurf);    
    tCSGList.append(_pSph);
    _pSph = new GTSphere(GPoint3(-2.55, 0, 0), 0.6);
    _pSph->setSurface(_pSurf);    
    tCSGList.append(_pSph);
    _pSph = new GTSphere(GPoint3(0, 2.55, 0), 0.6);
    _pSph->setSurface(_pSurf);    
    tCSGList.append(_pSph);
    _pSph = new GTSphere(GPoint3(0, -2.55, 0), 0.6);
    _pSph->setSurface(_pSurf);    
    tCSGList.append(_pSph);
    
    aSc.append(new GX_CSG_Difference(tCSGList) );

    //aSc.append(new GX_CSG_Intersection(tCSGList) );
    
/****/    
    
    
    
    
    GImage *aOutFile = 0;
    GPixmapDisplay  *aPix = 0;

    if(GXOptions::DisplayFlag) {	      
	aPix = new GPixmapDisplay("CSGTEST", 
				  GXOptions::PixWidth,
				  GXOptions::PixHeight);
	aRT.initRealization(*aPix);
    }

    if(GXOptions::OutputFlag) {
	if(tTargaFormat)
	  aOutFile = new GTargaImage("");
	else
	  aOutFile = new GSGIImage("");
	if(!strchr(tBuf, '.')) {
	    strcat(tBuf, ".");
	    strcat(tBuf, aOutFile->getExtension());
	}
	strcpy(GXOptions::OutFileName, tBuf);
	aOutFile->setFilename(GXOptions::OutFileName);
	aOutFile->setWidth(GXOptions::PixWidth);
	aOutFile->setHeight(GXOptions::PixHeight);
	aRT.setOutputImage(aOutFile);
    }
    
    aRT.setViewpoint(GPoint3(-5, -5, 10));
    aRT.setRefpoint(GPoint3(0, 0, 0));
    
    
    delete [] tVerts;
    
    //// and trace it now..
    aRT.rendering(aSc);
    
    if(GXOptions::PauseFlag) {
	cout << "press any key...\n" ;
	char c=getchar();
    }
    
    gkillPointer(aPix);
    gkillPointer(aOutFile);
    return (0);
}
