//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan 30 08:17:12 1996 - juhu@rz.tu-ilmenau.de
#include "render/graytrac.h"

#include "gx/gx_defs.h"
#include "gt_obj.h"

enum { X_AXIS=0, Y_AXIS=1, Z_AXIS=2 };

time_t beg, end;

////
//// @GTSphere@ 
////


// bool GTSphere::hitRay(const GRay3& pRay) {
// 
//     GFloat _d;
//     
//     // origin to center 
//     GVector3 _o2c(mCenter - pRay.getPoint());      
//     // square of tangential distance
//     GFloat _tan_sq = _o2c.dot(_o2c) - mRad * mRad;
//    
//     // projected distance of _o2c to rays direction  
//     GFloat _project = _o2c.dot(pRay.getVector());
//    
//     // if the rays origin is outside and the rays direction
//     // points away to the sphere, no intersection occurs.
//     if( _tan_sq >= 0 && _project < gEpsilon ) 
//       return (false);;
//     
//     _d = _project * _project - _tan_sq;
// 
//     //@WARNING@ : Checking for validity ???
//     return (_d >= gEpsilon);
// }

bool GTSphere::intersect(const GRay3& pRay, GInterSectionList& rInter) {
    bool _HIT = false;
    gstat(GXStatistics::Sphere.mTests++);
    
    GFloat _d, _tmin, _tmax;
    
    // origin to center 
    GVector3 _o2c(mCenter - pRay.getPoint());      
    // square of tangential distance
    GFloat _tan_sq = _o2c.dot(_o2c) - mRad * mRad;
   
    // projected distance of _o2c to rays direction  
    GFloat _project = _o2c.dot(pRay.getVector());
   
    // if the rays origin is outside and the rays direction
    // points away to the sphere, no intersection occurs.
    if( _tan_sq >= 0 && _project < gEpsilon ) 
	goto L_EXIT;
    
    _d = _project * _project - _tan_sq;
    
    if (_d < 0) 
	goto L_EXIT;
   
    _d = sqrt(_d);
    _tmax = _project + _d;
    _tmin = _project - _d;
    
    // OUT
    if (pRay.isValid(_tmax)) {
	rInter.add( GInterSection( self, _tmax, false));
	_HIT = true;
    } 
    // IN
    if(pRay.isValid(_tmin)) {
	rInter.add( GInterSection( self, _tmin, true));
	_HIT = true;
    } 
    
    L_EXIT:
    gstat(GXStatistics::Sphere.mHits += _HIT);
    
    return (_HIT);
}
    


////
//// @GTPolygon@CTor@
////
GTPolygon::GTPolygon(int pNumVerts,	
		     GPoint3 *pVertices)
: GTPrimitive()
{
    setFlags(getFlags() | GRayTraceable::PLANAR);

    mVertices = 0;
    mUV_Vertices = 0;
    mValid = true;
   
    mNumVerts = pNumVerts;		  
    mVertices = new GPoint3[mNumVerts];
    for (register int i = 0; i < mNumVerts; i++)
	mVertices[i] = pVertices[i];
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();
}

////
//// @GTPolygon@CTor@
////
GTPolygon::GTPolygon(const GVertexAggregate& pVertices) 
: GTPrimitive() 
{
    // im convex!
    setFlags(getFlags() | GRayTraceable::PLANAR);
    mVertices = 0;
    mUV_Vertices = 0;
    mValid = true;
    
    mNumVerts = pVertices.getNumber();		  
    mVertices = new GPoint3[mNumVerts];

    for(register int i = 0; i< mNumVerts; i++) 
	mVertices[i] = (pVertices[i]).getPoint();
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();

}

////
//// @GTPolygon@DTor@
////
GTPolygon::~GTPolygon(void)
{
  //gkillArray(mVertices);
    gkillArray(mUV_Vertices);
}


////
//// @GTPolygon@precompute@
////
void GTPolygon::precompute(void)
{
    //// Compute the plane equation of the polygon.
    //GVector3 tX((mVertices[2] - mVertices[1]).cross(mVertices[0] - mVertices[1]));
    
    GVector3 tX(0, 0, 0);
    
    for (register int i = 0; i < mNumVerts; i++) {
	int j = (i+1) % mNumVerts;
	GVector3 tT1 = mVertices[i] - mVertices[j];
	GVector3 tT2 = mVertices[i] + mVertices[j];
	tX += GVector3(tT1.getY() * tT2.getZ(),
		       tT1.getZ() * tT2.getX(),
		       tT1.getX() * tT2.getY());
    }
    

    if ( tX.sqdist() > gEpsilon2) {
	mValid = true;
	mNormal = tX.unitize();
    }
    else {
	mValid = false;
	return;
    }
    mDist = mNormal.dot(mVertices[0]);

    //// Determine the dominant axis of the normal
    mDominantAxis = X_AXIS;		
    GFloat tMax = fabs(mNormal.getX());
    if (fabs(mNormal.getY()) > tMax) { 
	tMax = fabs(mNormal.getY());	  
	mDominantAxis = Y_AXIS;
    }
    if (fabs(mNormal.getZ()) > tMax)	
	mDominantAxis = Z_AXIS;		

    getWorldBounds().min().project(mUmin, mVmin, mDominantAxis);
    getWorldBounds().max().project(mUmax, mVmax, mDominantAxis);
    mUV_Vertices = new GFloat2[mNumVerts];
    for (i = 0; i < mNumVerts; i++) 
	mVertices[i].project(mUV_Vertices[i][X_AXIS], 
			     mUV_Vertices[i][Y_AXIS], mDominantAxis);
    
    gkillArray(mVertices);
}


inline bool __intersectSegment(GFloat pX,  GFloat pY, 
			       GFloat pU1, GFloat pV1, 
			       GFloat pU2, GFloat pV2)
{

    pU1 -= pX; pU2 -= pX;
    pV1 -= pY; pV2 -= pY;

    //// check the intervall
    if ((pV1 > 0 && pV2 > 0) ||
	(pV1 < 0 && pV2 < 0) ||
	(pU1 < 0 && pU2 < 0))
	return false;

    if (pU1 > 0 && pU2 > 0)
	return true;

    GFloat tYdiff = pV2 - pV1;
    if (fabs(tYdiff) < gEpsilon ) {	
	if (((fabs(pV1) > gEpsilon) || (pU1 > 0) || (pU2 > 0)))
	    return false;
	return true;
    }

    return ((pU1 + pV1  * (pU1 - pU2) / tYdiff) > 0);
}

//// Polygon intersection routine.
bool GTPolygon::intersect(const GRay3& pRay, GInterSectionList& rInter) 
{
    bool _HIT = false;
    gstat(GXStatistics::Polygon.mTests++);

    GFloat tDot, tT;
    
    //// intersect w/ the plane
    if (!mValid)
	goto L_EXIT;

    tDot = mNormal.dot(pRay.getVector());
    if (fabs(tDot) < gEpsilon) 
	goto L_EXIT;
    
    // compute the t-parameter
    tT = (mDist - mNormal.dot(pRay.getPoint())) / tDot;
    if(!pRay.isValid(tT)) 
	goto L_EXIT;
    
    // project the I-point along the dominant axis
    GFloat tU, tV;
    pRay.extrap(tT).project(tU, tV, mDominantAxis);

    // is this point inside the bounding rechtangle ?
    if (tU < mUmin || tU > mUmax || tV < mVmin || tV > mVmax)
	goto L_EXIT;
      
#ifdef GTD_PIP_VER1
    //// test for odd number of intersections w/ polygon-segments
    register int i;
    int tCount;
    GFloat tU1, tV1, tU2, tV2;
    register GPoint3 *tVertexPtr;
    
    mVertices[mNumVerts - 1].project(tU1, tV1, mDominantAxis);
    
    for (i = 0, tVertexPtr = mVertices, tCount = 0; 
	 i < mNumVerts; 
	 i++, tVertexPtr++) 
    {
	tVertexPtr->project(tU2, tV2, mDominantAxis);
	tCount += __intersectSegment(tU, tV, tU1, tV1, tU2, tV2);
	
	tU1 = tU2;
	tV1 = tV2;
    }
    
    if(tCount & 1) {
	rInter.add( GInterSection( self, tT, (tDot < 0.0)));
	_HIT = true;
    }
#else
    
    register int    j, tV_flag0, tV_flag1, tInside_flag ;
    register GFloat *tVert0, *tVert1;

    tVert0 = mUV_Vertices[mNumVerts-1] ;
    tV_flag0 = ( tVert0[Y_AXIS] >= tV ) ;
    tVert1 = mUV_Vertices[0] ;

    tInside_flag = 0 ;
    for ( j = mNumVerts+1 ; --j ; ) {

	tV_flag1 = ( tVert1[Y_AXIS] >= tV ) ;
	if ( tV_flag0 != tV_flag1 ) {
	    if ( ((tVert1[Y_AXIS] - tV) * (tVert0[X_AXIS] - tVert1[X_AXIS]) >=
		  (tVert1[X_AXIS] - tU) * (tVert0[Y_AXIS] - tVert1[Y_AXIS])) 
		 == tV_flag1 ) 
		tInside_flag = !tInside_flag ;
	}

	tV_flag0 = tV_flag1 ;
	tVert0 = tVert1;
	tVert1 += 2 ;
    }
    
    if(tInside_flag) {
	rInter.add( GInterSection( self, tT, (tDot < 0.0)));
	_HIT = true;
    }
#endif
    
    L_EXIT:
    
    gstat(GXStatistics::Polygon.mHits += _HIT);
    
    return (_HIT);

}


////
//// @GTSmoothPolygon@
////
GTSmoothPolygon::GTSmoothPolygon(int pNumVerts,	
			   GPoint3  *pVertices,
			   GVector3 *pNormals)
: GTPolygon()
{
    mNumVerts = pNumVerts;		  
    mVertices = new GPoint3[mNumVerts];
    mNormals  = new GVector3[mNumVerts];
    for (register int i = 0; i < mNumVerts; i++) {
	mNormals[i] = pNormals[i].unitize();
	mVertices[i] = pVertices[i];
    }
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();
}

GTSmoothPolygon::GTSmoothPolygon(const GVertexAggregate& pVertices) 
: GTPolygon() 
{
    mNumVerts = pVertices.getNumber();		  
    mVertices = new GPoint3[mNumVerts];
    mNormals = new GVector3[mNumVerts];

    for(register int i = 0; i< mNumVerts; i++) { 
	mNormals[i] = ((pVertices[i]).getNormal())->unitize();
	mVertices[i] = (pVertices[i]).getPoint();
    }
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();

}

GTSmoothPolygon::~GTSmoothPolygon(void)
{
    gkillArray(mNormals);
}


static bool __computeUV(GFloat pX,  GFloat pY, 
			GFloat pU1, GFloat pV1, 
			GFloat pU2, GFloat pV2,
			GFloat& rU, GFloat& rV)
{

    pU1 -= pX; pU2 -= pX;
    pV1 -= pY; pV2 -= pY;

    GFloat tYdiff = pV1 - pV2;
    if (fabs(tYdiff) < gEpsilon ) 	
	if (((fabs(pV1) > gEpsilon) || (pU1 > 0) || (pU2 > 0)))
	    return false;
    rU = pV1 / tYdiff;
    rV = pU1 + rU  * (pU2 - pU1);
    
    return true;
}


GVector3 GTSmoothPolygon::getNormal(const GPoint3& pPoint) 
{
    GFloat tU, tV;
    pPoint.project(tU, tV, mDominantAxis);

    GVector3 tNormals[2];
    GFloat tVs[2];
    
    
    register GFloat *tVert0, *tVert1;
    tVert0 = mUV_Vertices[mNumVerts-1] ;
    tVert1 = mUV_Vertices[0] ;
    register int i, j;
    for (i = 0, j = 0; i < mNumVerts, j < 2; i++) 
    {
	GFloat u, v;
	if (__computeUV(tU, tV, tVert0[X_AXIS], tVert0[Y_AXIS], 
			tVert1[X_AXIS], tVert1[Y_AXIS], u, v)) 
	{
	    tVs[j] = v;
	    if (i) 
		tNormals[j] = glerp(u, mNormals[i-1], mNormals[i]);
	    else 
		tNormals[j] = glerp(u, mNormals[mNumVerts-1], mNormals[0]);
	    j++;
	}
	  
	tVert0 = tVert1;
	tVert1 += 2 ;

    }
    tV = tVs[0] / ( tVs[0] - tVs[1]);
    return (glerp(tV, tNormals[0], tNormals[1]).unitize());

}
 

////
//// @GTCone@
////
GTCone::GTCone(const GPoint3& pBase, GFloat pBaseRad,
	       const GPoint3& pApex, GFloat pApexRad)
: GTPrimitive()
{
    setFlags(getFlags() | GRayTraceable::CONVEX | GRayTraceable::BOUNDED);
    GVector3 tBase, tApex;
    
    //// switch base & apex-point if needed.
    if(pApexRad < pBaseRad) {
	tBase = pApex;
	tApex = pBase;
	gswap(pBaseRad, pApexRad);
    }
    else {
	tBase = pBase;
	tApex = pApex;
    }
    

    //// Find the axis and axis length.
    GVector3 tAxis = tApex - tBase;
    GFloat tLenght = tAxis.abs();
    gcheck(tLenght < gEpsilon, "the cone is degenerated !\n");
    tAxis /= tLenght;
    

    //// compute the modelling transformation 
    GFloat tantheta = (pApexRad - pBaseRad) / tLenght;
    if(tantheta < gEpsilon) {
	mCylFlag = true;
    }
    else {
	mCylFlag = false;
	GFloat lprime = pBaseRad / tantheta;
	tBase -= (lprime * tAxis) ;
	tLenght += lprime;
	mStartDist = lprime / tLenght;
    }
    
    GMatrix4 tMat;
    tMat = tMat.scale(GVector3(pApexRad, pApexRad, tLenght));
    
    GVector3 tRotAxis;
    if (1. - fabs(tAxis.getZ()) < gEpsilon) 
	tRotAxis = GVector3(1, 0, 0);
    else 
	tRotAxis = GVector3(tAxis.getY(), -tAxis.getX(), 0);
    
    tRotAxis = tRotAxis.unitize();
    GFloat n1 = tRotAxis.getX(); 
    GFloat n2 = tRotAxis.getY(); 
    GFloat n3 = tRotAxis.getZ();
    
    GFloat costheta = tAxis.getZ();
    GFloat sintheta = -sqrt(1 - costheta*costheta);

    GMatrix4 m1;
    m1.set(0,0, (n1*n1 + (1. - n1*n1)*costheta));
    m1.set(1,0, (n1*n2*(1 - costheta) + n3*sintheta));
    m1.set(2,0, (n1*n3*(1 - costheta) - n2*sintheta));
    m1.set(0,1, (n1*n2*(1 - costheta) - n3*sintheta));
    m1.set(1,1, (n2*n2 + (1 - n2*n2)*costheta));
    m1.set(2,1, (n2*n3*(1 - costheta) + n1*sintheta));
    m1.set(0,2, (n1*n3*(1 - costheta) + n2*sintheta));
    m1.set(1,2, (n2*n3*(1 - costheta) - n1*sintheta));
    m1.set(2,2, (n3*n3 + (1 - n3*n3)*costheta));
    
    m1.set(0,3, tBase.getX());
    m1.set(1,3, tBase.getY());
    m1.set(2,3, tBase.getZ());
    
    
    
    tMat = m1 * tMat;
    mModel.setMatrix(tMat);

    // compute a bounding box for BVH  
    computeBounds();

}

bool GTCone::intersect(const GRay3& pRay, GInterSectionList& rInter) {
    bool _HIT = false;

    gstat(GXStatistics::Cone.mTests++);

    GVector3 tPos = mModel.wc2mc(pRay.getPoint());
    GVector3 tDir = mModel.wc2mc(pRay.getPoint() + pRay.getVector()) - tPos;
    
    
    GFloat tLen = tDir.abs();
    tDir /= tLen;

    GFloat a, b, c, zmin;
    GFloat t1, t2, disc, zpos;
    
    if(mCylFlag) {
	a = tDir.getX() * tDir.getX() + tDir.getY() * tDir.getY();
	if(a < gEpsilon*gEpsilon)
	    goto L_EXIT;
	
	b = tDir.getX() * tPos.getX() + tDir.getY() * tPos.getY() ;
	c = tPos.getX() * tPos.getX() + tPos.getY() * tPos.getY() - 1;
	zmin = 0;
    }
    else {
	a = tDir.getX() * tDir.getX() + tDir.getY() * tDir.getY() 
	    - tDir.getZ()*tDir.getZ();
	b = tDir.getX() * tPos.getX() + tDir.getY() * tPos.getY() 
	    - tDir.getZ()*tPos.getZ();
	c = tPos.getX() * tPos.getX() + tPos.getY() * tPos.getY() 
	    - tPos.getZ()*tPos.getZ();
	
	zmin = mStartDist;
	if (fabs(a) < gEpsilon) {
	    
	    // Only one intersection point...
	    t1 = -0.5 * c / b;
	    zpos = tPos.getZ() + t1 * tDir.getZ();
	    if (zpos >= zmin && zpos <= 1.) {
		t1 /= tLen;
		// IN/OUT
		if (pRay.isValid(t1)) {
		    rInter.add( GInterSection( self, t1, true));
		    _HIT = true;
		}
		    
	    }
	    goto L_EXIT;
	}
    }
    
    
    disc = b*b - a*c;
    if (disc < 0.)
	goto L_EXIT;		
    
    disc = sqrt(disc);
    t1 = (-b - disc) / a;
    t2 = (-b + disc) / a;
    
    if (t1 > t2) 
	gswap(t1, t2);
    
    zpos = tPos.getZ() + t1 * tDir.getZ();
    if (zpos >= zmin && zpos <= 1.) {
	t1 /= tLen;
	if (pRay.isValid(t1)) {
	    rInter.add( GInterSection( self, t1, true));
	    _HIT = true;
	}
    }
    
    zpos = tPos.getZ() + t2 * tDir.getZ();
    if (zpos >= zmin && zpos <= 1.) {
	t2 /= tLen;
	if (pRay.isValid(t2)) {
	    rInter.add( GInterSection( self, t2, false));
	    _HIT = true;
	}
    }
	
    
    
    L_EXIT:

    gstat(GXStatistics::Cone.mHits += _HIT);
    
    return (_HIT);

}


// 
// bool GTCone::hitRay(const GRay3& pRay) {
//     bool _HIT = false;
// 
//     gstat(GXStatistics::Cone.mTests++);
// 
//     GVector3 tPos = mModel.wc2mc(pRay.getPoint());
//     GVector3 tDir = mModel.wc2mc(pRay.getPoint() + pRay.getVector()) - tPos;
//     //GVector3 tDir = mModel.getWorldMatrix().transpose() * pRay.getVector(); 
// 
//     GFloat tLen = tDir.abs();
//     tDir /= tLen;
// 
//     GFloat a, b, c, zmin;
//     GFloat t1, t2, disc, zpos;
//     
//     if(mCylFlag) {
// 	a = tDir.getX() * tDir.getX() + tDir.getY() * tDir.getY();
// 	if(a < gEpsilon*gEpsilon)
// 	    goto L_EXIT;
// 	
// 	b = tDir.getX() * tPos.getX() + tDir.getY() * tPos.getY() ;
// 	c = tPos.getX() * tPos.getX() + tPos.getY() * tPos.getY() - 1;
// 	zmin = 0;
//     }
//     else {
// 	a = tDir.getX() * tDir.getX() + tDir.getY() * tDir.getY() 
// 	    - tDir.getZ()*tDir.getZ();
// 	b = tDir.getX() * tPos.getX() + tDir.getY() * tPos.getY() 
// 	    - tDir.getZ()*tPos.getZ();
// 	c = tPos.getX() * tPos.getX() + tPos.getY() * tPos.getY() 
// 	    - tPos.getZ()*tPos.getZ();
// 	
// 	zmin = mStartDist;
// 	if (fabs(a) < gEpsilon) 
// 	    goto L_HIT;
// 	
//     }
//     
//     
//     disc = b*b - a*c;
//     if (disc < 0.)
// 	goto L_EXIT;		
//  
//     
//     
//   L_HIT:
//     _HIT = true;
//     gstat(GXStatistics::Cone.mHits += 1);
//   
//   L_EXIT:
//     return (_HIT);
// 
// }


GVector3 GTCone::getNormal(const GPoint3& pPoint) {
    // Transform intersection point to cone space.
    GPoint3 tPoint = mModel.wc2mc(pPoint);
    
    GVector3 tNormal(tPoint.getX(), tPoint.getY(), -tPoint.getZ());
    if(mCylFlag)
	tNormal.setZ(0);
    
    // Transform normal back to world space.
    return ((mModel.getInvWorldMatrix().transpose() * tNormal).unitize());
}
    



const GBounds3& GTCone::computeBounds(void) {
    if(mCylFlag) 
      setWorldBounds() = GBounds3(GPoint3(-1, -1, 0), GPoint3(1, 1, 1));
    else
      setWorldBounds() = GBounds3(GPoint3(-1, -1, mStartDist), GPoint3(1, 1, 1));
    
    setWorldBounds() =  mModel.getWorldMatrix() * getWorldBounds();

    return (getWorldBounds());
}
    




////
//// @GTTriangle@CTor@
////
GTTriangle::GTTriangle(GPoint3 *pVertices)
: GTPolygon()
{
    mNumVerts = 3;		  
    mVertices = new GPoint3[3];
    for (register int i = 0; i < 3; i++)
	mVertices[i] = pVertices[i];
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    precompute();
}




////
//// @GTTriangle@DTor
////
GTTriangle::~GTTriangle(void) {
    gkillArray(mEdges);
}


void GTTriangle::precompute(void)
{
    GFloat tSizes[3] ;
    

    //// compute UV-coords.	
    GTPolygon::precompute();
    if(!isValid())
	return;
    
    //// seting up the list of edges...
    mEdges = new EdgeSet[3];
    EdgeSet *tPtr = mEdges;
    
    // P0-P1
    tPtr->index = 2;
    tPtr->A = mUV_Vertices[0][Y_AXIS] - mUV_Vertices[1][Y_AXIS] ;
    tPtr->B = mUV_Vertices[1][X_AXIS] - mUV_Vertices[0][X_AXIS] ;
    tPtr->C = tPtr->A * mUV_Vertices[0][X_AXIS] 
	+ tPtr->B * mUV_Vertices[0][Y_AXIS] ;
    //tSizes[0] = tPtr->A * tPtr->A + tPtr->B * tPtr->B ;
    tSizes[0] = fabs(tPtr->A * tPtr->B);

    // P1-P2
    tPtr++ ;
    tPtr->index = 0;
    tPtr->A = mUV_Vertices[1][Y_AXIS] - mUV_Vertices[2][Y_AXIS] ;
    tPtr->B = mUV_Vertices[2][X_AXIS] - mUV_Vertices[1][X_AXIS] ;
    tPtr->C = tPtr->A * mUV_Vertices[1][X_AXIS] 
	+ tPtr->B * mUV_Vertices[1][Y_AXIS] ;
    //tSizes[1] = tPtr->A * tPtr->A + tPtr->B * tPtr->B ;
    tSizes[1] = fabs(tPtr->A * tPtr->B);

    // P2-P0
    tPtr++ ;
    tPtr->index = 1;
    tPtr->A = mUV_Vertices[2][Y_AXIS] - mUV_Vertices[0][Y_AXIS] ;
    tPtr->B = mUV_Vertices[0][X_AXIS] - mUV_Vertices[2][X_AXIS] ;
    tPtr->C = tPtr->A * mUV_Vertices[2][X_AXIS] 
	+ tPtr->B * mUV_Vertices[2][Y_AXIS] ;
    //tSizes[2] = tPtr->A * tPtr->A + tPtr->B * tPtr->B ;
    tSizes[2] = fabs(tPtr->A * tPtr->B);

    
    // find an average point which must be inside of the triangle 
    GFloat tX = (  mUV_Vertices[0][X_AXIS] 
		 + mUV_Vertices[1][X_AXIS] 
		 + mUV_Vertices[2][X_AXIS] ) / 3.0 ;
    GFloat tY = (  mUV_Vertices[0][Y_AXIS] 
		 + mUV_Vertices[1][Y_AXIS] 
		 + mUV_Vertices[2][Y_AXIS] ) / 3.0 ;
    
    // check sense and reverse the vertices-order if needed
    register int i, j;
    if ( tPtr->A * tX + tPtr->B * tY >= tPtr->C ) {
	tPtr -= 2 ;
	for ( i = 0 ; i < 3 ; i++ ) {
	    tPtr->A = -tPtr->A ;
	    tPtr->B = -tPtr->B ;
	    tPtr->C = -tPtr->C ;
	    tPtr++ ;
	}
    } 
    
    // sorting the edges based on the size of its bounding rectangles 
    
    tPtr = mEdges ;
    for ( i = 0 ; i < 2 ; i++ ) {
	for ( j = i+1 ; j < 3 ; j++ ) {
	    if ( tSizes[i] < tSizes[j] ) {
		gswap(tPtr[i], tPtr[j]);
		gswap(tSizes[i], tSizes[j]);
	    }
	}
    }
    
}
    

bool GTTriangle::intersect(const GRay3& pRay, GInterSectionList& rInter) {
    bool _HIT = false;
 
    gstat(GXStatistics::Polygon.mTests++);

    GFloat tT, tDot;
    register EdgeSet *tPtr;
    
    //// intersect w/ the plane
    if (!mValid)
	goto L_EXIT;

    tDot = mNormal.dot(pRay.getVector());
    if (fabs(tDot) < gEpsilon) 
	goto L_EXIT;
    
    // compute the t-parameter
    tT = (mDist - mNormal.dot(pRay.getPoint())) / tDot;
    if(!pRay.isValid(tT)) 
	goto L_EXIT;
    
    // project the I-point along the dominant axis
    GFloat tU, tV;
    pRay.extrap(tT).project(tU, tV, mDominantAxis);

    // is this point inside the bounding rechtangle ?
    if (tU < mUmin || tU > mUmax || tV < mVmin || tV > mVmax)
	goto L_EXIT;

    
    
    tPtr = mEdges;
    if ( tPtr->A * tU + tPtr->B * tV < tPtr->C ) {
	tPtr++ ;
	if ( tPtr->A * tU + tPtr->B * tV < tPtr->C ) {
	    tPtr++ ;
	    if ( tPtr->A * tU + tPtr->B * tV <= tPtr->C ) {
		rInter.add( GInterSection( self, tT, (tDot < 0.0)));
		_HIT = true ;
	    }
	} 
    } 
    

    L_EXIT:

    gstat(GXStatistics::Polygon.mHits += _HIT);
    
    return (_HIT);

}


////
//// @GTSmoothTriangle@Ctor@
////
GTSmoothTriangle::GTSmoothTriangle(GPoint3  *pVertices, GVector3 *pNormals)
: GTTriangle()
{
    mNumVerts = 3;		  
    mVertices = new GPoint3[mNumVerts];
    mNormals  = new GVector3[mNumVerts];
    for (register int i = 0; i < mNumVerts; i++) {
	mNormals[i] = pNormals[i].unitize();
	mVertices[i] = pVertices[i];
    }
    
    // compute a bounding box for BVH  
    computeBounds();
    // compute the plane equation, the dominant axis and the projection 
    // rectangle along this axis.
    GTTriangle::precompute();
}

////
//// @GTSmoothTriangle@DTor
////
GTSmoothTriangle::~GTSmoothTriangle(void) {
    gkillArray(mNormals);
}



GVector3 GTSmoothTriangle::getNormal(const GPoint3& pPoint) 
{
#ifndef GTD_INTERP_VER1
    GFloat tU, tV;
    pPoint.project(tU, tV, mDominantAxis);
    
    GFloat u0, u1, u2, v0, v1, v2;
    
    u0 = mUV_Vertices[0][X_AXIS] - tU;
    u1 = mUV_Vertices[1][X_AXIS] - tU;
    u2 = mUV_Vertices[2][X_AXIS] - tU;
    v0 = mUV_Vertices[0][Y_AXIS] - tV;
    v1 = mUV_Vertices[1][Y_AXIS] - tV;
    v2 = mUV_Vertices[2][Y_AXIS] - tV;
    
    GFloat s0 = u1*v2 - u2*v1;
    GFloat s1 = u2*v0 - u0*v2;
    GFloat s2 = u0*v1 - u1*v0;
    
    GVector3 tNormal = mNormals[0]*s0 + mNormals[1]*s1 + mNormals[2]*s2;
    tNormal /= (s0 + s1 + s2);
    
    return (tNormal.unitize());
#else
    GFloat tU, tV;
    pPoint.project(tU, tV, mDominantAxis);

    GVector3 tNormals[2];
    GFloat tVs[2];
    
    
    register GFloat *tVert0, *tVert1;
    tVert0 = mUV_Vertices[mNumVerts-1] ;
    tVert1 = mUV_Vertices[0] ;
    register int i, j;
    for (i = 0, j = 0; i < mNumVerts, j < 2; i++) 
      {
	  GFloat u, v;
	  if (__computeUV(tU, tV, tVert0[X_AXIS], tVert0[Y_AXIS], 
			  tVert1[X_AXIS], tVert1[Y_AXIS], u, v)) 
	    {
		tVs[j] = v;
		if (i) 
		  tNormals[j] = GTM_LERP(u, mNormals[i-1], mNormals[i]);
		else 
		  tNormals[j] = GTM_LERP(u, mNormals[mNumVerts-1], mNormals[0]);
		j++;
	    }
	  
	  tVert0 = tVert1;
	  tVert1 += 2 ;

      }
    tV = tVs[0] / ( tVs[0] - tVs[1]);
    return (GTM_LERP(tV, tNormals[0], tNormals[1]).unitize());

#endif
}
 

GTDisc::GTDisc(const GVector3& pC, const GVector3& pN, GFloat pR) 
: GTPrimitive(), mCenter(pC), mNormal(pN), mRad(pR)
{
    setFlags(getFlags() | GRayTraceable::PLANAR);
    
    gcheck(mRad < gEpsilon, "GTDisc: radius = 0!");
    gcheck(mNormal.abs() < gEpsilon*gEpsilon, "GTDisc: normal unvalid!");
    
    mNormal = mNormal.unitize();
    mDist = mCenter.dot(mNormal);

    computeBounds();
}
 
bool GTDisc::intersect(const GRay3& pRay, GInterSectionList& rInter) {

    //// intersect w/ the plane
    GFloat tDot = mNormal.dot(pRay.getVector());
    if (fabs(tDot) < gEpsilon) 
	return (false);
    
    // compute the t-parameter
    GFloat tT = (mDist - mNormal.dot(pRay.getPoint())) / tDot;
    if(!pRay.isValid(tT)) 
	return (false);
    
    // in/outside the disc ?
    GPoint3 tHitP = pRay.extrap(tT); 
    if((tHitP - mCenter).sqdist() > mRad * mRad)
	return false;    
    
    rInter.add( GInterSection( self, tT, (tDot < 0.)));
    return (true);
}


const GBounds3& GTDisc::computeBounds(void) {
    GVector3 tExtent(mRad * sqrt(1 - mNormal.getX() * mNormal.getX()),
		     mRad * sqrt(1 - mNormal.getY() * mNormal.getY()),
		     mRad * sqrt(1 - mNormal.getZ() * mNormal.getZ()));
    
    setWorldBounds().setMin(mCenter - tExtent);
    setWorldBounds().setMax(mCenter + tExtent);

    return(getWorldBounds());
}


