//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef SMOOTHER_H
#define SMOOTHER_H


#include "util/gvector.h"
#include "util/gdyna.h"



// hash table size; related to HASH 
#define HASH_TABLE_SIZE	       1000

// quantization increment 
#define QSIZE	 1000.0

#define QUANT(x)     (((int)((x)*QSIZE))/QSIZE)
#define ABSQUANT(x)  (((int)((gabs(x))*QSIZE))/QSIZE)


enum {WAITING,  WORKING, DONE };

class Polygon {
  friend class PolySmoother;
  
  public:
    Polygon(int pNumVerts, GPoint3 *pVerts, void *pUser) {
	mUser = pUser;
	mNumVerts = pNumVerts;
	mVertices = new GPoint3[pNumVerts];
	for(register int i=0; i<pNumVerts;i++)
	    mVertices[i] = pVerts[i];
	
	mNormals  = new GPoint3[pNumVerts];
	
	mValid = true;
	computeNormal();
    }
    
    ~Polygon(void) {
	gkillArray(mVertices);
	gkillArray(mNormals);
    }
    

    GPoint3	*mVertices;  // polygon vertices
    GVector3	*mNormals;   // normal at each vertex 
    GVector3	mNormal;     // normal for polygon 
    int		mNumVerts;   // number of vertices 
    void        *mUser;      // user-defined (i.e Material)
    bool        mValid;
    
    
  private:    
    void computeNormal(void);
    
};   

class HashNode {
  friend class PolySmoother;

  public:
    HashNode(Polygon *pPoly, int pNum) {
	mPolygon = pPoly;
	mVertexNum = pNum;
	mMarked = WAITING;
    }
    
  private:
    Polygon   *mPolygon;     // polygon for this vertex 
    int	      mVertexNum;    // which vertex this is 
    int	      mMarked;	     // vertex status 
};



class PolySmoother {
  public:
    enum {WAITING,  WORKING, DONE };
    
    PolySmoother(int pHashTableSize = HASH_TABLE_SIZE);
    ~PolySmoother(void);
    
    void addPolygon(Polygon *pPolygon);
    void computeNormals(void);
    void setFuzzFraction(GFloat fuzzFraction);
    void enableEdgePreservation(GFloat minDot);
    void disableEdgePreservation(void);

    GDynaList<Polygon*> *getPolygons(void) {
	return(&mPolygonTable);
    }
    
    
    
  private:
    GDynaList<HashNode  *> *mHashTable;
    GDynaList<Polygon   *> mPolygonTable;
    GFloat    mFuzz;	      // distance for vertex equality 
    GFloat    mFuzzFraction;  // fraction of model size for fuzz 
    bool      mEdgeTest;      // apply edging test using minDot 
    GFloat    mMinDot;	      // if > this, make sharp edge 
    int       mHashTableSize;
    
    

    bool compareVerts(GPoint3 *v0, GPoint3 *v1);
    void computeFuzz(void);
    void processHashNode(int pHomeI, int pBucketI);
    
};




#endif
