//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_HERSHEY_FONT
#define G_HERSHEY_FONT

#include "util/gident.h"
#include "util/gvector.h"

extern const int gHersheyFirstChar;
extern const char* gHersheyDefaultFont;

//@ManMemo: Encapsulation of Hershey font data.
class GHersheyFontData {
  public:
    GHersheyFontData(short, signed char*);
    ~GHersheyFontData(void);

    int getLines(void) const {
	return(mLines);
    }
    int getWidth(void) const {
	return(mROff - mLOff);
    }
    int getNumberOfPoints(int pLine) const {
	return(mPline[pLine]);
    }
    GPoint3 getPoint(int pLine, int pPoint) const {
	return(GPoint3(mPx[pLine][pPoint], mPy[pLine][pPoint], 0));
    }
  private:
    short* mPline;
    signed char** mPx;
    signed char** mPy;
    short mLines;
    signed char mLOff, mROff;
};

//@ManMemo: 
class GHersheyFont: public GIdentification {
/*@Doc: 
Abstract and rendering independent representation of Hershey font.
Must be subclassed for usage.
*/
  protected:
    GHersheyFont(const GString&);
  public:
    virtual ~GHersheyFont(void);

    int getLines(int pCode) const {
	return((pCode < 256 && mData[pCode]) ?
	       mData[pCode]->getLines() : 0);
    }

    int getNumberOfPoints(int pCode, int pLine) const {
	return((pCode < 256 && mData[pCode]) ?
	       mData[pCode]->getNumberOfPoints(pLine) : 0);
    }

    GFloat getCharWidth(int pCode) const {
	return((pCode < 256 && mData[pCode]) ? 
	       (GFloat)mData[pCode]->getWidth() / (mCapHeight) : 0);
    }

    GFloat getSpacing(void) const {
	return(0); 
    }

    GPoint3 getPoint(int pCode, int pLine, int pPoint) const {
	return((pCode < 256 && mData[pCode]) ?
	       mData[pCode]->getPoint(pLine, pPoint) * (1.0 /(mCapHeight)) : GPoint3(gOrigin3));
    }
    int getMaxPoints(void) const {
	return(mMaxPoints); 
    }

    int getTotalLines(void) const { 
	return(mMaxLines); 
    }

    int getType(void) const {
	// line font:
	return(1);
    } 
    GFloat getBaseLineDist(void) const { 
	return((GFloat)(mBaseLineDist)/mCapHeight);
    }
    GFloat getCapHeight(void) const {
	return(1);
    }
  protected:
    bool load(FILE*);
  private:
    int mNchars, mMaxPoints, mMaxLines, mBaseLineDist, mCapHeight;
    GHersheyFontData* mData[256];
};

#endif
