//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_GENERIC_H
#define G_GENERIC_H

#ifdef GSINGLE_PRECISION 
typedef float GFloat;
#else
typedef double GFloat;
#endif

#include <iostream.h>
#include <assert.h>
#include <math.h>

#ifdef GWINDOWS
#include <windows.h>
#define M_PI       3.141592654
#define M_PI_2     1.570796327
#define M_PI_4     0.785398163
#define M_E		2.718282	
#define M_SQRT2		1.414214	
#define M_SQRT3		1.732051	
#define fsqrt(arg) ((GFloat)sqrt(arg))
#undef max(a, b)
#undef min(a, b)
#else
#include <stdio.h>
#endif

#define gmax(a, b)     (((a)>(b))?(a):(b))
#define gmin(a, b)     (((a)<(b))?(a):(b))

#define gmax3(a, b, c)     (((a)>(b))?((a)>(c)?(a):(c)):(((b)>(c))?(b):(c)))
#define gmin3(a, b, c)     (((a)<(b))?((a)<(c)?(a):(c)):(((b)<(c))?(b):(c)))

#define self *this

#define gEpsilon  1E-10
#define gEpsilon2 1E-20
#define MAXFLT	 1E+20
#ifndef HUGE
#define HUGE MAXFLT
#endif

#ifdef GPRE_ANSI
typedef int bool;
#define true 1
#define false 0
#endif

// initialization (true if sucessful):
extern bool gInitialize(void);
// IMPLEMENTED PLATFORM DEPENDENT

// 10 axes of the 5D-tuple
enum { G_LO_X, G_HI_X, G_LO_Y, G_HI_Y, G_LO_Z, G_HI_Z, G_LO_U, G_HI_U, G_LO_V, G_HI_V };

// 5 dimensions
enum { G_D_X, G_D_Y, G_D_Z, G_D_U, G_D_V};


// debug ON/OFF
#ifdef GDEBUG
#define gcheck(cond, msg) if((cond)){ gOutput->fatal( (msg) ); }
#else
#define gcheck(cond, msg)  ;;;
#endif

// statistics ON/OFF
#ifdef GSTAT
#define gstat(action) {(action);}
#else
#define gstat(action) ;;;
#endif
   
// macros for mem-killing!
#define gkillPointer(x) if ((x)) { delete (x); (x) = 0;}
#define gkillArray(x) if ((x)) { delete [] (x); (x) = 0;}

// absolute value of a 
#define gabs(a)		(((a)<0) ? -(a) : (a))

// round a to nearest int 
#define ground(a)	floor((a)+0.5)

// take sign of a, either -1, 0, or 1 
#define gzsgn(a)		(((a)<0) ? -1 : (a)>0 ? 1 : 0)

// take binary sign of a, either -1, or 1 if >= 0 
#define gsgn(a)		(((a)<0) ? -1 : 1)

// square a 
#define gsqrt(a)		((a)*(a))

// swap a and b (see Gem by Wyvill) 
#define gmswap(a,b)	{ a^=b; b^=a; a^=b; }

// From Ron Guillmette; apparently needed for Hansens code 
template<class T>
inline void gswap(T& a, T& b) { T c = a; a = b; b = c; }

// linear interpolation from l (when a=0) to h (when a=1)
// (equal to (a*h)+((1-a)*l) 
#define glerp(a,l,h)	((l)+(((h)-(l))*(a)))

// clamp the input to the specified range 
#define gclamp(v,l,h)	((v)<(l) ? (l) : (v) > (h) ? (h) : v)

//// some constants

#ifndef M_GOLDEN
#define M_GOLDEN	1.618034	
#endif

// convert degrees to radians 
#ifndef M_D2R
#define M_D2R		0.017453	
#endif

// convert radians to degrees 
#ifndef M_R2D
#define M_R2D		57.29578       
#endif



#endif
