//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_ERROR_H
#define G_ERROR_H

#ifdef GWINDOWS
#include <windows.h>
#else
#include <stdarg.h>
#endif

#include "util/gstring.h"

//@ManMemo: Message-based output device.
class GOutput {
/*@DOC:
Abstract class for the output of messages. The class must be subclassed in order to drive a concrete output device.
*/
  public:
    GOutput(): prog("GENERIC") {}
    
    virtual void special(const char*) = 0;
    virtual void specialVar(const char* ...) = 0;
    // special messages will not presented 
    // to the standard user
    // they can be used for log files etc.

    virtual void message(const char*) = 0;
    virtual void messageVar(const char* ...) = 0;

    virtual void warning(const char*) = 0;
    virtual void warningVar(const char* ...) = 0;

    virtual void error(const char*) = 0;
    virtual void errorVar(const char* ...) = 0;

    // a fatal error leads to the exit
    virtual void fatal(const char*) = 0;
    virtual void fatalVar(const char* ...) = 0;

    void setProgram(const char* pProg) {
	prog = pProg;
    }
  protected:
    GString prog;
};

//@ManMemo: Concrete output class printing messages to stdout.
class GSystemOutput: public GOutput {
  public:
    void special(const char*) {}
    void specialVar(const char* ...) {}

    void message(const char*);
    void messageVar(const char* ...);

    void warning(const char*);
    void warningVar(const char* ...);
    
    void error(const char*);
    void errorVar(const char*...);

    void fatal(const char*);
    void fatalVar(const char*...);
};

extern GOutput* gOutput;

#endif
