//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_BBOX_H__
#define __GX_BBOX_H__

#include "gx/gx_obj.h"


/// Axis aligned box as bounding volume 
class GXBBox : public GBounds3, public GXObject, public GXAccelCell {
/**
   Implements ABVHs, using the Goldsmith && Salmon -algorithm
  */
 
    
  public:
    /// CTor: default
    GXBBox(void)
    : GXObject(GRayTraceable::BOUND | GRayTraceable::CONVEX),
      GBounds3(GPoint3(MAXFLT), GPoint3(-MAXFLT))  {}
   
    /// Ctor:  construct the bbox w/ 2 min. and max. extents 
    GXBBox(const GPoint3& pMin, const GPoint3& pMax)
    : GXObject(GRayTraceable::BOUND | GRayTraceable::CONVEX),
      GBounds3(pMin, pMax) {}
   
    /// Ctor: Copy
    GXBBox(const GBounds3& pBox)
    : GXObject(GRayTraceable::BOUND | GRayTraceable::CONVEX),
      GBounds3(pBox)  {}
   
    /// Dtor
    virtual ~GXBBox(void);
   
    
    /// intersect the given ray w/ all children of the box
    bool intersect(const GRay3& pRay, GInterSectionList& rInter) {
	return (intersectShadow(pRay, rInter));
    }

    /// overload the function of GBounds3
    bool findMinMax(const GRay3& ray, GFloat &tmin, GFloat &tmax) const;
    
    /// intersect the given SHADOW-ray w/ all children of the box
    bool intersectShadow(const GRay3&, GInterSectionList&); 

    /// intersect the given SHADING-ray w/ all children of the box
    bool intersectShading(GRay3 &, GInterSectionList &);

    /// should be never called!
    GVector3 getNormal(const GPoint3&) {
	return (GVector3(0, 0, 0));
    }
    
    const GBounds3& getBounds(void) const { return(self); }
    const GBounds3& getWorldBounds(void) const { return(self); }

    /// return the heuristic for the intersection\'s cost of this bbox
    GFloat   getCost(void) const;

    /// insert an object to the bbox
    void     add(GXObject *);
   
    /// overlapping-test w/ the given point
    bool     isOverlap(const GPoint3& point) const;
    
    /// overlapping-test w/ the given box
    bool     isOverlap(const GBounds3& box) const;
    
    /// build the ABVH recusively...
    void           buildHierarchy(void);
    
    GXObjectList  *getListOfChildren(void);
    int            getNumOfChildren(void) const; 


  protected:

    /// list of children
    GXObjectList mListChildren;
   
};

inline GFloat GXBBox::getCost(void) const {
    return (GBounds3::getSurfaceArea());
}

inline GVector3 getNormal(const GPoint3&) {
    gstat(cerr << "Error: GXBox::getNormal was called!\n");
    return GVector3(0, 0, 1);
}

inline void GXBBox::add(GXObject *pObj) {
    // expand the actual bounding box
    (self)+= pObj->getWorldBounds();
   
    // Add the new object to the list of children
    mListChildren.append(pObj);
}

inline GXObjectList* GXBBox::getListOfChildren(void) {
    return (&mListChildren);
}

inline int GXBBox::getNumOfChildren(void) const { 
   return (mListChildren.count());
}

   
inline bool GXBBox::isOverlap(const GPoint3& pPoint) const {
    return (GBounds3::isOverlap(pPoint));
}

inline bool GXBBox::isOverlap(const GBounds3&  pBox) const {
    return (GBounds3::isOverlap(pBox));
}

inline bool GXBBox::findMinMax(const GRay3& ray, GFloat &tmin, GFloat &tmax) const {
    gstat(GXStatistics::BBox.mTests++);

    if(GBounds3::findMinMax(ray, tmin, tmax)) {
	gstat(GXStatistics::BBox.mHits++);
	return (true);
    }
    else
	return(false);
}


#endif
