//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_BSP_H__
#define __GX_BSP_H__

#include "gx/gx_bbox.h"

#define OPTI_SUBDIVISION    


#ifdef OPTI_SUBDIVISION
class Segment;
typedef GDynaList<Segment *> SegmentList;
typedef GDynaListIterator<Segment *>    SegmentItr;
struct SortedSegmentLists {
    SegmentList Mins[3];
    SegmentList Maxs[3];
};

class Segment {
  public:
    GFloat min, max;
    GXObject *obj;
};

class SortedSegmentLists;
#endif


/// BSP-node, derived from GXBBox
class GX_BSP_Node : public GXBBox {
/** A BSP-node is also a bounding box.
  This is neccessary for an `efficient` implementation of
  a combined version of BSP and BVH.
  */

  public:
    ///
    GX_BSP_Node(void) {
	mChild[0] = mChild[1] = 0;
	mCandidatesLists[G_LO_X] = 
	    mCandidatesLists[G_HI_X] = 
	    mCandidatesLists[G_LO_Y] = 
	    mCandidatesLists[G_HI_Y] = 
	    mCandidatesLists[G_LO_Z] = 
	    mCandidatesLists[G_HI_Z] = 0; 
	
	gstat(mNumOfCandidates = 0);
    }

    virtual         ~GX_BSP_Node(void);
  
    
  private:

    GX_BSP_Node*     mChild[2];
    int              mAxis;
    GFloat           mPlane;

    // because the primary list of candidate-objects can be
    // changed to BVH, the returned value of getNumOfChildren()
    // is always 1. This variable stores the old value before
    // BVHing.
#ifdef GSTAT
    long mNumOfCandidates;
#endif


    // Yup, if ABVH doesnt help for speeding-up, use these 6 candidates
    // lists, whose children are sorted after the 6 dominant axes
    GXObjectList    *mCandidatesLists[6];
    

    void             classify(const GRay3&    pRay,
			      GFloat&         rDist,
			      GX_BSP_Node*&   rNear,
			      GX_BSP_Node*&   rFar);


#ifdef OPTI_SUBDIVISION    
  friend void __subdivide(GX_BSP_Node*        rNode, 
			  SortedSegmentLists* pSortedLists,
			  int                 pDepth,        
			  int                 pMaxDepth,     
			  int                 pMaxListLength );
    
#else
  friend void __subdivide(GX_BSP_Node*  rNode,
			  int           pDepth,        
			  int           pMaxDepth,     
			  int           pMaxListLength );
#endif
    
  friend void __buildBVH(GX_BSP_Node *rNode);
  friend class GX_BSP_Tree;
};


/// BSP-Implementation 
class GX_BSP_Tree  : public GXAccelMethode {
/** Implements the BSP
 */    

  public:
    /// candidates-list of the whole enviroment
    GX_BSP_Tree(const GXPrimitiveList&);
    virtual ~GX_BSP_Tree(void);
    
    /// intersect the given SHADING ray w/ the BSP-Hierarchy.
    bool   intersectShading(GRay3& rRay, GInterSectionList& rInter);
    /** This routine is based on the "linear time voxel traversing"
      algorithm from Jim Arvo. Have a look on RTNews2 to see how it works!
      */
    
    /// intersect the given SHADING ray w/ the BSP-Hierarchy.
    bool   intersectShadow(GRay3 &rRay, 
			   GInterSectionList &rInter,
			   GRayTraceable *&rLastOpaque,
			   GXAccelCell *&rLastVoxel);
    /** This routine is based on the "linear time voxel traversing"
      algorithm from Jim Arvo. Have a look on RTNews2 to see how it works!
      */
    
    /// Build the BSP-hierarchy and contruct ABVHs for its nodes 
    void   buildHierarchy(void);

  private:

    GX_BSP_Node*    mRoot;
};

#endif 
