//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_CAM_H__
#define __GX_CAM_H__

#include "render/graycam.h"
#include "render/gpixmap.h"
#include "render/gimage.h"

#include "gx/gx_defs.h"
#include "gx/gx_obj.h"
#include "gx/gx_bsp.h"
#include "gx/gx_hyper.h"
#include "gx/gx_shade.h"
#include "gx/gx_light.h"
#include "gx/gx_sampl.h"




/// GX-RayTracer
class GXRayTracer: public GLookatRayTracer<GXObject, GPixmapDisplay> {
/**  Implements the main task of an "optimized" raytracer
 */
  public:
   
    /// CTor
    GXRayTracer(void) : GLookatRayTracer<GXObject, GPixmapDisplay>(),
    mPrimaryAccel(0), mShadowAccel(0), mSecondaryAccel(0), mOutputImage(0) {}
    
    ///
    virtual ~GXRayTracer(void);
    
    /// trace the scene! 
    virtual void rendering(const GXPrimitiveList&);
    
    /// trace a SHADING-ray
    GColorRGB traceShadingRay(int       LEVEL, 
			      GFloat    WEIGHT, 
			      GRay3&    rWC_RAY);
   
    /// trace a SHADOW-ray
    GColorRGB traceShadowRay(GRay3&          rWC_RAY,
			     GRayTraceable*& rLastOpaque,
			     GXAccelCell*& rLastVoxel );
    /** trace a shadow ray and return the light attenuation.
      If an opaque object was found, save it in pLastOpaque
      for the shadow cache.
      
      Notes: Many optimizations (and simplifications, too:) are impl.
      here: We save all transparent objects in a list for later
      testing. For the worse case, if no opaque object was found, 
      check this list and multiple all attenuation factors of 
      transparent-objects along the ray. This simplification is 
      neccessary for shadow-calculation, because refraction 
      calculation is expensive for this task. 
      */
   
    /// for compatibility w/ GLookatRayTracer
    GColorRGB trace(int    LEVEL, 
		    GFloat WEIGHT, 
		    GRay3& rWC_RAY);
    
    /// compute the color at the given intersection
    GColorRGB shade(int    LEVEL, 
		    GFloat WEIGHT, 
		    GRay3& rWC_RAY, 
		    const GInterSection& pInter);

    /// spezify the output-file
    void setOutputImage(GImage* pImage) { mOutputImage = pImage;};

  private:

    // acceleration methodes
    GXAccelMethode 
    *mPrimaryAccel,    // for eye-rays
    *mShadowAccel,     // for shadow-rays
    *mSecondaryAccel;  // for refl/refr-rays
    
    // list of lights
    GXLightList     mLightList;

    GImage *mOutputImage;
};

#endif

