//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_DEFS_H__
#define __GX_DEFS_H__

#include <time.h>

#include "util/gvector.h"
#include "util/gtempl.h"
#include "util/gdyna.h"

/// Acceleration modes :
enum  {
    /// Automatic Bounding Hierarchy
    GX_BVH, 
	/// Binary Space Partition
	GX_BSP, 
	/// Ray\'s Classification (HyperOctree:)
	GX_HOT, 
	/// Light/Vista-Buffer
	GX_BUF 
};


class GXObject;
class GXLight;
class GXPrimitive;
class GXCamera;
class GInterSection;
class GInterSectionList;

/// GXObjectList
typedef GDynaList<GXObject *>    GXObjectList;

/// GXLightList
typedef GDynaList<GXLight *>     GXLightList;

/// GXPrimitiveList
typedef GDynaList<GXPrimitive *> GXPrimitiveList;

/// GXObjectItr
typedef GDynaListIterator<GXObject *>    GXObjectItr;

/// GXLightItr
typedef GDynaListIterator<GXLight *>     GXLightItr;

/// GXPrimitiveItr
typedef GDynaListIterator<GXPrimitive *> GXPrimitiveItr;


typedef struct {
    long mNum;
    long mTests;
    long mHits;
    time_t mTime;
} GXIntStat;


/// A class for statistics
class GXStatistics {
 public:
    ///
    static GXIntStat BBox;
    ///
    static GXIntStat Sphere;
    ///
    static GXIntStat Polygon;
    ///
    static GXIntStat Cone;

    /// EYE-rays count
    static long EyeRays;
    /// Refl. rays count
    static long ReflectionRays;
    /// Refr. rays count
    static long RefractionRays;
    /// SHADOW-rays count
    static long ShadowRays;

    static long ShadowHits;
    static long ShadingHits;
    
    static long ShadowLastObjectHits;
    static long ShadowLastVoxelHits;
    static long ShadingLastObjectHits;
    static long ShadingLastVoxelHits;
    
    static long MailBoxHits;
    
    static long HOT_NumShadingVoxels;
    static long HOT_NumObjsInShadingVoxels;
    static long HOT_NumShadowVoxels;
    static long HOT_NumObjsInShadowVoxels;
    static long BSP_NumShadingVoxels;
    static long BSP_NumObjsInShadingVoxels;
    static long BSP_NumShadowVoxels;
    static long BSP_NumObjsInShadowVoxels;

    static long LastNearestHits;
};


/// A class for GX\'s configuration 
class GXOptions {
 public:
    /// image\'s width
    static int PixWidth;
    /// image\'s height
    static int PixHeight;
    /// image\'s filename
    static char OutFileName[255];
    /// antialiasing flag
    static bool Antialiasing;
    /// max. num. of pixels to supersample
    static int MaxSamples;
    /// min. num. of pixels to supersample
    static int MinSamples;
    /// max. num. of randomly generated samples (Mitchell)
    static int NumCandidates;
    /// threshold for the contrast between samples
    static GVector3 ContrastThreshold;
    
    /// max. level of BSP-subdivision
    static int BSP_MaxPartitionLevel; 
    /// max. number of objects in a BSP-node
    static int BSP_MaxObjInNode; 
    /// max. level of 5D-subdivision    
    static int HOT_MaxPartitionLevel;  
    /// max. number of objects in a 5D-cube
    static int HOT_MaxObjInNode; 

    /// Acceleration methode for primary rays
    static int PrimaryAcceleration;
    /// Acceleration methode for secondary rays
    static int SecondaryAcceleration;
    /// Acceleration methode for shadow  rays
    static int ShadowAcceleration;

    /// Use ABVH in BSP-Node ?
    static bool ABVHFlag;
    /// display the pix to screen ?
    static bool DisplayFlag;
    /// save results to the output-image ?
    static bool OutputFlag;
    /// verbose infos ?
    static bool VerboseFlag;
    /// waiting for a key after computation ?
    static bool PauseFlag;
};


#endif
