//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_HYPER_H__
#define __GX_HYPER_H__




#include "gx/gx_beam.h"


/// The 5D Hypercube
class GXHyperCube  : public GXBeam {
  public:
    ///
    GXHyperCube(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList);
    
    virtual ~GXHyperCube(void);
 
  private:
    int mLevel;

    // flag for candidates list inheritance
    bool mInherited;    
 
    // the 32 children, generated by partitioning the cube along an axis
    GXHyperCube *mChild[32]; 
    
    void subdivide(void);

    void reclassifyCandidatesList(void) {
	classifyCandidatesList();
	mInherited = false;
    }
    
    
    GXHyperCube *findLeaf(GFloat *p5DCoord);
 
    //// friend
    friend class GXHyperOctree;
};



/// The 5D-Hyperoctree
class GXHyperOctree : public GXAccelMethode {
/** This class implements the 5D-HyperOctree after Arvo && Kirk
 */
  public:
    /// CTor
    GXHyperOctree(/// the 6 sorted list of candidates
		  GXObjectList *pObjLists[6], 
		  /// the eye\'s position
		  const GVector3 &pEye);
    virtual ~GXHyperOctree(void);
    
    ///
    bool   intersectShading(GRay3& rRay, GInterSectionList& rInter);
    ///
    bool   intersectShadow(GRay3 &rRay, 
			   GInterSectionList &rInter,
			   GRayTraceable *&rLastOpaque,
			   GXAccelCell *&rLastVoxel);

    
  private:
    bool mEyeInEnvBox;
    GBounds3 mEnvBox;
    
    
    GXHyperCube *mRootCubes[6];
    GXCache *mCache;    

};


#endif
