//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GX_LBUF_H__
#define __GX_LBUF_H__

#include "gx/gx_hyper.h"


/// Base class for view volumes
class GXViewVolume : public GXBeam {
/** Derived this class to implement Vista/Light-buffer!
  Subdivision occurs only in 2 dimensions  U and V.
 */

  public:
    GXViewVolume(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList);
    virtual ~GXViewVolume(void);

  private:
    /// subdivision depth
    int mLevel;

    /// flag for candidates list inheritance
    bool mInherited;    
 
    /// the 4 children, generated by partitioning the pyramid along U and V
    GXViewVolume *mChild[4]; 
   
    ///
    void reclassifyCandidatesList(void) {
	classifyCandidatesList();
	mInherited = false;
    }
    
    /// recursively subdivision 
    void subdivide(void);
    
    /// looking for the leaf, which contains the 5D-ray
    GXViewVolume *findLeaf(GFloat *p5D);
 
    /// friend
    friend class GXViewBuffer;
    
};


/// Acceleration methode for Vista/Light-Buffer
class GXViewBuffer : public GXAccelMethode {

  public:
    /// CTor
    GXViewBuffer(/// 6 sorted list of candidates
		 GXObjectList *pObjLists[6], 
		 /// the origin of the cube
		 const GBounds3& pViewer);

    virtual ~GXViewBuffer(void);
    
    ///
    bool   intersectShadow(GRay3 &rRay, 
			   GInterSectionList &rInter,
			   GRayTraceable *&rLastOpaque,
			   GXAccelCell *&rLastVoxel);
    ///
    bool   intersectShading(GRay3& rRay, GInterSectionList& rInter);
    
    /// looking for the view-volume containing the given ray
    GXViewVolume* classify(const GRay3 &pRay); 
    
  private:
       
    GXViewVolume *mRootCubes[6];
    
};


#endif
