//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_RESOLUTION_H
#define G_RESOLUTION_H

#include "goutput.h"
#include "model/gattrib.h"

//@ManMemo: Spatial resolution entity.
class GResolution {
  public:
    GResolution(GFloat pVal = 0.1) {
	setResolution(pVal);
    }

    void setResolution(GFloat pVal) { 
	if (pVal >= 0.0 && pVal <= 1.0) mVal = pVal; 
	else gOutput->error( "Invalid resolution value. Resolution remains unchanged." );
    }
    GFloat getResolution(void) const { return mVal; }
    bool operator==(const GResolution& pResolution) const {
	return(mVal == pResolution.mVal ? true : false);
    }
  private:
    GFloat mVal;
};

inline ostream& operator<<(ostream& pStream, const GResolution& pResol) {
    pStream << pResol.getResolution();
    return(pStream);
}

//@ManMemo: Resolution attribute.
class GResolutionAttribute: public GAttribute {
  public:
    GResolutionAttribute(void);
    GResolutionAttribute(GFloat);
    
    bool operator==(const GResolutionAttribute&) const;

    void setResolution(const GResolution&);
    const GResolution& getResolution(void) const;

    void reference(const GResolutionAttribute&);
  protected:
    virtual void resolutionChanged(void) {}
  private:
    // a dynamically associated data instance:
    GResolution* mResolution;
    // the default instance of this attribute type:
    static const GResolutionAttribute attribute;
};

inline bool GResolutionAttribute::operator==(const GResolutionAttribute& pResolution) const {
    return(getResolution().getResolution() == pResolution.getResolution().getResolution() ? true : false);
}

#endif

