//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_LIGHT_H
#define G_LIGHT_H

#include "util/gvector.h"
#include "util/gcolor.h"

//@ManMemo: Rendering-independent base class of light sources.
class GLight {
  protected:
    GLight(void): mColor(gWhite), mPosition(gOrigin3), mState(true) {}
  public:
    void setColor(const GColorRGB& pColor) {
	mColor = pColor;
    }
    const GColorRGB& getColor(void) const {
	return(mColor);
    }

    void setPosition(const GPoint3& pPosition) {
	mPosition = pPosition;
    }
    const GPoint3& getPosition(void) const {
	return(mPosition);
    }

    void on(void) {
	mState = true;
    }
    void off(void) {
	mState = false;
    }
    bool isOn(void) const {
	return(mState);
    }
  private:
    GColorRGB mColor;
    GPoint3 mPosition;
    bool mState;
};

//@ManMemo: Simple point light source.
class GPointLight: public GLight {
  protected:
    GPointLight(void): mDirection(GVector3(0.0, 0.0, -1.0)) {}
  public:
    void setDirection(const GVector3& pDirection) {
	mDirection = pDirection;
    }
    const GVector3& getDirection(void) const {
	return(mDirection);
    }
  private:
    GVector3 mDirection;
};

#endif
