//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// FILE: gLookat.h
//
// CONTENT: Definition of OpenGL's lookat camera model.

#ifndef G_LOOKAT_H
#define G_LOOKAT_H

#include "util/gvector.h"
#include "util/gmatrix.h"

class GLookatCamera {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An abstract renderer
    // basing on the lookat() camera model of OpenGL.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 

  public:
    GLookatCamera(void): mView(GPoint3(0.0, 0.0, 10.0)), 
    mRef(GPoint3(0.0, 0.0, 0.0)), mFar(100.0), mNear(0.01), 
    mTwist(0.0), mZoom(40.0), mPersp(true) {}

    void setViewpoint(const GPoint3& pView) { 
	mView = pView; 
	computeMatrix();
    }
    const GPoint3& getViewpoint(void) const { return(mView); }

    void setRefpoint(const GPoint3& pRef) { 
	mRef = pRef; 
	computeMatrix();
    }
    const GPoint3& getRefpoint(void) const { return(mRef); }

    void setFar(GFloat pFar) { mFar = pFar; }
    GFloat getFar(void) const { return(mFar); }

    void setNear(GFloat pNear) { mNear = pNear; }
    GFloat getNear(void) const { return(mNear); }

    void setTwist(GFloat pTwist) { 
	mTwist = pTwist; 
	computeMatrix();
    }
    GFloat getTwist(void) const { return(mTwist); }

    void setZoom(GFloat pZoom) { 
	mZoom = pZoom; 
	computeMatrix();
    }
    GFloat getZoom(void) const { return(mZoom); }
    
    void perspectivic(void) {
	mPersp = true;
    }
    void orthographic(void) {
	mPersp = false;
    }
    bool isPerspectivic(void) const {
	return(mPersp);
    }
  public:
    GRay3 inverseTransform(const GPoint2&, GFloat = 1.0) const;
  protected:
    void computeMatrix(void);
    const GMatrix4& getMatrix(void) const {
	return(mMat);
    }
  private:
    GPoint3 mView, mRef;
    GFloat mFar, mNear, mTwist, mZoom;
    GMatrix4 mMat;
    bool mPersp;
};

#endif
