//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// FILE: gPCXImg.h
//
// CONTENT: Definition of a PCX image as a subclass
// of GImage.

#ifndef G_PCXIMG_H
#define G_PCXIMG_H

#include "render/grealiz.h"
#include "util/gcolor.h"
#include "util/gstring.h"
#include "goutput.h" 
#include "render/gimage.h"   

typedef unsigned char byte;

typedef struct {
    // CONCEPT : Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specific data structure
    // for creating PCX image files.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS:
    char*           name;
    unsigned short  width,height;
    byte	    colorplanes;        // 1 for 256 colors   
} GPCXFileStruct;			// 3 for 16.7 Mio colors

typedef struct {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specific data structure
    // for PCX image files.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS:
    
    byte     	   id;              // should be $0Ah 
    byte	   version;
    byte	   compressed;
    byte	   bitppix;

    unsigned short xmin;	
    unsigned short ymin;
    unsigned short xmax;
    unsigned short ymax;
    
    unsigned short hdpi;
    unsigned short vdpi;

    byte           colormap[16][3];
    byte	   dummy;
    byte	   colorplanes;     // max. 4

    unsigned short byteperline;     // even number
    unsigned short paltype;         // 1 = color  2 = grey

    byte           dummys[58];    

} GPCXImageStruct;

typedef struct {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: VGA Color Map
    // for PCX image files.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
    
    byte           map[256][3];       // entrys

} GPCXVGAColorMap;


class GPCXImage: public GImage {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Encapsulation of ZSoft's
    // PCX image file format.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GPCXImage(const char*);
    GPCXImage(GPCXFileStruct);

    virtual ~GPCXImage(void) {};

    virtual bool read(void); 
    virtual bool write(void);
            bool getValid(void);

    virtual void setColor(int, int, const GColorRGB&);
    virtual GColorRGB getColor(int, int) const;
    virtual const char* getExtension() const { return("pcx"); }
    
  private: 
    virtual void newGeometry(void) {
	if (mBase) delete mBase;
	mBase = new unsigned long [mXSize * mYSize];
    }
    
    static void putShort(FILE*, unsigned short);
    static unsigned short getShort(FILE*);
    
    void readHeader(FILE*);
    int  writeHeader(FILE*);
    void uncompressLine(FILE*,byte*);
    void compressLine(FILE*,byte*);

  private:
    GPCXImageStruct image;
    GPCXVGAColorMap colormap;
    int		    mapcnt;      
    bool	    valid;
};

#endif
