//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GRAY_2D_H
#define GRAY_2D_H

#include "render/graytrac.h"
#include "util/gparam2D.h"
#include <fstream.h>

//@ManMemo: A ray-traceable triangle.
class GTrianglePatch: public GRayTraceable, virtual public GGeometry {
  public:
    GTrianglePatch(void);
    GTrianglePatch
	(const GSurfaceVertex&, const GSurfaceVertex&, const GSurfaceVertex&);
    
    GFloat getCutM(void) const { return mCutM; }
    GFloat getCutN(void) const { return mCutN; }
    GFloat getCutT(void) const { return mCutT; }

    const GPoint3& getCutPoint(void) const { return mCutPoint; }
  public:
    // get tc of a point in/on triangle
    bool wc2tc(const GPoint3&, GFloat&, GFloat&);
    // get point in/on triangle in wc
    GPoint3 tc2wc(GFloat, GFloat);
   
    // get normal value in wc
    GVector3 normal2wc(GFloat, GFloat);
    GVector3 normal2wc(const GPoint3&);

    // tc will be transformed to uw space, based on the provided initial values
    bool tc2uw(GFloat, GFloat, GFloat&, GFloat&);

    // distance plane - point
    GFloat getDistance(const GPoint3&);

    GVector3 getNormal(const GPoint3&);
    bool& wasHit(void) { return mTrianHit; }

    bool intersect(const GRay3&, GInterSectionList&);

  protected:
    bool interpolate(const GPoint3&, GVector3&);
    void computeBounds(void);

  private:
    bool validate(GFloat m, GFloat n) {

	m = gabs(m) < gEpsilon ? 0.0 : m;
	n = gabs(n) < gEpsilon ? 0.0 : n;

	m = (m < 1.0+gEpsilon && m > 1.0-gEpsilon) ? 1.0 : m;
	n = (n < 1.0+gEpsilon && n > 1.0-gEpsilon) ? 1.0 : n;

	return(m >= 0.0 && m <= 1.0 &&
	       n >= 0.0 && n <= 1.0 && (m + n) < 1.0);
    }

    bool validate(GFloat& t) {
	t = fabs(t) < gEpsilon ? 0.0 : t;
	return(t > -HUGE && t < HUGE);
    }

  public:
    GSurfaceVertex mX, mY, mZ;
    GVector3 mXZ, mXY, mN;
  private:
    GMatrix mCutMat, mCoorMat, mCoorMatTrans;
    GFloat mCutM, mCutN, mCutT;
    GPoint3 mCutPoint;
    bool mTrianHit, isBadTriangle;
};

#endif
