//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_RAY_LIGHT_H
#define G_RAY_LIGHT_H

#include "util/gcolor.h"
#include "util/gsurface.h"
#include "util/gvector.h"
#include "util/gtempl.h"
#include "render/glight.h"

class GRayInterSector;

class GRayLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A class defining the 
    // protocol for "ray-tracing" light sources.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual GColorRGB shade(const GPoint3& P, const GVector3& N, const GVector3& R, 
			    const GSurface &SURFACE, GRayInterSector& CAMERA) = 0;
    // P - surface point
    // N - normal vector
    // R - reflexion vector
    // V - view vector
};

extern GAggregate<GRayLight>* gRayLights;

template <class P>
class GRayLightShadeIterator: public GIterator<P> {
  public:
    GRayLightShadeIterator(GColorRGB& pColor, const GPoint3& pPoint, const GVector3& pNormal, 
			   const GVector3& pRefl, const GSurface& pSurface, GRayInterSector& pCamera): 
    mColor(pColor), mPoint(pPoint), mNormal(pNormal), mRefl(pRefl), mSurface(pSurface), mCamera(pCamera) {}
    virtual bool exec(P& p) { 
	mColor = mColor + p.shade(mPoint, mNormal, mRefl, mSurface, mCamera);
	return(false);
    }
  private:
    GColorRGB& mColor;
    const GPoint3& mPoint;
    const GVector3& mNormal;
    const GVector3& mRefl;
    const GSurface mSurface;
    GRayInterSector& mCamera;
};

class GRayAmbientLight: public GRayLight, virtual public GLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A ray-tracing-compatible ambient light.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual GColorRGB shade(const GPoint3&, const GVector3&, const GVector3&, const GSurface &SURFACE, GRayInterSector&);

};

class GRayPointLight: public GRayLight, virtual public GPointLight {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: A ray-tracing-compatible point light.
    //
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual GColorRGB shade(const GPoint3&, const GVector3&, const GVector3&, const GSurface &SURFACE, GRayInterSector&);

};

#endif
