//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_CONSTRAINT_H
#define G_CONSTRAINT_H

#include "util/gtempl.h"
#include "util/gstring.h"

#define gDbOutVar(c) (c->variables[c->methodOuts[c->whichMethod]])
#define gDbWeaker(a, b) (a > b)

typedef enum {
    G_REQUIRED,
    G_STRONG_PREFERRED,
    G_PREFERRED,
    G_STRONG_DEFAULT,
    G_DEFAULT,
    G_WEAK_DEFAULT,
    G_WEAKEST
} GDbStrength;

#define gDbMaxVar 10
#define gDbNoMethod (-1)

extern const GDbStrength gDbStrengths[];
extern GDbStrength gDbStrength;
extern const char* gDbStrengthNames[];
extern long gDbCurrentMark;

class GDbVariable;
class GDbConstraintList;

//@ManMemo: A generic DeltaBlue constraint.
class GDbConstraint {
    friend class GDbVariable;
    friend class GDbConstraintList;
  public:
    GDbConstraint(int, GDbStrength);
    virtual ~GDbConstraint(void);
    
    GDbVariable* variables[gDbMaxVar];
    GDbConstraintList* extractPlanFromConstraint(void);
    
    void incrementalAdd(void);
    bool satisfied(void) const { 
	return(whichMethod != gDbNoMethod); 
    }
  public:
    bool inputFlag;
    GDbStrength strength;
    int whichMethod;
    int methodCount;
    int varCount;
    int methodOuts[gDbMaxVar];
  public:
    virtual void execute(void) = 0;
  protected:
    void add(void);
  private:
    bool addPropagate(void);
    GDbConstraint* satisfy(void);
    int chooseMethod(void);
    void recalculate(void);
    bool constantOutput(void);
    GDbStrength outputWalkStrength(void);
    bool inputsKnown(void);
    void incrementalRemove(void);
};

//@ManMemo: An aggregate of constraints.
class GDbConstraintList {
    friend class GDbVariable;
    friend class GDbConstraint;
  public:
    GDbConstraintList(void): mList(false) {}

    void add(GDbConstraint* pConstr) {
	mList.append(pConstr);
    }
    void remove(GDbConstraint* pConstr) {
	mList.remove(pConstr);
    }
    int getNumber(void) const {
	return(mList.getNumber());
    }

    void executePlan(void);

    void doWithElements(GIterator<GDbConstraint>& pIter) const {
	mList.doWithElements(pIter);
    }

    GDbConstraint* getFirst(void) const {
	if (getNumber()) return(&(mList[0]));
	else return(0);
    }
  private:
    GDbConstraint* removeFirst(void);  
    GDbConstraint* listRemoveFirst(void);
    GDbConstraintList* extractPlanFromConstraints(void);
    static GDbConstraintList* makePlan(void);
    static GDbConstraintList* extractPlan(void);
    GDLList<GDbConstraint> mList;
};

extern GDbConstraintList* gDbHot;
extern GDbConstraintList* gDbTodo1;
extern GDbConstraintList* gDbTodo2;

///// standard constraints:

//@ManMemo: A constraint keeping the variable constant. 
class GDbStayC: public GDbConstraint {
  public:
    GDbStayC(GDbVariable* v, GDbStrength strength): GDbConstraint(1, strength) {
	variables[0] = v;
	methodCount = 1;
	methodOuts[0] = 0;
	add();
    }
    virtual void execute(void) {}
};

//@ManMemo: A constraint allowing editing the variable.
class GDbEditC: public GDbConstraint {
  public:
    GDbEditC(GDbVariable* v, GDbStrength strength): GDbConstraint(1, strength) {
	inputFlag = true;
	variables[0] = v;
	methodCount = 1;
	methodOuts[0] = 0;
	add();
    }
    virtual void execute(void) {}
};

///// some utilities:

const char* gDbStrength2string(GDbStrength);
GDbStrength gDbString2strength(const char*);
// returns -1 if bad strength

#endif

