//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_FACET_H
#define G_FACET_H

#include "ggeneric.h"

extern const int gNoVertexIndex;

//@ManMemo: A list consisting of integers.
class GFacet {
    friend ostream& operator<<(ostream&, const GFacet&);
  public:
    // Default initializer:
    GFacet(void): mIndex(0), mArray(0) {}    
    // Create facet from integer list:
    GFacet(int pIndex, const int* pArray) { set(pIndex, pArray); }

    GFacet(const GFacet& pFacet) {
	set(pFacet.mIndex, pFacet.mArray);
    }

    ~GFacet(void) { if (mArray) delete mArray; }
    
    // Assignment operator:
    void operator=(const GFacet &f) {
	set(f.mIndex, f.mArray);
    }
	
    // Set facet completely from data:
    void set(int, const int*);
	
    // Return the index of the pointer in facet or gNoVertexIndex:
    int getVertexIndex(int) const;
    
    // Return pointer to vertex list:
    int getPointerToVertex(int ni) const { return(mArray[ni]); }
    
    // Return number of vertices in facet:
    int getNumber(void) const { return(mIndex); }
    
    // Return true if vertex pointer is in this facet:
    bool isVertexInFacet(int) const;
    
    // Return true if vertex pointer ci is the first in facet:
    bool isFirstVertex(int ci) const { return(ci == mArray[0]); }
    
    // Return true if vertex pointer ci is the lase in facet:
    bool isLastVertex(int ci) const { return(ci == mArray[mIndex - 1]); }

    // Return true if facet contains less than three points:
    bool hasNotEnoughPoints(void) const {
	if (mIndex < 3) return(true);
	else return(false);
    }

    // Return true if one index is greater than the specified:
    bool hasBadPointer(int) const;
  private:
    int mIndex;
    int* mArray;
};

#endif
