//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GENERAL_MATRIX_H__
#define __GENERAL_MATRIX_H__

#include "util/ggenvec.h"
#include "util/gmatrix.h"

//@ManMemo: A generic matrix of GFloats.
class GMatrix {
/*@Doc: 
A generic matrix of GFloats. The matrix is always initialized
with all elements equal zero. 
*/
    friend ostream& operator<<(ostream&, const GMatrix&);
    friend istream& operator>>(istream&, GMatrix&);
  public:
    GMatrix(void) { rows = cols = 0; v = 0; }
    GMatrix(int, int);
    GMatrix(const GMatrix&);
    GMatrix(const GVector&);
    GMatrix(int,int,GFloat**);
    GMatrix(int,const GVector*);
    GMatrix(const GMatrix4&);
    
    GMatrix& operator=(const GMatrix&);
    
    ~GMatrix();
    
    int      Rows(void)  const  {  return rows; }
    int      Cols(void)  const  {  return cols; }
    
    void     newGeometry(int,int);
    
    GVector  getRow(int) const;
    void     setRow(int, const GVector&);
    
    GVector  getCol(int) const;
    void     setCol(int, const GVector&);
    
    GMatrix  transpose(void) const;
    GMatrix  inverse(bool&)   const;
    GFloat   det(void)   const;
    
    GMatrix  solve(const GMatrix&, bool&) const;
    GVector  solve(const GVector& b, bool& sing) const { return GVector(solve(GMatrix(b), sing)); }
    
    // type operators..
    operator GVector(void) const; 
    operator GMatrix4(void) const;
    
    //GVector& operator[](int i)    const { return row(i); }
    
    GFloat& operator()(int, int);
    GFloat operator()(int, int) const;
    
    bool operator==(const GMatrix&)    const;
    bool operator!=(const GMatrix& x)  const { return !(*this == x); }
    
    GMatrix operator+(const GMatrix&);
    
    GMatrix operator-(const GMatrix&);
    GMatrix operator-(); // unary
    
    GMatrix& operator-=(const GMatrix&);
    GMatrix& operator+=(const GMatrix&);
    
    GMatrix operator*(const GMatrix&) const;
    GVector operator*(const GVector&);
    GMatrix operator*(const GFloat&);
  private:
    GVector** v;
    int rows;
    int cols;

    void flip_rows(int,int);
    void check_dimensions(const GMatrix&) const; 
    GFloat& elem(int, int) const;
    GFloat** triang(const GMatrix&, int&) const;
};

#endif





