//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef __GENERAL_VECTOR_H__
#define __GENERAL_VECTOR_H__

#include "ggeneric.h"
#include "util/gvector.h"
#include "util/gcolor.h"

#define sqr(a) ((a)*(a))

class GMatrix;

//@ManMemo: A generic vector of GFloats.
class GVector {
/*@Doc: 
A generic vector of GFloats. All elements are initialized with zero.
*/
    friend class GMatrix;
    friend ostream& operator<<(ostream&, const GVector&);
    friend istream& operator>>(istream&, GVector&);
  public:
    GVector(void) { d = 0; v = 0; } 
    GVector(int d);
    GVector(GFloat,GFloat,GFloat);
    GVector(GFloat,GFloat);
    GVector(const GVector&);
    GVector(const GVector2&);
    GVector(const GVector3&);

    GVector& operator=(const GVector&);

    ~GVector() { if (v) delete [] v; }

    int dim(void) const { return d; }
    GFloat abs(void) const { return sqrt(*this * *this); }

    void newGeometry(int);

    GVector unitize(void) const { return *this/abs(); }
    
    GFloat& operator[](int i);
    GFloat operator[](int) const;

    GVector& operator+=(const GVector&);
    GVector& operator-=(const GVector&);
  
    GVector operator+(const GVector& v1) const;
    GVector operator-(const GVector& v1) const;

    GFloat operator*(const GVector& v1) const;
    GVector operator*(GFloat r) const;

    GVector operator-(void) const;
    GVector operator/(GFloat) const;
  
    bool operator==(const GVector&) const;
    bool operator!=(const GVector& w) const { return !(*this == w); }

    operator GVector3(void) { 
	if(d==3) 
	    return GVector3(v[0],v[1],v[2]);
	else
	    return GVector3();
    }
    operator GVector2(void) {
	if(d==2)
	    return GVector2(v[0],v[1]);
	else
	    return GVector2();
	
    }
    operator GColorRGB(void) {
	if(d==3) 
	    return GColorRGB(v[0],v[1],v[2]);
	else
	    return GColorRGB();
    }
	
    GFloat max();
    GFloat min();
    
    friend bool compare(const GVector&, const GVector&);
    friend GFloat chord(const GVector&, const GVector&);
    friend GVector cross(const GVector&,const GVector&);
    friend GFloat angle(const GVector&, const GVector&);
  private:
    GFloat* v;
    int d;
    bool check_dimensions(const GVector& vec) const {
	if (d!=vec.d) 
	    return false;
	else
	    return true;
    }
};

#endif




