//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GHEAP_H
#define GHEAP_H

template<class T>
//@ManMemo: A parametric heap.
class GHeap {  
  public:  
    inline GHeap(int pSize=2);  
    inline GHeap(const GHeap<T>&);
    inline ~GHeap(void);  
    
    inline void     insert(const T& x);  
    inline T        remove(void);  
    inline const T& head(void) const;
    inline void     reset(void);  
    inline int      count(void) const;  
  private:  
    T*       mItems;  
    int      mNumber;  
    int      mSize;  
    
    inline void     upHeap(int i);  
    inline void     downHeap(int i);  
};  
 
template<class T> 
inline GHeap<T>::GHeap(int pSize)  {  
    mItems  = new T[pSize];  
    mNumber = 0;  
    mSize   = pSize;  
}  

template<class T> 
inline GHeap<T>::GHeap(const GHeap<T>& pHeap)  {  
    mItems  = new T[pHeap.mSize];  
    mNumber = pHeap.mNumber;  
    mSize   = pHeap.mSize;
    for (register int i = 0; i < mSize; i++)  
      mItems[i] = pHeap.mItems[i];  
}  

template<class T>
inline GHeap<T>::~GHeap()  {  
    delete[] mItems;  
}  
 
template<class T> 
inline void GHeap<T>::insert(const T& pItem)  {  
    if (++mNumber >= mSize)   {  
        T *temp = new T[mSize <<= 1];  
        for (register int i = 0; i < mNumber-1; i++)  
          temp[i] = mItems[i];  
        delete[] mItems;  
        mItems = temp;  
    }  
    mItems[mNumber-1] = pItem;  
    upHeap(mNumber-1);  
}  
 
template<class T>
inline T GHeap<T>::remove(void)  {  
    T ret = mItems[0];  
    if (--mNumber > 0)   {  
        mItems[0] = mItems[mNumber];  
        downHeap(0);  
    }  
    return(ret);  
}  

template<class T>
inline const T& GHeap<T>::head(void) const {  
    return(mItems[0]);  
}  

template<class T>
inline void  GHeap<T>::reset(void)  {  
    mNumber = 0;  
}  
 
template<class T> 
inline int GHeap<T>::count(void) const  {  
    return(mNumber);  
}  
 
template<class T>
inline void  GHeap<T>::upHeap(int i)  {  
    int p;  
    T temp = mItems[i];  
    
    while (i)   {  
        p = (i-1) >> 1;  
        if ( mItems[p] < temp )  
          break;  
        mItems[i] = mItems[p];  
        i = p;  
    }  
    mItems[i] = temp;  
}  
 
template<class T>
inline void GHeap<T>::downHeap(int i)  {  
    int c;  
    T temp = mItems[i];  
    
    while((c = ((i<<1)+1)) < mNumber)   {  
        if (c+1 < mNumber)  
          c += (mItems[c+1] < mItems[c]) ;  
        if (temp < mItems[c] )  
          break;  
        mItems[i] = mItems[c];  
        i = c;  
    }  
    mItems[i] = temp;  
}

#endif
