//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_IDENTIFICATION_H
#define G_IDENTIFICATION_H

#include "util/gstring.h"
#include "util/gtempl.h"
#include "goutput.h"

//@ManMemo: String-based identification of objects.
class GIdentification {
/*@Doc: 
An identification consisting of an ASCII name.
Further semantics, e.g., support for hierarchies, may be implemented
in derived classes.
*/
  public:
    GIdentification(const char* = 0);
    // CTOR: Constructing an identification from a string.

    virtual ~GIdentification(void);
    // DTOR: Virtual destructor.

    const GString& getIdentification(void) const;
    // METHOD: Return identification as string.

    const GString& getName(void) const {
	return(getIdentification());
    }
    // METHOD: Return identification as string.
  public:
    static bool identificationExists(const char*);
    // METHOD: Return true if the passed string already
    // exists for an object.
    static GIdentification* getObjectByIdentification(const char*);
    // METHOD: Return the related object or zero.
  private:
    // a dynamically-associated string:
    GString* mName;
    // the static storage of all identification
    // objects (i.e. mName != 0):
    static GDLList<GIdentification> sList;
};

#endif
