//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_PARAMETRIC_1D_H
#define G_PARAMETRIC_1D_H

#include "util/gparam.h"
#include "util/gspline.h"

//@ManMemo: Realization of a b-spline curve.
class GApproxSplineCurve: public GCurveTopology {
  public:
    GApproxSplineCurve(int, const GResolution&, const GSplineMode&);
    GApproxSplineCurve(const GCurveVertexAggregate&, const GResolution&, const GSplineMode&);

    int getSegments(void) const { return mSpline.getSegments(); }

    void setOrder(int);
    const int getOrder(void) { return mSpline.getSpline().order; }
    
    void setMode(const GSplineMode&);
    const GSplineMode& getMode(void) { return mSpline.getSpline().mode; }
    
    void setEnd(const GSplineEnd&);
    const GSplineEnd& getEnd(void) { return mSpline.getSpline().end; }

    const GVector& getParameterVector(void);
    const GCurveVertex getCurveVertexOn(GFloat);
  protected:
    void geometryChanged(void);
    // overload global method and set geometry of the spline curve
  private:
    GApproxSpline mSpline;
};

//@ManMemo: A fitting curve based on splines.
class GFittingSplineCurve: public GCurveTopology {
  public:
    GFittingSplineCurve(int, const GResolution&, const GSplineMode&);
    GFittingSplineCurve(const GCurveVertexAggregate&, const GResolution&, const GSplineMode&);

    int getSegments(void) const { return mSpline.getSegments(); }

    void setMode(const GSplineMode&);
    const GSplineMode& getMode(void) { return mSpline.getMode(); }

    const GVector& getParameterVector(void);
    const GCurveVertex getCurveVertexOn(GFloat);
  protected:
    void geometryChanged(void);
  private:
    GNaturalFittingSpline mSpline;
};

//@ManMemo: A general fitting curve.
class GGeneralFittingSplineCurve: public GCurveTopology {
  public:
    GGeneralFittingSplineCurve(int, const GResolution&, const GSplineMode&, const GVector3&, const GVector3&);
    GGeneralFittingSplineCurve(const GCurveVertexAggregate&, const GResolution&, const GSplineMode&, const GVector3&, const GVector3&);

    int getSegments(void) const { return mSpline.getSegments(); }

    void setFirstTangent(const GVector3&);
    void setLastTangent(const GVector3&);

    void setMode(const GSplineMode&);
    const GSplineMode& getMode(void) { return mSpline.getMode(); }

    const GVector& getParameterVector(void);
    const GCurveVertex getCurveVertexOn(GFloat);
  protected:
    void geometryChanged(void);
  private:
    GGeneralFittingSpline mSpline;
};
   
    
#endif
