//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_PARAMETRIC_2D_H
#define G_PARAMETRIC_2D_H

#include "util/gparam.h"
#include "util/gspline.h"

//@ManMemo: Realization of a approximating b-spline surface.
class GApproxSplineSurface: public GSurfaceTopology {
  protected:
    GApproxSplineSurface(void): GSurfaceTopology(), uSpline(), wSpline() {}
  public:
    GApproxSplineSurface(int, int, const GResolution&, const GSplineMode&, const GSplineMode&);
    GApproxSplineSurface(int, int, const GSurfaceVertexAggregate&, const GResolution&, 
			 const GSplineMode&, const GSplineMode&);
    
    int getSegmentsX(void) const { return uSpline.getSegments(); }
    int getSegmentsY(void) const { return wSpline.getSegments(); }

    const GVector& getParameterVectorU(void);
    const GVector& getParameterVectorW(void);

    const GSurfaceVertex getSurfaceVertexOn(GFloat, GFloat);

    void setOrderU(int);
    void setOrderW(int);

    const int getOrderU(void) { return uSpline.getSpline().order; }
    const int getOrderW(void) { return wSpline.getSpline().order; }

    void setModeU(const GSplineMode&);
    void setModeW(const GSplineMode&);

    const GSplineMode& getModeU(void) { return uSpline.getSpline().mode; }
    const GSplineMode& getModeW(void) { return wSpline.getSpline().mode; }

    void setEndU(const GSplineEnd&);
    void setEndW(const GSplineEnd&);

    const GSplineEnd& getEndU(void) { return uSpline.getSpline().end; }
    const GSplineEnd& getEndW(void) { return wSpline.getSpline().end; }
    // set/get spline values
  public:
    void geometryChanged(void);
    // overload global method and set geometry of the spline surface
  private:
    GApproxSpline uSpline, wSpline;
    // mathematic b-spline instances
};

//@ManMemo: A fitting surface based on splines.
class GFittingSplineSurface: public GSurfaceTopology {
  protected:
    GFittingSplineSurface(void): GSurfaceTopology(), uSpline(), wSpline() {}

    void computePhantomTopology(void);
  public:
    GFittingSplineSurface(int, int, const GResolution&, const GSplineMode&, 
			  const GSplineMode&);
    GFittingSplineSurface(int, int, const GSurfaceVertexAggregate&, const GResolution&, 
			  const GSplineMode&, const GSplineMode&);

    ~GFittingSplineSurface(void) { if(mControlNet) delete mControlNet; }

    int getSegmentsX(void) const { return uSpline.getSegments(); }
    int getSegmentsY(void) const { return wSpline.getSegments(); }

    const GVector& getParameterVectorU(void);
    const GVector& getParameterVectorW(void);

    const GSurfaceVertex getSurfaceVertexOn(GFloat, GFloat);

    void setModeU(const GSplineMode&);
    void setModeW(const GSplineMode&);

    const GSplineMode& getModeU(void) const { return uSpline.getMode(); }
    const GSplineMode& getModeW(void) const { return wSpline.getMode(); }
    // set/get spline values
  public:
    void geometryChanged(void);
    // overload global method and set geometry of the spline surface
  public:
    const GPhantomSurfaceTopology& getPhantomTopology(void) const {
	return *mControlNet;
    }
  private:
    GNaturalFittingSpline uSpline, wSpline;
    // mathematic b-spline instances
    GPhantomSurfaceTopology *mControlNet;
};

#endif

