//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GSTACK_H
#define GSTACK_H

#ifndef GDEFAULT_STACK_SIZE
#define GDEFAULT_STACK_SIZE 64
#endif

template<class T> 
//@ManMemo: A parametric stack.
class GStack { 
  public: 
    GStack(int sz = GDEFAULT_STACK_SIZE); 
    GStack(const GStack<T>&); 
    ~GStack(void); 
      
    void                operator =(const GStack<T>&); 
    void                push(const T& item); 
    T                   pop(void); 
    T&                  top(void); 
    void                delTop(void); 
      
    int                 length(void) const; 
    int                 empty(void) const; 
    int                 full(void) const; 
    void                reset(void); 
    void                resize(int sz); 

 protected: 
    int                 mSize; 
    int                 mPtr; 
    T*                  mItems; 
}; 

template<class T>
inline GStack<T>::GStack(int pSize) { 
    mItems = new T[mSize = pSize]; 
    mPtr = 0; 
} 

template<class T>
inline GStack<T>::GStack(const GStack<T>& b):
mSize(b.mSize), mPtr(b.mPtr), mItems(new T[b.mSize]) { 
    for (int i = 0; i < mPtr; ++i) 
      mItems[i] = b.mItems[i]; 
} 

template<class T>
inline GStack<T>::~GStack(void) { 
    delete [] mItems; 
} 

template<class T>
inline void GStack<T>::reset(void) {
    mPtr = 0; 
} 

template<class T>
inline int GStack<T>::empty(void) const { 
    return(mPtr == 0); 
} 

template<class T>
inline int GStack<T>::full(void) const { 
    return(mPtr == mSize); 
} 

template<class T>
inline int GStack<T>::length(void) const { 
    return(mPtr); 
} 

template<class T>
inline void GStack<T>::push(const T& pItem) {
    gcheck(full(), "Full stack in GStack<T>::push()");
    mItems[mPtr++] = pItem; 
} 

template<class T>
inline T GStack<T>::pop(void) { 
    gcheck(empty(), "Empty stack in GStack<T>::pop()");
    return(mItems[--mPtr]); 
} 

template<class T>
inline void GStack<T>::delTop(void) { 
    gcheck(empty(), "Empty stack in GStack<T>::delTop()");
    --mPtr; 
} 

template<class T>
inline T& GStack<T>::top(void) { 
    gcheck(empty(), "Empty stack in GStack<T>::top()");
    return(mItems[mPtr-1]); 
}


template<class T>
inline void GStack<T>::operator = (const GStack<T>& b) { 
    if (&b == this) return; 
    if (mSize < b.mPtr) { 
        delete [] mItems; 
        mItems = new T[b.mSize]; 
        mSize = b.mSize; 
    } 
    mPtr = b.mPtr; 
    for (int i = 0; i < mPtr; ++i) 
	mItems[i] = b.mItems[i]; 
} 

template<class T>
inline void GStack<T>::resize(int newsz) { 
    gcheck( (newsz < mPtr), "Size to small in GStack<T>::resize()");
    T* news = new T[newsz]; 
    for (int i = 0; i < mPtr; ++i) 
	news[i] = mItems[i]; 
    delete [] mItems; 
    mItems = news; 
    mSize = newsz; 
} 

#endif
