//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_STRING_H
#define G_STRING_H

#include "gdata.h"
#include <string.h>

//@ManMemo: Encapsulating C strings.
class GString: public GStorable {
    friend GString operator+(const GString&, const char*);
    friend GString operator+(const char*, const GString&);
    friend GString operator+(const GString&, const GString&);
    friend ostream& operator<<(ostream&, const GString&);
  public:
    //@ManMemo: Construct string from C string.
    GString(const char*);

    //@ManMemo: Copy constructor.
    GString(const GString&);

    //@ManMemo: Default constructor with an optional length reservation.
    GString(int = 10);

    virtual ~GString(void) { delete mChars; }

    // Assignment operator:
    GString& operator=(const GString&);

    // Aquivalency/antivalency:
    bool operator==(const GString&) const;
    bool operator==(const char*) const;
    bool operator!=(const GString&) const;
    bool operator!=(const char*) const;

    // relation operator
    bool operator< (const GString&) const;
    bool operator<=(const GString&) const;
    bool operator> (const GString&) const;
    bool operator>=(const GString&) const;
  
    // Append a string/character on the end:
    const GString& operator+=(const char*);
    const GString& operator+=(const GString& s);
    const GString& operator+=(char);

    // return the buffer from i-th -> j-th position of the old string
    GString operator()(int i, int j) const;
    
    operator char* (void) const { return(mChars); }
    // METHOD: Overloaded type-cast for interfacing with C:

    GString substr(int i1, int i2);
    // METHOD: Compute sub string from index $$i1$$ to $$i2$$.

    void del(int i1, int i2);
    // METHOD: Delete sub string of from index $$i1$$ to $$i2$$.

    int getLength(void) const { return(strlen(mChars)); }
    // METHOD: Return current string length.

    char operator[](int) const;
    char& operator[](int);
    // METHOD: Get character at specified position.

    GString car(void) const;
    // METHOD: Return a String with the first word.

    GString cdr(void) const;
    // METHOD: Return a String with all but the first word.
    
    const char* tolower(void) const;
    // METHOD: Convert string to lower cases. Returns a reference
    // to a static string that must be "used" immediately.

    int getPos(const char*) const;
    // METHOD: Get the index of the first position of the string or -1. 


    unsigned int hash(void) const;
    // METHOD: return a hash-value for this string
	
  private:

    int mLength;
    // Size of reserved memory.
    char *mChars;

    static char sBuffer[100];
};

#endif
 
