//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_SURFACE_H
#define G_SURFACE_H

#include "util/gcolor.h"

//@ManMemo: A standard surface type.
class GSurface {
    friend ostream& operator<<(ostream&, const GSurface&);
  public:
    GSurface(void);
    // CTOR: Create a surface with default parameters.
 
    GSurface(const GColorRGB&);

    void setAmbient(const GColorRGB& pAmbient) { mAmbient = pAmbient; }
    const GColorRGB& getAmbient(void) const { return mAmbient; }

    void setDiffuse(const GColorRGB& pDiffuse) { mDiffuse = pDiffuse; }
    const GColorRGB& getDiffuse(void) const { return mDiffuse; }

    void setSpecular(const GColorRGB& pSpecular) { mSpecular = pSpecular; }
    const GColorRGB& getSpecular(void) const { return mSpecular; }

    void setShininess(GFloat pShininess) { mShininess = pShininess; }
    const GFloat getShininess(void) const { return mShininess; }

    void setRefraction(GFloat pRefraction) { mRefraction = pRefraction; }
    const GFloat getRefraction(void) const { return mRefraction; }

    void setTransparency(GFloat pTransparency) { mTransparency = pTransparency; }
    const GFloat getTransparency(void) const { return mTransparency; }

    bool operator==(const GSurface&) const;
  private:
    GColorRGB mAmbient, mDiffuse, mSpecular;
    GFloat mShininess, mRefraction, mTransparency;
};

extern const GSurface gRedSurface, gBlueSurface, gGreenSurface;

inline GSurface::GSurface(void) {
    // default surface:
    setAmbient(GColorRGB(1.0, 1.0, 1.0));
    setDiffuse(GColorRGB(1.0, 1.0, 1.0));
    setSpecular(GColorRGB(0.0, 0.0, 0.0));
    setShininess(30);
    setRefraction(1.0);
    setTransparency(0.0);
}
    

#endif
