//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef G_TCL_H
#define G_TCL_H

#ifdef GWITH_TCL

#include "util/gtclutil.h"

//@ManMemo: Base class for interfacing with Tcl.
class GTclInterface {
  protected:
    //@ManMemo: Initialization of Tcl command.
    GTclInterface(const char*);

    //@ManMemo: Remove Tcl command.
    virtual ~GTclInterface(void);

    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
  public:
    //@ManMemo: Inquire object's name.
    const GString& getName(void) const { return(mName); }

    //@ManMemo: Map a Tcl "method" onto C++.
    virtual bool objectCMD(char*[]) = 0;

    //@ManMemo: To be called from static object command.
    static int objectCMD(ClientData, Tcl_Interp*, int, char*[]);

    //@ManMemo: Check existency.
    static bool exists(const char*);
  protected:
    static void setResult(const char* pResult) { sResult = pResult; }
    static const char* getResult(void) { return(sResult); }
  private:
    static GString sResult;
    GString mName;
};

typedef int (*GTclLoadProc)(const char*);

//@ManMemo: Base class for interpretative classes.
class GTclClass: public GTclInterface {
/*@DOC:
Base class for interpretative classes. Has to be used via (multiple) 
inheritance in concrete sub classes.
*/
    friend void gTclInitClassCommands(Tcl_Interp*);
  protected:
    GTclClass(const char*, GTclClassDesc*);
  public:
    virtual ~GTclClass(void);

    bool isA(const char*) const;
    const char* getKeywords(void) const { return(mDesc->getKeywords()); }
    const char* getClass(void) const { return(mDesc->getClass()); }
    const char* getDescription(void) const { return(mDesc->getDescription()); } 
    
    operator const GTclClassDesc*(void) const { return(mDesc); }
  protected:
    // Try to autoload class description.
    // Returns true if successful.
    static bool loadClass(const char*pClass) {
	return(sLoadProc(pClass));
    }
    static int classCMD(ClientData, Tcl_Interp*, int, char*[]);
    virtual bool objectCMD(char*[]);

    static int addClassCMD(ClientData, Tcl_Interp*, int, char*[]);
  public:
    static int addMethodCMD(ClientData, Tcl_Interp*, int, char*[]);
    static const GDLList<GTclClassDesc>& getClasses(void){
	return(sClasses);
    }
  public:
    // interface for coupling a class browser:
    static int getClassesCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int deleteAllClassesCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int deleteClassCMD(ClientData, Tcl_Interp*, int, char*[]);

    static int getMethodsCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int deleteMethodCMD(ClientData, Tcl_Interp*, int, char*[]);

    static int getParameterCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int getDescriptionCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int getBodyCMD(ClientData, Tcl_Interp*, int, char*[]);
    static int getParentClassCMD(ClientData, Tcl_Interp*, int, char*[]);

    static void setLoadProc(GTclLoadProc pLoadProc) {
	sLoadProc = pLoadProc;
    }

    static GTclClassDesc* getClassByName(const char*);
  protected:
    void addMembers(const char*);
    // Add entries for members into the members list.
    // The string must contain null, one or more names
    // of members separated by spaces.

    bool callMethod(const GTclClassDesc* DESC, char** ARGS, bool CONSTR = false);
    // Call a Tcl method by evaluating the ARGS.
    // If CONSTR is false it is a standard method, else it is the constructor.
    // Returns true if a method could be performed successfully.

    static GTclMethodDesc* find(Tcl_Interp*, int, char *[]);
    // Find a method description of class argv[1] and method argv[2]
    // Returns 0 if nothing was found.

    const GTclMemberList& getMemberList(void) {
	return(mMembers);
    }
  private:
    static int localLoadClass(const char*);
  protected:  
    GString mName;

    GTclClassDesc* mDesc;
    // a link to the related TclClassClassDescription

    GTclMemberList mMembers;
  protected:
    static GDLList<GTclClassDesc> sClasses;
    // a list containing all class descriptions
  private:
    static GTclLoadProc sLoadProc;
};

// initialize Tcl class system:
void gTclInitClassCommands(Tcl_Interp*);

#endif

#endif

