//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifndef GTCL_PARSE_H
#define GTCL_PARSE_H

#ifdef GWINDOWS
#include <windows.h>
#endif

#include <stdlib.h>
#include <tcl.h>

#include "util/gvector.h"
#include "util/gcolor.h"
#include "util/gsurface.h"
#include "util/gmatrix.h"
#include "util/gstring.h"
#include "util/gtempl.h"

// the next free enum value to be used in extended kernels: 
extern const int gTclNextParseEnum;

extern const char* gTclHelp;

extern Tcl_Interp* gTclTmpInterp;
// This static temporary Tcl interpreter may be used for
// splitting of argument list.

extern GString gTclTmpString;

extern Tcl_Interp* gGlobalInterp;

void gTclAppendResult(const char*);
int gTclGlobalEval(const char*);
const char* gTclGetResult(void);

typedef enum { 
    gTclEnd = 1, 
    gTclNone, 
    gTclString, 
    gTclBool,
    gTclInt, 
    gTclFloat,
    gTclVector2,
    gTclVector3,
    gTclMatrix4,
    gTclSurface,
    gTclSpecial 
} GTclParseType;

typedef struct {
    char* mName;
    GTclParseType mType;
    char* mParam;
    bool* mIsSet;
    char* mInfo;
    char* mTclTypes;
} GTclParseEntry;

int gTclFindString(char** ARGS, const char *NAME);
// Return position number of NAME in ARGS or -1.

void gTclClearArgv(char** ARGS, int A, int N);
// Remove N strings from position A in ARGS (including the A position).

void gTclParseTable(char** ARGS, GTclParseEntry*);

// The utility functions return true if successful:

bool gTclFromString(const char**, bool&, int&);
bool gTclFromString(const char**, int&, int&);
bool gTclFromString(const char**, GFloat&, int&);
bool gTclFromString(const char**, const char*&, int&);
bool gTclFromString(const char**, GVector2&, int&);
bool gTclFromString(const char**, GVector3&, int&);
bool gTclFromString(const char**, GMatrix4&, int&);
bool gTclFromString(const char**, GColorRGB&, int&);
bool gTclFromString(const char**, GSurface&, int&);

bool gTclFromString(const char*, bool&);
bool gTclFromString(const char*, int&);
bool gTclFromString(const char*, GFloat&);
bool gTclFromString(const char*, GVector2&);
bool gTclFromString(const char*, GVector3&);
bool gTclFromString(const char*, GMatrix4&);
bool gTclFromString(const char*, GColorRGB&);
bool gTclFromString(const char*, GSurface&);

template<class T>
inline bool gTclFromString(const char*, GDLList<T>&);

// The return value of the following functions has to be
// copied immediately:
const char* gTclToString(bool);
const char* gTclToString(int);
const char* gTclToString(GFloat);
const char* gTclToString(const GVector2&);
const char* gTclToString(const GVector3&);

// template implementation(s):

template<class T>
inline bool gTclFromString(const char* s, GDLList<T>& pList) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
	bool ok = false;
	if (argc > 0) {
	    ok = true;
	    for (int i = 0; i < argc; i++) {
		T p;
		if (gTclFromString(argv[i], p)) {
		    pList.append(new T(p));
		}
		else break;
	    }
	}
	free((char*)argv);
	return(ok);
    }
    else return(false);
}

#endif
