//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <limits.h>
#include "format/ghfont.h"

const int gHersheyFirstChar = ' ';
const char* gHersheyDefaultFont = "timesr"; 

#define	ghersheyRelativ(x) ((x) - (signed char)'R')

GHersheyFontData::~GHersheyFontData(void) {
    if (mLines) {
	if (mPline) delete mPline;
	for (int t = 0; t < mLines; t++) {
	    if (mPx[t]) delete mPx[t];
	    if (mPy[t]) delete mPy[t];
	}
	if (mPx) delete mPx; 
	if (mPy) delete mPy;
    }
}

GHersheyFontData::GHersheyFontData(short n, signed char *pBuf):
mPline(0), mPx(0), mPy(0), mLines(0), mLOff(0), mROff(0) {
    if (n < 2 || !pBuf) return;
    mLOff = ghersheyRelativ(pBuf[0]);
    mROff = ghersheyRelativ(pBuf[1]);
    if (n < 6) return;
    for (int t = 6; t < n; t+= 2) if (pBuf[t] == ' ') mLines++;
    if (pBuf[t-2] != ' ') mLines++;
    mPline = new short [mLines];
    mPx = new signed char* [mLines];
    mPy = new signed char* [mLines];
    for (t = 0; t < mLines; t++) {
	mPline[t] = 0; 
	mPx[t] = mPy[t] = 0;
    }
    int p = 2;
    for (t = 0; t < mLines; t++) {	    
	for (int tt = p+4; tt < n && pBuf[tt] != ' '; tt+=2);
	mPline[t] = (tt - p) / 2;
	mPx[t] = new signed char[mPline[t]];
	mPy[t] = new signed char[mPline[t]];
	for (int ttt = p; ttt < tt; ttt+=2) {
	    mPx[t][(ttt-p)/2] = ghersheyRelativ(pBuf[ttt]) - mLOff;
	    mPy[t][(ttt-p)/2] = -ghersheyRelativ(pBuf[ttt+1]); // there tv coords
	}
	p = tt+2;
    }
}

GHersheyFont::GHersheyFont(const GString& pName): GIdentification(pName),
mMaxPoints(0), mMaxLines(0), mBaseLineDist(0), mCapHeight(0) {}

GHersheyFont::~GHersheyFont(void) {
    for (int t = 0; t < 256; t++) 
	if (mData[t]) delete mData[t]; 
}

#ifndef GWINDOWS
#if BYTE_ORDER == BIG_ENDIAN 
#define ntohs(x)        (x)
#else
#if BYTE_ORDER == LITTLE_ENDIAN 
static unsigned short ntohs(unsigned short x) { return (x << 8) | (x >> 8); }
#endif
#endif
#endif

bool GHersheyFont::load(FILE* pFile) {
    unsigned short d;
    int t;
    for (t = 0; t < 256; t++) mData[t] = 0;
    fread(&d, sizeof(short), 1, pFile);
    mNchars = ntohs(d);

    // not needed:
    fread(&d, sizeof(short), 1, pFile); // number of vectors in font
    fread(&d, sizeof(short), 1, pFile); // max ascender
    fread(&d, sizeof(short), 1, pFile); // max descender
    fread(&d, sizeof(short), 1, pFile); // max width
    int tBufSize = 100;
    signed char* tBuf = new signed char[tBufSize];
    for (t = gHersheyFirstChar; t < mNchars + gHersheyFirstChar && t < 256; t++) {
	fread(&d, sizeof(short), 1, pFile);
	d = ntohs(d);
	if (d > tBufSize) {
	    tBufSize = d;
	    delete [] tBuf;
	    tBuf = new signed char[tBufSize];
	}
	fread(tBuf, sizeof(char), d, pFile);
	mData[t] = new GHersheyFontData(d, tBuf);

	// set maxLines and maxPoints:
	int tLines = mData[t]->getLines();
	if (tLines > mMaxLines) mMaxLines = tLines;
	for (int l = 0; l < tLines; l++) {
	    int p = mData[t]->getNumberOfPoints(l);
	    if (p > mMaxPoints) mMaxPoints = p;
	}	    
    }
    delete tBuf;

    // set base line distance and cap height:
    t = 'A';
    if (!mData[t])
	for (t = gHersheyFirstChar + 1; 
	     t < mNchars + gHersheyFirstChar && t < 256 && !mData[t]; 
	     t++) ;
    if (mData[t]) {
	mBaseLineDist = SCHAR_MAX; 
	mCapHeight = SCHAR_MIN;
	for (int tt = 0; tt < mData[t]->getLines(); tt++) {
	    for (int ttt = 0; ttt < mData[t]->getNumberOfPoints(tt); ttt++) {
		GPoint3 v = mData[t]->getPoint(tt,ttt);
		int tY = (int)v.getY();
		if (tY < mBaseLineDist) mBaseLineDist = tY;
		if (tY > mCapHeight) mCapHeight = tY;
	    }
	}
	mCapHeight -= mBaseLineDist;
    }
    return(true);
}

