//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////

#include "render/graytrac.h"
#include "util/gstack.h"

#include "gx/gx_defs.h"
#include "gx/gx_bbox.h"

#include "gx/gx_beam.h"



//////////////////////////////////////////////////////////////////////////////
// Face	: POS_X
// U	: Y
// V	: Z
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesPOS_X(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] > 0) {
	rPlanes[0] = new 
	    GXBeamPlane(GVector3(pExts[G_LO_U], -1, 0),
			pExts[G_LO_U] * pExts[G_HI_X] - pExts[G_LO_Y]);
	rPlanes[0]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    else {
	rPlanes[0] = new 
	    GXBeamPlane(GVector3(pExts[G_LO_U], -1, 0),
			pExts[G_LO_U] * pExts[G_LO_X] - pExts[G_LO_Y]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] < 0) {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(-pExts[G_HI_U], 1, 0),
		      -pExts[G_HI_U] * pExts[G_HI_X] + pExts[G_HI_Y]);
	rPlanes[1]->setNearIndex(G_LO_X, G_LO_Y, G_HI_Z);
    }

    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(-pExts[G_HI_U], 1, 0),
		      -pExts[G_HI_U] * pExts[G_LO_X] + pExts[G_HI_Y]);
	rPlanes[1]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] > 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(pExts[G_LO_V], 0, -1),
		      pExts[G_LO_V] * pExts[G_HI_X] - pExts[G_LO_Z]);
	rPlanes[2]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(pExts[G_LO_V], 0, -1),
		      pExts[G_LO_V] * pExts[G_LO_X] - pExts[G_LO_Z]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] < 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(-pExts[G_HI_V], 0, 1),
		      -pExts[G_HI_V] * pExts[G_HI_X] + pExts[G_HI_Z]);
	rPlanes[3]->setNearIndex(G_LO_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(-pExts[G_HI_V], 0, 1),
		      -pExts[G_HI_V] * pExts[G_LO_X] + pExts[G_HI_Z]);
	rPlanes[3]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
}


//////////////////////////////////////////////////////////////////////////////
// Face	: NEG_X
// U	: Y
// V	: Z
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesNEG_X(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] < 0) {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(-pExts[G_LO_U], -1, 0),
		      -pExts[G_LO_U] * pExts[G_HI_X] - pExts[G_LO_Y]);
	rPlanes[0]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    else {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(-pExts[G_LO_U], -1, 0),
		      -pExts[G_LO_U] * pExts[G_LO_X] - pExts[G_LO_Y]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] > 0) {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(pExts[G_HI_U], 1, 0),
		      pExts[G_HI_U] * pExts[G_HI_X] + pExts[G_HI_Y]);
	rPlanes[1]->setNearIndex(G_LO_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(pExts[G_HI_U], 1, 0),
		      pExts[G_HI_U] * pExts[G_LO_X] + pExts[G_HI_Y]);
	rPlanes[1]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] < 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-pExts[G_LO_V], 0, -1),
		      -pExts[G_LO_V] * pExts[G_HI_X] - pExts[G_LO_Z]);
	rPlanes[2]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-pExts[G_LO_V], 0, -1),
		      -pExts[G_LO_V] * pExts[G_LO_X] - pExts[G_LO_Z]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] > 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(pExts[G_HI_V], 0, 1),
		      pExts[G_HI_V] * pExts[G_HI_X] + pExts[G_HI_Z]);
	rPlanes[3]->setNearIndex(G_LO_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(pExts[G_HI_V], 0, 1),
		      pExts[G_HI_V] * pExts[G_LO_X] + pExts[G_HI_Z]);
	rPlanes[3]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
}



//////////////////////////////////////////////////////////////////////////////
// Face	: POS_Y
// U	: Z
// V	: X
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesPOS_Y(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] > 0) {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(0, pExts[G_LO_U], -1),
		      pExts[G_LO_U] * pExts[G_HI_Y] - pExts[G_LO_Z]);
	rPlanes[0]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(0, pExts[G_LO_U], -1),
		      pExts[G_LO_U] * pExts[G_LO_Y] - pExts[G_LO_Z]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] < 0) {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(0, -pExts[G_HI_U], 1),
		      -pExts[G_HI_U] * pExts[G_HI_Y] + pExts[G_HI_Z]);
	rPlanes[1]->setNearIndex(G_HI_X, G_LO_Y, G_LO_Z);
    }
    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(0, -pExts[G_HI_U], 1),
		      -pExts[G_HI_U] * pExts[G_LO_Y] + pExts[G_HI_Z]);
	rPlanes[1]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] > 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-1, pExts[G_LO_V], 0),
		      pExts[G_LO_V] * pExts[G_HI_Y] - pExts[G_LO_X]);
	rPlanes[2]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-1, pExts[G_LO_V], 0),
		      pExts[G_LO_V] * pExts[G_LO_Y] - pExts[G_LO_X]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] < 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(1, -pExts[G_HI_V], 0),
		      -pExts[G_HI_V] * pExts[G_HI_Y] + pExts[G_HI_X]);
	rPlanes[3]->setNearIndex(G_LO_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(1, -pExts[G_HI_V], 0),
		      -pExts[G_HI_V] * pExts[G_LO_Y] + pExts[G_HI_X]);
	rPlanes[3]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
}


//////////////////////////////////////////////////////////////////////////////
// Face	: NEG_Y
// U	: Z
// V	: X
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesNEG_Y(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] < 0) {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(0, -pExts[G_LO_U], -1),
		      -pExts[G_LO_U] * pExts[G_HI_Y] - pExts[G_LO_Z]);
	rPlanes[0]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(0, -pExts[G_LO_U], -1),
		      -pExts[G_LO_U] * pExts[G_LO_Y] - pExts[G_LO_Z]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] > 0) {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(0, pExts[G_HI_U], 1),
		      pExts[G_HI_U] * pExts[G_HI_Y] + pExts[G_HI_Z]);
	rPlanes[1]->setNearIndex(G_HI_X, G_LO_Y, G_LO_Z);
    }
    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(0, pExts[G_HI_U], 1),
		      pExts[G_HI_U] * pExts[G_LO_Y] + pExts[G_HI_Z]);
	rPlanes[1]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] < 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-1, -pExts[G_LO_V], 0),
		      -pExts[G_LO_V] * pExts[G_HI_Y] - pExts[G_LO_X]);
	rPlanes[2]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(-1, -pExts[G_LO_V], 0),
		      -pExts[G_LO_V] * pExts[G_LO_Y] - pExts[G_LO_X]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] > 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(1, pExts[G_HI_V], 0),
		      pExts[G_HI_V] * pExts[G_HI_Y] + pExts[G_HI_X]);
	rPlanes[3]->setNearIndex(G_LO_X, G_LO_Y, G_HI_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(1, pExts[G_HI_V], 0),
		      pExts[G_HI_V] * pExts[G_LO_Y] + pExts[G_HI_X]);
	rPlanes[3]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
}



//////////////////////////////////////////////////////////////////////////////
// Face	: POS_Z
// U	: X
// V	: Y
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesPOS_Z(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] > 0) {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(-1, 0, pExts[G_LO_U]),
		      pExts[G_LO_U] * pExts[G_HI_Z] - pExts[G_LO_X]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[0] = new 
	    GXBeamPlane(GVector3(-1, 0, pExts[G_LO_U]),
			pExts[G_LO_U] * pExts[G_LO_Z] - pExts[G_LO_X]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] < 0) {
	rPlanes[1] = new 
	    GXBeamPlane(GVector3(1, 0, -pExts[G_HI_U]),
			-pExts[G_HI_U] * pExts[G_HI_Z] + pExts[G_HI_X]);
	rPlanes[1]->setNearIndex(G_LO_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(1, 0, -pExts[G_HI_U]),
		      -pExts[G_HI_U] * pExts[G_LO_Z] + pExts[G_HI_X]);
	rPlanes[1]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] > 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(0, -1, pExts[G_LO_V]),
		      pExts[G_LO_V] * pExts[G_HI_Z] - pExts[G_LO_Y]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(0, -1, pExts[G_LO_V]),
		      pExts[G_LO_V] * pExts[G_LO_Z] - pExts[G_LO_Y]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] < 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(0, 1, -pExts[G_HI_V]),
		      -pExts[G_HI_V] * pExts[G_HI_Z] + pExts[G_HI_Y]);
	rPlanes[3]->setNearIndex(G_HI_X, G_LO_Y, G_LO_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(0, 1, -pExts[G_HI_V]),
		      -pExts[G_HI_V] * pExts[G_LO_Z] + pExts[G_HI_Y]);
	rPlanes[3]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
}


//////////////////////////////////////////////////////////////////////////////
// Face	: NEG_Z
// U	: X
// V	: Y
//////////////////////////////////////////////////////////////////////////////
void _constructBeamPlanesNEG_Z(GFloat *pExts, GXBeamPlane *rPlanes[4] ) { 
    
    //// construct the 2 planes bounding the U-extensions 
    if(pExts[G_LO_U] < 0) {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(-1, 0, -pExts[G_LO_U]),
		      -pExts[G_LO_U] * pExts[G_HI_Z] - pExts[G_LO_X]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[0] = new 
	  GXBeamPlane(GVector3(-1, 0, -pExts[G_LO_U]),
		      -pExts[G_LO_U] * pExts[G_LO_Z] - pExts[G_LO_X]);
	rPlanes[0]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_U] > 0) {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(1, 0, pExts[G_HI_U]),
		      pExts[G_HI_U] * pExts[G_HI_Z] + pExts[G_HI_X]);
	rPlanes[1]->setNearIndex(G_LO_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[1] = new 
	  GXBeamPlane(GVector3(1, 0, pExts[G_HI_U]),
		      pExts[G_HI_U] * pExts[G_LO_Z] + pExts[G_HI_X]);
	rPlanes[1]->setNearIndex(G_LO_X, G_HI_Y, G_HI_Z);
    }
    
    //// construct the 2 planes bounding the V-extensions 
    if(pExts[G_LO_V] < 0) {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(0, -1, -pExts[G_LO_V]),
		      -pExts[G_LO_V] * pExts[G_HI_Z] - pExts[G_LO_Y]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_LO_Z);
    }
    else {
	rPlanes[2] = new 
	  GXBeamPlane(GVector3(0, -1, -pExts[G_LO_V]),
		      -pExts[G_LO_V] * pExts[G_LO_Z] - pExts[G_LO_Y]);
	rPlanes[2]->setNearIndex(G_HI_X, G_HI_Y, G_HI_Z);
    }
    
    if(pExts[G_HI_V] > 0) {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(0, 1, pExts[G_HI_V]),
		      pExts[G_HI_V] * pExts[G_HI_Z] + pExts[G_HI_Y]);
	rPlanes[3]->setNearIndex(G_HI_X, G_LO_Y, G_LO_Z);
    }
    else {
	rPlanes[3] = new 
	  GXBeamPlane(GVector3(0, 1, pExts[G_HI_V]),
		      pExts[G_HI_V] * pExts[G_LO_Z] + pExts[G_HI_Y]);
	rPlanes[3]->setNearIndex(G_HI_X, G_LO_Y, G_HI_Z);
    }
}





GFloat _getMinDistPOS_X(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.min().getX() - pRay.getPoint().getX()) / pRay.getVector().getX());
}
GFloat _getMinDistNEG_X(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.max().getX() - pRay.getPoint().getX()) / pRay.getVector().getX());
}
GFloat _getMinDistPOS_Y(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.min().getY() - pRay.getPoint().getY()) / pRay.getVector().getY());
}
GFloat _getMinDistNEG_Y(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.max().getY() - pRay.getPoint().getY()) / pRay.getVector().getY());
}
GFloat _getMinDistPOS_Z(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.min().getZ() - pRay.getPoint().getZ()) / pRay.getVector().getZ());
}
GFloat _getMinDistNEG_Z(const GRay3& pRay, const GBounds3 &pB) {
    return ((pB.max().getZ() - pRay.getPoint().getZ()) / pRay.getVector().getZ());
}


bool _isBehindVoxelPOS_X(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.max().getX() < pVoxel.min().getX());
}
bool _isBehindVoxelNEG_X(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.min().getX() > pVoxel.max().getX());
}
bool _isBehindVoxelPOS_Y(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.max().getY() < pVoxel.min().getY());
}
bool _isBehindVoxelNEG_Y(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.min().getY() > pVoxel.max().getY());
}
bool _isBehindVoxelPOS_Z(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.max().getZ() < pVoxel.min().getZ());
}
bool _isBehindVoxelNEG_Z(const GBounds3 &pVoxel, const GBounds3 &pBound) {
    return (pBound.min().getZ() > pVoxel.max().getZ());
}




//////////////////////////////////////////////////////////////////////////////
// @GXBeam@
//////////////////////////////////////////////////////////////////////////////

GXBeam::GXBeam(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList) {

    // setup face-dependent routines
    mFace = pFace;
    switch (pFace) {
      case POS_X: 
	_constructBeamPlanes = _constructBeamPlanesPOS_X; 
	_getMinDist = _getMinDistPOS_X;
	_isBehindVoxel = _isBehindVoxelPOS_X;
	break;
      
      case NEG_X: 
	_constructBeamPlanes = _constructBeamPlanesNEG_X; 
	_getMinDist = _getMinDistNEG_X;
	_isBehindVoxel = _isBehindVoxelNEG_X;
	break;
      
      case POS_Y: 
	_constructBeamPlanes = _constructBeamPlanesPOS_Y; 
	_getMinDist = _getMinDistPOS_Y;
	_isBehindVoxel = _isBehindVoxelPOS_Y;
	break;
	
      case NEG_Y: 
	_constructBeamPlanes = _constructBeamPlanesNEG_Y;
	_getMinDist = _getMinDistNEG_Y;
	_isBehindVoxel = _isBehindVoxelNEG_Y;
	break;
      
      case POS_Z: 
	_constructBeamPlanes = _constructBeamPlanesPOS_Z; 
	_getMinDist = _getMinDistPOS_Z;
	_isBehindVoxel = _isBehindVoxelPOS_Z;
	break;
      
      default:   
	_constructBeamPlanes = _constructBeamPlanesNEG_Z; 
	_getMinDist = _getMinDistNEG_Z;
	_isBehindVoxel = _isBehindVoxelNEG_Z;
	break;
    }

    // setup the 10 extents in 5D
    for( register int i=0; i< 10; i++)
	mExtents[i] = pExtents[i];
    
    // setup the 4 beam-planes
    for(i=0; i< 4; i++)
	mPlanes[i] = 0;
    
    // copy initital list of candidates
    mCandidatesList = pCandidatesList;
}

GXBeam::~GXBeam(void) {
    gkillPointer(mCandidatesList);
    
    for(register int i = 0; i< 4; i++)
	gkillPointer(mPlanes[i]);
}

    
bool GXBeam::isInside(GFloat *p5D) {
    return ((p5D[G_D_X] >= mExtents[G_LO_X]) && (p5D[G_D_X] <= mExtents[G_HI_X]) &&
	    (p5D[G_D_Y] >= mExtents[G_LO_Y]) && (p5D[G_D_Y] <= mExtents[G_HI_Y]) &&
	    (p5D[G_D_Z] >= mExtents[G_LO_Z]) && (p5D[G_D_Z] <= mExtents[G_HI_Z]) &&
	    (p5D[G_D_U] >= mExtents[G_LO_U]) && (p5D[G_D_U] <= mExtents[G_HI_U]) &&
	    (p5D[G_D_V] >= mExtents[G_LO_V]) && (p5D[G_D_V] <= mExtents[G_HI_V]));
}


bool GXBeam::isOverlap(const GBounds3& pB) const {
   
    gcheck(!mPlanes[0], "GXBeam: cliping-planes not generated yet!\n");
    
    //// create a 3D-box, which is the 3D-voxel of the 5D-hyper-cube
    GPoint3 tMin(mExtents[G_LO_X], mExtents[G_LO_Y], mExtents[G_LO_Z]);
    GPoint3 tMax(mExtents[G_HI_X], mExtents[G_HI_Y], mExtents[G_HI_Z]);
    GBounds3 tVoxel(tMin, tMax);
    
    // culling this object, if its whole bound lies behind the rays origin
    if(_isBehindVoxel(tVoxel, pB))
	return(false);
	
	
    // First, include this object to the candidates list, 
    // if it overlapes the 3D-Voxel
    if(tVoxel.isOverlap(pB)) 
	return (true);
	
    // Second, overlapping-test w/ the cutting-planes
    GFloat tExts[6];
    tExts[0] = pB.getMin().getX(); tExts[1] = pB.getMax().getX();
    tExts[2] = pB.getMin().getY(); tExts[3] = pB.getMax().getY();
    tExts[4] = pB.getMin().getZ(); tExts[5] = pB.getMax().getZ();
    if(mPlanes[0]->isOverlap(tExts) && mPlanes[1]->isOverlap(tExts) &&
       mPlanes[2]->isOverlap(tExts) && mPlanes[3]->isOverlap(tExts))
	return (true);
    
    return (false);   
}


void GXBeam::classifyCandidatesList(void) { 
    
    //// construct the 4 shaft-planes
    _constructBeamPlanes(mExtents, mPlanes);
   
 
        
    //// create a new list
    GXObjectList *tList = new GXObjectList();
    
    //// create a 3D-box, which is the 3D-voxel of the 5D-hyper-cube
    GPoint3 tMin(mExtents[G_LO_X], mExtents[G_LO_Y], mExtents[G_LO_Z]);
    GPoint3 tMax(mExtents[G_HI_X], mExtents[G_HI_Y], mExtents[G_HI_Z]);
    GBounds3 tVoxel(tMin, tMax);
    
    

    //// Intersecting this beam w/ all candidates of the list.
    
    for(GXObjectItr tItr(*mCandidatesList); tItr.more(); tItr.next()) {
	GBounds3 tB = tItr.cur()->getWorldBounds();
	//// isOverlap(tB), inlined version
	
	// culling this object, if its whole bound lies behind the rays origin
	if(_isBehindVoxel(tVoxel, tB))
	    continue;
	
	// include this object to the candidates list, if it overlapes the 3D-Voxel
	if(tVoxel.isOverlap(tB)) {
	    tList->append(tItr.cur());
	    continue;
	}
	
	GFloat tExts[6];
	tExts[0] = tB.getMin().getX(); tExts[1] = tB.getMax().getX();
	tExts[2] = tB.getMin().getY(); tExts[3] = tB.getMax().getY();
	tExts[4] = tB.getMin().getZ(); tExts[5] = tB.getMax().getZ();
	
	// if object also inside the 4 shaft-planes, add it to the candidates list.
	if(mPlanes[0]->isOverlap(tExts) && mPlanes[1]->isOverlap(tExts) &&
	   mPlanes[2]->isOverlap(tExts) && mPlanes[3]->isOverlap(tExts))
	    tList->append(tItr.cur());
    }
    
    // A new candidates list created and classified.
    mCandidatesList = tList;

}



void GXBeam::getClippedCandidatesList(const GRay3& pRay, 
				      GXObjectList &rList) const
{
    rList.reset();
    
    for(GXObjectItr _Itr(*mCandidatesList); _Itr.more(); _Itr.next()) {
	GXObject *tPrim = _Itr.cur();
	
	GBounds3 tB = tPrim->getWorldBounds();
	if(pRay.getTmax() < _getMinDist(pRay, tB)) {
	    break;
	}
	else
	    rList.append(tPrim);
	
    }
}



bool GXBeam::intersectShadow(const GRay3& pRay, GInterSectionList& rInter) {
 
    GFloat tMin, tMax;
    static GXObjectList sTransList;


    for(GXObjectItr _Itr(*mCandidatesList); _Itr.more(); _Itr.next()) {

	GXObject *tPrim = _Itr.cur();
	
	GBounds3 tB = tPrim->getWorldBounds();
	if(pRay.getTmax() < _getMinDist(pRay, tB)) {
	    break;
	}


	// check the mailbox and the bbox, before I-testing w/ the prim. 
	if(tPrim->checkID(pRay)) {
	    gstat(GXStatistics::MailBoxHits++);
	    continue;
	}
	
	if(tPrim->isBounded()) {
	    gstat(GXStatistics::BBox.mTests++);
	    
	    // miss the bbox, continuing w/ the next prim.
	    if(!tB.findMinMax(pRay, tMin, tMax))
		continue;
	    
	    gstat(GXStatistics::BBox.mHits++);
	}


	
	// If this prim is transparent, append it to a temporal list for
	// later processing.
	if(tPrim->getTransmission()) {
	    sTransList.append(tPrim);
	    continue;
	}

	if(tPrim->intersect(pRay, rInter))
	    return (true);
    
    }

    // no opaque object was found, so continue w/ transparent-objects
    if(sTransList.count()) {
	GOneInterSectionList tInter;
	bool tHit = false;
	for(GXObjectItr Itr(sTransList); Itr.more(); Itr.next()) {
	    GXObject *tPrim = Itr.cur();
	    
	    if(tPrim->intersect(pRay, tInter))
		tHit = true;
	}
	if(tHit) rInter.updateTransmission(tInter.getTransmission());
    
	sTransList.reset();
    }
    
    return (false);
}


bool GXBeam::intersectShading(GRay3& pRay, GInterSectionList& rInter) {
    bool tHIT = rInter.getNumber();
    GFloat tMinT = pRay.getTmax();
    GFloat tMin, tMax;

    for(GXObjectItr _Itr(*mCandidatesList); _Itr.more(); _Itr.next()) {

	GXObject *tPrim = _Itr.cur();
	GBounds3 tB = tPrim->getWorldBounds();
	
	if(tHIT && tMinT < _getMinDist(pRay, tB))
	    break;
	
	// check the mailbox and the bbox, before I-testing w/ the prim. 
	if(tPrim->checkID(pRay)) {
	    gstat(GXStatistics::MailBoxHits++);
	    continue;
	} 
	
	if(tPrim->isBounded()) {
	    gstat(GXStatistics::BBox.mTests++);
	    
	    // miss the bbox, continuing w/ the next prim.
	    if(!tB.findMinMax(pRay, tMin, tMax))
		continue;
	    
	    gstat(GXStatistics::BBox.mHits++);
	}
	

	if(tPrim->intersect(pRay, rInter)) {
	    tHIT = true;
	    tMinT = rInter.getFirst().getT();
	}
    }
    
    return (tHIT);
}


