//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <stdlib.h>
#include "render/graytrac.h"
#include "util/gstack.h"

#include "gx/gx_defs.h"
#include "gx/gx_bbox.h"
#include "gx/gx_bsp.h"


extern GFloat _getMinDistPOS_X(const GRay3& pRay, const GBounds3 &pB); 
extern GFloat _getMinDistNEG_X(const GRay3& pRay, const GBounds3 &pB);
extern GFloat _getMinDistPOS_Y(const GRay3& pRay, const GBounds3 &pB);
extern GFloat _getMinDistNEG_Y(const GRay3& pRay, const GBounds3 &pB);
extern GFloat _getMinDistPOS_Z(const GRay3& pRay, const GBounds3 &pB);
extern GFloat _getMinDistNEG_Z(const GRay3& pRay, const GBounds3 &pB);





#ifdef OPTI_SUBDIVISION

int __compareMinSeg(const void *pA, const void *pB )   {
    Segment *tA  = (Segment *) pA;
    Segment *tB  = (Segment *) pB;
    if ( tA->min == tB->min ) 
	return( 0 ) ;
    else   
	return( tA->min < tB->min ? -1 : 1 ) ;
}

int __compareMaxSeg(const void *pA, const void *pB )   {
    Segment *tA  = (Segment *) pA;
    Segment *tB  = (Segment *) pB;
    if ( tA->max == tB->max ) 
	return( 0 ) ;
    else   
	return( tA->max > tB->max ? -1 : 1 ) ;
}



//////////////////////////////////////////////////////////////////////////////
// This routine should be faster than __subdivide, but this istn the case 
// for the computer i used (Indy 2)! Give it a try if __subdivide()
// runs too slow on your computer.
// The main idea is based on the observation, that __subdivide calls qsort() 
// every time to sort the min. and max. extents of the bounding boxes. We can
// avoid this extra work by clever use the already sorted lists.
//////////////////////////////////////////////////////////////////////////////
void __subdivide(GX_BSP_Node*        rNode, 
		 SortedSegmentLists* pSortedLists,
		 int                 pDepth,        
		 int                 pMaxDepth,     
		 int                 pMaxListLength )
{
    register int i, j, k;
    GX_BSP_Node*  tCur;
    int           tAxis;
    GFloat        tDiv;
   

   
    GXObjectList *tListChildren = rNode->getListOfChildren();
    int tNum = tListChildren->count();
   
    ////
    //// Contruct the 6 sorted lists for 6 dominant axes
    ////
    
    for(j=0; j< 6; j++)
	rNode->mCandidatesLists[j] = new GXObjectList;
    
    for (j = 0; j < tNum; j++) {
	rNode->mCandidatesLists[POS_X]->append(pSortedLists->Mins[G_D_X][j]->obj);
	rNode->mCandidatesLists[NEG_X]->append(pSortedLists->Maxs[G_D_X][j]->obj);
	rNode->mCandidatesLists[POS_Y]->append(pSortedLists->Mins[G_D_Y][j]->obj);
	rNode->mCandidatesLists[NEG_Y]->append(pSortedLists->Maxs[G_D_Y][j]->obj);
	rNode->mCandidatesLists[POS_Z]->append(pSortedLists->Mins[G_D_Z][j]->obj);
	rNode->mCandidatesLists[NEG_Z]->append(pSortedLists->Maxs[G_D_Z][j]->obj);
    }
    
    // init. 2 children
    rNode->mChild[1] = rNode->mChild[0] = 0;
    
    if ((tNum  > pMaxListLength) && (pDepth < pMaxDepth))  {
      
	//// generate 2 children w/ randomly div-axis! 
	for (i = 0; i < 2; i++)  {
	    tCur = rNode->mChild[i] = new GX_BSP_Node;
	 
	    tCur->min(rNode->min());
	    tCur->max(rNode->max());
	    tCur->mAxis = rand() % 3;
	}
	
	//// testing for optimal subdivision 
	bool tFOUND = false;

/**/
	for (j = 0; j < tNum && !tFOUND; j++) {
		
	    // for each axis.. 
	    for (tAxis = rNode->mAxis, k = 0; k < 3; tAxis = (tAxis+1)%3, k++) {
	       
		if (pSortedLists->Mins[tAxis][j]->min >= pSortedLists->Maxs[tAxis][j]->max) {
// 		    
// 		    tDiv = 0.5 * (pSortedLists->Mins[tAxis][j]->min + 
// 				  pSortedLists->Maxs[tAxis][j]->max);
// 		    
		    tDiv = pSortedLists->Mins[tAxis][j]->min - gEpsilon;
		    rNode->mChild[0]->max()[tAxis] = 
			rNode->mChild[1]->min()[tAxis] = tDiv;
		    
		    rNode->mAxis = tAxis;
		    tFOUND = true;
		    
		    break;
		}
	    }
	}
/**/
/**
	
	// @Yuk!
	tAxis = rNode->mAxis;
	GFloat tDiv = (rNode->min()[tAxis] + rNode->max()[tAxis])/2;
	rNode->mChild[0]->max()[tAxis] = 
	  rNode->mChild[1]->min()[tAxis] = tDiv;
		    
	tFOUND = TRUE;
**/


	//// contruct the 2 children
	if (tFOUND)  {
	    // cerr << "<" << rNode->mAxis << ">";
	    tAxis = rNode->mAxis;
	    
	    
	    SortedSegmentLists tLeft, tRight;
	    
	    ////
	    //// construct the sorted lists of extents along the subdivision axis
	    ////
	    
	    //// process the MIN-List of the current voxel.
	    for(SegmentItr _minI(pSortedLists->Mins[tAxis]); _minI.more(); _minI.next()) {
		Segment *tSeg = _minI.cur();
		
		if(tSeg->min <= tDiv) {
		    // Add this object to the left child
		    rNode->mChild[0]->getListOfChildren()->append(tSeg->obj);
		    
		    // Add the segment to the MIN-list of the left child.
		    tLeft.Mins[tAxis].append(tSeg);
		    
		    // Add the segment to the MIN-List of the right child.
		    if(tSeg->max >= tDiv)
			tRight.Mins[tAxis].append(tSeg);
		}
		else
		    // Add the segment to the MIN-list of the right child. 
		    tRight.Mins[tAxis].append(tSeg);
	    
	    }
	    
	    //// process the MAX-List of the current voxel.
	    for(SegmentItr _maxI(pSortedLists->Maxs[tAxis]); _maxI.more(); _maxI.next()) {

		Segment *tSeg = _maxI.cur();
		
		if(tSeg->max >= tDiv) {
		    // Add this object to the right child
		    rNode->mChild[1]->getListOfChildren()->append(tSeg->obj);
		 
		    // Add the segment to the MAX-list of the right child.
		    tRight.Maxs[tAxis].append(tSeg);

		    // Add the segment to the MAX-List of the left child.
		    if(tSeg->min <= tDiv)
			tLeft.Maxs[tAxis].append(tSeg);
		}
		else
		    // Add the segment to the MAX-list of the left child. 
		    tLeft.Maxs[tAxis].append(tSeg);
	    
	    }

	    ////
	    //// construct the sorted list of extents for 2 remaining axes
	    ////

	    {
		for( tAxis = (rNode->mAxis+1)%3; tAxis != rNode->mAxis; tAxis = (tAxis+1)%3) {
		    for(SegmentItr _minI(pSortedLists->Mins[tAxis]); _minI.more(); _minI.next()) {
			Segment *tSeg = _minI.cur();
			
			if(tSeg->obj->getWorldBounds().getMin()[rNode->mAxis] <= tDiv) 
			    tLeft.Mins[tAxis].append(tSeg);
			if(tSeg->obj->getWorldBounds().getMax()[rNode->mAxis] >= tDiv)
			    tRight.Mins[tAxis].append(tSeg);
		    }

		    for(SegmentItr _maxI(pSortedLists->Maxs[tAxis]); _maxI.more(); _maxI.next()) {
			Segment *tSeg = _maxI.cur();
		
			if(tSeg->obj->getWorldBounds().getMin()[rNode->mAxis] <= tDiv) 
			    tLeft.Maxs[tAxis].append(tSeg);
			if(tSeg->obj->getWorldBounds().getMax()[rNode->mAxis] >= tDiv)
			    tRight.Maxs[tAxis].append(tSeg);
		    }
		}
	    }
	    
	    
	    
	    //// process the 2 chidren recursivly... 
	    __subdivide(rNode->mChild[0], &tLeft, pDepth+1, pMaxDepth, pMaxListLength);
	    __subdivide(rNode->mChild[1], &tRight, pDepth+1, pMaxDepth, pMaxListLength);
	
	} // subdivision not found
	
	//// stop subdividing if it doesnt help anymore! 
	else {
	    gkillPointer(rNode->mChild[0]);
	    gkillPointer(rNode->mChild[1]);
	    
	    // @DEBUG@
	    cerr << "Subdivision stop at depth " << pDepth << "\n";;
	}
	
    }
}

#else // OPTI_SUBDIVISION


int __compareGFloat(const void *pA, const void *pB )   {
    GFloat *tA  = (GFloat *) pA;
    GFloat *tB  = (GFloat *) pB;
    if ( *tA == *tB ) 
	return( 0 ) ;
    else   
	return( *tA < *tB ? -1 : 1 ) ;
}



// @FUNCT@: @__subdivide@
// @DESCR@: recursive divide a BSP-Node 
// @NOTES@: boys, this routine is really slow ! But dont bother about it
// because it's pre-processing time. The idea, how to make a balanced
// BSP-Tree, is simple, but very hard to implement without bounding boxes.
void __subdivide(GX_BSP_Node*  rNode,
		 int           pDepth,        
		 int           pMaxDepth,     
		 int           pMaxListLength )
{
    int i, j, k;
    GX_BSP_Node*  tCur;
    GFloat*       tMins[3];
    GFloat*       tMaxs[3];
    int           tNum;
    int           tAxis;
   
    rNode->mChild[1] = rNode->mChild[0] = 0;
   
    GXObjectList *tListChildren = rNode->getListOfChildren();
    tNum = tListChildren->count();
   
    if ((tNum  > pMaxListLength) && (pDepth < pMaxDepth))  {
      
	// initialize 3 arrays for later sorting.. 
	for(i = 0; i < 3; i++) {
	    tMins[i] = new GFloat[tNum];
	    tMaxs[i] = new GFloat[tNum];
	}
      
	i = 0;
	for (GXObjectItr _Itr(*tListChildren); _Itr.more(); _Itr.next()) {
	    GBounds3  tBox(_Itr.cur()->getWorldBounds());
	 
	    tMins[D_X][i] = tBox.min().getX();
	    tMins[D_Y][i] = tBox.min().getY();
	    tMins[D_Z][i] = tBox.min().getZ();
	    tMaxs[D_X][i] = tBox.max().getX();
	    tMaxs[D_Y][i] = tBox.max().getY();
	    tMaxs[D_Z][i] = tBox.max().getZ();
	    i++;
	}
      
	// Sorting.. 
	for(i = 0; i < 3 ; i++) {
	    qsort( tMins[i], tNum, sizeof(GFloat), __compareGFloat) ;
	    qsort( tMaxs[i], tNum, sizeof(GFloat), __compareGFloat) ;
	}
      
	// generate 2 children w/ randomly div-axis! 
	for (i = 0; i < 2; i++)  {
	    tCur = rNode->mChild[i] = new GX_BSP_Node;
	 
	    tCur->min(rNode->min());
	    tCur->max(rNode->max());
	    tCur->mAxis = rand() % 3;
	}
	
	// testing for optimal subdivision 
	bool tFOUND = FALSE;


	for (j = 0; j < tNum && !tFOUND; j++) {
		
	    // for each axis.. 
	    for (tAxis = rNode->mAxis, k = 0; k < 3; 
		 tAxis = (tAxis+1)%3, k++) {
	       
		if (tMins[tAxis][j] >= tMaxs[tAxis][tNum - j - 1]) {
		    GFloat tDiv = 0.5 * (tMins[tAxis][j] + 
					 tMaxs[tAxis][tNum -j - 1]);
		    rNode->mChild[0]->max()[tAxis] = 
		      rNode->mChild[1]->min()[tAxis] = tDiv;
		    
		    rNode->mAxis = tAxis;
		    tFOUND = TRUE;
		    break;
		}
	    }
	}
	 



	// stop subdividing if it doesnt help anymore! 
	if (tFOUND)  {
	    cerr << "<" << rNode->mAxis << ">";
	    
	    for (i = 0; i < 2; i++)  {
		tCur = rNode->mChild[i];
		for(GXObjectItr _Itr(*tListChildren);
		    _Itr.more(); _Itr.next()) 
		  {
		      GXObject *tObj = _Itr.cur();
		      
		      // simple test for overlaping 
		      // Note: There is no need for expanding the 
		      // parent BSP-bound!
		      if( tObj->isOverlap(*tCur))
			tCur->getListOfChildren()->append(tObj);
		  }
	    
		__subdivide(tCur, pDepth+1, pMaxDepth, pMaxListLength);
	    }
	}
	else {
	    gkillPointer(rNode->mChild[0]);
	    gkillPointer(rNode->mChild[1]);
	}
	
	for( i = 0; i < 3; i++) {
	    delete [] tMins[i];
	    delete [] tMaxs[i];
	}
    }
}

#endif // OPTI_SUBDIVISION    





class NodeLevel {
  public:
    GX_BSP_Node *mNode;
    int mLevel;


    NodeLevel(void) 
    :mNode(0), mLevel(0) {}
    
    NodeLevel(GX_BSP_Node *pNode, int pLevel) 
    :mNode(pNode), mLevel(pLevel){}
    
    NodeLevel(const NodeLevel& pN) {
	mNode = pN.mNode;
	mLevel = pN.mLevel;
    }
    
}; 

//////////////////////////////////////////////////////////////////////////////
// Build the ABVHs for all BSP-nodes
//
//////////////////////////////////////////////////////////////////////////////
void __buildBVH(GX_BSP_Node *rNode) {
    long tNumVoxels = 0L;
    long tNumCandidates = 0L;
    long tLevels = 0L;
    
    
    

    GStack<NodeLevel> tStk;
    tStk.push(NodeLevel(rNode, 0));
    while(!tStk.empty()) {
	NodeLevel tNodeLevel = tStk.pop();
	GX_BSP_Node *tNode = tNodeLevel.mNode;
	int tLevel = tNodeLevel.mLevel;

	if(!(tNode->mChild[0])) {
	    gstat(tNode->mNumOfCandidates = tNode->getNumOfChildren());
	    
	    tNumVoxels ++;
	    tNumCandidates += tNode->getNumOfChildren();
	    tLevels += tLevel;
	    
	
	    tNode->buildHierarchy();
	
	} else {
	    tStk.push(NodeLevel(tNode->mChild[0], tLevel +1));
	    tStk.push(NodeLevel(tNode->mChild[1], tLevel +1));
	    tNode->getListOfChildren()->free();
	}
    }

    cerr << "Total # of voxels : " <<  tNumVoxels << "\n";
    cerr << "Average # of candidates per voxel : " 
	 << (GFloat)tNumCandidates / (GFloat)tNumVoxels << "\n";  
	
    cerr << "Average depth of the BSP-Tree : " 
	 <<  (GFloat)tLevels/ (GFloat)tNumVoxels << "\n";
    
}

//////////////////////////////////////////////////////////////////////////////
// @GX_BSP_Node::~GX_BSP_Node@
// DTor
//////////////////////////////////////////////////////////////////////////////
GX_BSP_Node::~GX_BSP_Node(void) {
    gkillPointer(mChild[0]);
    gkillPointer(mChild[1]);
}


//////////////////////////////////////////////////////////////////////////////
// @GX_BSP_Node::classify@
// 
// @DESCR@: find out, in wich child-node the rays origin lies.
// Modify: dist  -> distance from rays origin to the div-plane
//         pNear -> the node, in wich the rays origin lies
//         pFar  -> the other node.
//
// @NOTES@:
//  There was a bug in the BSP-Code in GGemsIII ! If the rays
//  origin lies in the division plane, we must test the rays
//  direction to classify the right NEAR- and FAR child.
//  
//  The actual BSP-code works in object-space. Modification for
//  ray-space is neccessary by ray-classification-algorithm. 
//
//////////////////////////////////////////////////////////////////////////////
void GX_BSP_Node::classify(const GRay3&  pRay,
			   GFloat&       rDist,
			   GX_BSP_Node*& rNear,
			   GX_BSP_Node*& rFar) 
{
    
    GFloat tOriginToDivPlane = mChild[0]->getMax()[mAxis] 
      - (pRay.getPoint())[mAxis];
    GFloat tDivDirection     = (pRay.getVector())[mAxis];
    
    if ( tOriginToDivPlane > 0 ){
	rNear = mChild[0];
	rFar  = mChild[1];
    } else if(tOriginToDivPlane < 0){
	rFar  = mChild[0];
	rNear = mChild[1];
    } 
    // This bug was really hard to find...
    else {
	 if ( tDivDirection < 0) {
	     rNear = mChild[0];
	     rFar  = mChild[1];
	 } else {
	     rFar  = mChild[0];
	     rNear = mChild[1];
	 }
     }     
    rDist =  (tDivDirection) ? tOriginToDivPlane / tDivDirection : MAXFLT;
}




//////////////////////////////////////////////////////////////////////////////
// define a stack for recursion-elimination
//////////////////////////////////////////////////////////////////////////////
class BSP_StkNode  {
 public:
    GX_BSP_Node*  mNode;
    GFloat        mMin, mMax;
   
    BSP_StkNode(GX_BSP_Node* pNode = 0, GFloat pMin = 0, GFloat pMax = 0) {
	mNode = pNode;
	mMin = pMin;
	mMax = pMax;
    }
};
 


//////////////////////////////////////////////////////////////////////////////
// @FUNCT@: @GX_BSP_Tree::intersect@
// 
// @DESCR@: intersect the ray w/ the BSP-Hierarchy.
// 
// @NOTES@: This routine is based on the "linear time voxel traversing"
// algorithm from Jim Arvo. Have a look on RTNews2 to see how it works!
// 
//////////////////////////////////////////////////////////////////////////////
bool GX_BSP_Tree::intersectShadow(GRay3 &rRay, 
				  GInterSectionList &rInter,
				  GRayTraceable *&rLastOpaque,
				  GXAccelCell *&rLastVoxel)
{ 
    // define a static stack for traversing 
    static GStack<BSP_StkNode>  sTraversStack;   
    
    GX_BSP_Node     *tCur, *tNear, *tFar;
    GFloat          tDist, tTmin, tTmax;
    
    bool tHIT = false;
    rLastOpaque = 0;
    rLastVoxel = 0;
    
    if (!mRoot->findMinMax(rRay, tTmin, tTmax))
	return(false);
   
    // update tmin and tmax of the ray 
    rRay.minmax(tTmin, tTmax);
   
    //  begin w/ the root node... 
    tCur = mRoot;
    sTraversStack.push(BSP_StkNode(0, 0, 0));

    // Iterate, while the stack not empty...
    while ( tCur )  {
      
	// Is the current node a leaf ? 
	while ( tCur->mChild[0] )  {
	  
	    // calculate the distance from the ray-origin to the subdiv.
	    // plane. This rountine also determines the 2 FAR and NEAR
	    // children. 
	    tCur->classify(rRay, tDist, tNear, tFar);
	 
	    // Only intersection w/ NEAR child possible! 
	    if ((tDist > tTmax) || (tDist <= 0) ) {
		tCur = tNear;
	    } 
	    // the same w/ FAR! 
	    else if (tDist < tTmin)   {
		tCur = tFar;
	    } 
	    // the ray can intersect w/ both children 
	    else  {
	    
		// PUSH the FAR child into the stack 
		sTraversStack.push(BSP_StkNode(tFar, tDist, tTmax));
	    
		// The next node to work with is the NEAR child ! 
		tCur   = tNear;
		tTmax  = tDist;
	    }
	}
	
	
	gstat(GXStatistics::BSP_NumShadowVoxels++);
	gstat(GXStatistics::BSP_NumObjsInShadowVoxels += tCur->mNumOfCandidates);
	
	// @TEST@
	// intersect w/ all children of the current node. 
	if( !tCur->checkID(rRay) && tCur->intersectShadow(rRay, rInter)) {
	    rLastOpaque = &(rInter.getFirst().getPrimitive());
	    rLastVoxel = tCur;
	    break; 
	} 
	
	// POP the last FAR child 
	BSP_StkNode   stkNode = sTraversStack.pop();
	tCur   = stkNode.mNode;
	tTmin  = stkNode.mMin;
	tTmax  = stkNode.mMax;
    }
    
    sTraversStack.reset();
    return tHIT;
}
 



//////////////////////////////////////////////////////////////////////////////
// @FUNCT@: @GX_BSP_Tree::intersect@
// 
// @DESCR@: intersect the ray w/ the BSP-Hierarchy.
// 
// @NOTES@: This routine is based on the "linear time voxel traversing"
// algorithm from Jim Arvo. Have a look on RTNews2 to see how it works!
// 
//////////////////////////////////////////////////////////////////////////////
bool GX_BSP_Tree::intersectShading(GRay3&              rRay, 
				   GInterSectionList&  rInter)
{ 
    // define a static stack for traversing 
    static GStack<BSP_StkNode>  sTraversStack;   

    static GAllInterSectionList sInterHeap;

    GX_BSP_Node     *tCur, *tNear, *tFar;
    GFloat          tDist, tTmin, tTmax;
    
    bool tHIT = false;
   
    if (!mRoot->findMinMax(rRay, tTmin, tTmax))
	return(false);
   
    // update tmin and tmax of the ray 
    rRay.minmax(tTmin, tTmax);
   
    //  begin w/ the root node... 
    tCur = mRoot;
    sTraversStack.push(BSP_StkNode(0, 0, 0));


    int tAxis = rRay.getVector().getDominantAxis();
    GFloat (*_getMinDist)(const GRay3& pRay, const GBounds3 &pB);

    switch (tAxis) {
      case POS_X: 
	_getMinDist = _getMinDistPOS_X;
	break;
      
      case NEG_X: 
	_getMinDist = _getMinDistNEG_X;
	break;
      
      case POS_Y: 
	_getMinDist = _getMinDistPOS_Y;
	break;
      
      case NEG_Y: 
	_getMinDist = _getMinDistNEG_Y;
	break;
      
      case POS_Z: 
	_getMinDist = _getMinDistPOS_Z;
	break;
      
      default:   
	_getMinDist = _getMinDistNEG_Z;
	break;
    }


    // Iterate, util the stack not empty...
    while ( tCur )  {
      
	// Is the current node a leaf ? 
	while ( tCur->mChild[0] )  {
	  
	    // calculate the distance from the ray-origin to the subdiv.
	    // plane. This rountine also determines the 2 FAR and NEAR
	    // children. 
	    tCur->classify(rRay, tDist, tNear, tFar);
	 
	    // Only intersection w/ NEAR child possible! 
	    if ((tDist > tTmax) || (tDist <= 0) ) {
		tCur = tNear;
	    } 
	    // the same w/ FAR! 
	    else if (tDist < tTmin)   {
		tCur = tFar;
	    } 
	    // the ray can intersect w/ both children 
	    else  {
	    
		// PUSH the FAR child into the stack 
		sTraversStack.push(BSP_StkNode(tFar, tDist, tTmax));
	    
		// The next node to work with is the NEAR child ! 
		tCur   = tNear;
		tTmax  = tDist;
	    }
	}
	
	gstat(GXStatistics::BSP_NumShadingVoxels++);
	gstat(GXStatistics::BSP_NumObjsInShadingVoxels += tCur->mNumOfCandidates);
	
	
	if(GXOptions::ABVHFlag) 
	    // intersect w/ all children of the current node. 
	    tHIT = tCur->intersectShading(rRay, sInterHeap);
	
	else {
	
	    // check the root-bbox of this voxel.
	    GXObject *tObj = tCur->mListChildren[0];
      	    
	    if (!tObj->isBound()) {
		tHIT = tObj->intersect(rRay, sInterHeap);
	    }
	    else {
				
		GFloat tTmin, tTmax;
		
		if(((GXBBox *)tObj)->findMinMax(rRay, tTmin, tTmax)) {
		    
		    GFloat tMinT = MAXFLT;
		    for(GXObjectItr _Itr(*tCur->mCandidatesLists[tAxis]); 
			_Itr.more(); 
			_Itr.next()) {

			GXObject *tPrim = _Itr.cur();
			GBounds3 tB = tPrim->getWorldBounds();
			
			// intervall check ...
			if(tHIT && tMinT < _getMinDist(rRay, tB))
			    break;
			
			if(tPrim->isBounded()) {
			    gstat(GXStatistics::BBox.mTests++);
	    
			    // miss the bbox, continuing w/ the next prim.
			    if(!tB.findMinMax(rRay, tTmin, tTmax))
				continue;
	    
			    gstat(GXStatistics::BBox.mHits++);
			}
			
			
		
			if(!tPrim->checkMailBox(tHIT, rRay, sInterHeap) &&
			   tPrim->intersect(rRay, sInterHeap)) 
			{
			    tHIT = true;
			    tMinT = sInterHeap.getFirst().getT();
			}
		    }
		}
	    }
	}
	
	if(tHIT) {
	    tHIT = false;
	    
	    while(sInterHeap.count()) {
		GInterSection tInter = sInterHeap.remove();
		// checking for validity. If TRUE, We are done.
		if(tTmin <= tInter.getT() && tInter.getT() <= tTmax) {
		    rInter.add(tInter);
		    tHIT = true;
		    sInterHeap.reset();
		    break;
		}
		// opps..this intersection was outside the actual voxel.
		// So send it back to the primitive and avoid multiply 
		// intersection test.
		tInter.getPrimitive().sendMail(tInter);
	    }
	}
	
	if(tHIT) break; 



	// POP the last FAR child 
	BSP_StkNode   stkNode = sTraversStack.pop();
	tCur   = stkNode.mNode;
	tTmin  = stkNode.mMin;
	tTmax  = stkNode.mMax;
    }
    
    sTraversStack.reset();
    return tHIT;
}




////////////////////////////////////////////////////////////////////////////// 
// Add the new primitives to the children list and
// expand the root-node to bound them all !
//////////////////////////////////////////////////////////////////////////////
GX_BSP_Tree::GX_BSP_Tree(const GXPrimitiveList& pScene) 
{
    mRoot = new GX_BSP_Node;
    for(GXPrimitiveItr tItr(pScene); tItr.more(); tItr.next())
	mRoot->add(tItr.cur());
}
 



//////////////////////////////////////////////////////////////////////////////
// @GX_BSP_Tree::buildHierarchy@
// Build the BSP-hierarchy and contruct ABVHs for its nodes 
//
//////////////////////////////////////////////////////////////////////////////
void  GX_BSP_Tree::buildHierarchy(void) {
//     
//     if(!GXOptions::MaxObjInNode) GXOptions::MaxObjInNode = 1;
//     
//     GXOptions::MaxPartitionLevel = (int) ceil(log((GFloat) (mRoot->getNumOfChildren())/
// 			       (GFloat)  GXOptions::MaxObjInNode)
// 			   / log(2.0)) + 1;
//  

    cout << "Building BSP-Tree w/ max. depth= " 
	 << GXOptions::BSP_MaxPartitionLevel
	 << ", max. prims in a node= " \
	 << GXOptions::BSP_MaxObjInNode << "\n";
    
    mRoot->mAxis = G_D_X;

#ifndef OPTI_SUBDIVISION    
    
    // subdivision , qsort-version
    __subdivide(mRoot, 0, GXOptions::BSP_MaxPartitionLevel, GXOptions::BSP_MaxObjInNode);

#else
    // subdivision , version 2 (optimized)
    SortedSegmentLists tSortedLists;
    GXObjectList *tListChildren = mRoot->getListOfChildren();
    int tNum = tListChildren->count();
    
    Segment*       tMins[3];
    Segment*       tMaxs[3];
   
    for(int i = 0; i < 3; i++) {
	tMins[i] = new Segment[tNum];
	tMaxs[i] = new Segment[tNum];
	
	register int j=0;
	for(GXObjectItr _Itr(*tListChildren); _Itr.more(); _Itr.next(), j++) {
	    GBounds3  tBox(_Itr.cur()->getWorldBounds());
	    
	    tMins[i][j].min = tBox.min()[i];
	    tMins[i][j].max = tBox.max()[i];
	    tMins[i][j].obj = _Itr.cur();
	    tMaxs[i][j].min = tBox.min()[i];
	    tMaxs[i][j].max = tBox.max()[i];
	    tMaxs[i][j].obj = _Itr.cur();
	}
	
	// Sorting.. 
	qsort( tMins[i], tNum, sizeof(Segment), __compareMinSeg) ;
	qsort( tMaxs[i], tNum, sizeof(Segment), __compareMaxSeg) ;
	
	for(j = 0; j < tNum; j++) {
	    tSortedLists.Mins[i].append(&(tMins[i][j]));
	    tSortedLists.Maxs[i].append(&(tMaxs[i][j]));
	}
    }    
    
    __subdivide(mRoot, &tSortedLists, 0, 
		GXOptions::BSP_MaxPartitionLevel, 
		GXOptions::BSP_MaxObjInNode);
    
    
    for( i = 0; i < 3; i++) {
	delete [] tMins[i];
	delete [] tMaxs[i];
    }

#endif // OPTI_SUBDIVISION    
    
    cerr << "\nBuilding BVH-Hierarchies for each BSP-Node...\n"; 
    
    __buildBVH(mRoot);
 
}


//////////////////////////////////////////////////////////////////////////////
// @GX_BSP_Tree::~GX_BSP_Tree@
// Dtor
//////////////////////////////////////////////////////////////////////////////
GX_BSP_Tree::~GX_BSP_Tree(void)   {
    // Destroy our BSP-Tree, incl. BVHs in each BSP-node!
    gkillPointer(mRoot);
}



