//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <stdlib.h>
#include "render/graytrac.h"
#include "util/gstack.h"

#include "gx/gx_defs.h"
#include "gx/gx_bbox.h"
#include "gx/gx_hyper.h"



//////////////////////////////////////////////////////////////////////////////
// @GXHyperCube::GXHyperCube@
//////////////////////////////////////////////////////////////////////////////
GXHyperCube::GXHyperCube(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList) 
:GXBeam(pFace, pExtents, pCandidatesList)
{
    // 2^5 = 32 chidlren of the 5D hyper-cube
    
    for(register int i=0; i< 32; i++)
	mChild[i] = 0;
    
    mInherited = true;
}
  
GXHyperCube::~GXHyperCube(void) {
    for(register int i = 0; i< 32; i++)
	gkillPointer(mChild[i]);
    gkillArray(mChild);
}



//////////////////////////////////////////////////////////////////////////////
// @GXHyperCube::subdivide@
//////////////////////////////////////////////////////////////////////////////
void GXHyperCube::subdivide(void) {
    
    //// partition this hyper-cube along each axis into 32 children!
    for(register int i=0; i< 32; i++) {
	
	
	GFloat tExts[10];
	
    	//// X ////
	if ( i & 1) {
	    tExts[G_LO_X] = 0.5 * (mExtents[G_LO_X] + mExtents[G_HI_X]);
	    tExts[G_HI_X] = mExtents[G_HI_X];
	}
	else {
	    tExts[G_HI_X] = 0.5 * (mExtents[G_LO_X] + mExtents[G_HI_X]);
	    tExts[G_LO_X] = mExtents[G_LO_X];
	}
	
	//// Y ////
	if ( i & 2) {
	    tExts[G_LO_Y] = 0.5 * (mExtents[G_LO_Y] + mExtents[G_HI_Y]);
	    tExts[G_HI_Y] = mExtents[G_HI_Y];
	}
	else {
	    tExts[G_HI_Y] = 0.5 * (mExtents[G_LO_Y] + mExtents[G_HI_Y]);
	    tExts[G_LO_Y] = mExtents[G_LO_Y];
	}
	
	//// Z
	if ( i & 4) {
	    tExts[G_LO_Z] = 0.5 *(mExtents[G_LO_Z] + mExtents[G_HI_Z]);
	    tExts[G_HI_Z] = mExtents[G_HI_Z];
	}
	else {
	    tExts[G_HI_Z] = 0.5 *(mExtents[G_LO_Z] + mExtents[G_HI_Z]);
	    tExts[G_LO_Z] = mExtents[G_LO_Z];
	}
	
	//// U
	if ( i & 8) {
	    tExts[G_LO_U] = 0.5 *(mExtents[G_LO_U] + mExtents[G_HI_U]);
	    tExts[G_HI_U] = mExtents[G_HI_U];
	}
	else {
	    tExts[G_HI_U] = 0.5 *(mExtents[G_LO_U] + mExtents[G_HI_U]);
	    tExts[G_LO_U] = mExtents[G_LO_U];
	}
	
	//// V
	if ( i & 16) {
	    tExts[G_LO_V] = 0.5 *(mExtents[G_LO_V] + mExtents[G_HI_V]);
	    tExts[G_HI_V] = mExtents[G_HI_V];
	}
	else {
	    tExts[G_HI_V] = 0.5 *(mExtents[G_LO_V] + mExtents[G_HI_V]);
	    tExts[G_LO_V] = mExtents[G_LO_V];
	}

	// generate the child-hypercube
	mChild[i] = new GXHyperCube(mFace, tExts, mCandidatesList);
	mChild[i]->mLevel = mLevel + 1;
	mChild[i]->mInherited = true;
    }
}


//////////////////////////////////////////////////////////////////////////////
// @GXHyperCube::findLeaf@
//////////////////////////////////////////////////////////////////////////////
GXHyperCube* GXHyperCube::findLeaf(GFloat *p5D) {
    static GStack<GXHyperCube*> sStk;
    
    //// i dont own any child ! 
    if(!mChild[0]) 
	return (this);
    
    sStk.push(this);
    for(;;) {
	GXHyperCube *tHCube = sStk.pop();
	
	if(!tHCube->mChild[0]) {
	    sStk.reset();
	    return (tHCube);
	}
	
	register int i = 0;
	GFloat *tExtents = tHCube->mExtents;
	
	//// X
	if ( p5D[G_D_X] > 0.5 * (tExtents[G_LO_X] + tExtents[G_HI_X]))
	    i |= 1;

	//// Y
	if ( p5D[G_D_Y] > 0.5 * (tExtents[G_LO_Y] + tExtents[G_HI_Y]))
	    i |= 2;
    
	//// Z
	if ( p5D[G_D_Z] > 0.5 * (tExtents[G_LO_Z] + tExtents[G_HI_Z]))
	    i |= 4;
    
	//// U
	if ( p5D[G_D_U] > 0.5 * (tExtents[G_LO_U] + tExtents[G_HI_U]))
	    i |= 8;
	
	//// V
	if ( p5D[G_D_V] >  0.5 * (tExtents[G_LO_V] + tExtents[G_HI_V]))
	    i |= 16;
	
	sStk.push(tHCube->mChild[i]);
    }
    
}










//////////////////////////////////////////////////////////////////////////////
// @GXHyperOctree@ ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////



class SortNode {
  public:
    GFloat mExt;
    GXObject *mObj;
};

int __compareMin(const void *pA, const void *pB )   {
    SortNode *tA  = (SortNode *) pA;
    SortNode *tB  = (SortNode *) pB;
    if ( tA->mExt == tB->mExt ) 
	return( 0 ) ;
    else   
	return( tA->mExt < tB->mExt ? -1 : 1 ) ;
}

int __compareMax(const void *pA, const void *pB )   {
    SortNode *tA  = (SortNode *) pA;
    SortNode *tB  = (SortNode *) pB;
    if ( tA->mExt == tB->mExt ) 
	return( 0 ) ;
    else   
	return( tA->mExt > tB->mExt ? -1 : 1 ) ;
}


void sortObjects(const GXPrimitiveList& pScene, GXObjectList *rObjLists[6]) {
    SortNode *tSortArray[6];
    int tNum = pScene.count();

    //// sort the whole candidates w/ respect to each dominant axis
    for(register int i=0; i< 6; i++) 
	tSortArray[i] = new SortNode[tNum];
    
    
    register int j = 0;
    for(GXPrimitiveItr _Itr(pScene); _Itr.more(); _Itr.next(), j++) {
	GBounds3  tB(_Itr.cur()->getWorldBounds());
	
	tSortArray[POS_X][j].mExt = tB.min().getX();
	tSortArray[POS_X][j].mObj = _Itr.cur();
    	
	tSortArray[NEG_X][j].mExt = tB.max().getX();
	tSortArray[NEG_X][j].mObj = _Itr.cur();

	tSortArray[POS_Y][j].mExt = tB.min().getY();
	tSortArray[POS_Y][j].mObj = _Itr.cur();
    	
	tSortArray[NEG_Y][j].mExt = tB.max().getY();
	tSortArray[NEG_Y][j].mObj = _Itr.cur();

	tSortArray[POS_Z][j].mExt = tB.min().getZ();
	tSortArray[POS_Z][j].mObj = _Itr.cur();
    	
	tSortArray[NEG_Z][j].mExt = tB.max().getZ();
	tSortArray[NEG_Z][j].mObj = _Itr.cur();
    }

    qsort( tSortArray[POS_X], tNum, sizeof(SortNode), __compareMin) ;
    qsort( tSortArray[NEG_X], tNum, sizeof(SortNode), __compareMax) ;
    qsort( tSortArray[POS_Y], tNum, sizeof(SortNode), __compareMin) ;
    qsort( tSortArray[NEG_Y], tNum, sizeof(SortNode), __compareMax) ;
    qsort( tSortArray[POS_Z], tNum, sizeof(SortNode), __compareMin) ;
    qsort( tSortArray[NEG_Z], tNum, sizeof(SortNode), __compareMax) ;
	
      
    for(i=0; i< 6; i++) {
	//// initialize the lists of candidates.
	rObjLists[i] = new GXObjectList();
	for(j=0; j< tNum; j++) 
	    rObjLists[i]->append(tSortArray[i][j].mObj);

    }
    
       
    //// free up temp. arrays 
    for(i=0; i< 6; i++) 
	gkillArray(tSortArray[i]);
    
}



//////////////////////////////////////////////////////////////////////////////
// @GXHyperOctree@GXHyperOctree@ 
//////////////////////////////////////////////////////////////////////////////
GXHyperOctree::GXHyperOctree(GXObjectList *pObjLists[6], const GVector3 & pEye) {
    if(!GXOptions::HOT_MaxObjInNode) GXOptions::HOT_MaxObjInNode = 1;
//     // log(32) ?
//     GXOptions::MaxPartitionLevel = 
// 	(int) ceil(log((GFloat) (tNum)/ (GFloat) GXOptions::MaxObjInNode ) / log(8)) + 1;
//  
    cout << "Building 5D Hyper-Octree w/ max. depth= " 
	 <<  GXOptions::HOT_MaxPartitionLevel
	 << ", max. prims in a hyper-cube= " 
	 << GXOptions::HOT_MaxObjInNode << "\n";
    
   
    
    
    //// initialize the first hyper-cubes

    GFloat tExts[10];
    tExts[G_LO_X] = (*(pObjLists[G_LO_X]))[0]->getWorldBounds().min().getX();
    tExts[G_HI_X] = (*(pObjLists[G_HI_X]))[0]->getWorldBounds().max().getX();
    tExts[G_LO_Y] = (*(pObjLists[G_LO_Y]))[0]->getWorldBounds().min().getY();
    tExts[G_HI_Y] = (*(pObjLists[G_HI_Y]))[0]->getWorldBounds().max().getY();
    tExts[G_LO_Z] = (*(pObjLists[G_LO_Z]))[0]->getWorldBounds().min().getZ();
    tExts[G_HI_Z] = (*(pObjLists[G_HI_Z]))[0]->getWorldBounds().max().getZ();

    tExts[G_LO_U] = -1;
    tExts[G_HI_U] = 1;
    tExts[G_LO_V] = -1;
    tExts[G_HI_V] = 1;

    //// initialize the lists of candidates.
    for(register int i=0; i< 6; i++) {
	mRootCubes[i] = new GXHyperCube(i, tExts, pObjLists[i]);
	mRootCubes[i]->mLevel = 0;
	mRootCubes[i]->mInherited = false;
    }
   
         
    //// generate an enviroment-box (for spatial subdivision)
    mEnvBox = GBounds3(GPoint3(tExts[G_LO_X], tExts[G_LO_Y], tExts[G_LO_Z]), 
		       GPoint3(tExts[G_HI_X], tExts[G_HI_Y], tExts[G_HI_Z]));
   
    // is the eye inside the enviroment box ?
    mEyeInEnvBox = (mEnvBox.isOverlap(pEye)) ? true : false;
    

    //// cache for last seen cube.
    mCache = new GXCache();

}


//////////////////////////////////////////////////////////////////////////////
// @GXHyperOctree::~GXHyperOctree@
//
//////////////////////////////////////////////////////////////////////////////
GXHyperOctree::~GXHyperOctree() {
    for(register int i=0; i< 6; i++)
	gkillPointer(mRootCubes[i]);
    
    gkillPointer(mCache);
}

    


//////////////////////////////////////////////////////////////////////////////
// @GXHyperOctree::intersectShading@
//////////////////////////////////////////////////////////////////////////////
bool GXHyperOctree::intersectShading(GRay3& rRay, GInterSectionList& rInter) {
    
    // We dont include the enviroment-box w/ the eye. This is because including
    // the eye to it makes the hyper-octree unnecessary unbalanced. We recompute
    // the eye-ray's origin at the intersection point of the enviroment box. 
    
    if(rRay.isEyeRay() && !mEyeInEnvBox) {
	GFloat tTmin, tTmax;
	if(mEnvBox.findMinMax(rRay, tTmin, tTmax)) 
	    rRay.setPoint(rRay.extrap(tTmin));
	else
	    return (false);
    }
    
    // compute the dominant axis of this ray and its 5D-tuple
    GFloat t5D[5];
    int tAxis = rRay.compute5D(t5D);
    
    //////////////////////////////////////////////////////////////////////////
    //// finding the last-object and the last-voxel hited by this ray\'s level
    //////////////////////////////////////////////////////////////////////////
    GXCache *tCache = mCache->getNode(rRay); 
        
    // First, check the last hited object for intersection 
    if (tCache->mLastObject)  {
	
	tCache->mLastObject->updateID(rRay);

	if (tCache->mLastObject->intersect(rRay, rInter))  {
	    gstat(GXStatistics::ShadingLastObjectHits++);
	    
	    // update the maximum distance for this ray
	    rRay.setTmax(rInter.getFirst().getT());
	}
	// last object missed!
	else
	    tCache->mLastObject = 0;
    }
    // Second, check the last voxel intersected by this ray-level
    if ((tCache->mLastVoxel) && 
	(tAxis == ((GXHyperCube *)tCache->mLastVoxel)->getDominantAxis()) &&
	((GXHyperCube *)tCache->mLastVoxel)->isInside(t5D)) 
    {
	if(tCache->mLastVoxel->intersectShading(rRay, rInter)) {
	    gstat(GXStatistics::ShadingLastVoxelHits++);
	    return(true);
	}
	else
	    return(false);
    }
	

    //////////////////////////////////////////////////////////////////////
    //// Cache missed!..Traversing from the root..
    //////////////////////////////////////////////////////////////////////

    // get the root hyper-cube associating w/ tAxis
    GXHyperCube *tRoot = mRootCubes[tAxis];
    
    // get the leaf hyper-cube containing this ray (5D-tuple)
    // @CACHE@
    GXHyperCube *tLeaf = tRoot->findLeaf(t5D);
    
    if(tLeaf->mInherited) {
	//@DEBUG@
	//cerr << "Inherited/Shading/";

	tLeaf->reclassifyCandidatesList();
    }
    
    // subdividing, while the candidates list is too large and the size of 
    // this cube is not too small
    while(tLeaf->getNumOfChildren() >  GXOptions::HOT_MaxObjInNode
	  && tLeaf->mLevel < GXOptions::HOT_MaxPartitionLevel ) 
    {
	// subdivide the cube along each of the 5 dimensions
	tLeaf->subdivide();
	  
	// determine the child-cube, which contains the ray.
	tLeaf = tLeaf->findLeaf(t5D);
	  

	//@DEBUG@
	//cerr << "Shading/";

	// recompute the list of candidates for this child cube
	tLeaf->reclassifyCandidatesList();
    }

#ifdef GSTAT
    GXStatistics::HOT_NumShadingVoxels++;
    GXStatistics::HOT_NumObjsInShadingVoxels += tLeaf->getNumOfChildren();
#endif
    

    // update the voxel-cache
    tCache->mLastVoxel = tLeaf;

    // intersect w/ all candidates of the current cube. 
    if( tLeaf->intersectShading(rRay, rInter)) {
	tCache->mLastObject = &(rInter.getFirst().getPrimitive());
	return(true);
    }
    else {
	tCache->mLastObject = 0;
	return(false);
    }
    
}


//////////////////////////////////////////////////////////////////////////////
// @GXHyperOctree::intersectShadow@
//////////////////////////////////////////////////////////////////////////////
bool GXHyperOctree::intersectShadow(GRay3&              rRay, 
				    GInterSectionList&  rInter,
				    GRayTraceable*&     rLastOpaque,
				    GXAccelCell*&     rLastVoxel ) {
    
    
    // compute the dominant axis of this ray and its 5D-tuple
    GFloat t5D[5];
    int tAxis = rRay.compute5D(t5D);
    
    
    
    // get the root hyper-cube associating w/ tAxis
    GXHyperCube *tRoot = mRootCubes[tAxis];
    
    // get the leaf hyper-cube containing this ray (5D-tuple)
    GXHyperCube *tLeaf = tRoot->findLeaf(t5D);
    
    if(tLeaf->mInherited) {
	
	//@DEBUG@
	//cerr << "Inherited/Shadow/";

	tLeaf->reclassifyCandidatesList();
    }
    
    // subdividing, while the candidates list is too large and the size of 
    // this cube is not too small
    while(tLeaf->getNumOfChildren() > GXOptions::HOT_MaxObjInNode && 
	  tLeaf->mLevel < GXOptions::HOT_MaxPartitionLevel ) {
	// subdivide the cube along each of the 5 dimensions
	tLeaf->subdivide();
	  
	// determine the child-cube, which contains the ray.
	tLeaf = tLeaf->findLeaf(t5D);
	  

	//@DEBUG@
	//cerr << "Shadow/";
	
	// recompute the list of candidates for this child cube
	tLeaf->reclassifyCandidatesList();
	  
    }

#ifdef GSTAT
    GXStatistics::HOT_NumShadowVoxels++;
    GXStatistics::HOT_NumObjsInShadowVoxels += tLeaf->getNumOfChildren();
#endif
    

    rLastVoxel = tLeaf;
    
    // intersect w/ all candidates of the current cube. 
    if( tLeaf->intersectShadow(rRay, rInter)) {
	rLastOpaque = &(rInter.getFirst().getPrimitive());
	return (true);
    }
    
    return (false);
    
}























