//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////

#include <stdlib.h>
#include "render/graytrac.h"
#include "util/gstack.h"

#include "gx/gx_defs.h"
#include "gx/gx_bbox.h"

#include "gx/gx_vbuf.h"



//////////////////////////////////////////////////////////////////////////////
// @GXViewVolume::GXViewVolume@
//////////////////////////////////////////////////////////////////////////////
GXViewVolume::GXViewVolume(int pFace, GFloat *pExtents, GXObjectList *pCandidatesList) 
:GXBeam(pFace, pExtents, pCandidatesList)
{
    // 2^2 = 4 chidlren of the 5D hyper-cube
    
    for(register int i=0; i< 4; i++)
	mChild[i] = 0;
    
    mInherited = true;
}
  
GXViewVolume::~GXViewVolume(void) {
    for(register int i = 0; i< 4; i++)
	gkillPointer(mChild[i]);
    gkillArray(mChild);
}



//////////////////////////////////////////////////////////////////////////////
// @GXViewVolume::subdivide@
//////////////////////////////////////////////////////////////////////////////
void GXViewVolume::subdivide(void) {
    
    //// partition this hyper-cube along each axis into 32 children!
    for(register int i=0; i< 4; i++) {
	
	
	GFloat tExts[10];
	
	tExts[G_HI_X] = mExtents[G_HI_X];
	tExts[G_LO_X] = mExtents[G_LO_X];
	tExts[G_HI_Y] = mExtents[G_HI_Y];
	tExts[G_LO_Y] = mExtents[G_LO_Y];
	tExts[G_HI_Z] = mExtents[G_HI_Z];
	tExts[G_LO_Z] = mExtents[G_LO_Z];
	
	
	//// U
	if ( i & 1) {
	    tExts[G_LO_U] = 0.5 *(mExtents[G_LO_U] + mExtents[G_HI_U]);
	    tExts[G_HI_U] = mExtents[G_HI_U];
	}
	else {
	    tExts[G_HI_U] = 0.5 *(mExtents[G_LO_U] + mExtents[G_HI_U]);
	    tExts[G_LO_U] = mExtents[G_LO_U];
	}
	
	//// V
	if ( i & 2) {
	    tExts[G_LO_V] = 0.5 *(mExtents[G_LO_V] + mExtents[G_HI_V]);
	    tExts[G_HI_V] = mExtents[G_HI_V];
	}
	else {
	    tExts[G_HI_V] = 0.5 *(mExtents[G_LO_V] + mExtents[G_HI_V]);
	    tExts[G_LO_V] = mExtents[G_LO_V];
	}

	// generate the child-hypercube
	mChild[i] = new GXViewVolume(mFace, tExts, mCandidatesList);
	mChild[i]->mLevel = mLevel + 1;
	mChild[i]->mInherited = true;
    }
}


//////////////////////////////////////////////////////////////////////////////
// @GXViewVolume::findLeaf@
//////////////////////////////////////////////////////////////////////////////
GXViewVolume* GXViewVolume::findLeaf(GFloat *p5D) {
    static GStack<GXViewVolume*> sStk;
    
    //// i dont own any child ! 
    if(!mChild[0]) 
	return (this);
    
    sStk.push(this);
    for(;;) {
	GXViewVolume *tHCube = sStk.pop();
	
	if(!tHCube->mChild[0]) {
	    sStk.reset();
	    return (tHCube);
	}
	
	register int i = 0;
	GFloat *tExtents = tHCube->mExtents;
	
	//// U
	if ( p5D[G_D_U] > 0.5 * (tExtents[G_LO_U] + tExtents[G_HI_U]))
	    i |= 1;
	
	//// V
	if ( p5D[G_D_V] >  0.5 * (tExtents[G_LO_V] + tExtents[G_HI_V]))
	    i |= 2;
	
	sStk.push(tHCube->mChild[i]);
    }
    
}
 
GXViewBuffer::GXViewBuffer(GXObjectList *pObjLists[6], const GBounds3& pViewer) {
   
    //// initialize the first hyper-cubes
    GFloat tExts[10];
    tExts[G_LO_X] = pViewer.min().getX();
    tExts[G_HI_X] = pViewer.max().getX();
    tExts[G_LO_Y] = pViewer.min().getY();
    tExts[G_HI_Y] = pViewer.max().getY();
    tExts[G_LO_Z] = pViewer.min().getZ();
    tExts[G_HI_Z] = pViewer.max().getZ();
    
    tExts[G_LO_U] = -1;
    tExts[G_HI_U] = 1;
    tExts[G_LO_V] = -1;
    tExts[G_HI_V] = 1;

    for(register int i=0; i<6; i++) {
	mRootCubes[i] = new GXViewVolume(i, tExts, pObjLists[i] );
	mRootCubes[i]->mLevel = 0;
	mRootCubes[i]->mInherited = false;
    }
    

}

GXViewBuffer::~GXViewBuffer(void) {
    for(register int i=0; i< 6; i++)
	gkillPointer(mRootCubes[i]);
}

GXViewVolume* GXViewBuffer::classify(const GRay3 &pRay) {
     // compute the dominant axis of this ray and its 5D-tuple
    GFloat t5D[5];
    int tAxis = pRay.compute5D(t5D);

    // get the root hyper-cube associating w/ tAxis
    GXViewVolume *tRoot = mRootCubes[tAxis];
    
    // get the leaf hyper-cube containing this ray (5D-tuple)
    GXViewVolume *tLeaf = tRoot->findLeaf(t5D);
    
    if(tLeaf->mInherited) {
	tLeaf->reclassifyCandidatesList();
    } 

    // subdividing, while the candidates list is too large and the size of 
    // this cube is not too small
    while(tLeaf->getNumOfChildren() > GXOptions::HOT_MaxObjInNode && 
	  tLeaf->mLevel < GXOptions::HOT_MaxPartitionLevel ) 
    {
	tLeaf->subdivide();
	tLeaf = tLeaf->findLeaf(t5D);
	tLeaf->reclassifyCandidatesList();
    }

    
    return(tLeaf);
}
    
bool GXViewBuffer::intersectShadow(GRay3 &rRay, 
				   GInterSectionList &rInter,
				   GRayTraceable *&rLastOpaque,
				   GXAccelCell *&rLastVoxel) 
{
    
    GXViewVolume *tLeaf = classify(rRay);
    
#ifdef GSTAT
    GXStatistics::HOT_NumShadowVoxels++;
    GXStatistics::HOT_NumObjsInShadowVoxels += tLeaf->getNumOfChildren();
#endif
    

    rLastVoxel = tLeaf;
    
    // intersect w/ all candidates of the current cube. 
    if( tLeaf->intersectShadow(rRay, rInter)) {
	rLastOpaque = &(rInter.getFirst().getPrimitive());
	return (true);
    }
    
    return (false);
}



bool GXViewBuffer::intersectShading(GRay3& rRay, GInterSectionList& rInter) {
    
    // compute the dominant axis of this ray and its 5D-tuple
    GFloat t5D[5];
    int tAxis = rRay.compute5D(t5D);
    static GXViewVolume *sLastVoxel = 0;
    static GRayTraceable *sLastObject = 0;
    

     // First, check the last hited object for intersection 
    if (sLastObject)  {
	
	sLastObject->updateID(rRay);

	if (sLastObject->intersect(rRay, rInter))  {
	    gstat(GXStatistics::ShadingLastObjectHits++);
	    
	    // update the max distance for this ray
	    rRay.setTmax(rInter.getFirst().getT());
	}
	// last object missed!
	else
	    sLastObject = 0;
    }

    if ( sLastVoxel && tAxis == sLastVoxel->getDominantAxis() &&
	 sLastVoxel->isInside(t5D)) 
    {
	if (sLastVoxel->intersectShading(rRay, rInter)) {
	    gstat(GXStatistics::ShadingLastVoxelHits++);
	    return(true);
	}
	else
	    return(false);
	
    }
	



    GXViewVolume *tLeaf = classify(rRay);
    
#ifdef GSTAT
    GXStatistics::HOT_NumShadingVoxels++;
    GXStatistics::HOT_NumObjsInShadingVoxels += tLeaf->getNumOfChildren();
#endif
    

    // update the voxel-cache
    sLastVoxel = tLeaf;

      // intersect w/ all candidates of the current cube. 
    if( tLeaf->intersectShading(rRay, rInter)) {
	sLastObject = &(rInter.getFirst().getPrimitive());
	return(true);
    }
    else {
	sLastObject = 0;
	return(false);
    }
}


