//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "model/gmodel.h"

const GModeling gNoModeling;

void GModeling::setFatherModeling(const GModeling& pModel) {
    individualize();
    pFatherModeling = &pModel;
    concat();
}

const GMatrix4& GModeling::getGlobalMatrix(void) const {
    assert(pFatherModeling);
    return(pFatherModeling->getWorldMatrix());
}

const GMatrix4& GModeling::getInvGlobalMatrix(void) const {
    assert(pFatherModeling);
    return(pFatherModeling->getInvWorldMatrix());
}

void GModeling::individualize(void) {
    if (!local) local = new GMatrix4Pair;
    if (!world) world = new GMatrix4Pair;
}

void GModeling::concat(void) {
    assert((int)local * (int)world);
    world->mat = pFatherModeling->getWorldMatrix() * local->mat;
    world->imat = local->imat * pFatherModeling->getInvWorldMatrix();
}

void GModeling::reset(void) {
    if (local) delete local;
    if (world) delete world;
    local = world = 0;
}

void GModeling::rotate(const GVector3& v) {
    individualize();
    local->mat = local->mat.rotate(v);
    local->imat = local->imat.invRotate(v);
    concat();
}

void GModeling::scale(const GVector3& v) {
    individualize();
    local->mat = local->mat.scale(v);
    local->imat = local->imat.invScale(v);
    concat();
}

void GModeling::translate(const GVector3& v) {
    individualize();
    local->mat = local->mat.translate(v);
    local->imat = local->imat.invTranslate(v);
    concat();
}

void GModeling::setTranslation(const GPoint3& v) {
    individualize();
    local->mat.set(0, 3, v.getX());
    local->mat.set(1, 3, v.getY());
    local->mat.set(2, 3, v.getZ());
    local->imat = local->mat.inverse();
    concat();
}

void GModeling::setMatrix(const GMatrix4& pMatrix) {
    individualize();
    local->mat = pMatrix;
    local->imat = pMatrix.inverse();
    concat();
}

GVector3 GModeling::mc2wc(const GVector3& v) const { 
    return(getWorldMatrix() * v); 
}

GVector3 GModeling::wc2mc(const GVector3& v) const { 
    return(getInvWorldMatrix() * v); 
}

const GMatrix4& GModeling::getWorldMatrix(void) const {
    if (world) return(world->mat);
    if (this == &gNoModeling) return(gUnityMatrix4);
    return(pFatherModeling->getWorldMatrix());
}

const GMatrix4& GModeling::getInvWorldMatrix(void) const {
    if (world) return(world->imat);
    if (this == &gNoModeling) return(gUnityMatrix4);
    return(pFatherModeling->getInvWorldMatrix());
}

const GModelingAttribute GModelingAttribute::attribute;

void GModelingAttribute::setModeling(const GModeling& pModel) {
    if (!getExist()) {
	mModeling = new GModeling();
	setExist();
    }
    *mModeling = pModel;
    modelingChanged();
}

const GModeling& GModelingAttribute::getModeling(void) const {
    if (getExist()) return(*mModeling);
    else return(((const GModelingAttribute&)getReference()).getModeling());
}

GModeling& GModelingAttribute::getPrivateModeling(void) {
    if (!getExist()) {
	mModeling = new GModeling();
	setExist();
    }
    return(*mModeling);
}

void GModelingAttribute::reference(const GModelingAttribute& pAttr) {
    if (getExist()) {
	assert(mModeling);
	mModeling->setFatherModeling(pAttr.getModeling());
    }
    else setReference(pAttr);
    modelingChanged();
}

GModelingAttribute::GModelingAttribute(): mModeling(0), GAttribute(attribute) {
    if (this == &attribute) {
	mModeling = new GModeling; 
	setExist();
    }
}


