//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [window handling and interaction]
// Windows/NT + OpenGL 1.0
// This version only supports direct color (RGB) display drivers.

#include <windows.h>

#include <GL/glaux.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gdevice.h"
#include "input/gevent.h"
#include "util/gtcl.h"

#include <tk.h>
#include <unistd.h>

#define WM_INIT		WM_USER

Tk_Window gTkWin;

class GWinNTInputDevice: public GInputDevice {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An input device for 
    // handling Windows Events
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GWinNTInputDevice(void): GInputDevice((GInputDevice*)gNoInputDevice) {}
    bool event(GEvent* pEvent = 0);
};

bool GWinNTInputDevice::event(GEvent*) {
    if (!isOn()) return(false);
    bool ret = false;
    if (Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT)) {
	callCBs();
	ret = true;
    }
    return(ret);
}

GWinNTInputDevice* gWinNTInputDevice;

bool gInitialize(void) {
    if (!gGlobalInterp) gOutput->fatal("Global Interpreter missing.");
    
    gTkWin = Tk_MainWindow(gGlobalInterp);
    if (!gTkWin) gOutput->fatal("Toplevel window missing.");

    // create a input device:
    gWinNTInputDevice = new GWinNTInputDevice;
    gWinNTInputDevice->on();
    return(true);
}

class GPixmapDisplayData {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specific data of a Windows NT
    // coupled pixmap display.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    // Supports only RGB displays 
    //
    // CONSTANTS: 
  public:
    GPixmapDisplayData(void);

    static bool WndClassIsRegistered;
    static HGLRC context;
    static int PixmapWindowCount;	
    bool first;
    HWND hWnd;
    HDC hdc;
    PIXELFORMATDESCRIPTOR PixelFormatDescriptor;
    int width, height;
    GEvent *mEvent; // the last event
};

bool GPixmapDisplayData::WndClassIsRegistered = FALSE;
HGLRC GPixmapDisplayData::context = 0;
int GPixmapDisplayData::PixmapWindowCount;

GPixmapDisplayData::GPixmapDisplayData(void) {
    first = true;
    width = height = 0;
    mEvent = 0;
    hWnd = 0;
    hdc  = 0;
    PixelFormatDescriptor.nSize           = sizeof(PIXELFORMATDESCRIPTOR);
    PixelFormatDescriptor.nVersion        = 1;
    PixelFormatDescriptor.dwFlags         = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    PixelFormatDescriptor.iPixelType      = PFD_TYPE_RGBA;
    PixelFormatDescriptor.cColorBits      = 24;
    PixelFormatDescriptor.cRedBits        = 0;
    PixelFormatDescriptor.cRedShift       = 0;
    PixelFormatDescriptor.cGreenBits      = 0;
    PixelFormatDescriptor.cGreenShift     = 0;
    PixelFormatDescriptor.cBlueBits       = 0;
    PixelFormatDescriptor.cBlueShift      = 0;
    PixelFormatDescriptor.cAlphaBits      = 0;
    PixelFormatDescriptor.cAlphaShift     = 0;
    PixelFormatDescriptor.cAccumBits      = 0;
    PixelFormatDescriptor.cAccumRedBits   = 0;
    PixelFormatDescriptor.cAccumGreenBits = 0;
    PixelFormatDescriptor.cAccumBlueBits  = 0;
    PixelFormatDescriptor.cAccumAlphaBits = 0;
    PixelFormatDescriptor.cDepthBits      = 32;
    PixelFormatDescriptor.cStencilBits    = 0;
    PixelFormatDescriptor.cAuxBuffers     = 0;
    PixelFormatDescriptor.iLayerType      = PFD_MAIN_PLANE;
    PixelFormatDescriptor.bReserved       = 0;
    PixelFormatDescriptor.dwLayerMask     = 0;
    PixelFormatDescriptor.dwVisibleMask   = 0;
    PixelFormatDescriptor.dwDamageMask    = 0;
}

LRESULT CALLBACK GPixmapDisplayWndProc(HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam ) {

    GPixmapDisplay *tPixmap = (GPixmapDisplay *)GetWindowLong(hWnd, 0);
    if(tPixmap == NULL)
      return DefWindowProc(hWnd, wMsg, wParam, lParam);
    GPixmapDisplayData &tData = tPixmap->getData();
    
    switch (wMsg) {

      case WM_INIT: {
	  tData.hdc = GetDC(hWnd);
        int nPixelFormat = ChoosePixelFormat(tData.hdc, &tData.PixelFormatDescriptor);
	  if(nPixelFormat) {
          SetPixelFormat(tData.hdc, nPixelFormat, &tData.PixelFormatDescriptor); 
          if(!tData.context)
            tData.context = wglCreateContext(tData.hdc);
	    if(tData.context)
	      tData.PixmapWindowCount++;
	  }
	  break;
	}

      case WM_PAINT: 
        tData.first = true;
	  break;

	case WM_SIZE: 
	  tPixmap->setWidth(tData.width = LOWORD(lParam));
	  tPixmap->setHeight(tData.height = HIWORD(lParam));
	  tData.first = true;
	  break;

	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP: 
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP: 
	case WM_MBUTTONDOWN:
	case WM_MBUTTONUP: 
	{
	  GMouseButtonEvent* pev = new GMouseButtonEvent;
	  if(wParam & MK_LBUTTON) pev->left   = true;
	  if(wParam & MK_MBUTTON) pev->middle = true;
	  if(wParam & MK_RBUTTON) pev->right  = true;
	  if(wParam & MK_CONTROL) pev->ctrl   = true;
	  if(wParam & MK_SHIFT)   pev->shift  = true;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = LOWORD(lParam);  
	  pev->y = tPixmap->getHeight() - HIWORD(lParam);
	  tData.mEvent = pev;
	  break;
	}

	case WM_MOUSEMOVE: {
	  GMouseMotionEvent* pev = new GMouseMotionEvent;
	  if(wParam & MK_LBUTTON) pev->left   = true;
	  if(wParam & MK_MBUTTON) pev->middle = true;
	  if(wParam & MK_RBUTTON) pev->right  = true;
	  if(wParam & MK_CONTROL) pev->ctrl   = true;
	  if(wParam & MK_SHIFT)   pev->shift  = true;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = LOWORD(lParam);  
	  pev->y = tPixmap->getHeight() - HIWORD(lParam);
	  tData.mEvent = pev;
	  break;
      }

      case WM_DESTROY: {
	  tData.PixmapWindowCount--;
	  wglMakeCurrent(NULL, NULL);
	  if(!tData.PixmapWindowCount)
	    wglDeleteContext(tData.context);
	  ReleaseDC(hWnd, tData.hdc);
	  tData.hdc = 0;
	  tData.hWnd = 0;
	  break;
	}
    }

    if (tData.first && tData.context && !tData.mEvent) {
	tData.first = false; 
	GUpdatePixmapEvent* uev = new GUpdatePixmapEvent;
	uev->pixmap = tPixmap;
	tData.mEvent = uev;
    }
    return DefWindowProc( hWnd, wMsg, wParam, lParam );
}

extern HINSTANCE Main_hInstance;

GPixmapDisplay::GPixmapDisplay(const GString& pName, int pWidth, int pHeight): mWidth(pWidth), mHeight(pHeight) {
    mData = new GPixmapDisplayData;

    if(!mData.WndClassIsRegistered) {
      WNDCLASS wndclass;
      wndclass.lpszClassName = "GPixmapDisplayWindow";
      wndclass.hInstance     = Main_hInstance;
      wndclass.lpfnWndProc   = GPixmapDisplayWndProc;
      wndclass.hCursor       = LoadCursor( NULL, IDC_ARROW );
      wndclass.hIcon         = NULL;
      wndclass.lpszMenuName  = NULL;
      wndclass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
      wndclass.style         = CS_HREDRAW | CS_VREDRAW | 
                               CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW |
                               CS_DBLCLKS;
      wndclass.cbClsExtra    = 0;
      wndclass.cbWndExtra    = sizeof(LONG);

      RegisterClass(&wndclass );

      mData.WndClassIsRegistered = TRUE;
    }

    mData->hWnd = CreateWindow( "GPixmapDisplayWindow",
      (char*)pName,
      WS_OVERLAPPEDWINDOW |
      WS_CLIPCHILDREN |
      WS_CLIPSIBLINGS,
      CW_USEDEFAULT,
      CW_USEDEFAULT,
      mWidth,
      mHeight,
      Tk_WindowId(gTkWin),
      NULL,
      Main_hInstance,
      NULL);

    if (!mData->hWnd) 
      gOutput->fatal("Cannot create pixmap window.");

    SetWindowLong(mData->hWnd, 0, (LONG)this);

    SendMessage(mData->hWnd, WM_INIT, 0, 0L);

    PIXELFORMATDESCRIPTOR pfd;
    int nPixelFormat = GetPixelFormat(mData->hdc);
    if(!nPixelFormat)
      gOutput->fatal("The given pixel format specification is not supported by this device context.");
    
    DescribePixelFormat(mData->hdc, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
    mSglBuf = !(pfd.dwFlags & PFD_DOUBLEBUFFER);

    ShowWindow(mData->hWnd, SW_SHOWNORMAL);
    UpdateWindow(mData->hWnd);
}

GPixmapDisplay::~GPixmapDisplay(void) {
    SendMessage(mData->hWnd, WM_DESTROY, 0, 0);
    delete mData;
}  

void GPixmapDisplay::activate() {
    if (!mData->hWnd)
      gOutput->fatal("Missing output window.");
    if (!mData->hdc)
      gOutput->fatal("Missing device context.");
    if (!mData->context)
      gOutput->fatal("Missing OpenGL rendering context.");
    if (!wglMakeCurrent(mData->hdc, mData->context))
      gOutput->fatal("Can't connect device context and rendering context.");
    glViewport(0, 0, mData->width, mData->height);
}

GEvent* GPixmapDisplay::getEvent(void) {    
    GEvent* tEvent = mData->mEvent;
    mData->mEvent = 0;
    return(tEvent);
}    

void GPixmapDisplay::putPixel(int x, int y, const GColorRGB &colr) {
    GLfloat f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB();
    glColor3fv(f);
    glRecti(x, y, x + 1, y + 1);
    glFlush();
}

void GPixmapDisplay::clear(const GColorRGB &colr) {
    glDrawBuffer(GL_BACK);
    float f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB(); 
    glClearColor(f[0], f[1], f[2], 0.0);
    glClear((GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));

    // setup OpenGL for a pixel-oriented output:
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    if (mSglBuf == false) glDrawBuffer(GL_FRONT_AND_BACK);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, (GLdouble)mData->width , 0.0, (GLdouble)mData->height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GPixmapDisplay::swapBuffers(void) {
    if (inSingleBufferMode()) 
      glFlush();
    else 
      SwapBuffers(mData->hdc);
}
