//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [shading]
// OpenGL 1.0

#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gevent.h"

static int gpLightCnt;

// class GShadable:

GShadable::~GShadable(void) {
    if (getID()) {
	if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
	gFreeID(getID());
    }
}

void GShadable::setMatrix(const GMatrix4& pMat) {
    GLfloat m[16];
    int ind = 0;
    int i, j;
    for (i = 0; i < 4; i++)
	for (j = 0; j < 4; j++)
	    m[ind++] = pMat.get(j, i);
    glPushMatrix();
    glMultMatrixf(m);
}

void GShadable::unsetMatrix(void) { 
    glPopMatrix(); 
}

void GShadable::setSurface(const GSurface& pSurface) {
    glColorMaterial(GL_FRONT, GL_AMBIENT);
    glColor3f(pSurface.getAmbient().getR(), pSurface.getAmbient().getG(), pSurface.getAmbient().getB());
    glColorMaterial(GL_FRONT, GL_DIFFUSE);
    glColor3f(pSurface.getDiffuse().getR(), pSurface.getDiffuse().getG(), pSurface.getDiffuse().getB());
    glColorMaterial(GL_FRONT, GL_SPECULAR);
    glColor3f(pSurface.getSpecular().getR(), pSurface.getSpecular().getG(), pSurface.getSpecular().getB());
}

void GShadable::openDisplayList(void) {
    if (!mID) mID = gGetID();
    if (glIsList((GLuint)getID())) glDeleteLists((GLuint)getID(), (GLsizei)1);
    glNewList((GLuint)getID(), GL_COMPILE);
}

void GShadable::closeDisplayList(void) { 
    glEndList(); 
}

void GShadable::callDisplayList(void) const {
    if (!getID()) gOutput->fatal("Attempt to call non-existing display list.");
    glCallList((GLuint)getID());
} 

// class GShadableQuadmesh:

void GShadableQuadmesh::shade(void) {
    GVector3 defNormal(0, 0, 1);
    int rows = getX();
    int cols = getY();
    float f[3];

    register int i, j, p, r;
    if (fillstyle) {
	for (j = 0; j < cols - 1; j++) {
	    glBegin(GL_QUAD_STRIP);
	    for (i = 0; i < rows - 1 ; i++) {
		for(p = 0; p < 2; p++)
		    for(r = 0; r < 2; r++) {
			const GVertex& vt = getVertex(i + p, j + r);

			if (vt.getNormal()) { 
			    vt.getNormal()->get(f); 
			}
			else defNormal.get(f);
		
			glNormal3fv(f);
			vt.getPoint().get(f);
			glVertex3fv(f);
		    }
	    }
	    glEnd();
	}
    } else {
/*
    if (fillstyle) {
	for (i = 0; i < rows - 1; i++) {
	    glBegin(GL_QUAD_STRIP);
	    for (j = 0; j < cols; j++) {
		for (p = 0; p < 2; p++) {
		    const GVertex& vt = getVertex(i + p, j);
		    if (vt.getNormal()) { 
			vt.getNormal()->get(f); 
		    }
		    else defNormal.get(f);
		    glNormal3fv(f);
		    vt.getPoint().get(f);
		    glVertex3fv(f);
		}
	    }
	    glEnd();
	}
    }
*/
	for (i = 0; i < rows; i++) {
	    glBegin(GL_LINE_STRIP);
	    for(j = 0; j < cols; j++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
	for (j = 0; j < cols; j++) {
	    glBegin(GL_LINE_STRIP);
	    for (i = 0; i < rows; i++) {
		const GVertex& vt = getVertex(i, j);  
		vt.getPoint().get(f);
		glVertex3fv(f);
	    }
	    glEnd();
	}
    }
}

template<class V>
class GShadeTopology1DIterator: public GIterator<V> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Shading iterator for subclasses
    // of 1D topology.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(V& v) { 
	GLfloat p[3];
	const GVector3* n = v.getNormal();
	if (n) {
	    n->get(p); 
	    glNormal3fv(p);
	}
	GVector3 po = v.getPoint();
	p[0] = po.getX();
	p[1] = po.getY();
	p[2] = po.getZ();
	glVertex3fv(p);
	return(false);
    }
  private:
    const GVector3* mNormal;
};

void GShadablePolyline::shade(void) {
    glBegin(GL_LINE_STRIP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

void GShadablePolygon::shade(void) {
    if (fillstyle) {
	glBegin(GL_POLYGON); 
	GLfloat n[3];
	getNormal().get(n);
	glNormal3fv(n); 
    }
    else glBegin(GL_LINE_LOOP);
    GShadeTopology1DIterator<GVertex> tIter;
    getVertexList().doWithElements(tIter);
    glEnd();
}

template<class F>
class GShadeTopologyPhIterator: public GIterator<F> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator for shading polyhedron-compatible
    // shadable primitives.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: The iterator is only useable to render
    // exactly one polyhedron.
    // CONSTANTS: 
  public:
    GShadeTopologyPhIterator(const GTopologyPh& pPolyhedron): mPolyhedron(pPolyhedron), mNumber(0) {}
    virtual bool exec(F& f) { 
	GLfloat sFloat[3]; 
    	if (GShadable::fillstyle) glBegin(GL_POLYGON);
   	else glBegin(GL_LINE_LOOP);	
	int i;
	for (i = 0; i < f.getNumber(); i++) {
	    const GVertex& v = mPolyhedron.getVertex(mNumber, i);
	    if (v.getNormal()) {
		v.getNormal()->get(sFloat);
		glNormal3fv(sFloat); 
	    }
	    v.getPoint().get(sFloat);
	    glVertex3fv(sFloat);
	}
	glEnd();
	mNumber++;
	return(false);
    }
  private:
    const GTopologyPh& mPolyhedron;
    int mNumber;
};

void GShadablePolyhedron::shade(void) {
    GShadeTopologyPhIterator<GFacet> tIter(*this);
    getFacetList().doWithElements(tIter);
}

void gShaderSetLookatMatrix(const GLookatCamera& r, const GPixmapDisplay& p) {
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_COLOR_MATERIAL); 
    glEnable(GL_NORMALIZE);  
    glShadeModel(GL_SMOOTH);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glDrawBuffer(GL_BACK);
    glFrontFace(GL_CW);

    glPushMatrix();
    // setup projection:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (r.isPerspectivic()) {
	gluPerspective((GLdouble)r.getZoom(), (GLdouble)p.getWidth()/(GFloat)p.getHeight(), (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }
    else {
	GFloat tWidth = tan((GLdouble)r.getZoom() * M_PI_2 / 90.0);
	GFloat tHeight = tWidth;
	if (p.getWidth() > p.getHeight()) {
	    tWidth *= (GLdouble)p.getWidth()/(GFloat)p.getHeight();
	}
	if (p.getHeight() > p.getWidth()) {
	    tHeight *= (GLdouble)p.getHeight()/(GFloat)p.getWidth();
	}

	glOrtho(-tWidth, tWidth, -tHeight, tHeight, (GLdouble)r.getNear(), (GLdouble)r.getFar());
    }

    GVector3 vup = gCalcVUP(r.getViewpoint(), r.getRefpoint(), r.getTwist() * M_PI / 180.0);
    gluLookAt(r.getViewpoint().getX(), r.getViewpoint().getY(), r.getViewpoint().getZ(),
	      r.getRefpoint().getX(), r.getRefpoint().getY(), r.getRefpoint().getZ(), 
	      -vup.getX(), -vup.getY(), -vup.getZ()
	      );    
    
    // switch in model mode:
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // reset light sources:
    gpLightCnt = 0;

    int i;
    for (i = 0; i < 8; i++) glDisable((GLenum)(GL_LIGHT0 + i));

    // setup the light sources:
    GShadingIterator<GShadeLight> tIter;
    if (gShadeLights) gShadeLights->doWithElements(tIter);
} 

int gGetID(void) {
    static sNumber = 1;
    return(sNumber++);
}

void gFreeID(int) {}

void gShaderUnsetMatrix(const GLookatCamera&, const GPixmapDisplay&) {
    glPopMatrix();
}

void GShadePointLight::shade(void) {
    if (!isOn()) return;

    setNumber(gpLightCnt);

    GLenum tEnum = (GLenum)(GL_LIGHT0 + gpLightCnt++);

    GLfloat tF4[4]; 

    int n = 0; 
    tF4[n++] = 0.0; 
    tF4[n++] = 0.0; 
    tF4[n++] = 0.0; 
    tF4[n++] = 1.0;
    glLightfv(tEnum, GL_AMBIENT, tF4);

    n = 0;
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0;
    glLightfv(tEnum, GL_DIFFUSE, tF4);

    n = 0;
    tF4[n++] = 1.0;
    tF4[n++] = 1.0;
    tF4[n++] = 1.0;
    tF4[n++] = 1.0;
    glLightfv(tEnum, GL_SPECULAR, tF4);

    n = 0;
    tF4[n++] = getPosition().getX();
    tF4[n++] = getPosition().getY();
    tF4[n++] = getPosition().getZ();
    tF4[n++] = 1.0;
    glLightfv(tEnum, GL_POSITION, tF4);

    GFloat tF;
    tF = 0.0;
    glLightf(tEnum, GL_CONSTANT_ATTENUATION, tF); 

    tF = 1.0;
    glLightf(tEnum, GL_LINEAR_ATTENUATION, tF); 

    glEnable(tEnum);
}

void GShadeAmbientLight::shade(void) {
    if (!isOn()) return;

    setNumber(gpLightCnt);

    GLenum tEnum = (GLenum)(GL_LIGHT0 + gpLightCnt++);
    
    GLfloat tF4[4]; 
    int n = 0; 
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0; 
    tF4[n++] = 1.0;
    glLightfv(tEnum, GL_AMBIENT, tF4);

    glEnable(tEnum);
}

GColorRGB GPixmapDisplay::getPixel(int pX, int pY) {
    GLfloat c[3];
    if (pX < 0 || pX >= getWidth() || pY < 0 || pY >= getHeight() )
	gOutput->error("PixmapDisplay::getPixel(): Coordinates out of range.");
    else glReadPixels(pX, pY, 1, 1, GL_RGB, GL_FLOAT, c);
    return(GColorRGB(c[0], c[1], c[2]));
}

void GShadeLight::disable(void) {
    glDisable((GLenum)(GL_LIGHT0 + getNumber()));
}
