//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [window handling and interaction]
// Windows/NT + Tk + OpenGL 1.0
// This version only supports direct color (RGB) display drivers.

#include <windows.h>

#include <GL/glaux.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gdevice.h"
#include "input/gevent.h"
#include "util/gtcl.h"

#include <tk.h>
#include <unistd.h>

#define WM_INIT		WM_USER

Tk_Window gTkWin;

class GWinNTInputDevice: public GInputDevice {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An input device for 
    // handling Windows Events
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GWinNTInputDevice(void): GInputDevice((GInputDevice*)gNoInputDevice) {}
    bool event(GEvent* pEvent = 0);
};

bool GWinNTInputDevice::event(GEvent*) {
    if (!isOn()) return(false);
    bool ret = false;
    if (Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT)) {
	callCBs();
	ret = true;
    }
    return(ret);
}

GWinNTInputDevice* gWinNTInputDevice;

bool gInitialize(void) {
    if (!gGlobalInterp) 
      gOutput->fatal("Global Interpreter missing.");
    
    gTkWin = Tk_MainWindow(gGlobalInterp);
    if (!gTkWin) 
      gOutput->fatal("Toplevel window missing.");

    // create a input device:
    gWinNTInputDevice = new GWinNTInputDevice;
    gWinNTInputDevice->on();
    return(true);
}

class GPixmapDisplayData {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specific data of a Windows NT
    // coupled pixmap display.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    // Supports only RGB displays 
    //
    // CONSTANTS: 
  public:
    GPixmapDisplayData(void);

    static bool WndClassIsRegistered;
    static HGLRC context;
    static int PixmapWindowCount;	
    bool first;
    HWND hWnd;
    HDC hdc;
    PIXELFORMATDESCRIPTOR PixelFormatDescriptor;
    int width, height;
    GEvent *mEvent; // the last event

	Tk_Window  tkwin;
	int border;
	int borderWidth;
};

bool GPixmapDisplayData::WndClassIsRegistered = FALSE;
HGLRC GPixmapDisplayData::context = 0;
int GPixmapDisplayData::PixmapWindowCount;

GPixmapDisplayData::GPixmapDisplayData(void) {
    first = true;
    width = height = 0;
    mEvent = 0;
    hWnd = 0;
    hdc  = 0;
	border = 2;
	borderWidth = 2;
 
    PixelFormatDescriptor.nSize           = sizeof(PIXELFORMATDESCRIPTOR);
    PixelFormatDescriptor.nVersion        = 1;
    PixelFormatDescriptor.dwFlags         = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    PixelFormatDescriptor.iPixelType      = PFD_TYPE_RGBA;
    PixelFormatDescriptor.cColorBits      = 24;
    PixelFormatDescriptor.cRedBits        = 0;
    PixelFormatDescriptor.cRedShift       = 0;
    PixelFormatDescriptor.cGreenBits      = 0;
    PixelFormatDescriptor.cGreenShift     = 0;
    PixelFormatDescriptor.cBlueBits       = 0;
    PixelFormatDescriptor.cBlueShift      = 0;
    PixelFormatDescriptor.cAlphaBits      = 0;
    PixelFormatDescriptor.cAlphaShift     = 0;
    PixelFormatDescriptor.cAccumBits      = 0;
    PixelFormatDescriptor.cAccumRedBits   = 0;
    PixelFormatDescriptor.cAccumGreenBits = 0;
    PixelFormatDescriptor.cAccumBlueBits  = 0;
    PixelFormatDescriptor.cAccumAlphaBits = 0;
    PixelFormatDescriptor.cDepthBits      = 32;
    PixelFormatDescriptor.cStencilBits    = 0;
    PixelFormatDescriptor.cAuxBuffers     = 0;
    PixelFormatDescriptor.iLayerType      = PFD_MAIN_PLANE;
    PixelFormatDescriptor.bReserved       = 0;
    PixelFormatDescriptor.dwLayerMask     = 0;
    PixelFormatDescriptor.dwVisibleMask   = 0;
    PixelFormatDescriptor.dwDamageMask    = 0;
}

void eventProc(ClientData cd, register XEvent *ev)
{
    GPixmapDisplay *tPixmap = (GPixmapDisplay*)cd;
    GPixmapDisplayData &tData = tPixmap->getData();

    if (tData.first) {
	  tData.first = false; 
	  GUpdatePixmapEvent* uev = new GUpdatePixmapEvent;
	  uev->pixmap = tPixmap;
	  tData.mEvent = uev;
	  tPixmap->activate(); // connect the context to the window:
	  return;
    }
    switch(ev->type) {
      case ButtonPress: { 
	  XButtonEvent* evb = (XButtonEvent*)ev;
	  GMouseButtonEvent* pev = new GMouseButtonEvent;
	  if (evb->state & Button1Mask) pev->left = true; 
	  if (evb->state & Button2Mask) pev->middle = true; 
	  if (evb->state & Button3Mask) pev->right = true; 
	  if (evb->state & ShiftMask) pev->shift = true;
	  if (evb->state & ControlMask) pev->ctrl = true;
	  if (evb->state & Mod1Mask) pev->alt = true;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = evb->x;  
	  pev->y = tPixmap->getHeight()-evb->y;
	  tData.mEvent = pev;
	  return;
      }
      case ButtonRelease: { 
	  XButtonEvent* evb = (XButtonEvent*)ev;
	  GMouseButtonEvent* pev = new GMouseButtonEvent;
	  if (evb->state & Button1Mask) pev->left = true; 
	  if (evb->state & Button2Mask) pev->middle = true; 
	  if (evb->state & Button3Mask) pev->right = true; 
	  if (evb->state & ShiftMask) pev->shift = true;
	  if (evb->state & ControlMask) pev->ctrl = true;
	  if (evb->state & Mod1Mask) pev->alt = true;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = evb->x;  
	  pev->y = tPixmap->getHeight()-evb->y;
	  tData.mEvent = pev;
	  return;
      }
      case MotionNotify: {
	  XPointerMovedEvent* evm = (XPointerMovedEvent*)ev;
	  GMouseMotionEvent* pev = new GMouseMotionEvent;
	  if (evm->state & Button1Mask) pev->left = true; 
	  if (evm->state & Button2Mask) pev->middle = true; 
	  if (evm->state & Button3Mask) pev->right = true; 
	  if (evm->state & ShiftMask) pev->shift = 1;
	  if (evm->state & ControlMask) pev->ctrl = 1;
	  if (evm->state & Mod1Mask) pev->alt = 1;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = evm->x;  
	  pev->y = tPixmap->getHeight()-evm->y;
	  tData.mEvent = pev;
	  return;
      }
      case Expose:
	GUpdatePixmapEvent* uev = new GUpdatePixmapEvent;
	uev->pixmap = tPixmap;
	tData.mEvent = uev;
	return;
    } 
}


extern HINSTANCE Main_hInstance;

GPixmapDisplay::GPixmapDisplay(const GString& pName, int pWidth, int pHeight): mWidth(pWidth), mHeight(pHeight) {
    mData = new GPixmapDisplayData;
	
	mData->width = pWidth;
	mData->height = pHeight;

    mData->tkwin =  Tk_CreateWindowFromPath(gGlobalInterp, gTkWin, (char*)pName, 0);
    if (!mData->tkwin) gOutput->fatal("Can't create main window.");
    Tk_SetClass(mData->tkwin, "TkPixmap");
    Tk_CreateEventHandler(mData->tkwin, ButtonPressMask|ButtonReleaseMask|PointerMotionMask|ButtonMotionMask|StructureNotifyMask|ExposureMask|EnterWindowMask|LeaveWindowMask,
	    eventProc, (ClientData) this);
	
	/*
	Tk_SetBackgroundFromBorder(mData->tkwin, mData->border);
    Tk_SetInternalBorder(mData->tkwin, mData->borderWidth);
    */


    Tk_GeometryRequest(mData->tkwin, pWidth, pHeight);
    
}

GPixmapDisplay::~GPixmapDisplay(void) {
	mData->PixmapWindowCount--;
	wglMakeCurrent(NULL, NULL);
	if(!mData->PixmapWindowCount)
	    wglDeleteContext(mData->context);
	ReleaseDC(mData->hWnd, mData->hdc);
	delete mData;
}  

void GPixmapDisplay::activate() {
    if(Tk_IsMapped(mData->tkwin)) {
      if(!mData->hWnd) {
	    mData->hWnd = Tk_WindowId(mData->tkwin);
	    mData->hdc = GetDC(mData->hWnd);
        int nPixelFormat = ChoosePixelFormat(mData->hdc, &mData->PixelFormatDescriptor);
	    if(nPixelFormat) {
          SetPixelFormat(mData->hdc, nPixelFormat, &mData->PixelFormatDescriptor); 
          if(!mData->context)
            mData->context = wglCreateContext(mData->hdc);
          if(mData->context)
	        mData->PixmapWindowCount++;
	    }

        PIXELFORMATDESCRIPTOR pfd;
        nPixelFormat = GetPixelFormat(mData->hdc);
        if(!nPixelFormat)
          gOutput->fatal("The given pixel format specification is not supported by this device context.");
    
        DescribePixelFormat(mData->hdc, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
        mSglBuf = !(pfd.dwFlags & PFD_DOUBLEBUFFER);
	
        if (!mData->hWnd)
          gOutput->fatal("Missing output window.");
        if (!mData->hdc)
          gOutput->fatal("Missing device context.");
        if (!mData->context)
          gOutput->fatal("Missing OpenGL rendering context.");
  	  }
      if (!wglMakeCurrent(mData->hdc, mData->context))
        gOutput->fatal("Can't connect device context and rendering context.");
      glViewport(0, 0, mData->width, mData->height);
	}
}

GEvent* GPixmapDisplay::getEvent(void) {    
    GEvent* tEvent = mData->mEvent;
    mData->mEvent = 0;
    return(tEvent);
}    

void GPixmapDisplay::putPixel(int x, int y, const GColorRGB &colr) {
    GLfloat f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB();
    glColor3fv(f);
    glRecti(x, y, x + 1, y + 1);
    glFlush();
}

void GPixmapDisplay::clear(const GColorRGB &colr) {
    glDrawBuffer(GL_BACK);
    float f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB(); 
    glClearColor(f[0], f[1], f[2], 0.0);
    glClear((GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));

    // setup OpenGL for a pixel-oriented output:
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    if (mSglBuf == false) glDrawBuffer(GL_FRONT_AND_BACK);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, (GLdouble)mData->width , 0.0, (GLdouble)mData->height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void GPixmapDisplay::swapBuffers(void) {
    if (inSingleBufferMode()) 
      glFlush();
    else 
      SwapBuffers(mData->hdc);
}
