//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
// G driver [window handling and interaction]
// Tk 3.6, 4.0, 4.1 + System V + OpenGL 1.0

#include <stdlib.h>
#include <signal.h>

#include <tk.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "goutput.h"
#include "render/gshade.h"
#include "render/gpixmap.h"
#include "input/gdevice.h"
#include "input/gevent.h"
#include "util/gtcl.h"

Tk_Window gTkWin;

// the first pixmap will create a new graphics context
// all other will link this
static GLXContext gGlobalContext = 0;

class GTkInputDevice: public GInputDevice {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An input device for 
    // handling Tk Events
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTkInputDevice(void): GInputDevice((GInputDevice*)gNoInputDevice) {}
    bool event(GEvent* pEvent = 0);
};

bool GTkInputDevice::event(GEvent*) {
    if (!isOn()) return(false);
    bool ret = false;
    if (Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT)) {
	callCBs();
	ret = true;
    }
    return(ret);
}

GTkInputDevice* gTkInputDevice;

bool gInitialize(void) { 
    signal(SIGFPE, SIG_IGN);

    if (!gGlobalInterp) gGlobalInterp = Tcl_CreateInterp();
    Tcl_SetVar(gGlobalInterp, "tcl_interactive", isatty(0) ? "1" : "0",
	       TCL_GLOBAL_ONLY);
#ifdef GSYSTEM_V
    Tcl_SetVar(gGlobalInterp, "gSystem", "System V", TCL_GLOBAL_ONLY);
#endif

#if (TK_MAJOR_VERSION<4)
    gTkWin = Tk_CreateMainWindow(gGlobalInterp, 0, "gsh", "GSH");
#endif 
	
    // initialize Tk:
    if (Tcl_Init(gGlobalInterp) == TCL_ERROR)
	gOutput->fatal ("Can't initialize Tcl.");
    if (Tk_Init(gGlobalInterp) == TCL_ERROR)
	gOutput->fatal ("Can't initialize Tk.");
    gTkWin = Tk_MainWindow(gGlobalInterp);
    if (!gTkWin)
	gOutput->fatal("Can't create toplevel window.");

    // create a Tk input device:
    gTkInputDevice = new GTkInputDevice;
    gTkInputDevice->on();
    return(true);
}

class GPixmapDisplayData {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Specific data of a Tk
    // coupled pixmap display.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GPixmapDisplayData(void): first(true) {}

    Tk_Window tkwin;
    Display* display;
    XVisualInfo* visualInfo;
    GLXContext context;
    Colormap cmap;
    int width, height;
    bool first;
    GEvent *mEvent; // the last event
};

void staticEventProc(ClientData cd, XEvent* ev) {
    GPixmapDisplay *tPixmap = (GPixmapDisplay*)cd;
    GPixmapDisplayData &tData = tPixmap->getData();

    if (tData.first) {
	tData.first = false; 
	tData.mEvent = new GUpdatePixmapEvent(tPixmap);
	tPixmap->activate(); // connect the context to the window:
	return;
    }
    int tPress = false;
    switch(ev->type) {
      case ButtonPress:  
	tPress = true;
      case ButtonRelease: { 
	  XButtonEvent* evb = (XButtonEvent*)ev;
	  GMouseButtonEvent* pev = new GMouseButtonEvent;
	  if ((evb->button == 1) && Button1Mask && tPress) pev->left = true; 
	  if ((evb->button == 2) && Button2Mask && tPress) pev->middle = true; 
	  if ((evb->button == 3) && Button3Mask && tPress) pev->right = true; 
	  if (evb->state & ShiftMask) pev->shift = true;
	  if (evb->state & ControlMask) pev->ctrl = true;
	  if (evb->state & Mod1Mask) pev->alt = true;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = evb->x;  
	  pev->y = tPixmap->getHeight()-evb->y;
	  tData.mEvent = pev;
	  return;
      }
      case MotionNotify: {
	  XPointerMovedEvent* evm = (XPointerMovedEvent*)ev;
	  GMouseMotionEvent* pev = new GMouseMotionEvent;
	  if (evm->state & Button1Mask) pev->left = true; 
	  if (evm->state & Button2Mask) pev->middle = true; 
	  if (evm->state & Button3Mask) pev->right = true; 
	  if (evm->state & ShiftMask) pev->shift = 1;
	  if (evm->state & ControlMask) pev->ctrl = 1;
	  if (evm->state & Mod1Mask) pev->alt = 1;
	  pev->w = tPixmap->getWidth();
	  pev->h = tPixmap->getHeight();
	  pev->x = evm->x;  
	  pev->y = tPixmap->getHeight()-evm->y;
	  tData.mEvent = pev;
	  return;
      }
      case Expose:
	tData.mEvent = new GUpdatePixmapEvent(tPixmap);
	return;
      case ConfigureNotify: {
	tData.width = Tk_Width(tData.tkwin);
	tData.height = Tk_Height(tData.tkwin);
	tPixmap->setWidth(tData.width);
	tPixmap->setHeight(tData.height);
	tData.mEvent = new GResizePixmapEvent(tPixmap);
	return;
      }
   }
}

GPixmapDisplay::GPixmapDisplay(const GString& pName, int pWidth, int pHeight): mWidth(pWidth), mHeight(pHeight) {
    static int sDbConfiguration[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_DOUBLEBUFFER,
        GLX_DEPTH_SIZE, 1,
        (int)None,
    };

    // config for singlebuffer mode
    static int sSbConfiguration[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_DEPTH_SIZE, 1,
        (int)None,
    };
    
    int erb, evb; // error base and event base
    mData = new GPixmapDisplayData;
    mData->width = mWidth;
    mData->height = mHeight;
    mData->mEvent = 0;
    
    mData->tkwin =  Tk_CreateWindowFromPath(gGlobalInterp, gTkWin, (char*)pName, 0);

    if (!mData->tkwin) gOutput->fatal("Can't create main window.");
    Tk_SetClass(mData->tkwin, "TkPixmap");
    mData->display = Tk_Display(mData->tkwin);
    
    // is the GLX extension present ?
    if (!glXQueryExtension(mData->display, &erb, &evb)) 
	gOutput->fatal ("X server hasn't an OpenGL GLX extension.");
    mSglBuf = false;

    mData->visualInfo = glXChooseVisual(Tk_Display(mData->tkwin), Tk_ScreenNumber(mData->tkwin), sDbConfiguration);
    if (!mData->visualInfo) {
        mSglBuf = true;
        mData->visualInfo = glXChooseVisual(Tk_Display(mData->tkwin), Tk_ScreenNumber(mData->tkwin), sSbConfiguration);
        if (!mData->visualInfo) gOutput->fatal("No appropriate RGB visual with depth buffer.");
    }
#ifdef GIRIX
    mData->cmap = XCreateColormap(Tk_Display(mData->tkwin),
				  RootWindow(Tk_Display(mData->tkwin), Tk_ScreenNumber(mData->tkwin)),
				  mData->visualInfo->visual, AllocNone);
    if (Tk_SetWindowVisual(mData->tkwin, mData->visualInfo->visual,
			   mData->visualInfo->depth, mData->cmap) == 0) 
	gOutput->fatal ("Can`t set Tk_Window visual.");
#else
    mData->cmap = None;
    if (Tk_SetWindowVisual(mData->tkwin, mData->visualInfo->visual, mData->visualInfo->depth,
			   DefaultColormap(Tk_Display(mData->tkwin), Tk_ScreenNumber(mData->tkwin))) == 0)
	gOutput->fatal ("Can`t set Tk_Window visual.");
#endif
    
    mData->context = glXCreateContext(Tk_Display(mData->tkwin), mData->visualInfo, gGlobalContext, GL_TRUE);
    if (!mData->context) gOutput->fatal ("Can't create a context."); 
    if (!gGlobalContext) gGlobalContext = mData->context;
    if (mWidth > 0 && mHeight > 0) Tk_GeometryRequest(mData->tkwin, mWidth, mHeight);
    
    Tk_CreateEventHandler(mData->tkwin,
			  KeyPressMask
			  |KeyReleaseMask
			  |ButtonPressMask
			  |ButtonReleaseMask
			  |PointerMotionMask
			  |PointerMotionHintMask
			  |ButtonMotionMask
			  |Button1MotionMask
			  |Button2MotionMask
			  |Button3MotionMask
			  |Button4MotionMask
			  |Button5MotionMask
			  |EnterWindowMask
			  |LeaveWindowMask
			  |FocusChangeMask
			  |FocusChangeMask
			  |KeymapStateMask
			  |ExposureMask
			  |VisibilityChangeMask
			  |StructureNotifyMask
			  |SubstructureRedirectMask
			  |PropertyChangeMask
			  |ResizeRedirectMask
			  |FocusChangeMask,
			  staticEventProc,(ClientData)this);


    GString tStr(20);
    tStr = "pack ";
    tStr += pName;
    tStr += " -fill both -expand yes; update";
    gTclGlobalEval(tStr);
    activate();
}

GPixmapDisplay::~GPixmapDisplay(void) {
    Tk_DestroyWindow(mData->tkwin);
    if (mData->display) {
	if (mData->cmap != None) XFreeColormap(mData->display, mData->cmap);
	if (mData->visualInfo != None) XFree((char*)mData->visualInfo);
	if (mData->context != 0) glXDestroyContext(mData->display, mData->context);
    }
    delete mData;
}  

void GPixmapDisplay::activate() {
    if (!Tk_IsMapped(mData->tkwin)) return;

    if (!glXMakeCurrent(Tk_Display(mData->tkwin), Tk_WindowId(mData->tkwin),
			mData->context))
	gOutput->fatal("Can't connect context and window.");
}

GEvent* GPixmapDisplay::getEvent(void) {    
    GEvent* tEvent = mData->mEvent;
    mData->mEvent = 0;
    return(tEvent);
}    

void GPixmapDisplay::putPixel(int x, int y, const GColorRGB &colr) {
    GLfloat f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB();
    glColor3fv(f);
    glRecti(x, y, x + 1, y + 1);
    glFlush();
}

void GPixmapDisplay::swapBuffers(void) {
    if (!Tk_IsMapped(mData->tkwin)) return;
    if (inSingleBufferMode()) glFlush();
    else glXSwapBuffers(Tk_Display(getData().tkwin), Tk_WindowId(getData().tkwin));
}

void GPixmapDisplay::clear(const GColorRGB &colr) {
    glDrawBuffer(GL_BACK);
    float f[3];
    f[0] = colr.getR(); 
    f[1] = colr.getG();
    f[2] = colr.getB(); 
    glClearColor(f[0], f[1], f[2], 0.0);
    if (GShadable::fillstyle) 
	glClear((GLbitfield)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
    else 
	glClear(GL_COLOR_BUFFER_BIT);

    // setup OpenGL for a pixel-oriented output:
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    if (mSglBuf == false) glDrawBuffer(GL_FRONT_AND_BACK);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, (GLdouble)mData->width , 0.0, (GLdouble)mData->height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}


