//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include "render/gimage.h"

bool GImage::checkIndices(int pX, int pY) const {
    if ((pX < 0)|| (pX>= mXSize) || (pY < 0) || (pY >= mYSize)) 
	return(false);
    else return(true);
}

GColorRGB GImage::getIntColor(GFloat x, GFloat y) const {
    // this method should be extended to perform an interpolation
    // between the four color values
    GColorRGB c(gBlack);

    if ((x < 0) || (y < 0)) {
	gOutput->error("Indices out of area!");
	return(c);
    }

    // repeating the texture:
    if (x > 1) x = x - (int)x;
    if (y > 1) y = y - (int)y;

    int xx = (int)(x * (mXSize-1)); 
    if (xx >= mXSize) xx = mXSize - 1;

    int yy = (int)(y * (mYSize-1));
    if (yy >= mYSize) yy = mYSize - 1;

    return(getColor(xx, yy)); 
}

// SGI Image:

GSGIImage::GSGIImage(const char* pFname): GImage(pFname) {
    mBase = 0; mZSize = 3;
    // in following, a virtual method will be called
    // thus, a subclass of SGIImage shouldn't overload the read() method
    gOutput->messageVar("Trying to autoload image ", getFilename(), ".", 0);
    read();
}

void GSGIImage::copyBw(long* lptr, int n) {
    while(n>=8) {
	lptr[0] = 0xff000000 + (0x010101 * (lptr[0]&0xff));
	lptr[1] = 0xff000000 + (0x010101 * (lptr[1]&0xff));
	lptr[2] = 0xff000000 + (0x010101 * (lptr[2]&0xff));
	lptr[3] = 0xff000000 + (0x010101 * (lptr[3]&0xff));
	lptr[4] = 0xff000000 + (0x010101 * (lptr[4]&0xff));
	lptr[5] = 0xff000000 + (0x010101 * (lptr[5]&0xff));
	lptr[6] = 0xff000000 + (0x010101 * (lptr[6]&0xff));
	lptr[7] = 0xff000000 + (0x010101 * (lptr[7]&0xff));
	lptr += 8;
	n-=8;
    }
    while(n--) {
	*lptr = 0xff000000 + (0x010101 * (*lptr&0xff));
	lptr++;
    }
}

void GSGIImage::setAlpha(unsigned char* lptr, int n) {
    while(n>=8) {
	lptr[0 * 4] = 0xff;
	lptr[1 * 4] = 0xff;
	lptr[2 * 4] = 0xff;
	lptr[3 * 4] = 0xff;
	lptr[4 * 4] = 0xff;
	lptr[5 * 4] = 0xff;
	lptr[6 * 4] = 0xff;
	lptr[7 * 4] = 0xff;
	lptr += 4 * 8;
	n -= 8;
    }
    while(n--) {
	*lptr = 0xff;
	lptr += 4;
    }
}

void GSGIImage::expandRow(unsigned char* optr, unsigned char* iptr, int z) {
    unsigned char pixel, count;
    optr += z;
    while(1) {
	pixel = *iptr++;
	if (!(count = (pixel & 0x7f)))
	    return;
	if(pixel & 0x80) {
	    while(count>=8) {
		optr[0 * 4] = iptr[0];
		optr[1 * 4] = iptr[1];
		optr[2 * 4] = iptr[2];
		optr[3 * 4] = iptr[3];
		optr[4 * 4] = iptr[4];
		optr[5 * 4] = iptr[5];
		optr[6 * 4] = iptr[6];
		optr[7 * 4] = iptr[7];
		optr += 8 * 4;
		iptr += 8;
		count -= 8;
	    }
	    while(count--) {
		*optr = *iptr++;
		optr+=4;
	    }
	} else {
	    pixel = *iptr++;
	    while(count>=8) {
		optr[0 * 4] = pixel;
		optr[1 * 4] = pixel;
		optr[2 * 4] = pixel;
		optr[3 * 4] = pixel;
		optr[4 * 4] = pixel;
		optr[5 * 4] = pixel;
		optr[6 * 4] = pixel;
		optr[7 * 4] = pixel;
		optr += 8 * 4;
		count -= 8;
	    }
	    while(count--) {
		*optr = pixel;
		optr+=4;
	    }
	}
    }
}

void GSGIImage::addLongImgTag(unsigned long* dptr, int xsize, int ysize) {
    dptr = dptr + (xsize * ysize);
    dptr[0] = 0x12345678;
    dptr[1] = 0x59493333;
    dptr[2] = 0x69434222;
    dptr[3] = xsize;
    dptr[4] = ysize;
}

unsigned short GSGIImage::getShort(FILE* inf) {
    unsigned char buf[2];
    fread(buf, 2, 1, inf);
    return((buf[0]<<8) + (buf[1]<<0));
}

unsigned long GSGIImage::getLong(FILE* inf) {
    unsigned char buf[4];
    fread(buf, 4, 1, inf);
    return((buf[0]<<24) + (buf[1]<<16) + (buf[2]<<8) + (buf[3]<<0));
}

void GSGIImage::putShort(FILE* outf, unsigned short val) {
    unsigned char buf[2];
    buf[0] = (val>>8);
    buf[1] = (val>>0);
    fwrite(buf, 2, 1, outf);
}

int GSGIImage::putLong(FILE* outf, unsigned long val) {
    unsigned char buf[4];
    buf[0] = (unsigned char)(val>>24);
    buf[1] = (unsigned char)(val>>16);
    buf[2] = (unsigned char)(val>>8);
    buf[3] = (unsigned char)(val>>0);
    return(fwrite(buf, 4, 1, outf));
}

void GSGIImage::readHeader(FILE* inf) {
    memset(&image, 0, sizeof(GSGIImageStruct));
    image.imagic = getShort(inf);
    image.type = getShort(inf);
    image.dim = getShort(inf);
    image.xsize = getShort(inf);
    image.ysize = getShort(inf);
    image.zsize = getShort(inf);
}

int GSGIImage::writeHeader(FILE* outf) {
    GSGIImageStruct t;
    memset(&t, 0, sizeof(GSGIImageStruct));
    // write out a zero field:
    fwrite(&t, sizeof(GSGIImageStruct), 1, outf);
    fseek(outf, 0, SEEK_SET);
    putShort(outf, image.imagic);
    putShort(outf, image.type);
    putShort(outf, image.dim);
    putShort(outf, image.xsize);
    putShort(outf, image.ysize);
    putShort(outf, image.zsize); 
    putLong(outf, (unsigned char)image.min);
    putLong(outf, (unsigned char)image.max);
    putLong(outf, 0);
    return(fwrite("no name", 8, 1, outf));
}

int GSGIImage::writeTab(FILE* outf, unsigned long* tab, int len) {
    int r;
    while(len) {
	r = putLong(outf, *tab++);
	len -= 4;
    }
    return(r);
}

void GSGIImage::readTab(FILE* inf, unsigned long* tab, int len) {
    while(len) {
	*tab++ = getLong(inf);
	len -= 4;
    }
}

void GSGIImage::lumRow(unsigned char* rgbptr, unsigned char* lumptr, int n) {
    lumptr += GSGI_CHANOFFSET(0);
    while(n--) {
	*lumptr = GSGI_ILUM(rgbptr[GSGI_OFFSET_R], rgbptr[GSGI_OFFSET_G], rgbptr[GSGI_OFFSET_B]);
	lumptr += 4;
	rgbptr += 4;
    }
}

int GSGIImage::compressRow(unsigned char* lbuf, unsigned char* rlebuf, int z, int cnt) {
    unsigned char* iptr, * ibufend, * sptr, * optr;
    short todo, cc;							
    long count;

    lbuf += z;
    iptr = lbuf;
    ibufend = iptr+cnt*4;
    optr = rlebuf;

    while(iptr<ibufend) {
	sptr = iptr;
	iptr += 8;
	while((iptr<ibufend)&& ((iptr[-8]!=iptr[-4])||(iptr[-4]!=iptr[0])))
	    iptr+=4;
	iptr -= 8;
	count = (iptr-sptr)/4;
	while(count) {
	    todo = count > 126 ? 126 : (short)count;
	    count -= todo;
	    *optr++ = 0x80|todo;
	    while(todo>8) {
		optr[0] = sptr[0*4];
		optr[1] = sptr[1*4];
		optr[2] = sptr[2*4];
		optr[3] = sptr[3*4];
		optr[4] = sptr[4*4];
		optr[5] = sptr[5*4];
		optr[6] = sptr[6*4];
		optr[7] = sptr[7*4];
		optr += 8;
		sptr += 8*4;
		todo -= 8;
	    }
	    while(todo--) {
		*optr++ = *sptr;
		sptr += 4;
	    }
	}
	sptr = iptr;
	cc = *iptr;
	iptr += 4;
	while((iptr<ibufend) && (*iptr == cc))
	    iptr += 4;
	count = (iptr-sptr)/4;
	while(count) {
	    todo = count > 126 ? 126 : (short)count;
	    count -= todo;
	    *optr++ = todo;
	    *optr++ = cc;
	}
    }
    *optr++ = 0;
    return(optr - (unsigned char *)rlebuf);
}

bool GSGIImage::sizeOfImage(void) {
    FILE* inf = fopen((char*)getFilename(), "r");
    if(!inf) {
	gOutput->warningVar("Can't open image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    readHeader(inf);
    fclose(inf);
    if(image.imagic != GSGI_IMAGIC) {
	gOutput->errorVar("Bad magic number in image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    mXSize = image.xsize;
    mYSize = image.ysize;
    return(true);
}

bool GSGIImage::read(void) {
    if (!sizeOfImage()) return(false);

    unsigned long* lptr;
    // the temporary pointer to the long array
    // will be initialized with this->base 
    
    FILE* inf = fopen((char*)getFilename(), "r");
    if(!inf) {
	gOutput->warningVar("Can't open image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    readHeader(inf);
    if(image.imagic != GSGI_IMAGIC) {
	gOutput->errorVar("Bad magic number in image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    int rle = GSGI_ISRLE(image.type);
    int bpp = GSGI_BPP(image.type);
    if(bpp != 1) {
	gOutput->errorVar("Image ", (char*)getFilename(), " must have 1 byte per pix chan.", 0);
	return(false);
    }
    mXSize = image.xsize;
    mYSize = image.ysize;
    mZSize = image.zsize;

    int y, z;

    if (rle) {
	int tablen = mYSize*mZSize*sizeof(long);
	long* starttab = new long[tablen];
	long* lengthtab = new long[tablen];
	int rlebuflen = (int)1.05*mXSize + 10;
	unsigned char* rledat = new unsigned char [rlebuflen];
	fseek(inf, 512, SEEK_SET);
 	readTab(inf, (unsigned long*)starttab, tablen);
	readTab(inf, (unsigned long*)lengthtab, tablen);

	// check data order:
	int cur = 0;
	int badorder = 0;
	for (y = 0; y < mYSize; y++) {
	    for(z = 0; z < mZSize; z++) {
		if(starttab[y + z * mYSize] < cur) {
		    badorder = 1;
		    break;
		}
		cur = (int)starttab[y + z*mYSize];
	    }
	    if (badorder) break;
	}

	fseek(inf, 512 + 2*tablen, SEEK_SET);
	cur = 512 + 2*tablen;

	if (mBase) delete mBase;
	mBase = new unsigned long [mXSize*mYSize + GSGI_TAGLEN];

	addLongImgTag(mBase, mXSize, mYSize);
  	if (badorder) {
	    for (z = 0; z < mZSize; z++) {
		lptr = mBase;
		for (y = 0; y < mYSize; y++) {
		    if(cur != starttab[y + z*mYSize]) {
			fseek(inf, starttab[y + z*mYSize], SEEK_SET);
			cur = (int)starttab[y + z*mYSize];
		    }
		    if (lengthtab[y + z*mYSize] > rlebuflen) {
			gOutput->error("Rlebuf is too small - bad poop.");
			fclose(inf);
			delete starttab;
			delete lengthtab;
			delete rledat;
			return(false);
		    }
		    fread(rledat, (unsigned int)lengthtab[y + z*mYSize], 1, inf);
		    cur += (int)lengthtab[y + z*mYSize];
		    expandRow((unsigned char*)lptr, rledat, 3-z);
		    lptr += mXSize;
		}
	    }
	} else {
	    lptr = mBase;
	    for (y = 0; y < mYSize; y++) {
		for (z = 0; z < mZSize; z++) {
		    if(cur != starttab[y + z*mYSize]) {
			fseek(inf, starttab[y + z*mYSize], SEEK_SET);
			cur = (int)starttab[y + z*mYSize];
		    }
		    fread(rledat, (unsigned int)lengthtab[y + z*mYSize], 1, inf);
		    cur += (int)lengthtab[y + z*mYSize];
		    expandRow((unsigned char*)lptr, rledat, 3-z);
		}
		lptr += mXSize;
	    }
    	}

	if (mZSize == 3) setAlpha((unsigned char*)mBase, mXSize*mYSize);
	else if(mZSize<3) copyBw((long*)mBase, mXSize*mYSize);

	fclose(inf);
	delete starttab;
	delete lengthtab;
	delete rledat;
	return(true);
    } else {
	if (mBase) delete mBase;
	mBase = new unsigned long[mXSize*mYSize + GSGI_TAGLEN];

	addLongImgTag(mBase, mXSize, mYSize);
	unsigned char* verdat = new unsigned char [mXSize];
	fseek(inf, 512, SEEK_SET);
	for (z = 0; z < mZSize; z++) {
	    lptr = mBase;
	    for (y = 0; y < mYSize; y++) {
		fread(verdat, mXSize, 1, inf);
		interleaveRow((unsigned char*)lptr, verdat, 3-z, mXSize);
		lptr += mXSize;
	    }
	}
	if(mZSize == 3) setAlpha((unsigned char*)mBase, mXSize*mYSize);
	else if(mZSize<3) copyBw((long*)mBase, mXSize*mYSize);
	fclose(inf);
	delete verdat;
	return(true);
    }
}

bool GSGIImage::write(void) {
    FILE* outf;
    int tablen, y, z, pos, len;
    long* starttab, * lengthtab;
    unsigned char* rlebuf;
    unsigned long* lptr, * lumbuf;
    int rlebuflen, goodwrite;

    lptr = mBase;
    goodwrite = 1;
    outf = fopen((char*)getFilename(), "w");
    if(!outf) {
	gOutput->errorVar("Can't open file ", (char*)getFilename(), " for writing.", 0);
	return(false);
    }
    tablen = mYSize*mZSize*sizeof(long);

    starttab = new long [tablen];
    lengthtab = new long [tablen];
    rlebuflen = (int)1.05*mXSize + 10;
    rlebuf = new unsigned char [rlebuflen];
    lumbuf = new unsigned long [mXSize];

    memset(&image, 0, sizeof(GSGIImageStruct));
    image.imagic = GSGI_IMAGIC; 

    image.type = GSGI_RLE(1); 
    if(mZSize>1) image.dim = 3;
    else image.dim = 2;
    image.xsize = mXSize;
    image.ysize = mYSize;
    image.zsize = mZSize;
    image.min = 0;
    image.max = 255;
    goodwrite *= writeHeader(outf);

    fseek(outf, 512 + 2*tablen, SEEK_SET);
    
    pos = 512 + 2*tablen;
    for(y=0; y<mYSize; y++) {
	for(z=0; z<mZSize; z++) {
	    if(mZSize == 1) {
		lumRow((unsigned char*)lptr, (unsigned char*)lumbuf, mXSize);
		len = compressRow((unsigned char*)lumbuf, rlebuf, GSGI_CHANOFFSET(z), mXSize);
	    } else {
		len = compressRow((unsigned char*)lptr, rlebuf, GSGI_CHANOFFSET(z), mXSize);
	    }
	    if(len>rlebuflen) {
		gOutput->error("Rlebuf is too small - bad poop.");
		delete starttab;
		delete lengthtab;
		delete rlebuf;
		delete lumbuf;
		return(false);
	    }
	    goodwrite *= fwrite(rlebuf, len, 1, outf);
	    starttab[y + z*mYSize] = pos;
	    lengthtab[y + z*mYSize] = len;
	    pos += len;
	}
	lptr += mXSize;
    }

    fseek(outf, 512, SEEK_SET);
    goodwrite *= writeTab(outf, (unsigned long*)starttab, tablen);
    goodwrite *= writeTab(outf, (unsigned long*)lengthtab, tablen);
    delete starttab;
    delete lengthtab;
    delete rlebuf;
    delete lumbuf;
    fclose(outf);
    if (goodwrite) return(true);
    else {
	gOutput->error("Not enough space for image!");
	return(false);
    }
}

void GSGIImage::setColor(int x, int y, const GColorRGB& c) {
    if (!checkIndices(x, y)) return;
    long l;
    unsigned char* cc = (unsigned char*)&l;
    cc[3] = (char) (c.getR() * 255);
    cc[2] = (char) (c.getG() * 255);
    cc[1] = (char) (c.getB() * 255);
    cc[0] = 255;
    mBase[y * mXSize + x] = l;
}

GColorRGB GSGIImage::getColor(int x, int y) const { 
    GColorRGB c;
    if (!checkIndices(x, y)) return(c);
    unsigned char* cc = (unsigned char*)&mBase[y * mXSize + x];
    c.setR(((GFloat)cc[3])/255.0);
    c.setG(((GFloat)cc[2])/255.0);
    c.setB(((GFloat)cc[1])/255.0);
    return(c); 
}

// TARGA IMAGE:

GTargaImage::GTargaImage(const char* pFname) : GImage(pFname) { 
  mBase = 0;
  unsigned int test = 1;
  unsigned char* ctest = (unsigned char*)&test;
  lh = *ctest;  //short integer : low-byte - high-byte order ?  
  
  // in following, a virtual method will be called
  // thus, a subclass of TargaImage shouldn't overload the read() method
  gOutput->messageVar("Trying to autoload image ", getFilename(), ".", 0);
  read();
}

unsigned short GTargaImage::getShort(FILE* inf) {
    unsigned char* cval, buf[2];
    unsigned int val;
    cval = (unsigned char*)&val;
    fread(buf, 2, 1, inf);
    if(lh) { 
       cval[0] = buf[0]; cval[1] = buf[1];
    } else {
       cval[1] = buf[0]; cval[0] = buf[1]; // swap low / high
    }
    return(val);   
}

void GTargaImage::putShort(FILE* outf, unsigned short val) {
  unsigned char* cval, buf[2];
  cval = (unsigned char*)&val;
  if(lh) { 
    buf[0] = cval[0]; buf[1] = cval[1];
  } else {
    buf[0] = cval[1]; buf[1] = cval[0]; // swap low / high
  }
  fwrite(buf, 2, 1, outf);
}

void GTargaImage::readHeader(FILE* inf) {
  fread(&Header, 1, 3, inf);
  Header.cmap_origin = getShort(inf);
  Header.cmap_len = getShort(inf);
  fread(&Header.centry_size, 1, 1, inf);              
  Header.image_xorg = getShort(inf);
  Header.image_yorg = getShort(inf); 
  Header.image_width = getShort(inf); 
  Header.image_height = getShort(inf); 
  fread(&Header.pixel_size, 2, 1, inf);      
}

int GTargaImage::writeHeader(FILE* outf) {
  Header.ident_len    = 8;   
  Header.cmap_type    = 0;   // no color-table
  Header.image_type   = 10;  // compressed RGB
  Header.cmap_origin  = 0;
  Header.cmap_len     = 0;
  Header.centry_size  = 0;
  Header.image_xorg   = 0;
  Header.image_yorg   = 0;
  Header.image_width  = mXSize;
  Header.image_height = mYSize;
  Header.pixel_size   = 24;  // 24 bit / pixel
  Header.image_discr  = 0;   // left -> right ; bottom -> up

  fwrite(&Header, 12, 1, outf); 
  putShort(outf, Header.image_width);
  putShort(outf, Header.image_height);
  fwrite(&Header.pixel_size, 2, 1, outf);        
  return(fwrite("no name", 8, 1, outf));
}

bool GTargaImage::sizeOfImage(void) {
    FILE* inf = fopen((char*)getFilename(),"rb");
    if(!inf) {
	gOutput->errorVar("Can't open image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    readHeader(inf);
    fclose(inf);
    mXSize = Header.image_width;
    mYSize = Header.image_height;
    return(true);
}

bool GTargaImage::read(void) {
    if (!sizeOfImage()) return(false);

    FILE* inf = fopen((char*)getFilename(),"r");
    if(!inf) {
      gOutput->errorVar("Can't open image file ", (char*)getFilename(), ".", 0);
      return(false);
    }
    readHeader(inf);
    if(Header.cmap_type != 0 || (Header.image_type != 2 && Header.image_type != 10) ||
       Header.pixel_size != 24 || (Header.image_discr & 0xc0) != 0) {
       gOutput->errorVar("Can't read ", (char*)getFilename(), ". (Today the TARGA image class is limited to RGB images (no color-table). This will be changed in later versions).", 0);    
       return(false);
    } 
    mXSize = Header.image_width;
    mYSize = Header.image_height;
	
    if (mBase) delete mBase;
    mBase = new unsigned long [mXSize * mYSize];
    
    if(Header.ident_len) {
      char ident[256];
      fread(ident, 1, Header.ident_len, inf); 
      ident[Header.ident_len] = '\0';
      gOutput->message(ident);
    } 
    
    long l; 
    unsigned char* cc = (unsigned char*)&l;
    int top_down = Header.image_discr & 0x20;  
    int h, x, y = top_down ? mYSize - 1 : 0;

    if(Header.image_type == 2) {            // uncompressed RGB
      GTargaRGB* pcnt,* pix_row = new GTargaRGB [mXSize];
      
      for(h = 0; h < mYSize; h++) {
        pcnt = pix_row; 
        fread(pcnt, sizeof(GTargaRGB), mXSize, inf); 
        for(x = 0; x < mXSize; x++) {
          cc[3] = pcnt->r; cc[2] = pcnt->g; cc[1] = pcnt->b; cc[0] = 255; 
          pcnt++;
          mBase[y * mXSize + x] = l;
        }
        if(top_down) y--; else y++;
      }
      delete pix_row;
    }

    if(Header.image_type == 10) {            // compressed RGB
      GTargaRGB pix; unsigned char cnt; int mode = 0;
      for(h = 0; h < mYSize; h++) {
        x = 0;
        while(x < mXSize) { 
          if(!mode) {   
            fread(&cnt, 1, 1, inf); //read length       
            if(cnt & 0x80) {     //runlength coded
              fread(&pix, sizeof(GTargaRGB), 1, inf);
              cc[3] = pix.r; cc[2] = pix.g; cc[1] = pix.b; cc[0] = 255;
              cnt &= 0x7f; mode = 2;
            } else mode = 1;
          } else {
            if(mode == 1) {
              fread(&pix, sizeof(GTargaRGB), 1, inf);
              cc[3] = pix.r; cc[2] = pix.g; cc[1] = pix.b; cc[0] = 255;
            }
            mBase[y * mXSize + x] = l;
            if(!cnt) mode = 0; else cnt--;
            x++;
          }
        }
        if(top_down) y--; else y++;   
      }
    }
    return(true);
}

int GTargaImage::compressRow(GTargaRGB* pix_row, unsigned char* rlebuf) {
  int off = 0, offp1 = 1, rleoff = 0; 

  while(off < mXSize) {
    if(offp1 < mXSize) {
      unsigned char cnt = 0;
      if(pix_row[off].r == pix_row[offp1].r && 
         pix_row[off].g == pix_row[offp1].g &&
         pix_row[off].b == pix_row[offp1].b) {
        while (pix_row[off].r == pix_row[offp1].r && 
               pix_row[off].g == pix_row[offp1].g &&
               pix_row[off].b == pix_row[offp1].b && cnt < 127) {
          off++; offp1++; cnt++; if(offp1 >= mXSize) break; 
        } 
        cnt |= 0x80;
        rlebuf[rleoff++] = cnt;
        rlebuf[rleoff++] = pix_row[off].b;
        rlebuf[rleoff++] = pix_row[off].g;
        rlebuf[rleoff++] = pix_row[off].r;
        off++; offp1++;   
      } else {
        int rleoff1 = rleoff++;
        while ((pix_row[off].r != pix_row[offp1].r ||
                pix_row[off].g != pix_row[offp1].g ||
                pix_row[off].b != pix_row[offp1].b) && cnt < 127) {
          rlebuf[rleoff++] = pix_row[off].b;
          rlebuf[rleoff++] = pix_row[off].g;
          rlebuf[rleoff++] = pix_row[off].r;
          off++; offp1++; cnt++; if(offp1 >= mXSize) break;
        }     
        rlebuf[rleoff1] = --cnt;
      }
    } else {
      rlebuf[rleoff++] = 0;
      rlebuf[rleoff++] = pix_row[off].b;
      rlebuf[rleoff++] = pix_row[off].g;
      rlebuf[rleoff++] = pix_row[off].r;
      off++; offp1++;
    }
  }
  return(rleoff);   
}

bool GTargaImage::write(void) {
    FILE* outf;
    unsigned long* lptr = mBase;
    int goodwrite = 1;

    outf = fopen((char*)getFilename(), "wb");
    if(!outf) {
      gOutput->errorVar("Can't open file ", (char*)getFilename(), " for writing.", 0);
      return(false);
    }

    goodwrite *= writeHeader(outf);

    GTargaRGB* pcnt,* pix_row = new GTargaRGB [mXSize];
    unsigned char* rlebuf = new unsigned char [mXSize * sizeof(GTargaRGB) + mXSize];      
    for(int y = 0; y < mYSize; y++) {
      pcnt = pix_row;
      for(int x = 0; x < mXSize; x++) {         
        unsigned char* cc = (unsigned char*)lptr++;
        pcnt->r = cc[3]; pcnt->g = cc[2]; pcnt->b = cc[1];
        pcnt++;
      } 
      int len = compressRow(pix_row, rlebuf);
      goodwrite *= fwrite(rlebuf, len, 1, outf);
    }     
    delete rlebuf; delete pix_row;
    fclose(outf);
 
    if (goodwrite) return(true);
    else {
      gOutput->error("Not enough space for image!");
      return(false);
    }
}

void GTargaImage::setColor(int x, int y, const GColorRGB& c) {
    if (!checkIndices(x, y)) return;
    long l;
    unsigned char *cc = (unsigned char*)&l;
    cc[3] = (char) (c.getR() * 255);
    cc[2] = (char) (c.getG() * 255);
    cc[1] = (char) (c.getB() * 255);
    cc[0] = 255;
    mBase[y * mXSize + x] = l;
}

GColorRGB GTargaImage::getColor(int x, int y) const { 
    GColorRGB c;
    if (!checkIndices(x, y)) return(c);
    unsigned char* cc = (unsigned char*)&mBase[y * mXSize + x];
    c.setR(((GFloat)cc[3])/255.0);
    c.setG(((GFloat)cc[2])/255.0);
    c.setB(((GFloat)cc[1])/255.0);
    return(c); 
}

// class RGAB Image:

unsigned short GRGBAImage::getShort(int fd) {
    unsigned char buf[2];
    ::read(fd, buf, 2);
    return((buf[0]<<8)+(buf[1]<<0));
}

unsigned long GRGBAImage::getLong(int fd) {
    unsigned char buf[4];
    ::read(fd, buf, 4);
    return((buf[0]<<24)+(buf[1]<<16)+(buf[2]<<8) + (buf[3]<<0));
}
 
int GRGBAImage::putShort(int fd, unsigned short val) {
    unsigned char buf[2];
    buf[0] = (val>>8);
    buf[1] = (val>>0);
    return(::write(fd, buf, 2));
}

int GRGBAImage::putLong(int fd,unsigned long val) {
    unsigned char buf[4];
    buf[0] = (unsigned char)(val>>24);
    buf[1] = (unsigned char)(val>>16);
    buf[2] = (unsigned char)(val>>8);
    buf[3] = (unsigned char)(val>>0);
    return(::write(fd, buf, 4));
}

void GRGBAImage::writeHeader(int fd) {
    putShort(fd, image.imagic);
    putShort(fd, image.xsize);
    putShort(fd, image.ysize);
}

void GRGBAImage::readHeader(int fd) {
    memset(&image, 0, sizeof(GRGBAImageHeader));
    image.imagic = getShort(fd);
    image.xsize = getShort(fd);
    image.ysize = getShort(fd);
}

bool GRGBAImage::read(void) {
    int fd, size, ret;

    if (mBase) delete (mBase);
    if ((fd = open((char*)getFilename(), O_RDONLY, 0666)) < 0) {
	gOutput->errorVar("Can't open file: ", getFilename(), ".");
	return(false);
    }
    
    // read header
    readHeader(fd);
    if(image.imagic != GRGBA_IMAGIC) {
	gOutput->errorVar("Bad magic number in image file ", (char*)getFilename(), ".", 0);
	return(false);
    }
    
    // allocate data memory and read data
    size = mXSize * mYSize * 4;
    mBase = (unsigned long*) new char [size];
    assert(mBase != 0);
    ret = ::read(fd, (char*)mBase, size);
    close(fd);
    if (ret < size) {
	gOutput->errorVar("Error while reading file: ", (char*)getFilename(), ".");
	return(false);
    }
    return(true);
}

bool GRGBAImage::write(void) {
    int fd;
    int size;
    int ret;

    if (!mBase) return(false);
    if ((fd = open((char*)getFilename(), O_WRONLY|O_CREAT | O_TRUNC, 0666)) < 0) {
	gOutput->errorVar("Can't create file: ", getFilename(), ".");
	return(false);
    }

    // set the header info and write header
    memset(&image, 0, sizeof(GRGBAImageHeader));
    image.imagic = GRGBA_IMAGIC;
    image.xsize = mXSize;
    image.ysize = mYSize;
    writeHeader(fd);

    // dump image data
    size = mXSize * mYSize * 4;
    ret = ::write(fd, (char*)mBase, size);
    close (fd);
    if (ret < size) {
	gOutput->errorVar("Error while writing file: ", (char*)getFilename(), ".");
	return(false);
    }
    return(true);
}

void GRGBAImage::setColor(int x, int y, const GColorRGB& c) {
    int offset;
    char* src;
    
    y = mYSize - y;
    if (!checkIndices(x, y)) return;
    offset = (y * mXSize + x) * 4;
    src = &((char*)mBase)[offset];
    *++src = (char) (c.getB() * 255);
    *++src = (char) (c.getG() * 255);
    *++src = (char) (c.getR() * 255);
    return;
}

GColorRGB GRGBAImage::getColor(int x, int y) const {
    GColorRGB c;
    char* src;
    int offset;

    y = mYSize - y;
    if (!checkIndices(x, y)) return(c);
    offset = (y * mXSize + x) * 4;
    src = &((char*)mBase)[offset];
    c.setB(((GFloat)*++src)/255.0); 
    c.setG(((GFloat)*++src)/255.0); 
    c.setR(((GFloat)*++src)/255.0); 
    return(c);
}



