//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "render/glookat.h"

void GLookatCamera::computeMatrix(void) {
    GVector3 dr = (getRefpoint() - getViewpoint()).unitize(); 
    // lookat direction
    GFloat twist = -getTwist() * M_PI / 180.0;
    // the z axis rotation angle 
    
    // compute y vector:
    GVector3 ye;
    ye.setX(0); 
    ye.setY(1); 
    ye.setZ(0);  
    // defaults
    
    // is dr kollinear with ye ?:
    if((dr.getX() == 0) && (dr.getY() == 1) && (dr.getZ() == 0)) { 
	ye.setX(0); 
	ye.setY(0); 
	ye.setZ(1);
    }
    
    if((dr.getX() == 0) && (dr.getY() == -1) && (dr.getZ() == 0)) {
	ye.setX(0); 
	ye.setY(0); 
	ye.setZ(-1);
    }
    
    // compute eye point xe vector:
    GVector3 xe;
    xe = ye.cross(dr); 
    
    // unitize xe vector:
    xe = xe.unitize();
    
    // compute ye vector:
    ye = dr.cross(xe); 
    
    GMatrix4 trRt; // translation and rotation matrix
    trRt.set(0, 0, xe.getX());
    trRt.set(1, 0, xe.getY());
    trRt.set(2, 0, xe.getZ());
    
    trRt.set(0, 1, ye.getX());
    trRt.set(1, 1, ye.getY());
    trRt.set(2, 1, ye.getZ()); 
    
    trRt.set(0, 2, dr.getX());
    trRt.set(1, 2, dr.getY());
    trRt.set(2, 2, dr.getZ());
    
    trRt.set(0, 3, getViewpoint().getX()); 
    trRt.set(1, 3, getViewpoint().getY()); 
    trRt.set(2, 3, getViewpoint().getZ());
    
    trRt.set(3, 0, 0); 
    trRt.set(3, 1, 0); 
    trRt.set(3, 2, 0); 
    trRt.set(3, 3, 1);
    
    GMatrix4 tw;
    tw.set(0, 0, cos(twist));
    tw.set(0, 1, -sin(twist));
    tw.set(1, 0, sin(twist));
    tw.set(1, 1, cos(twist));
    
    mMat = (trRt * tw);
}

GRay3 GLookatCamera::inverseTransform(const GPoint2& pPoint, GFloat pAspect) const {
    if (isPerspectivic()) {
	GVector3 vn;  
	// normalized vector
	vn.setX((-2 * pPoint.getX() + 1) * pAspect); 
	vn.setY(( 2 * pPoint.getY() - 1));
	vn.setZ(1/tan(getZoom()/2.0 * M_PI/180.0));
	
	GVector3 dir = (getMatrix() * vn) - getViewpoint();
	
	GRay3 tRay;
	tRay.setPoint(getViewpoint());
	tRay.setVector(dir.unitize());
	return(tRay);
    }
    else {
	// NOTE: currently twist not supported.
	GFloat tScale = tan((GFloat)getZoom() * M_PI_2 / 90.0);
	
	GFloat dx = (2 * pPoint.getX() - 1) * tScale;
	GFloat dy = (2 * pPoint.getY() - 1) * tScale;
	if (pAspect > 1) dx *= pAspect;
	else dy /= pAspect;

	GVector3 tDir = (getRefpoint() - getViewpoint()).unitize();

	GPoint3 tPoint;
	tPoint.setX(getViewpoint().getX() + dx);
	if (fabs(tDir.getY()) + gEpsilon > 1.0) {
	    tPoint.setY(getViewpoint().getY());
	    tPoint.setZ(getViewpoint().getZ() + tDir.getY() * dy);
	} 
	else {
	    tPoint.setY(getViewpoint().getY() + dy);
	    tPoint.setZ(getViewpoint().getZ());
	}
	
	GRay3 tRay;
	tRay.setPoint(tPoint);
	tRay.setVector(tDir);
	return(tRay);
    }
}
