//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "render/grayligh.h"
#include "render/graytrac.h"
#include "render/graycam.h"

GAggregate<GRayLight>* gRayLights = 0;

GColorRGB GRayPointLight::shade(const GPoint3& pPoint, const GVector3& pNormal, const GVector3&, const GSurface &pSurface, GRayInterSector& pCamera) {
    if (!isOn()) return(GColorRGB(gBlack));

    // lighting model:
    // c = tf * (ca + cd + cs)

    // tf - is the slowest transmission factor
    // 0 means that there is no transmission

    // #### ca - ambient light ####
    // ca = 0

    // #### cd - diffuse light #### 
    // cd = sd * lc * (n * l) * la;
    // l - light source position vector
    // la - attenuation factor
    // sd - diffuse surface parameters

    // #### cs - specular light #### 
    // cs = ss * lc * (e * p)^se * la
    // ss - specular surface parameter
    // se - specular exponent

    GVector3 l = getPosition() - pPoint; 
    GFloat d = l.abs();
    GFloat la = 1.0 / (1 + 0.4 * d);
    l = l.unitize();

    GRay3 tRay;
    tRay.setPoint(pPoint);
    tRay.setVector(l);
    tRay.setTmin(gEpsilon);
    tRay.setTmax(d - gEpsilon);

    if (pNormal.dot(l) <= 0) return(GColorRGB(gBlack));

    GFloat tf = 1.0; 

    GOneInterSectionList tList;
    pCamera.intersect(tRay, tList);
    if (tList.getNumber()) tf = tList.getTransmission();

    // if anything blocks the ray totally:
    if (!tf) return(GColorRGB(gBlack));

    GColorRGB cd = pSurface.getDiffuse() * pNormal.dot(l);
    GFloat dot = (pPoint.unitize()).dot(l);
    GColorRGB cs(gBlack);
    if (dot > 0) cs = pSurface.getSpecular() * pow(dot, pSurface.getShininess());
    return(getColor() * tf * la * (cd /*+ cs*/));
}

GColorRGB GRayAmbientLight::shade(const GPoint3&, const GVector3&, const GVector3&, const GSurface& pSurface, GRayInterSector&) {
    if (!isOn()) return(GColorRGB(gBlack));
    else return(getColor() * pSurface.getAmbient());
}
