//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan  9 14:33:17 1996 - juhu@rz.tu-ilmenau.de
#include "render/graytrac.h"

const GSurface gDummySurface;

bool GBounds3::findMinMax(const GRay3& pRay, GFloat &rTmin, GFloat &rTmax) const {
    GFloat   tT;
    GPoint3  tOrg(pRay.getPoint());
    GVector3 tDir(pRay.getVector());
    
    rTmin = pRay.getTmin();
    rTmax = pRay.getTmax();
   
    // X-Axis 
    if (tDir.getX() < -gEpsilon) {   
	tT = (mMin.getX() - tOrg.getX()) / tDir.getX();
	if (tT < rTmin)		     
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
	
	tT = (mMax.getX() - tOrg.getX()) / tDir.getX();
	if (tT >= rTmin) {
	    if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    } else if (tDir.getX() > gEpsilon) {
	tT = (mMax.getX() - tOrg.getX()) / tDir.getX();
	if (tT < rTmin)
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
	tT = (mMin.getX() - tOrg.getX()) / tDir.getX();
	if (tT >= rTmin){
	 if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    } else if (tOrg.getX() < mMin.getX() || tOrg.getX() > mMax.getX())
      return (false);
   
    // Y-Axis 
    if (tDir.getY() < -gEpsilon) {
	tT = (mMin.getY() - tOrg.getY()) / tDir.getY();
	if (tT < rTmin)
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
	tT = (mMax.getY() - tOrg.getY()) / tDir.getY();
	if (tT >= rTmin){
	    if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    } else if (tDir.getY() > gEpsilon) {
	tT = (mMax.getY() - tOrg.getY()) / tDir.getY();
	if (tT < rTmin)
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
	tT = (mMin.getY() - tOrg.getY()) / tDir.getY();
	if (tT >= rTmin){
	    if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    } else if (tOrg.getY() < mMin.getY() || tOrg.getY() > mMax.getY())
      return (false);
   
    // Z-Axis 
    if (tDir.getZ() < -gEpsilon) {
	tT = (mMin.getZ() - tOrg.getZ()) / tDir.getZ();
	if (tT < rTmin)
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
      
	tT = (mMax.getZ() - tOrg.getZ()) / tDir.getZ();
	if (tT >= rTmin){
	    if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    } else if (tDir.getZ() > gEpsilon) {
      
	tT = (mMax.getZ() - tOrg.getZ()) / tDir.getZ();
	if (tT < rTmin)
	  return (false);
	if (tT <= rTmax)
	  rTmax = tT;
      
	tT = (mMin.getZ() - tOrg.getZ()) / tDir.getZ();
	if (tT >= rTmin){
	    if (tT > rTmax)
	      return (false);
	    rTmin = tT;
	}
    }else if (tOrg.getZ() < mMin.getZ() || tOrg.getZ() > mMax.getZ())
      return (false);
    
    return (true);
}
 
// TRANSFORM
GBounds3 operator *(const GMatrix4& pM, const GBounds3& pB) {
    GBounds3 tBox(GPoint3(MAXFLT, MAXFLT, MAXFLT), 
		  GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
    tBox +=(pM * GVector3(pB.mMin.getX(), pB.mMin.getY(), pB.mMin.getZ()));
    tBox +=(pM * GVector3(pB.mMin.getX(), pB.mMin.getY(), pB.mMax.getZ()));
    tBox +=(pM * GVector3(pB.mMin.getX(), pB.mMax.getY(), pB.mMin.getZ()));
    tBox +=(pM * GVector3(pB.mMin.getX(), pB.mMax.getY(), pB.mMax.getZ()));
    tBox +=(pM * GVector3(pB.mMax.getX(), pB.mMin.getY(), pB.mMin.getZ()));
    tBox +=(pM * GVector3(pB.mMax.getX(), pB.mMin.getY(), pB.mMax.getZ()));
    tBox +=(pM * GVector3(pB.mMax.getX(), pB.mMax.getY(), pB.mMin.getZ()));
    tBox +=(pM * GVector3(pB.mMax.getX(), pB.mMax.getY(), pB.mMax.getZ()));
    
    return(tBox);
}


ostream& operator<<(ostream& pStream, const GBounds3& pBounds) {
    pStream << '{' << pBounds.mMin << "  " << pBounds.mMax << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GRayTraceableBounds& pBounds) {
    return(pStream << pBounds.getBounds());
}

void GRayTraceableBounds::set(GFloat pMinx, GFloat pMiny, GFloat pMinz, GFloat pMaxx, GFloat pMaxy, GFloat pMaxz) {
    setBounds().setMin(GVector3(pMinx, pMiny, pMinz));
    setBounds().setMax(GVector3(pMaxx, pMaxy, pMaxz));
}

GFloat GRayTraceableBounds::getNearest(const GRay3& pRay, GRayTraceable*& rPrim) {
    bool hit = false;
    GFloat nt = HUGE;
    GFloat tt;
    GFloat u;

    if (pRay.getVector().getX() != 0.0) {
	tt = (getBounds().getMin().getX() - pRay.getPoint().getX())/pRay.getVector().getX();
	u = pRay.getPoint().getY() + pRay.getVector().getY() * tt;
	if (!(getBounds().getMin().getY() > u || getBounds().getMax().getY() < u)) {
	    u = pRay.getPoint().getZ() + pRay.getVector().getZ() * tt;
	    if (!(getBounds().getMin().getZ() > u || getBounds().getMax().getZ() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
	tt = (getBounds().getMax().getX() - pRay.getPoint().getX())/pRay.getVector().getX();
	u = pRay.getPoint().getY() + pRay.getVector().getY() * tt;
	if (!(getBounds().getMin().getY() > u || getBounds().getMax().getY() < u)) {
	    u = pRay.getPoint().getZ() + pRay.getVector().getZ() * tt;
	    if (!(getBounds().getMin().getZ() > u || getBounds().getMax().getZ() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
    }	

    if (pRay.getVector().getY() != 0.0) {
	tt = (getBounds().getMin().getY() - pRay.getPoint().getY())/pRay.getVector().getY();
	u = pRay.getPoint().getX() + pRay.getVector().getX() * tt;
	if (!(getBounds().getMin().getX() > u || getBounds().getMax().getX() < u)) {
	    u = pRay.getPoint().getZ() + pRay.getVector().getZ() * tt;
	    if (!(getBounds().getMin().getZ() > u || getBounds().getMax().getZ() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
	tt = (getBounds().getMax().getY() - pRay.getPoint().getY())/pRay.getVector().getY();
	u = pRay.getPoint().getX() + pRay.getVector().getX() * tt;
	if (!(getBounds().getMin().getX() > u || getBounds().getMax().getX() < u)) {
	    u = pRay.getPoint().getZ() + pRay.getVector().getZ() * tt;
	    if (!(getBounds().getMin().getZ() > u || getBounds().getMax().getZ() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
    }	
    
    if (pRay.getVector().getZ() != 0.0) {
	tt = (getBounds().getMin().getZ() - pRay.getPoint().getZ())/pRay.getVector().getZ();
	u = pRay.getPoint().getX() + pRay.getVector().getX() * tt;
	if (!(getBounds().getMin().getX() > u || getBounds().getMax().getX() < u)) {
	    u = pRay.getPoint().getY() + pRay.getVector().getY() * tt;
	    if (!(getBounds().getMin().getY() > u || getBounds().getMax().getY() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
	tt = (getBounds().getMax().getZ() - pRay.getPoint().getZ())/pRay.getVector().getZ();
	u = pRay.getPoint().getX() + pRay.getVector().getX() * tt;
	if (!(getBounds().getMin().getX() > u || getBounds().getMax().getX() < u)) {
	    u = pRay.getPoint().getY() + pRay.getVector().getY() * tt;
	    if (!(getBounds().getMin().getY() > u || getBounds().getMax().getY() < u)) {
		if (tt > gEpsilon && tt < nt) {
		    nt = tt;
		    hit = true;
		}
	    }
	}
    }

    if (hit) rPrim = this;
    else rPrim = 0;
    
    return(nt);
}

bool GRayTraceableBounds::intersect(const GRay3&, GInterSectionList&) {
    return(false);
}

GVector3 GRayTraceableBounds::getNormal(const GPoint3&) {
    GVector3 tVector;
    return(tVector);
}
//// The Quadric:

GRayTraceableQuadric::GRayTraceableQuadric(const GVector3& pQuadric, const GVector3& pMixed, const GVector3& pLinear, GFloat pConst):
mQuadric(pQuadric), mMixed(pMixed), mLinear(pLinear), mConst(pConst) {
    coefficients2matrix();
}

void GRayTraceableQuadric::coefficients2matrix(void) {
    mMat.set(0, 0, mQuadric.getX());
    mMat.set(0, 1, mMixed.getX()); 
    mMat.set(0, 2, mMixed.getY()); 
    mMat.set(0, 3, mLinear.getX()); 

    mMat.set(1, 0, mMixed.getX());
    mMat.set(1, 1, mQuadric.getY());
    mMat.set(1, 2, mMixed.getZ()); 
    mMat.set(1, 3, mLinear.getY()); 

    mMat.set(2, 0, mMixed.getY());
    mMat.set(2, 1, mMixed.getZ());
    mMat.set(2, 2, mQuadric.getZ());
    mMat.set(2, 3, mLinear.getZ()); 

    mMat.set(3, 0, mLinear.getX());
    mMat.set(3, 1, mLinear.getY());
    mMat.set(3, 2, mLinear.getZ());
    mMat.set(3, 3, mConst);
}

void GRayTraceableQuadric::matrix2coefficients(void) {
    mQuadric.setX(mMat.get(0, 0));
    mQuadric.setY(mMat.get(1, 1));
    mQuadric.setZ(mMat.get(2, 2));

    mMixed.setX(mMat.get(0, 1));
    mMixed.setY(mMat.get(0, 2));
    mMixed.setZ(mMat.get(1, 2));

    mLinear.setX(mMat.get(0, 3));
    mLinear.setY(mMat.get(1, 3));
    mLinear.setZ(mMat.get(2, 3));

    mConst = mMat.get(3, 3); 
}

void GRayTraceableQuadric::translate(const GVector3& pVec) {
    GMatrix4 tMat = gUnityMatrix4.translate(pVec);
    mMat = tMat * mMat * tMat.transpose();
    matrix2coefficients(); 
}

void GRayTraceableQuadric::rotate(const GVector3& pVec) {
    GMatrix4 tMat = gUnityMatrix4.rotate(pVec);
    mMat = tMat * mMat * tMat.transpose();
    matrix2coefficients(); 
}

void GRayTraceableQuadric::scale(const GVector3& pVec) {
    GMatrix4 tMat = gUnityMatrix4.scale(pVec);
    mMat = tMat * mMat * tMat.transpose();
    matrix2coefficients(); 
}

GFloat GRayTraceableQuadric::getNearest(const GRay3&, GRayTraceable*&) {
    cerr << "Quadric get Nearest should never be called." << endl;
    return(0.0);
}

bool GRayTraceableQuadric::intersect(const GRay3& pRay, GInterSectionList& pList) {
    GFloat nb = 2 * mMixed.getX();
    GFloat nc = 2 * mMixed.getY();
    GFloat nd = 2 * mLinear.getX();
    GFloat nf = 2 * mMixed.getZ();
    GFloat ng = 2 * mLinear.getY();
    GFloat ni = 2 * mLinear.getZ();
    
    GFloat aq = pRay.getVector().getX() * (pRay.getVector().getX() * mQuadric.getX() + pRay.getVector().getY() * nb + pRay.getVector().getZ() * nc)
	+ pRay.getVector().getY() * (pRay.getVector().getY() * mQuadric.getY() + pRay.getVector().getZ() * nf) 
	    + pRay.getVector().getZ() * pRay.getVector().getZ() * mQuadric.getZ();

    GFloat nbq = pRay.getVector().getX() * (pRay.getPoint().getX() * mQuadric.getX() + pRay.getPoint().getY() * mMixed.getX() + pRay.getPoint().getZ() * mMixed.getY() + mLinear.getX())
	+ pRay.getVector().getY() * (pRay.getPoint().getX() * mMixed.getX() + pRay.getPoint().getY() * mQuadric.getY() + pRay.getPoint().getZ() * mMixed.getZ() + mLinear.getY())
	    + pRay.getVector().getZ() * (pRay.getPoint().getX() * mMixed.getY() + pRay.getPoint().getY() * mMixed.getZ() + pRay.getPoint().getZ() * mQuadric.getZ() + mLinear.getZ());

    GFloat cq = pRay.getPoint().getX() * (pRay.getPoint().getX() * mQuadric.getX() + pRay.getPoint().getY() * nb + pRay.getPoint().getZ() * nc + nd)
	+ pRay.getPoint().getY() * (pRay.getPoint().getY() * mQuadric.getY() + pRay.getPoint().getZ() * nf + ng) 
	    + pRay.getPoint().getZ() * (mQuadric.getZ() * pRay.getPoint().getZ() + ni) + mConst;

    bool ret = false;
    if (aq == 0.0) {
	GFloat t = -2 * cq/nbq;
	if (pRay.isValid(t)) {
	    pList.add(GInterSection(self, t, false));
	    ret = true;
	}
    } 
    else {
	GFloat ka = -nbq/aq;
	GFloat df = ka * ka - cq/aq; 
	if (df < 0) return(false);
	GFloat sq = sqrt(df);
	GFloat t1 = ka - sq;
	GFloat t2 = ka + sq;


	if (pRay.isValid(t1)) {
	    pList.add(GInterSection(self, t1, true));
	    ret = true;
	}
	if (pRay.isValid(t2)) {
	    pList.add(GInterSection(self, t2, false));
	    ret = true;
	}
    }
    return(ret);
}

GVector3 GRayTraceableQuadric::getNormal(const GPoint3& pPoint) {
    GVector3 tVector;
    tVector.setX(mMat.get(0, 0) * pPoint.getX() + mMat.get(0, 1) * pPoint.getY() +
		 mMat.get(0, 2) * pPoint.getZ() + mMat.get(0, 3));

    tVector.setY(mMat.get(1, 0) * pPoint.getX() + mMat.get(1, 1) * pPoint.getY() +
		 mMat.get(1, 2) * pPoint.getZ() + mMat.get(1, 3));

    tVector.setZ(mMat.get(2, 0) * pPoint.getX() + mMat.get(2, 1) * pPoint.getY() +
		 mMat.get(2, 2) * pPoint.getZ() + mMat.get(2, 3));

    return(tVector.unitize());
}

bool GInterSection::operator < (const GInterSection& pIS) const {
    return(mT < pIS.mT); 
}

void GOneInterSectionList::add(const GInterSection& pInter) {
    GFloat tFloat = pInter.getPrimitive().getTransmission();
    
    mTransmission =  mTransmission * tFloat;

    if (!mHit) {
	mHit = true;
	mInter = pInter;
	return;
    }
    
    if (pInter < mInter) mInter = pInter; 
}

void GAllInterSectionList::add(const GInterSection& pInter) {
    GHeap<GInterSection>::insert(pInter);
    GFloat tFloat = pInter.getPrimitive().getTransmission();
    mTransmission =  mTransmission * tFloat;

}



GRayTraceablePolygon::GRayTraceablePolygon(void): mPolygon(0), mPolySegs(0), mPolyValid(false) {
    geometryChanged();
}

GRayTraceablePolygon::GRayTraceablePolygon(const GVertexAggregate& pVertices): GPolygon(pVertices),  mPolygon(0), mPolySegs(0), mPolyValid(false) {
    geometryChanged();
    
}

bool gIntersectXY(const GPoint3& p, const GPoint3& s, const GPoint3& r) {
    if (r.getX() == 0.0) return(false);
    GFloat b = (p.getX() - s.getX())/r.getX();
    if (b <= 0.0 || b > 1.0) return(false);
    if ((s.getY() + (b * r.getY()) - p.getY()) < 0.0) return(false);
    return(true);
}

bool gIntersectXZ(const GPoint3& p, const GPoint3& s, const GPoint3& r) {
    if (r.getX() == 0.0) return(false);
    GFloat b = (p.getX() - s.getX())/r.getX();
    if (b <= 0.0 || b > 1.0) return(false);
    if ((s.getZ() + (b * r.getZ()) - p.getZ()) < 0.0) return(false);
    return(true);
}

bool gIntersectYZ(const GPoint3& p, const GPoint3& s, const GPoint3& r) {
    if (r.getY() == 0.0) return(false);
    GFloat b = (p.getY() - s.getY())/r.getY();
    if (b <= 0.0 || b > 1.0) return(false);
    if ((s.getZ() + (b * r.getZ()) - p.getZ()) < 0.0) return(false);
    return(true);
}

void GRayTraceablePolygon::geometryChanged(void) {
    if (mPolygon) {
	delete [] mPolygon; 
	mPolygon = 0;
    }
    mPolyValid = false;
    mPolySegs = getNumber();
    if (mPolySegs < 3) {
	mPolySegs = 0;
	return;
    }
    mPolygon = new GPoint3[mPolySegs];
    GPoint3* v = new GPoint3[mPolySegs];
    v[0] = getVertex(0).getPoint();
    int i = 1;
    while (i < mPolySegs) {
	mPolygon[i - 1] = (v[i] = getVertex(i).getPoint()) - v[i - 1];
	i++;
    }
    mPolygon[mPolySegs - 1] = v[0] - v[mPolySegs - 1];

    if (GPolygon::getNormal().abs() < gEpsilon) 
	setNormal(gCalcNormal(mPolySegs, v));
    GFloat nabs = GPolygon::getNormal().abs();
    delete [] v;
    // use a point in the plane to compute distance from 0:
    mDist = GPolygon::getNormal().dot(getVertex(0).getPoint()) / nabs ;
//    setNormal(GPolygon::getNormal() * (1/nabs));
    ifunc = &gIntersectYZ;
    if (fabs(GPolygon::getNormal().getY()) > fabs(GPolygon::getNormal().getX()))
	ifunc = &gIntersectXZ;
    if (fabs(GPolygon::getNormal().getZ()) > fabs(GPolygon::getNormal().getX())
	&& fabs(GPolygon::getNormal().getZ()) > fabs(GPolygon::getNormal().getY()))
	ifunc = &gIntersectXY;
    
    // compute the bounding box in local coord. system
    setBounds() = GBounds3(GPoint3(MAXFLT, MAXFLT, MAXFLT), 
			   GPoint3(-MAXFLT, -MAXFLT, -MAXFLT));
    for (i=0; i< mPolySegs; i++) 
	setBounds() += getVertex(i).getPoint();
    // adjust the bound 
    GVector3 tDiff(getBounds().min() - getBounds().max());
    for (i = 0; i < 3; i++)
	if (fabs(tDiff[i]) < gEpsilon) {
	    setBounds().min()[i] -= gEpsilon;
	    setBounds().max()[i] += gEpsilon;
	}
    
    mPolyValid = true;
}

template<class T>
class GIntersectPolygonIterator: public GIterator<T> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Intersect the polygon :-)
    // 
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES: 
    // 
    // CONSTANTS: 
  public:
    GIntersectPolygonIterator(int pNumber, GIFuncPtr pFun, const GPoint3& pPoint, GPoint3* pPolygon):
    mSteps(0), mNumber(pNumber), mHits(0), mFun(pFun), mPoint(pPoint), mPolygon(pPolygon) {}
    virtual bool exec(T& t) {
	if (mFun(mPoint, t.getPoint(), *mPolygon)) mHits++;
	mPolygon++;
	if (++mSteps == mNumber) return(true);
	else return(false);
    }
    int getNumberOfHits(void) const {
	return(mHits);
    }
  private:
    int mSteps;
    int mNumber;
    int mHits;
    GIFuncPtr mFun;
    GPoint3 mPoint;
    GPoint3* mPolygon;
};

bool GRayTraceablePolygon::intersect(const GRay3& pRay, GInterSectionList& pList) {
    if (!mPolyValid) return(false);
    // the distance of ray from the plane:
    GFloat t = GPolygon::getNormal().dot(pRay.getVector());
    if (fabs(t) < gEpsilon) return(false);
    t = (mDist - GPolygon::getNormal().dot(pRay.getPoint()))/t;
    if (t < 0.0) return(false);

    GIntersectPolygonIterator<GVertex> tIter(mPolySegs, ifunc, pRay.compute(t), mPolygon);
    getVertexList().doWithElements(tIter);
    if (tIter.getNumberOfHits() & 1) 
	if (pRay.isValid(t)) {
	    pList.add(GInterSection(self, t, GPolygon::getNormal().dot(pRay.getVector()) < 0));
	    return(true);
	}
    return(false);
}

GRayTraceablePolygon::~GRayTraceablePolygon(void) {
    if (mPolygon) delete [] mPolygon;
}
    
