//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifdef GWINDOWS
#include <windows.h>
#else
#include <stdlib.h>
#endif

#include "render/gsolidt.h"

const gMaxNoise = 20;
int ***GSolidTexture::noiseTable;

GSolidTexture::GSolidTexture(void): smode(G_REPLACE_SAMPLE), nmode(G_NO_NORMAL) {
    static bool sFirst = true;
    if (!sFirst) return;
    sFirst = false;

    noiseTable = new int**[gMaxNoise];
    assert(noiseTable);

    int x;
    for (x = 0; x < gMaxNoise; x++) {
	noiseTable[x] = new int*[gMaxNoise];
	assert(noiseTable[x]);
    }
    for (x = 0; x < gMaxNoise; x++) {
	for (int y = 0; y < gMaxNoise; y++) {
	    noiseTable[x][y] = new int[gMaxNoise];
	    assert(noiseTable[x][y]);
	}
    }
    
    srand(1);
    int xx, yy, zz;
    for(x = 0; x < gMaxNoise; x++)
	for(int y = 0; y < gMaxNoise; y++)
   	    for(int z = 0; z < gMaxNoise; z++) {
		noiseTable[x][y][z] = rand() % 10000;
		xx = (x == (gMaxNoise - 1)) ? 0 : x;
		yy = (y == (gMaxNoise - 1)) ? 0 : y;
		zz = (z == (gMaxNoise - 1)) ? 0 : z;
		noiseTable[x][y][z] = noiseTable[xx][yy][zz];
	    }
}

GFloat frac(GFloat r) {
    r -= (int)r; 
    return((r<0) ? r + 1 : r);
}

GFloat GSolidTexture::getNoise(const GPoint3& p) const {
    GFloat n0, n1, n00, n01, n10, n11;  

    GFloat a = frac(p.getX()) * (gMaxNoise - 1); 
    GFloat b = frac(p.getY()) * (gMaxNoise - 1); 
    GFloat c = frac(p.getZ()) * (gMaxNoise - 1); 

    int nx = (int)a;
    int ny = (int)b;
    int nz = (int)c;

    int nxi = nx + 1; if (nxi == gMaxNoise) nxi = 0;
    int nyi = ny + 1; if (nyi == gMaxNoise) nyi = 0;
    int nzi = nz + 1; if (nzi == gMaxNoise) nzi = 0;

    a = frac(a);
    b = frac(b);
    c = frac(c);
    
    a = a * a * (3 - 2 * a);
    b = b * b * (3 - 2 * b);
    c = c * c * (3 - 2 * c);

    int n  = noiseTable[nx][ny][nz];
    n00 = n + a * (noiseTable[nxi][ny][nz] - n);
    n   = noiseTable[nx][ny][nzi];
    n01 = n + a * (noiseTable[nxi][ny][nzi] - n);
    n   = noiseTable[nx][nyi][nz];
    n10 = n + a * (noiseTable[nxi][nyi][nz] - n);
    n   = noiseTable[nx][nyi][nzi];
    n11 = n + a * (noiseTable[nxi][nyi][nzi] - n);
    
    n0  = n00 + b * (n10 - n00);
    n1  = n01 + b * (n11 - n01);
    
    return((n0 + c * (n1 - n0))/10000.0);
}

GColorRGB GSolidTexture::getRGBNoise(const GPoint3& p, GFloat range) const {
    GFloat n = range * (1 - 2 * getNoise(p));  
    return(GColorRGB(n, n, n));
}

GFloat GSolidTexture::getTurbulence(const GPoint3& p, int iter) const {
    int i;
    GFloat pow = 2;
    GFloat sum = 0;
    
    for (i = 0; i < iter; i++, pow *= 2)
	sum += fabs(getNoise(p * pow))/pow;
    return(sum);
}

// utilitys:

int base(GFloat r) {
    r += gEpsilon;
    int n = (int)r;
    return (n>r) ? n-1 : n; 
}

GFloat dsqr(GFloat d) { return(d * d); }

GFloat dmod(GFloat n, GFloat p) {
    assert(p!=0);
    return(n - base(n/p) * p);
}

GColorRGB GPolyBlobTexture::getSample(const GPoint3& p) const {
    GFloat r = ((GFloat)(int)(getNoise(mMatrix * p) *4))/4;
 
    if (r < 0.25) return(GColorRGB(1, 0, 0));  // red
    if (r < 0.5) return(GColorRGB(0, 1, 0));  // green
    if (r < 0.75) return(GColorRGB(0, 0, 1));  // blue
    return(GColorRGB(1, 1, 0));  // yellow
}

GColorRGB GNoiseTexture::getSample(const GPoint3& p) const {
    return(col * getNoise(mMatrix * p));
}

GColorRGB GWoodTexture::getSample(const GPoint3& p) const {
    GPoint3 q = mMatrix * p;
    GFloat radius  = sqrt(dsqr(q.getX()) + dsqr(q.getZ()));
    radius -= sin (2 * M_PI * radius / 2) / 6;
    radius += 0.1 * getNoise(GPoint3(q.getX()/3, q.getY(), q.getZ()));
    
    GFloat grain = dmod(radius, 0.3) / 0.3; 
    grain = xlight ? dsqr(grain) : sqrt(grain);  
    
    return(tex1->getSample(p) * (1 - grain) + tex0->getSample(p) * (grain));
};

GColorRGB GCheckerTexture::getSample(const GPoint3& p) const {
    GPoint3 q = mMatrix * p;
    int checker = 
	fabs(base(q.getX())%2) == fabs(base(q.getY())%2) == fabs(base(q.getZ())%2);
    return(checker ? tex0->getSample(p) : tex1->getSample(p));
};

GColorRGB GGridTexture::getSample(const GPoint3& pnt) const {
    GPoint3 p = mMatrix * pnt;
    GPoint3 q = p + GPoint3(getNoise(p*0.1), getNoise(p*0.1), getNoise(p*0.1)) * 2;
    
    int x = (frac(q.getX()) < xwidth);
    int y = (frac(q.getY()) < xwidth);
    int z = (frac(q.getZ()) < xwidth);
    
    if (x || y || z) return(getRGBNoise(q , 0.2) + tex0->getSample(pnt));
    return(getRGBNoise(q , 0.2) + tex1->getSample(pnt));
};

GColorRGB GTurbulenceTexture::getSample(const GPoint3& p) const {
    GFloat d = getTurbulence(mMatrix * p, xiter);
    if (d < xlevel) return tex0->getSample(p);
    d = (d - xlevel) / (1 - xlevel) * 0.6 + 0.4;
    return(tex1->getSample(p) * d);
};

GColorRGB GInterpolationTexture::getSample(const GPoint3& p) const {
    GFloat r = getNoise(mMatrix * p);
    r = (GFloat)(((int)(r*2*(xsteps-1)) + 1) / 2) / (xsteps-1);
    return(tex0->getSample(p) * (1-r) + tex1->getSample(p) * r);
};

GColorRGB GTruncTexture::getSample(const GPoint3& p) const {
    GColorRGB c = tex->getSample(p);
    c.setR(base(c.getR() * xsteps) / (GFloat)xsteps);
    c.setG(base(c.getG() * xsteps) / (GFloat)xsteps);
    c.setB(base(c.getB() * xsteps) / (GFloat)xsteps);
    return(c);
};

GColorRGB GMarbleTexture::getSample(const GPoint3& p) const {
    GPoint3 q = mMatrix * p;
    GFloat d = frac(q.getX() + getNoise(q * 0.33)) * w[3];
    GColorRGB n = getRGBNoise(q, 0.2);

    if (d < w[0]) return(tex1->getSample(p) + n);
    if (d < w[1]) return((tex1->getSample(p) * (w[1]-d) + tex0->getSample(p) *  (d-w[0])) * ((GFloat)1.0/(w[1]-w[0])) + n);
    if (d < w[2]) return(tex0->getSample(p) + n);

    return((tex1->getSample(p) * (d-w[2]) + tex0->getSample(p) * (w[3]-d)) * ((GFloat)1.0/(w[3]-w[2])) + n);
};

GColorRGB GXLinearMarbleTexture::getSample(const GPoint3& p) const {
    GPoint3 q = mMatrix * p;
    GFloat d = frac(q.getX() + getNoise(q*0.63) + 0.3*getNoise(q*3)) * w[6];
    GColorRGB n = getRGBNoise(q,0.1);

    if(d < w[0]) return(tex0->getSample(p) + n);
    if(d < w[1]) return((tex0->getSample(p)*(w[1]-d) + tex1->getSample(p)*(d-w[0])) * ((GFloat)1.0/(w[1]-w[0])) + n);
    if(d < w[2]) return(tex1->getSample(p) + n);
    if(d < w[3]) return((tex1->getSample(p)*(w[3]-d) + tex2->getSample(p) * (d-w[2])) * ((GFloat)1.0/(w[3]-w[2])) + n);
    if(d < w[4]) return((tex2->getSample(p) *(w[4]-d) + tex3->getSample(p)*(d-w[3])) * ((GFloat)1.0/(w[4]-w[3])) + n);
    if(d < w[5]) return(tex3->getSample(p) + n);
    return((tex3->getSample(p)*(w[6]-d) + tex0->getSample(p) * (d-w[5])) * ((GFloat)1.0/(w[6]-w[5])) + n);
};


GColorRGB GXCubicMarbleTexture::getSample(const GPoint3& p) const {
    // Hermite interpolation of the colors c[0]-c[4] to the interval [0,w4]
    GPoint3 q = mMatrix * p;
    GFloat   d = frac(q.getX() + getNoise(q*0.63) + 0.3*getNoise(q*3)) * w[4];
    GColorRGB n(getRGBNoise(q,0.2));
    GFloat dd;

    if (d<w[0]) dd = 0;
    else if (d<w[1]) dd = (d-w[0])/(w[1]-w[0]);
    else if (d<w[2]) dd = (d-w[1])/(w[2]-w[1]);
    else if (d<w[3]) dd = (d-w[2])/(w[3]-w[2]);
    else dd = (d-w[3])/(w[4]-w[3]);

    dd=dd*dd*(3-2*dd);

    if (d<w[0]) return(tex0->getSample(p) + n);
    if (d<w[1]) return(tex1->getSample(p) * dd + tex0->getSample(p) * (1-dd) + n);
    if (d<w[2]) return(tex2->getSample(p) * dd + tex1->getSample(p) * (1-dd) + n);
    if (d<w[3]) return(tex3->getSample(p) * dd + tex2->getSample(p) * (1-dd) + n);
    return(tex0->getSample(p) * dd + tex3->getSample(p) * (1-dd) + n);
};

GVector3 GLatticeTexture::getNormal(const GPoint3& p1) const {
    GPoint3 p = mMatrix * p1;
    GFloat dx = frac(p.getX()); 
    GFloat dy = frac(p.getY()); 
    GPoint3 bm = GPoint3(0,0,1);
    if (dx < 0.1) { if (dx < dy && (1.0 - dy) > dx ) bm.setX(1); }
    else if (dx > 0.9) { if (dx > dy && (1.0 - dy) < dx )  bm.setX(-1); }
    if (dy < 0.1) { if (dy < dx && (1.0 - dx) > dy) bm.setY(1); }
    else if (dy > 0.9) { if (dy > dx && (1.0 - dx) < dy) bm.setY(-1); }
    return(bm);
}

