//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gcolor.h"

ostream& operator<<(ostream& pStream, const GColorRGB& pColor) {
    pStream << '{' << pColor.mR << ' ' << pColor.mG << ' ' << pColor.mB << '}';
    return(pStream);
}

void GColorRGB::correct(void) {
    if (mR < 0) mR = 0; if (mR > 1) mR = 1;
    if (mG < 0) mG = 0; if (mG > 1) mG = 1;
    if (mB < 0) mB = 0; if (mB > 1) mB = 1;
}

void GColorRGB::setHSL(const GFloat hsl[3]) {
    GFloat h = hsl[0];
    GFloat s = hsl[1];
    GFloat l = hsl[2];
    GFloat v = (l <= 0.5) ? (l * (1.0 + s)) : (l + s - l * s);
    if (v <= 0) { mR = mG = mB = 0.0; }
    else { 
	GFloat m = l + l - v;
	GFloat sv = (v - m ) / v;
	h *= 6.0;
	int sextant = (int)h;    
	GFloat fract = h - sextant;
	GFloat vsf = v * sv * fract;
	GFloat mid1 = m + vsf;
	GFloat mid2 = v - vsf;
	switch (sextant) {
	  case 0: mR = v; mG = mid1; mB = m; break;
	  case 1: mR = mid2; mG = v; mB = m; break;
	  case 2: mR = m; mG = v; mB = mid1; break;
	  case 3: mR = m; mG = mid2; mB = v; break;
	  case 4: mR = mid1; mG = m; mB = v; break;
	  case 5: mR = v; mG = m; mB = mid2; break;
	}
    }
}

void GColorRGB::getHSL(GFloat hsl[3]) const {
    GFloat& h = hsl[0];
    GFloat& s = hsl[1];
    GFloat& l = hsl[2];
    GFloat v = gmax3(mR, mG, mB);
    GFloat m = gmin3(mR, mG, mB);
    if ((l = (m + v) / 2.0) <= 0.0) return;
    GFloat vm;
    if ((s = vm = v - m) > 0.0) {
	s /= (l <= 0.5) ? (v + m ) : (2.0 - v - m) ;
    } else return;

    GFloat r2 = (v - mR) / vm;
    GFloat g2 = (v - mG) / vm;
    GFloat b2 = (v - mB) / vm;

    if (mR == v) h = (mG == m ? 5.0 + b2 : 1.0 - g2);
    else if (mG == v) h = (mB == m ? 1.0 + r2 : 3.0 - b2);
    else h = (mR == m ? 3.0 + g2 : 5.0 - r2);
    h /= 6;
}

bool GColorRGB::operator==(const GColorRGB& s) const {
    if (mR != s.mR) return(false);
    if (mG != s.mG) return(false);
    if (mB != s.mB) return(false);
    return(true);
}

GColorRGB::operator unsigned long(void) const {
    unsigned long l = ((long)(255 * mR)) +
	((long)(255 * mG) << 8) +
	    ((long)(255 * mB) << 16);
    return(l);
}

const GColorRGB gBlack(0.0, 0.0, 0.0);
const GColorRGB gWhite(1.0, 1.0, 1.0);

