//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/ggenmat.h"

ostream& operator<<(ostream& pStream, const GMatrix& pMatrix) {
    pStream << '{';
    for(register int i = 0; i < pMatrix.Rows(); i++)
	if(i < pMatrix.Rows()-1)
	    pStream << pMatrix.getRow(i) << ' ';
        else
	    pStream << pMatrix.getRow(i);
    return(pStream << '}');
}

istream& operator>>(istream& iStream, GMatrix& pMatrix) {
    char p;
    GVector v(pMatrix.Cols());

    if(iStream >> p)
	if(p != '{') return(iStream);

    for(register int i = 0; i < pMatrix.Rows(); i++) {
	iStream >> v;
	pMatrix.setRow(i,v);
    }

    if (iStream >> p)
	if(p != '}') return(iStream);

    return(iStream);
}

GFloat& GMatrix::elem(int i, int j) const {
    assert(i >= 0 && i < rows && j >= 0 && j < cols);
    return(v[i]->v[j]); 
}

void GMatrix::flip_rows(int i,int j) {
    GVector* p = v[i];
    v[i] = v[j];
    v[j] = p;
}

void GMatrix::newGeometry(int dim1, int dim2) {
    assert( dim1 > 0 && dim2 > 0);
    if (v) {   
	while(rows--) delete v[rows]; 
	delete v;
    }
	    
    rows = dim1; 
    cols = dim2; 
	    
    v = new GVector*[rows];
    for (register int i = 0; i < rows; i++) v[i] = new GVector(cols); 
}

GMatrix::GMatrix(int dim1, int dim2) {
    assert( dim1 > 0 && dim2 > 0);
    rows = dim1; 
    cols = dim2; 
	    
    v = new GVector*[rows];
    for (register int i = 0; i < rows; i++) v[i] = new GVector(cols);
}

GMatrix::GMatrix(const GMatrix& p) { 
    rows = p.rows;
    cols = p.cols;
    
    if (rows > 0 && cols > 0) {
	v = new GVector*[rows];
  
	for (register int i = 0; i < rows; i++) v[i] = new GVector(*p.v[i]);
    }
    else v = 0;
}

GMatrix::GMatrix(int dim1, int dim2, GFloat** p) {
    rows = dim1; 
    cols = dim2; 

    v = new GVector*[rows];

    for(register int i = 0; i < rows; i++) { 
	v[i] = new GVector(cols);
	for(register int j = 0; j < cols; j++) elem(i,j) = p[i][j];
    }
}

/*
GMatrix::GMatrix(int nr,const GVector3 *vec) {
    if(vec) {
	cols = 3;
	rows = nr;

	v = new GVector*[rows];
	for(int i=0; i<rows; i++) 
	    v[i] = new GVector(vec[i]); 
    }
    else
	{
	    v=nil;
	    rows=0;
	    cols=0;
	}
}
*/

GMatrix::GMatrix(int nr,const GVector *vec) {
    if(vec) {
	cols = vec[0].dim();
	rows = nr;

	v = new GVector*[rows];
	for(register int i = 0; i < rows; i++) v[i] = new GVector(vec[i]); 
    }
    else
	{
	    v = 0;
	    rows = 0;
	    cols = 0;
	}
}

GMatrix::GMatrix(const GMatrix4 &mat) {
    register int i,j;
    rows = cols = 4;

    v = new GVector*[rows];
    for( i = 0; i < rows; i++) v[i] = new GVector(cols);

    for( i = 0; i < rows; i++)
	for(j = 0; j < cols; j++)
	    elem(i,j)=mat.get(i,j);
}

GMatrix::GMatrix(const GVector& vec) {
    rows = vec.d;
    cols = 1;
    v = new GVector*[rows];
    for(register int i = 0; i < rows; i++) {
	v[i] = new GVector(1);
	elem(i,0) = vec[i];
    }
}

GMatrix::~GMatrix() {
    if (v) {
	while(rows--) delete v[rows]; 
	delete v;
	v = 0;
    }
}

void GMatrix::check_dimensions(const GMatrix& mat) const { 
    assert(rows == mat.rows || cols == mat.cols);
}

GMatrix& GMatrix::operator=(const GMatrix& mat) {
    register int i,j;
    
    if (rows != mat.rows || cols != mat.cols) {
	for(i = 0; i < rows; i++) delete v[i];
	delete v;
	rows = mat.rows;
	cols = mat.cols;
	v = new GVector*[rows];
	for(i = 0; i < rows; i++) v[i] = new GVector(cols);
    }

    for(i = 0; i < rows; i++)
	for(j = 0; j < cols; j++) elem(i,j) = mat.elem(i,j);

    return *this;
}

bool GMatrix::operator==(const GMatrix& x) const {
    register int i,j;

    if (rows != x.rows || cols != x.cols) return false;

    for(i = 0; i < rows; i++)
	for(j = 0; j < cols; j++)
	    if (elem(i,j) != x.elem(i,j)) return false;
    
    return true;
}

GVector GMatrix::getRow(int i) const { 
    assert( i >= 0 && i < rows );
    return *v[i];
}

void GMatrix::setRow(int i,const GVector &vec) {
    assert( i >= 0 && i < rows && vec.d == cols );
    *v[i] = vec;
}

GFloat& GMatrix::operator()(int i, int j) {
    assert( i >= 0 && i < rows && j >= 0 && j < cols);
    return elem(i,j);
}

GFloat GMatrix::operator()(int i, int j) const {
    assert( i >= 0 && i < rows && j >= 0 && j < cols);
    return elem(i,j);
}

GVector GMatrix::getCol(int i) const {
    assert( i >= 0 && i < cols );
    
    GVector result(rows);
    register int j = rows;
    while (j--) result.v[j] = elem(j,i);
    return result;
}

void  GMatrix::setCol(int i,const GVector &vec) {
    assert(vec.d == rows);

    int j = rows;
    while (j--) elem(j,i) = vec.v[j]; 
}

GMatrix::operator GMatrix4() const {
    GMatrix4 ret;
    assert(cols == 4 && rows == 4);

    for(register int i = 0; i < rows; i++)
	for(register int j = 0; j < cols; j ++)
	    ret.set(i,j,elem(i,j));

    return ret;
}

GMatrix::operator GVector() const {
    assert(cols == 1 || rows == 1);

    if (cols == 1) return getCol(0);
    if (rows == 1) return getRow(0);

    return GVector();
}

GMatrix GMatrix::operator+(const GMatrix& mat) {
    register int i,j;

    check_dimensions(mat);

    GMatrix result(rows,cols);
    for(i = 0; i < rows; i++)
	for( j = 0; j < cols; j++)
	    result.elem(i,j) = elem(i,j) + mat.elem(i,j);

    return result;
}

GMatrix& GMatrix::operator+=(const GMatrix& mat) {
    register int i,j;

    check_dimensions(mat);

    for(i = 0; i < rows; i++)
	for(j = 0; j < cols; j++)
	    elem(i,j) += mat.elem(i,j);

    return *this;
}

GMatrix& GMatrix::operator-=(const GMatrix& mat) {
    register int i,j;

    check_dimensions(mat);

    for(i = 0; i < rows; i++)
	for(j = 0; j < cols; j++)
	    elem(i,j) -= mat.elem(i,j);

    return *this;
}

GMatrix GMatrix::operator-(const GMatrix& mat) {
    register int i,j;

    check_dimensions(mat);

    GMatrix result(rows,cols);
    for(i = 0; i < rows; i++)
	for(j = 0; j < cols;j++)
	    result.elem(i,j) = elem(i,j) - mat.elem(i,j);

    return result;
}

GMatrix GMatrix::operator-() {
    register int i,j;

    GMatrix result(rows,cols);
    for(i = 0; i < rows; i++)
	for(j = 0; j < cols; j++)
	    result.elem(i,j) = -elem(i,j);

    return result;
}

GMatrix GMatrix::operator*(const GFloat& f) {
    register int i,j;

    GMatrix result(rows,cols);
    for(i = 0;i < rows; i++)
	for(j = 0;j < cols; j++)
	    result.elem(i,j) = elem(i,j) * f;

    return(result);
}

GVector GMatrix::operator*(const GVector& vec) {
    GMatrix tMat(vec);
    return(GVector((*this) * tMat));
}

GMatrix GMatrix::operator*(const GMatrix& mat) const {
    GMatrix result(rows, mat.cols);
     
    assert(cols == mat.rows);
    
    register int i,j;
    for (i = 0; i < mat.cols; i++)
	for (j = 0; j < rows; j++) result.elem(j,i) = *v[j] * mat.getCol(i);
    
    return result;
}

GFloat GMatrix::det() const {

    assert(rows == cols);

    int n = rows;
    GMatrix M(n,1);

    int flips;
    GFloat** A = triang(M,flips);

    if (A == NULL) return 0;

    GFloat Det = 1;

    for(register int i = 0; i < n; i++) Det *= A[i][i];

    for(i = 0; i < n; i++) delete A[i];
    delete A;

    return (flips % 2) ? -Det : Det;
}

GFloat** GMatrix::triang(const GMatrix& M, int& flips) const {
    register GFloat **p, **q;
    register GFloat *l, *r, *s;

    register GFloat pivot_el,tmp;
    
    register int i,j,col,row;

    int n = rows;
    int d = M.cols;
    int m = n+d;

    GFloat** A = new GFloat*[n];
    
    p = A;

    for(i = 0; i < n; i++) {
	*p = new GFloat[m];
	l = *p++;
	for(j = 0; j < n; j++) *l++ = elem(i,j);
	for(j = 0; j < d; j++) *l++ = M.elem(i,j);
    }

    flips = 0;

    for(col = 0, row = 0; row < n; row++, col++) { 
	// search for row j with maximal absolute entry in current col
	j = row;
	for(i = row + 1; i < n; i++)
	    if (fabs(A[j][col]) < fabs(A[i][col])) j = i; // IIIT WAS HEERE...fabs()
	
	if ( n > j && j > row) { 
	    GFloat* p = A[j];
	    A[j] = A[row];
	    A[row] = p;
	    flips++;
	}

	tmp = A[row][col];
	q  = &A[row];

	if (fabs(tmp) < gEpsilon) {
	    // cerr << "Matrix has not full rank!\n"; 
	    p = A;
	    for(i = 0;i < n; i++) delete *p++;
	    delete A;
	    return NULL;
	}
	
	for(p = &A[n-1]; p != q; p--) { 
	    l = *p+col;
	    s = *p+m;	
	    r = *q+col;
	    
	    if (*l != 0.0) {
		pivot_el = *l/tmp;
		while(l < s) *l++ -= *r++ * pivot_el;
	    }
	}
    }
    return A;
}

GMatrix GMatrix::inverse(bool& isSingular) const {
    assert(rows == cols);

    isSingular = false;

    int n = rows;
    GMatrix I(n,n);
    for (register int i = 0; i < n; i++) I(i,i) = 1;
    return solve(I, isSingular);
}

GMatrix GMatrix::solve(const GMatrix& M, bool& isSingular) const {
    assert(rows == cols && rows == M.rows);

    isSingular = false;

    register GFloat **p, ** q;
    register GFloat *l, *r, *s;

    int      n = rows;
    int      d = M.cols;
    int      m = n+d;
    int      row, col,i;

    GFloat** A = triang(M,i);
    
    if(!A) {
	isSingular = true;

	return GMatrix(n, d);
    }

    for (col = n-1, p = &A[n-1]; col>=0; p--, col--) { 
	s = *p+m;
	GFloat tmp = (*p)[col];
	for (l = *p + n; l < s; l++) *l /=tmp;
	    
	for(q = A; q != p; q++ ) {
	    tmp = (*q)[col];
	    l = *q+n;
	    r = *p+n;
	    while (r < s)  *l++ -= *r++ * tmp;
	}
    }
    
    GMatrix result(n, d);
    
    for (row = 0; row < n; row++) {
	l = A[row]+n;
	for (col = 0; col < d; col++) result.elem(row,col) = *l++;
	delete A[row];
    }
    
    delete A;
    
    return result;
}

GMatrix GMatrix::transpose() const {
    GMatrix result(cols,rows);
    for (register int i = 0; i < cols; i++)
	for (register int j = 0; j < rows; j++)
	    result.elem(i,j) = elem(j,i);

    return result;
}



