//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Dec  5 14:06:32 1995 by paul(@prakinf.tu-ilmenau.de)
#include "util/ggenvec.h"

GVector::GVector(int n) { 
    d = n; 
    v = 0;

    if (d > 0) { 
	v = new GFloat[d];
	assert(v);
	while (n--) v[n] = 0.0;
    }
}

GVector::GVector(const GVector& p) {
    d = p.d; 
    v = 0;
    if (d > 0) {
	v = new GFloat[d];
	assert(v);
	for(int i=0; i<d; i++) v[i] = p.v[i];
    }
}

GVector::GVector(GFloat x, GFloat y, GFloat z) {
    v = new GFloat [3];
    d = 3;
    v[0] = x;
    v[1] = y;
    v[2] = z;
}

GVector::GVector(const GVector3& vec) {
    v = new GFloat [3];
    d = 3;
    v[0] = vec.getX();
    v[1] = vec.getY();
    v[2] = vec.getZ();
}
GVector::GVector(const GVector2& vec) {
    v = new GFloat [2];
    d = 2;
    v[0] = vec.getX();
    v[1] = vec.getY();
}


void GVector::newGeometry(int i) {
    if (v)
        delete[] v;
    v = 0;
    
    d = i;
    if (d > 0) { 
	v = new GFloat[i];
	assert(v);
	while (i--) v[i] = 0.0;
    }
}

GFloat GVector::operator[](int i) const {
    assert(i > 0 || i < d); 
    return v[i];
}

GFloat& GVector::operator[](int i) {
    assert(i > 0 || i < d);
    return v[i]; 
}


GVector& GVector::operator+=(const GVector& vec) {
    assert(check_dimensions(vec));
    register int n = d;
    while (n--) v[n] += vec.v[n];
    return *this;
}

GVector& GVector::operator-=(const GVector& vec) {
    assert(check_dimensions(vec));
    register int n = d;
    while (n--) v[n] -= vec.v[n];
    return *this;
}

GVector GVector::operator+(const GVector& vec) const {
    assert(check_dimensions(vec));
    register int n = d;
    GVector result(n);
    while (n--) result.v[n] = v[n]+vec.v[n];
    return result;
}

GVector GVector::operator-(const GVector& vec) const {
    assert(check_dimensions(vec));
    register int n = d;
    GVector result(n);
    while (n--) result.v[n] = v[n]-vec.v[n];
    return result;
}

GVector GVector::operator-() const { // unary minus
    register int n = d;
    GVector result(n);
    while (n--) result.v[n] = -v[n];
    return result;
}

GVector GVector::operator*(GFloat x) const {
    int n = d;
    GVector result(n);
    while (n--) result.v[n] = v[n] * x;
    return result;
}

GVector GVector::operator/(GFloat x) const {
    int n = d;
    GVector result(n);
    while (n--) result.v[n] = v[n] / x;
    return result;
}

GVector operator*(GFloat f, const GVector& v) { return v*f; } 

GFloat GVector::operator*(const GVector& vec) const {
    assert(check_dimensions(vec));
    GFloat result=0;
    register int n = d;
    while (n--) result = result+v[n]*vec.v[n];
    return result;
}

GVector& GVector::operator=(const GVector& vec) {
    register int n = vec.d;
    
    if (n != d) { 
	if (v) { delete[] v; }
	v = new GFloat[n];    
	d = n;
    }
    while (n--) v[n] = vec.v[n];
    return *this;
}


bool GVector::operator==(const GVector& vec)  const {
    if (vec.d != d) return false;
    int i = 0;
    while ((i<d) && (v[i]==vec.v[i])) i++;
    return (i==d) ? true : false;
}

GFloat GVector::max() {
    register int i;
    GFloat ret = v[0];

    for( i = 0; i < d; i++)
        if(v[i] > ret) ret=v[i];

    return(ret);
}

GFloat GVector::min() {
    register int i;
    GFloat ret=v[0];

    for( i = 0; i < d; i++)
        if(v[i] < ret) ret = v[i];

    return(ret);
}

// friends:
ostream& operator<<(ostream& pStream, const GVector& pVector) {
    pStream << '{';

    for( int i = 0; i < pVector.dim(); i++ )
	if( i < pVector.dim()-1 )
	    pStream << pVector[i] << ' ';
	else
	    pStream << pVector[i];
    
    pStream << '}';

    return(pStream);
}

istream& operator>>(istream& iStream, GVector& pVector) {
    char p;

    if(iStream >> p)
	if(p != '{') return iStream;

    for(register int i = 0; i < pVector.dim(); i++ )
	iStream >> pVector[i];

    if (iStream >> p)
	if(p != '}') return iStream;

    return iStream;
}

GFloat angle(const GVector& vec1, const GVector& vec2) {
    GFloat r1 = vec1.abs();
    GFloat r2 = vec2.abs();
    
    assert(r1 > gEpsilon || r2 > gEpsilon );
    return  acos(vec1*vec2/(r1*r2));  
}

bool compare(const GVector& v1, const GVector& v2) { 
    register int i=0;

    assert(v1.check_dimensions(v2));    
    
    for( i = 0; i < v1.dim() && v1[i]==v2[i]; i++);
    if (i == v1.dim()) 
	return true;
    else
	return false;
}

GFloat chord(const GVector &v1,const GVector &v2) {
    register int i;
    GFloat res = 0;
 
    assert(v1.check_dimensions(v2));
    
    for(i = 0;i < v1.dim();i++ )
        res+=sqr(v1[i]-v2[i]);

    return sqrt(res);
}

GVector cross(const GVector& vec1,const GVector& vec2) {
    assert(vec1.d==3 && vec2.d==3);
    
    GVector result(vec1.v[1]*vec2.v[2]-vec1.v[2]*vec2.v[1],
		   vec1.v[2]*vec2.v[0]-vec1.v[0]*vec2.v[2],
		   vec1.v[0]*vec2.v[1]-vec1.v[1]*vec2.v[0]);
    return result;
}
