//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gident.h"

const GString& GIdentification::getIdentification(void) const {
    static GString sName("<internal object>");
    return(mName ? *mName : sName);
}

bool GIdentification::identificationExists(const char *pName) {
    return(getObjectByIdentification(pName) ? true : false);
}

template<class I> 
class GFindIdentificationIterator: public GIterator<I> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Find an object by identification.
    // 
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GFindIdentificationIterator(const char* pName): mName(pName), mIdent(0) {}

    bool exec(I& i) {
	if (!strcmp(i.getIdentification(), mName)) {
	    mIdent = &i;
	    return(true);
	}
	else return(false);
    }
    I* getObject(void) const { return(mIdent); }
  private:
    const char* mName;
    I* mIdent;
};

GIdentification* GIdentification::getObjectByIdentification(const char* pName) {
    if (!pName) return(0);
    GFindIdentificationIterator<GIdentification> tIter(pName);
    sList.doWithElements(tIter);
    return(tIter.getObject());
}

GIdentification::GIdentification(const char* pName): mName(0) {
    if (pName) {
	mName = new GString(pName);
	sList.append(this);
    }
}

GIdentification::~GIdentification(void) {
    if (mName) {
	sList.remove(this);
	delete mName;
    }
}

GDLList<GIdentification> GIdentification::sList(false);
