//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gparam.h"

const int gMaxPoints = 10;
// constant to set maximal extraction of a curve or surface segment

const GCurveVertex GCurveTopology::getCurveVertexOn(GFloat) {
    const static GCurveVertex sVertex;
    return(sVertex);
}

const GVector& GCurveTopology::getParameterVector(void) {
    const static GVector sVector;
    return(sVector);
}

// parametric curve constants and methods:
GCurveTopology::GCurveTopology(void): GTopology1D(), mResolution(0), mCurv(0) {}

GCurveTopology::GCurveTopology(int pNumber): GTopology1D(), mCurv(0) {
    for(int i = 0; i < pNumber; i++)
	setVertex(i, new GCurveVertex);
}

GCurveTopology::GCurveTopology(const GCurveVertexAggregate& pPoints): GTopology1D(), mCurv(0) {
    for(int i = 0; i < pPoints.getNumber(); i++)
	setVertex(i, new GCurveVertex(pPoints[i]));
}

void GCurveTopology::setVertex(int i, GCurveVertex* pVertex) {
    GTopology1D::setVertex(i, pVertex);
}

void GCurveTopology::setCurveVertex(int i, GCurveVertex pVertex) {
    mCurv->setElement(i, pVertex);
}

void GCurveTopology::newGeometry(int pDim) {
    if(mCurv) delete mCurv; mCurv = 0;
    // delete old elements if necessary
    
    mCurv = new GCurveVertexVector(pDim);
}

void GCurveTopology::setVertex(int i, const GCurveVertex& pVertex) {
    setVertex(i, new GCurveVertex(pVertex));

    resolutionChanged();
    geometryChanged();
    // change geometry of the surface
}

const GCurveVertex& GCurveTopology::getVertex(int i) const {
    return (const GCurveVertex&)GTopology1D::getVertex(i);
}

void GCurveTopology::resolutionChanged() {
    int dim = 
	(int)((gMaxPoints*getSegments() - getSegments())*
	      getResolutionValue()+getSegments());

    int minD = dim, maxD = dim;
    
    while(maxD%getSegments()) maxD++;
    while(minD%getSegments()) minD--;
    
    dim = (maxD - dim) > (dim - minD) ? minD : maxD;

    newGeometry(dim+1);
    // set new dimension to result vertices list
}

void GCurveTopology::computeNormals() {
// not implemented either
/*
    for(int i = 0; i < getNumber(); i++)
	mCurv->getElement(i).setNormal(new GVector3);
*/
};

const GVector& GSurfaceTopology::getParameterVectorU(void) {
    const static GVector sVector;
    return(sVector);
}

const GVector& GSurfaceTopology::getParameterVectorW(void) {
    const static GVector sVector;
    return(sVector);
}
// get parameter values in u or w direction
        
const GSurfaceVertex GSurfaceTopology::getSurfaceVertexOn(GFloat, GFloat) {
    GSurfaceVertex tVertex;
    return(tVertex);
}

// Parametric surface constants and methods:
GSurfaceTopology::GSurfaceTopology(int x, int y): GTopology2D(x, y), mSurf(0) {}
// create a topology with given dimensions

GSurfaceTopology::GSurfaceTopology(int x, int y, const GSurfaceVertexAggregate& pVertices):
GTopology2D(x, y), mSurf(0) {
    for(int i = 0; i < getX(); i++)
	for(int j = 0; j < getY(); j++)
	    setVertex(i, j, new GSurfaceVertex(pVertices.getElement(i+getX()*j)));
}

void GSurfaceTopology::setSurfaceVertex(int pU, int pW, GSurfaceVertex pVertex) {
    mSurf->setElement(pU+getSurfX()*pW, pVertex);
}
// set the vertex value on the indices pU, pW

void GSurfaceTopology::setVertex(int x, int y, GSurfaceVertex* pVertex) {
    GTopology2D::setVertex(x, y, pVertex);
}
// direct access

void GSurfaceTopology::newGeometry(void) {
    if(mSurf) { delete mSurf; mSurf = 0; }
    // delete old elements if necessary
}

void GSurfaceTopology::newGeometry(int newDimX, int newDimY) {
    newGeometry();
    // delete old elements
	
    mSurf = new GSurfaceVertexVector(newDimX*newDimY);
    dimSurfX = newDimX;
    dimSurfY = newDimY;
}
// set new geometry to the surface, 
// has to be called always when the surface changes

void GSurfaceTopology::setVertices(const GSurfaceVertexAggregate& pVertices) {
    for(int i = 0; i < getX(); i++)
	for(int j = 0; j < getY(); j++)
	    setVertex(i, j, new GSurfaceVertex(pVertices.getElement(i+getX()*j)));

    geometryChanged();
}

void GSurfaceTopology::setVertex(int x, int y, const GSurfaceVertex& pVertex) {
    setVertex(x, y, new GSurfaceVertex(pVertex));

    geometryChanged();
}

const GSurfaceVertex& GSurfaceTopology::getVertex(int i, int j) const {
    return (GSurfaceVertex&)GTopology2D::getVertex(i, j);
}

const GSurfaceVertex& GSurfaceTopology::getSurfaceVertex(int pU, int pW) const {
    return mSurf->getElement(pU+getSurfX()*pW);
}
// get vertex on surface with this indices

void GSurfaceTopology::resolutionChanged() {
    int dimU =
	(int)((gMaxPoints*getSegmentsX() - getSegmentsX())*
	      getResolutionValue()+getSegmentsX());
    int dimW = 
	(int)((gMaxPoints*getSegmentsY() - getSegmentsY())*
	      getResolutionValue()+getSegmentsY());
       
    int minU = dimU, maxU = dimU;
    int minW = dimW, maxW = dimW;

    while(maxU%getSegmentsX()) maxU++;
    while(minU%getSegmentsX()) minU--;
    
    while(maxW%getSegmentsY()) maxW++;
    while(minW%getSegmentsY()) minW--;

    dimU = (maxU - dimU) > (dimU - minU) ? minU : maxU;
    dimW = (maxW - dimW) > (dimW - minW) ? minW : maxW;

    newGeometry(dimU+1, dimW+1);
    // set new dimension to result vertices list
}

void GSurfaceTopology::computeNormals() {
    for(int i = 0; i < getSurfX(); i++)
	for(int j = 0; j < getSurfY(); j++) {
	    const GVector3* tTangU = getSurfaceVertex(i, j).getTangentU();
	    const GVector3* tTangW = getSurfaceVertex(i, j).getTangentW();
	    
	    if(tTangU->cross(*tTangW).abs() > gEpsilon) {
		mSurf->getElement(i+getSurfX()*j).setNormal
		    (new GVector3(tTangU->cross(*tTangW).unitize()));
		    
		continue;
	    }

	    mSurf->getElement(i+getSurfX()*j).setNormal(new GVector3);
	}
}
