//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gparam2D.h"

// for check purpose only:
//#include "input/gtime.h"

// approximating spline surface:
// create spline surface with default settings
GApproxSplineSurface::GApproxSplineSurface(int pX, int pY, const GResolution& pResol, const GSplineMode& modeU, const GSplineMode& modeW): GSurfaceTopology(pX, pY), uSpline(GSplineValues(pX, 0, pX >= 3 ? 3 : 1, modeU, G_CLAMPED)), wSpline(GSplineValues(pY, 0, pY >= 3 ? 3 : 1, modeW, G_CLAMPED)) {
    setResolution(pResol);
}

GApproxSplineSurface::GApproxSplineSurface(int pX, int pY, const GSurfaceVertexAggregate& pVertices, const GResolution& pResol, const GSplineMode& modeU, const GSplineMode& modeW): GSurfaceTopology(pX, pY, pVertices), uSpline(GSplineValues(pX, 0, pX >= 3 ? 3 : 1, modeU, G_CLAMPED)), wSpline(GSplineValues(pY, 0, pY >= 3 ? 3 : 1, modeW, G_CLAMPED)) {
    setResolution(pResol);
}

void GApproxSplineSurface::setOrderU(int ordVal) {
    uSpline.setSpline().order = ordVal;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineSurface::setOrderW(int ordVal) {
    wSpline.setSpline().order = ordVal;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineSurface::setModeU(const GSplineMode& modeVal) {
    uSpline.setSpline().mode = modeVal;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineSurface::setModeW(const GSplineMode& modeVal) {
    wSpline.setSpline().mode = modeVal;

    resolutionChanged();
    geometryChanged();
}

void GApproxSplineSurface::setEndU(const GSplineEnd& endVal) {
    uSpline.setSpline().end = endVal;
    geometryChanged();
}

void GApproxSplineSurface::setEndW(const GSplineEnd& endVal) {
    wSpline.setSpline().end = endVal;
    geometryChanged();
}

void GApproxSplineSurface::geometryChanged() {
    if(!checkVertices()) return;

    int i,j,u,w;

    if(uSpline.getSpline().numVal != getX())
	uSpline.setSpline().numVal = getX();
    
    if(wSpline.getSpline().numVal != getY())
	wSpline.setSpline().numVal = getY();

    if(uSpline.getSpline().dim != getSurfX())
	uSpline.setSpline().dim = getSurfX();
    
    if(wSpline.getSpline().dim != getSurfY())
	wSpline.setSpline().dim = getSurfY();
    // set new spline values

    GMatrix Nu = uSpline.getBaseMatrix();
    GMatrix Nw = wSpline.getBaseMatrix();

    GMatrix Nu1 = uSpline.getBaseMatrix1();
    GMatrix Nw1 = wSpline.getBaseMatrix1();

    for( u = 0; u < getSurfX(); u++)
	for( w = 0; w < getSurfY(); w++ ) {
	    GVector3 surfPoint, surfTangU, surfTangW;
	    
	    for( i = 0; i < getX(); i++ ) 
                for( j = 0; j < getY(); j++ ) {
		    surfPoint +=
			getVertex(i,j).getPoint() * Nu(u, i) * Nw(w, j);
		    surfTangU +=
			getVertex(i,j).getPoint() * Nu1(u, i) * Nw(w, j);
		    surfTangW +=
			getVertex(i,j).getPoint() * Nu(u, i) * Nw1(w, j);
		}
	    
	    setSurfaceVertex
		(u, w, GSurfaceVertex(surfPoint, surfTangU, surfTangW,
				      uSpline.getParVector()[u], wSpline.getParVector()[w]));
	}

    computeNormals();
}

const GVector& GApproxSplineSurface::getParameterVectorU(void) {
    return uSpline.getParVector();
}

const GVector& GApproxSplineSurface::getParameterVectorW(void) {
    return wSpline.getParVector();
}

const GSurfaceVertex GApproxSplineSurface::getSurfaceVertexOn(GFloat pU, GFloat pW) {
    GVector3 tPoint, tTangU, tTangW, tNorm;
    GVector tTransU, tTransU1, tTransW, tTransW1;

    tTransU = uSpline.getBaseTranslation(pU);
    tTransU1 = uSpline.getBaseTranslation1(pU);

    tTransW = wSpline.getBaseTranslation(pW);
    tTransW1 = wSpline.getBaseTranslation1(pW);

    for(int i = 0; i < getX(); i++ ) 
	for(int j = 0; j < getY(); j++ ) {
	    tPoint +=
		getVertex(i,j).getPoint() * tTransU[i] * tTransW[j];
	    tTangU +=
		getVertex(i,j).getPoint() * tTransU1[i] * tTransW[j];
	    tTangW +=
		getVertex(i,j).getPoint() * tTransU[i] * tTransW1[j];
	}
	    
    tNorm = tTangU.cross(tTangW);
    // compute normal
	    
    return GSurfaceVertex(tPoint, tNorm, tTangU, tTangW, pU, pW);
}

// fitting surface mthods:
// create spline surface with default settings
GFittingSplineSurface::GFittingSplineSurface(int pX, int pY, const GResolution& pResol, const GSplineMode& modeU, const GSplineMode& modeW): GSurfaceTopology(pX, pY), uSpline(GSplineValues(pX, 0, 4, modeU, G_FREE)), wSpline(GSplineValues(pY, 0, 4, modeW, G_FREE)), mControlNet(0) {
    setResolution(pResol);
}

GFittingSplineSurface::GFittingSplineSurface(int pX, int pY, const GSurfaceVertexAggregate& pVertices, const GResolution& pResol, const GSplineMode& modeU, const GSplineMode& modeW): GSurfaceTopology(pX, pY, pVertices), uSpline(GSplineValues(pX, 0, 4, modeU, G_FREE)), wSpline(GSplineValues(pY, 0, 4, modeW, G_FREE)), mControlNet(0) {
    setResolution(pResol);
}

void GFittingSplineSurface::setModeU(const GSplineMode& modeVal) {
    uSpline.setMode(modeVal);

    resolutionChanged();
    geometryChanged();
}

void GFittingSplineSurface::setModeW(const GSplineMode& modeVal) {
    wSpline.setMode(modeVal);

    resolutionChanged();
    geometryChanged();
}

void GFittingSplineSurface::computePhantomTopology() {
    int i,j,u,w;

    if(uSpline.getDimension() != getSurfX())
	uSpline.setDimension(getSurfX());

    if(wSpline.getDimension() != getSurfY())
	wSpline.setDimension(getSurfY());

    // pre-compute the values of the control mesh
    // in each column (co-ordinate w in object space):
    GMatrix *wNet = new GMatrix[getX()];

    for(u = 0; u < getX(); u++) {
	wNet[u].newGeometry(getY(), 3);

	for(w = 0; w < getY(); w++) 
	    wNet[u].setRow(w, getVertex(u, w).getPoint());
	
	wSpline.setValuesMatrix(wNet[u]);

	// set up the u-th matrix with the control 
	// mesh values for the column u in object space:
	wNet[u] = wSpline.getVertexMatrix();
    }

    int newDimY = wSpline.getVertexMatrix().Rows();

    // compute values of entire control mesh
    // based on pre computed values for the columns
    // stored in matrices W:
    GMatrix *uNet = new GMatrix[newDimY];

    for(w = 0; w < newDimY; w++) {
	uNet[w].newGeometry(getX(), 3);

	for(u = 0; u < getX(); u++)
	    uNet[w].setRow(u, wNet[u].getRow(w));
	
	uSpline.setValuesMatrix(uNet[w]);
	// set up the w-th matrix with the control 
	// mesh values for the row w in object space:
	uNet[w] = uSpline.getVertexMatrix();
    }
    
    int newDimX =  uSpline.getVertexMatrix().Rows();

    // resize the control mesh:
    if(mControlNet) delete mControlNet;
    mControlNet = 0;
    
    mControlNet = new GPhantomSurfaceTopology(newDimX, newDimY);

    for(i = 0; i < newDimX; i++)
	for(j = 0; j < newDimY; j++) {
	    mControlNet->setVertex(i, j, GSurfaceVertex(uNet[j].getRow(i)));
	}

    delete [] uNet; delete [] wNet;    
}

void GFittingSplineSurface::geometryChanged() {
    // GTime tTime;
    if(!checkVertices()) return;

    // compute the "phantom" net for the surface:
    computePhantomTopology();

    for(int u = 0; u < getSurfX(); u++)
	for(int w = 0; w < getSurfY(); w++ )
	    setSurfaceVertex
		(u, w, getSurfaceVertexOn(uSpline.getParVector()[u], wSpline.getParVector()[w]));

    computeNormals();
    // cout << "construction time: " << GTime::getTime() - tTime << " secs" << endl;
}

const GVector& GFittingSplineSurface::getParameterVectorU(void) {
    return uSpline.getParVector();
}

const GVector& GFittingSplineSurface::getParameterVectorW(void) {
    return wSpline.getParVector();
}

const GSurfaceVertex GFittingSplineSurface::getSurfaceVertexOn(GFloat pU, GFloat pW) {
    GVector3 tPoint, tTangU, tTangW, tNorm;
    GVector tTransU, tTransU1, tTransW, tTransW1;

    tTransU = uSpline.getBaseTranslation(pU);
    tTransU1 = uSpline.getBaseTranslation1(pU);

    tTransW = wSpline.getBaseTranslation(pW);
    tTransW1 = wSpline.getBaseTranslation1(pW);

    for(int i = 0; i < mControlNet->getX(); i++ ) 
	for(int j = 0; j < mControlNet->getY(); j++ ) {
	    tPoint +=
		mControlNet->getVertex(i, j).getPoint()*tTransU[i]*tTransW[j];
	    tTangU +=
		mControlNet->getVertex(i, j).getPoint()*tTransU1[i]*tTransW[j];
	    tTangW +=
		mControlNet->getVertex(i, j).getPoint()*tTransU[i]*tTransW1[j];
	}
	    
    tNorm = tTangU.cross(tTangW);
    // compute normal
	    
    return GSurfaceVertex(tPoint, tNorm, tTangU, tTangW, pU, pW);
}
