//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gparam3D.h"

const int gNumberOfFacets = 4;

GCurve2SurfaceTopology::GCurve2SurfaceTopology(const GCurveVertexAggregate& pLine, 
const GResolution& pResol, const GSplineMode& pModeU, const GSplineMode& pModeW): 
GPhantomCurveTopology(pLine), mSurf(0), mModeU(pModeU), mModeW(pModeW) {}

void GCurve2SurfaceTopology::newGeometry(int newDimX, int newDimY) {
    if(mSurf) delete mSurf; mSurf = 0;
    mSurf = new GFittingSplineSurface
	(newDimX, newDimY, getResolution(), mModeU, mModeW);
}

GRotationSurfaceTopology::GRotationSurfaceTopology(const GVector3& pAxis, 
const GCurveVertexAggregate& pLine, const GResolution& pResol, 
const GSplineMode& pMode): 
GCurve2SurfaceTopology(pLine, pResol, G_CLOSED, pMode), mAxis(pAxis) {
    recompute();
    setResolution(pResol);
}

void GRotationSurfaceTopology::resolutionChanged() {
    int newDimX, newDimY;
    
    newDimX = gNumberOfFacets;
    newDimY = getNumber();
    
    newGeometry(newDimX, newDimY);
}

GFloat GRotationSurfaceTopology::getDistanceFromAxis(const GPoint3& pPoint) {
    return((pPoint.cross(mAxis)).abs()/mAxis.abs());
}

void GRotationSurfaceTopology::recompute() {
    mB.unity();
    mC.unity();
    
    GVector3 d(mAxis); d = d.unitize();

    GFloat dX, dY, dZ, v;
    // turn vector about z-axis to get it to plane x = 0
    d.project(dX, dY, POS_Z);
    if(gabs(dX) > gEpsilon) {
	v = sqr(dX) + sqr(dY);
	// cout << "Z-rotation: " << 360*acos(dY/sqrt(v))/(2*M_PI) << endl;
	mB = mB.rotateZ(acos(dY/sqrt(v)));
    }

    mBInv = mB.inverse(); 
    
    // turn vector about x-axis to get it to z = 0, x = 0 plane
    d.project(dY, dZ, POS_X);
    if(gabs(dZ) > gEpsilon) {
	v = sqr(dY) + sqr(dZ);
	// cout << "X-rotation: " << acos(dZ/sqrt(v)) << endl;
	mC = mC.rotateX(asin(dZ/sqrt(v)));
    }
    
    mCInv = mC.inverse();
};

GPoint3 GRotationSurfaceTopology::wc2rc(const GPoint3& pPoint, GFloat pAng) {
    mR.unity(); mR = mR.rotateY(pAng);  
    return (mB*mC*mR*mBInv*mCInv)*pPoint;
}

void GRotationSurfaceTopology::geometryChanged() {
    if(getX() != gNumberOfFacets || getY() != getNumber())
	newGeometry(gNumberOfFacets, getNumber());

    GFloat dAngle = 2*M_PI/getX();

    for(int j = 0; j < getY(); j++)
	for(int i = 0; i < getX(); i++) {
	    setNetVertex(i, j, GSurfaceVertex
			 (wc2rc(getVertex(j).getPoint(), (GFloat)i*dAngle)));
	}

    // change the geometry of the part surface finally:
    GCurve2SurfaceTopology::geometryChanged();
}

GSweepSurfaceTopology::GSweepSurfaceTopology(GFloat pLength, const GVector3& pSweepVector, 
const GCurveVertexAggregate& pLine, const GResolution& pResol, const GSplineMode& pMode): 
GCurve2SurfaceTopology(pLine, pResol, pMode, G_OPEN), mLength(pLength), mSweepVector(pSweepVector.unitize()) {
    setResolution(pResol);
}

void GSweepSurfaceTopology::resolutionChanged() {
    int newDimX, newDimY;
    
    newDimX = getNumber();
    newDimY = 2;

    newGeometry(newDimX, newDimY);
}

void GSweepSurfaceTopology::geometryChanged() {
    if(getX() != getNumber() && getY() != 2)
	newGeometry(getNumber(), 2);
    
    for(int j = 0; j < getY(); j++)
	for(int i = 0; i < getX(); i++) {
	    setNetVertex(i, j, GSurfaceVertex
			 (getVertex(i).getPoint()+(j*1.0)*mLength*mSweepVector));
	}
    // change the geometry of the part surface finally:
    GCurve2SurfaceTopology::geometryChanged();
}
    


