//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Wed Feb  7 10:21:14 1996 by paul(@prakinf.tu-ilmenau.de)
#include "util/gspline.h"
#include "goutput.h"

// GSpline methods:
int GSpline::getSegments() const {
    switch(values.mode) {
      case G_OPEN:
	return values.numVal-1-values.order+2;
      case G_CLOSED:
	return values.numVal;
      case G_NO_MODE:
	return 0;
      default:
	return 0;
    }
}

GSpline::GSpline(void): values(0, 0, 0, G_OPEN, G_CLAMPED), changed(false) {}

GSpline::GSpline(const GSplineValues& v): values(v), changed(true) {}

/*
GVector GSpline::compParVector() {
    GVector tmp(values.dim);
    GFloat tMax = 1.0*getSegments();
    int k = 0;

    for(int i = 0; i < tmp.dim(); i++) {
	tmp[i] = i*tMax/(1.0*(values.dim-1));
	
	// get next bigger knot value
	while(getPrivKnotVector()[k] <= tmp[i]) k++;

	while(tmp[i] < getPrivKnotVector()[k]) {
	    // get next parameter value
	    GFloat nextT = (i+1)*tMax/(1.0*(values.dim-1));

	    if(nextT >= getPrivKnotVector()[k]) {
		if(!(i+1 < tmp.dim())) break;
		tmp[i+1] = getPrivKnotVector()[k];
		i++;
	    }
	    else {
		if(!(i+1 < tmp.dim())) break;
		tmp[i+1] = nextT;
		i++;
	    }
	}
    }

    if(values.mode == G_OPEN && values.end == G_CLAMPED) {
	tmp[tmp.dim()-1] -= gEpsilon;
	tmp[0] += gEpsilon;
    }
    
    return tmp;
}
*/
    
GVector GSpline::compParVector() {
    GVector tmp(values.dim);
    for(int i = 0; i < tmp.dim(); i++)
        tmp[i] =
	    i*1.0*getSegments()/(values.dim-1);

    if(values.mode == G_OPEN && values.end == G_CLAMPED) {
        tmp[tmp.dim()-1] -= gEpsilon;
        tmp[0] += gEpsilon;
    }
    
    return tmp;
}

GFloat GSpline::blend(int j,int f,GFloat t) {
    GFloat value;
    if( f == 1 ) {
	if(K[j] <= t && t < K[j + 1])
	    value = 1;
	else
	    value = 0;
    }
    else {
	// check for divide by zero
	if(K[j+f-1] == K[j] && K[j+f] == K[j+1])
	    value = 0.0;
	else
	    if(K[j+f-1] == K[j]) // if denominator = 0, get deeper...
		value = 
		    (GFloat)( (K[j+f]-t)/(K[j+f]-K[j+1])*blend(j+1,f-1,t) );
	    else
		if(K[j+f] == K[j+1])
		    value =
			(GFloat)( ( (t-K[j])/(K[j+f-1]-K[j]) )*blend(j,f-1,t) );
		else
		    value =
			(GFloat)( (t-K[j])/(K[j+f-1]-K[j])*blend(j,f-1,t)
				 + (K[j+f]-t)/(K[j+f]-K[j+1])*blend(j+1,f-1,t) );
    }
    return value;
}

GFloat GSpline::blend1(int j,int f,GFloat t) {
    GFloat value;
    if( f == 1 ) 
	value = 0;
    else {
	// check for divide by zero
	if(K[j+f-1] == K[j] && K[j+f] == K[j+1])
	    value = 0.0;   
	else
	    // if denominator = 0, get deeper...  
	    if(K[j+f-1] == K[j])
		value = 
		    -blend(j+1,f-1,t)/(K[j+f]-K[j+1])+( (K[j+f]-t)/(K[j+f]-K[j+1]) )*
			blend1(j+1,f-1,t);
	    else
		if(K[j+f] == K[j+1])
		    value =
			blend(j,f-1,t)/(K[j+f-1]-K[j])+( (t-K[j])/(K[j+f-1]-K[j]) )*
			    blend1(j,f-1,t);
		else
		    value = 
			-blend(j+1,f-1,t)/(K[j+f]-K[j+1])+( (K[j+f]-t)/(K[j+f]-K[j+1]) )*
			    blend1(j+1,f-1,t)+
				blend(j,f-1,t)/(K[j+f-1]-K[j])+
				    ( (t-K[j])/(K[j+f-1]-K[j]) )*blend1(j,f-1,t);
    }
    return value;
}

GFloat GSpline::blend2(int j,int f,GFloat t) {
    GFloat value;
    if(f == 1 || f == 2) 
	value = 0;
    else {
	// check for divide by zero
	if(K[j+f-1] == K[j] && K[j+f] == K[j+1])
	    value = 0.0;
	else
	    if(K[j+f-1] == K[j]) // if denominator = 0, get deeper...
		value = 
		    -2*blend1(j+1,f-1,t)/(K[j+f]-K[j+1])+
			( (K[j+f]-t)/(K[j+f]-K[j+1]) )*blend2(j+1,f-1,t);
	    else
		if(K[j+f] == K[j+1])
		    value = 
			2*blend1(j,f-1,t)/(K[j+f-1]-K[j])+
			    ( (t-K[j])/(K[j+f-1]-K[j]) )*blend2(j,f-1,t);
	        else
		    value = 
			2*blend1(j,f-1,t)/(K[j+f-1]-K[j])+
			    ( (t-K[j])/(K[j+f-1]-K[j]) )*blend2(j,f-1,t)+
				-2*blend1(j+1,f-1,t)/(K[j+f]-K[j+1])+
				    ( (K[j+f]-t)/(K[j+f]-K[j+1]) )*blend2(j+1,f-1,t);
    }
    return value;
}

void GSpline::update() {
    if(isChanged()) {
	setKnotVector(compKnotVector());
	setParVector(compParVector());
	
	setChanged() = false;

	mBase.isUpToDate() = false;
	mBase1.isUpToDate() = false;
	mBase2.isUpToDate() = false;
    }
}

void GSpline::updateMatrices() {    
    if(mBase.isWanted())
	if(!mBase.isUpToDate()) {
	    compBase();
	    mBase.isUpToDate() = true;
	}
    
    if(mBase1.isWanted())
	if(!mBase1.isUpToDate()) {
	    compBase1();
	    mBase1.isUpToDate() = true;
	}

    if(mBase2.isWanted())
	if(!mBase2.isUpToDate()) {
	    compBase2();
	    mBase.isUpToDate() = true;
	}
}

GVector GSpline::getBaseTranslation(GFloat pT) {
    GVector tVector;
    // update values if necessary
    update();

    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, k;
	tVector.newGeometry(K.dim() - values.order);

	if(pT < getPrivParVector()[0] || pT > getPrivParVector()[getPrivParVector().dim()-1])
	    return tVector;

	// set up vector:
	for(i = 0, k = 0; i < tVector.dim(); i++, k++ )
	    if(K[k] <= pT && pT < K[k + values.order])
		tVector[i] = blend(i, values.order, pT);
	    else
		continue;
    }

    return tVector;
}

GVector GSpline::getBaseTranslation1(GFloat pT) {
    GVector tVector;
    // update values if necessary
    update();

    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, k;
	tVector.newGeometry(K.dim() - values.order);

	if(pT < getPrivParVector()[0] || pT > getPrivParVector()[getPrivParVector().dim()-1])
	    return tVector;

	// set up vector:
	for(i = 0, k = 0; i < tVector.dim(); i++, k++ )
	    if(K[k] <= pT && pT < K[k + values.order])
		tVector[i] = blend1(i, values.order, pT);
	    else
		continue;
    }

    return tVector;
}

GVector GSpline::getBaseTranslation2(GFloat pT) {
    GVector tVector;
    // update values if necessary
    update();

    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, k;
	tVector.newGeometry(K.dim() - values.order);

	if(pT < getPrivParVector()[0] || pT > getPrivParVector()[getPrivParVector().dim()-1])
	    return tVector;

	// set up vector:
	for(i = 0, k = 0; i < tVector.dim(); i++, k++ )
	    if(K[k] <= pT && pT < K[k + values.order])
		tVector[i] = blend2(i, values.order, pT);
	    else
		continue;
    }

    return tVector;
}

GMatrix& GSpline::getBaseMatrix() {
    update();

    mBase.isWanted() = true;
    updateMatrices();
    mBase.isWanted() = false;

    return mBase.mMatrix;
}

void GSpline::compBase() {    
    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, j, k;
	// set dimension of matrix:
	mBase.mMatrix.newGeometry(T.dim(), K.dim()-values.order);
	
	// set up matrix:
	for(i = 0; i < mBase.mMatrix.Rows(); i++) 
	    for(j = 0, k = 0; j < mBase.mMatrix.Cols(); j++,k++ )
		if(K[k]<= T[i] && T[i] < K[k+values.order])
		    mBase.mMatrix(i,j) = blend(j, values.order, T[i] );
		else
		    continue;
    }
}

GMatrix& GSpline::getBaseMatrix1() {
    update();

    mBase1.isWanted() = true;
    updateMatrices();
    mBase1.isWanted() = false;
    return mBase1.mMatrix;
}

void GSpline::compBase1() {    
    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, j, k;
	// set dimension of matrix:
	mBase1.mMatrix.newGeometry(T.dim(), K.dim()-values.order);
	
	// set up matrix:
	for(i = 0; i < mBase1.mMatrix.Rows(); i++) 
	    for(j = 0, k = 0; j < mBase1.mMatrix.Cols(); j++,k++ )
		if(K[k]<= T[i] && T[i] < K[k+values.order])
		    mBase1.mMatrix(i,j) = blend1(j, values.order, T[i] );
		else
		    continue;
    }
}

GMatrix& GSpline::getBaseMatrix2() {
    update();

    mBase2.isWanted() = true;
    updateMatrices();
    mBase2.isWanted() = false;
    return mBase2.mMatrix;
}

void GSpline::compBase2() {    
    if(values.order <= 0 || values.order > values.numVal) 
	gOutput->error("wrong order value, can't compute spline!");
    else {
	int i, j, k;
	// set dimension of matrix:
	mBase2.mMatrix.newGeometry(T.dim(), K.dim()-values.order);
	
	// set up matrix:
	for(i = 0; i < mBase2.mMatrix.Rows(); i++) 
	    for(j = 0, k = 0; j < mBase2.mMatrix.Cols(); j++,k++ )
		if(K[k]<= T[i] && T[i] < K[k+values.order])
		    mBase2.mMatrix(i,j) = blend2(j, values.order, T[i] );
		else
		    continue;
    }
}

// b-spline methods:
GApproxSpline::GApproxSpline(): GSpline() {
    fullFlag = isRebuilt = isRebuilt1 = isRebuilt2 = false;
}

GApproxSpline::GApproxSpline(const GSplineValues& v): GSpline(v) {}

GVector GApproxSpline::compKnotVector() {
    int i;
    GVector tVector;
    switch(getSpline().end) {
      case G_CLAMPED:
	switch(getSpline().mode) {
	  case G_OPEN:
	    tVector.newGeometry(getSpline().numVal+getSpline().order);
	    for(i = 0;i < tVector.dim();i++ )
		if(i<getSpline().order) tVector[i]=0;
		else
		    if(getSpline().order<=i && i<getSpline().numVal)
			tVector[i] = i-getSpline().order+1;
		    else
			if(i>=getSpline().numVal)
			    tVector[i]=getSpline().numVal-1-getSpline().order+2; 
	    
	    break;
	  case G_CLOSED:
	    tVector.newGeometry(getSpline().numVal+2*getSpline().order-1);
	    tVector[0]=1-getSpline().order;
	    for(i=1; i<tVector.dim(); i++) tVector[i]=tVector[i-1]+1;
	    break;
	  case G_NO_MODE:
	    tVector.newGeometry(0);
	    break;
	}
	break;
      case G_FREE:
	switch(getSpline().mode) {
	  case G_OPEN:
	    tVector.newGeometry(getSpline().numVal+getSpline().order);
	    tVector[0] = 1-getSpline().order;
	    for(i=1; i<tVector.dim(); i++) tVector[i]=tVector[i-1]+1;
	    break;
	  case G_CLOSED:
	    tVector.newGeometry(getSpline().numVal+2*getSpline().order-1);
	    tVector[0] = 1-getSpline().order;
	    for(i=1; i<tVector.dim(); i++) tVector[i]=tVector[i-1]+1;
	    break;
	  case G_NO_MODE:
	    tVector.newGeometry(0);
	    break;
	}
	break;
      case G_NO_END:
	tVector.newGeometry(0);
	break;
    }
   
    return tVector;
}

void GApproxSpline::rebuild(GMatrix &pMatrix) {
    // if G_CLOSED, rebuild the matrix like N.rows = numVal;
    if(getSpline().order > 1) {
	GMatrix tMatrix(getSpline().dim, getSpline().numVal);
	for(int i = 0; i < tMatrix.Rows(); i++) {
	    GVector tRow = pMatrix.getRow(i);

	    rebuildVector(tRow);
	    tMatrix.setRow(i, tRow);
	}
	
	pMatrix = tMatrix;
    }
}

void GApproxSpline::rebuildVector(GVector& pVector) {
    int i;
    GVector tVector(getSpline().numVal);

    for(i = 0; i < tVector.dim(); i++) tVector[i] = pVector[i];

    for(i = 0; i < pVector.dim() - tVector.dim(); i++)
	tVector[i] += pVector[i + tVector.dim()];

    pVector = tVector;
}

GVector GApproxSpline::getBaseTranslation(GFloat pT) {
    GVector tVector = GSpline::getBaseTranslation(pT);

    if(getSpline().mode == G_CLOSED) rebuildVector(tVector);
    return tVector;
}

GVector GApproxSpline::getBaseTranslation1(GFloat pT) {
    GVector tVector = GSpline::getBaseTranslation1(pT);

    if(getSpline().mode == G_CLOSED) rebuildVector(tVector);
    return tVector;
}

GVector GApproxSpline::getBaseTranslation2(GFloat pT) {
    GVector tVector = GSpline::getBaseTranslation2(pT);

    if(getSpline().mode == G_CLOSED) rebuildVector(tVector);
    return tVector;
}

void GApproxSpline::update() {
    if(isChanged())
	isRebuilt = isRebuilt1 = isRebuilt2 = false;
    
    GSpline::update();
}

const GMatrix& GApproxSpline::getBaseMatrix() {
    update();

    if(getSpline().mode == G_CLOSED)
	if(!fullFlag)
	    if(!isRebuilt) {
		rebuild(GSpline::getBaseMatrix());
		isRebuilt = true;
	    }

    return GSpline::getBaseMatrix();
}

const GMatrix& GApproxSpline::getBaseMatrix1() {
    update();

    if(getSpline().mode == G_CLOSED)
	if(!fullFlag)
	    if(!isRebuilt1) {
		rebuild(GSpline::getBaseMatrix1());
		isRebuilt1 = true;
	    }

    return GSpline::getBaseMatrix1();
}

const GMatrix& GApproxSpline::getBaseMatrix2() {
    update();
    
    if(getSpline().mode == G_CLOSED)
	if(!fullFlag)
	    if(!isRebuilt2) {
		rebuild(GSpline::getBaseMatrix2());
		isRebuilt2 = true;
	    }
    
    return GSpline::getBaseMatrix2();
}

// NaturalFittingSpline methods:
GNaturalFittingSpline::GNaturalFittingSpline(void): GSpline(), mResult() {}
GNaturalFittingSpline::GNaturalFittingSpline(const GSplineValues& v): mResult(GSplineValues(0, 0, 4, G_OPEN, G_FREE)) {
    setSpline().mode = v.mode;
    setSpline().order = v.order;
    setSpline().end = v.end;

    switch(getSpline().mode) {
      case G_CLOSED:
	// the control polygon has to be closed
	setSpline().dim = v.numVal + 1;
	setSpline().numVal = v.numVal + 3;

	mResult.setSpline().numVal = v.numVal + 3;
	break;
      case G_OPEN:
	setSpline().dim = v.numVal;
	setSpline().numVal = v.numVal + 2;

	mResult.setSpline().numVal = v.numVal + 2;
	break;
      default:
	break;
    }
}

void GNaturalFittingSpline::setMode(const GSplineMode& pMode) {
    setSpline().mode = pMode;
    update();
}

const GSplineMode& GNaturalFittingSpline::getMode(void) const {
    return getSpline().mode;
}

void GNaturalFittingSpline::computeConstraints(void) {
    switch(getSpline().mode) {
      case G_OPEN:
	setConstraints().mTranslation1 =
	    GSpline::getBaseTranslation2(getPrivParVector()[0]);
	setConstraints().mTranslation2 = 
	    GSpline::getBaseTranslation2(getPrivParVector()[getPrivParVector().dim() - 1]);
	break;
      case G_CLOSED:
	setConstraints().mTranslation1 = 
	    GSpline::getBaseTranslation1(getPrivParVector()[0]) -
		GSpline::getBaseTranslation1(getPrivParVector()[getPrivParVector().dim()-1]);
	setConstraints().mTranslation2 =
	    GSpline::getBaseTranslation2(getPrivParVector()[0]) -
		GSpline::getBaseTranslation2(getPrivParVector()[getPrivParVector().dim()-1]);
	break;
      default:
	break;
    }
    
    setConstraints().mValue1.newGeometry(mInput.Cols());
    setConstraints().mValue2.newGeometry(mInput.Cols());
}

void GNaturalFittingSpline::setValuesMatrix(const GMatrix& pInput) {
    assert(pInput.Rows() > 0 && pInput.Cols() > 0);

    if(pInput.Rows() < 2) {
	gOutput->error("Cannot interpolate less then two points, setting default!");
	mInput.newGeometry(2, pInput.Cols());
    }
    else
	mInput = pInput;

    switch(getSpline().mode) {
      case G_CLOSED:
	// the control polygon has to be closed
	setSpline().dim = mInput.Rows() + 1;
	setSpline().numVal = mInput.Rows() + 3;

	mResult.setSpline().numVal = mInput.Rows() + 3;
	break;
      case G_OPEN:
	setSpline().dim = mInput.Rows();
	setSpline().numVal = mInput.Rows() + 2;

	mResult.setSpline().numVal = mInput.Rows() + 2;
	break;
      default:
	break;
    }
}

void GNaturalFittingSpline::setDimension(int pDim) {
    mResult.setSpline().dim = pDim;
}

const int GNaturalFittingSpline::getDimension() const { return mResult.getSpline().dim; }

GVector GNaturalFittingSpline::compKnotVector() {
    int i;
    GFloat kVal;
    GVector tVector(getSpline().numVal + getSpline().order); // 2*(getOrd()-1)

    for(i = 0, kVal = -3.0; i < tVector.dim(); i++, kVal += 1.0) tVector[i] = kVal;

    return tVector;
}

GVector GNaturalFittingSpline::compParVector() {
    int i;
    GVector tVector(getSpline().dim);

    for(i = 0; i < tVector.dim(); i++ )
	tVector[i] = getPrivKnotVector()[i+3]; //getOrd()-1

    return tVector;
}

void GNaturalFittingSpline::update() {
    assert(mInput.Rows() > 0 && mInput.Cols() > 0);

    switch(getSpline().mode) {
      case G_CLOSED:
	// the control polygon has to be closed
	if(getSpline().dim != mInput.Rows() + 1)
	    setSpline().dim = mInput.Rows() + 1;
	if(getSpline().numVal != mInput.Rows() + 3)
	    setSpline().numVal = mInput.Rows() + 3;
	if(mResult.getSpline().numVal != mInput.Rows() + 3)
	    mResult.setSpline().numVal = mInput.Rows() + 3;
	break;
      case G_OPEN:
	if(getSpline().dim != mInput.Rows())
	    setSpline().dim = mInput.Rows();
	if(getSpline().numVal != mInput.Rows() + 2)
	    setSpline().numVal = mInput.Rows() + 2;
	if(mResult.getSpline().numVal != mInput.Rows() + 2)
	    mResult.setSpline().numVal = mInput.Rows() + 2;
	break;
      default:
	break;
    }
    
    if(isChanged()) {
	GSpline::update();
	
	// compute new control polygon
	updateVertexMatrix();
    }
}
    
const GMatrix& GNaturalFittingSpline::getBaseMatrix() {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();

    return mResult.getBaseMatrix();
}

const GMatrix& GNaturalFittingSpline::getBaseMatrix1() {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();
    
    return mResult.getBaseMatrix1();
}

const GMatrix& GNaturalFittingSpline::getBaseMatrix2() {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();

    return mResult.getBaseMatrix2();
}

GVector GNaturalFittingSpline::getBaseTranslation(GFloat pT) {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();

    return mResult.getBaseTranslation(pT);
}

GVector GNaturalFittingSpline::getBaseTranslation1(GFloat pT) {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();

    return mResult.getBaseTranslation1(pT);
}

GVector GNaturalFittingSpline::getBaseTranslation2(GFloat pT) {
    if(mResult.getSpline().numVal != getVertexMatrix().Rows())
	mResult.setSpline().numVal = getVertexMatrix().Rows();

    return mResult.getBaseTranslation2(pT);
}

const GMatrix& GNaturalFittingSpline::getPrivateBase() {
    update();
    return GSpline::getBaseMatrix();
}

const GMatrix& GNaturalFittingSpline::getPrivateBase1() {
    update();
    return GSpline::getBaseMatrix1();
}

const GMatrix& GNaturalFittingSpline::getPrivateBase2() {
    update();
    return GSpline::getBaseMatrix2();
}
    
const GMatrix& GNaturalFittingSpline::getVertexMatrix() {
    update();
    return mOutput;
}

void GNaturalFittingSpline::updateVertexMatrix() {
    // compute them:
    computeConstraints();

    GMatrix TMP1, TMP2, TMP3;

    TMP1 = getPrivateBase();
    // get current spline base to compute the control polygon
    
    TMP2.newGeometry(TMP1.Rows() + 2, TMP1.Cols());

    // set up the right side with the appropriate constraints values:
    TMP2.setRow(0, getConstraints().mTranslation1);
    TMP2.setRow(TMP2.Rows()-1, getConstraints().mTranslation2);

    int i, k;
    // set the rest of the right-side matrix
    for(i = 1; i < TMP2.Rows()-1; i++) TMP2.setRow(i, TMP1.getRow(i - 1));
    
    bool isSingular;
    TMP2 = TMP2.inverse(isSingular);
    
    assert(!isSingular);

    // set up the left side:
    TMP3.newGeometry(TMP2.Rows(), mInput.Cols());

    // set the appropriate values
    TMP3.setRow(0, getConstraints().mValue1);
    TMP3.setRow(TMP3.Rows()-1, getConstraints().mValue2);

    for(i = 1, k = 0; i < TMP3.Rows()-1; i++, k++)
	if(!(k < mInput.Rows())) {
	    // set the last input value = first
	    TMP3.setRow(i, mInput.getRow(0));
	}
	else
	    TMP3.setRow(i, mInput.getRow(k));


    mOutput = TMP2 * TMP3;
}

GGeneralFittingSpline::GGeneralFittingSpline(): GNaturalFittingSpline() {}
GGeneralFittingSpline::GGeneralFittingSpline(const GSplineValues& v, const GVector& constr1, const GVector& constr2): GNaturalFittingSpline(v) {
    mResult.setSpline().mode = getSpline().mode;

    setConstraints().mValue1 = constr1;
    setConstraints().mValue2 = constr2;
}

void GGeneralFittingSpline::setMode(const GSplineMode& pMode) {
    mResult.setSpline().mode = pMode;
    update();
}

const GSplineMode& GGeneralFittingSpline::getMode() const {
    return mResult.getSpline().mode;
}

void GGeneralFittingSpline::setTangentBeg(const GVector& tng) {
    setConstraints().mValue1 = tng;
    setChanged() = true;
}

void GGeneralFittingSpline::setTangentEnd(const GVector& tng) {
    setConstraints().mValue2 = tng;
    setChanged() = true;
}

void GGeneralFittingSpline::computeConstraints() {
    setConstraints().mTranslation1 = 
	GSpline::getBaseTranslation1(getPrivParVector()[0]);

    setConstraints().mTranslation2 =
	GSpline::getBaseTranslation1(getPrivParVector()[getPrivParVector().dim()-1]);
}

/*
// GNaturalFittingSplineSurface methods...
GNaturalFittingSplineSurface::GNaturalFittingSplineSurface(const GSplineMode& pModeX, const GSplineMode& pModeY): uSpline(pModeX), wSpline(pModeY), dimX(0), dimY(0) {}

void GNaturalFittingSplineSurface::setValuesMatrix(int pX, int pY, const GMatrix& pInput ) {
    assert(pInput.Rows() > 0 && pInput.Cols() > 0);

    // check the right dimension
    assert((pX * pY) == pInput.Rows());

    if(pInput.Rows() < 4) {
	gOutput->error("Can't interplolate a surface with less then 4 corners, setting default!");
	mInput.newGeometry(4, pInput.Cols());
	dimX = 2; dimY = 2;
    }
    else {
	mInput = pInput;
	dimX = pX; dimY = pY;
    }

    uSpline.isChanged() = true;
    wSpline.isChanged() = true;
}


GMatrix& GNaturalFittingSplineSurface::getVertexMatrix(int &pX, int &pY) {
    assert(mInput.Rows() > 0 && mInput.Cols() > 0);

    updateVertexMatrix();
    
    pX = getVertexMatrix().
	
	pX = dimX + 2; pY = dimY + 2;
	return mOutput;
    }
    else
	{
	    update();
	    pX = dimX + 2; pY = dimY + 2;
	    return mOutput;
	}
}

void GNaturalFittingSplineSurface::update() {
    
void GNaturalFittingSplineSurface::updateVertexMatrix() {   
    register int i, j, p, u, w;

    GMatrix *wNet = new GMatrix [dimX];
    GMatrix *uNet = new GMatrix [dimY + 2];

    // get vertices in u direction..(horizontal)..
    for(u = 0; u < dimX; u++) {
	p = u;

	wNet[u].newGeometry(dimY, mInput.Cols());
	for(w = 0; w < dimY; w++, p += dimX) 
	    wNet[u].setRow(w, mInput.getRow(p));

	uSpline.setValuesMatrix(wNet[u]);
	wNet[u] = uSpline.getVertexMatrix();
    }

    // get vertices in w direction..(vertical)..
    for(w = 0; w < dimY + 2; w++) {
	uNet[w].newGeometry(dimX, mInput.Cols());

	for(u = 0; u < dimX; u++) 
	    uNet[w].setRow(u, wNet[u].getRow(w));

	wSpline.setValuesMatrix(uNet[w]);
	uNet[w] = wSpline.getVertexMatrix();
    }
    
    mOutput.newGeometry((dimX+2)*(dimY+2), mInput.Cols());

    p = 0;
    while(p < mOutput.Rows()) {
	for(i = 0; i < dimY + 2; i++)
	    for(j = 0; j < dimX + 2; j++,p++)
		mOutput.setRow(p, uNet[i].getRow(j));
    }

    delete [] uNet; delete [] wNet;
}
*/








