//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "goutput.h"
#include "util/gstring.h"

ostream& operator<<(ostream& pStream, const GString& pString) {
    pStream << pString.mChars;
    return(pStream);
}

const GString& GString::operator+=(const GString& pString) {
    return(*this += pString.mChars);
}

const GString& GString::operator+=(char c) {	
    int tLength = strlen(mChars) + 2;
    if (tLength > mLength) {
	mLength = tLength + tLength % 16;
	char* tChars = mChars;
	mChars = new char[mLength];
	strcpy(mChars, tChars);
	delete tChars;
    }
    mChars[tLength - 2] = c;
    mChars[tLength - 1] = 0;
    assert(mLength > getLength());
    return(*this);
}

const GString& GString::operator+=(const char* pChars) {
    int tLength = strlen(mChars) + strlen(pChars) + 1;
    if (tLength > mLength) {
	mLength = tLength + tLength % 16;
	char* tChars = mChars;
	mChars = new char[mLength];
	strcpy(mChars, tChars);
	delete tChars;
    }
    strcat(mChars, pChars);
    assert(mLength > getLength());
    return(*this);
}

GString::GString(int pLength): mLength(pLength) {
    // Correct invalid length value:
    if (mLength < 1) mLength = 10;
    mChars = new char[mLength];
    *mChars = 0;
}

GString::GString(const char *pChars) {
    if (pChars) {
	mLength = strlen(pChars) + 1;
	mChars = new char[mLength];
	strcpy(mChars, pChars);
    }
    else {
	mLength = 1;
	mChars = new char[mLength];
	mChars[0] = 0;
    }
} 

GString::GString(const GString& pString) {
    assert(pString.mLength > strlen(pString.mChars));
    mLength = pString.mLength;
    mChars = new char[mLength];
    strcpy(mChars, pString.mChars);
}

bool GString::operator==(const GString& pString) const {
    // check identity first:
    if (this == &pString) return(true);
    return(strcmp(mChars, pString) ? false : true);
}

bool GString::operator!=(const GString& pString) const {
    return (!(self == pString));
}

bool GString::operator==(const char* pString) const {
    return(strcmp(mChars, pString) ? false : true);
}

bool GString::operator!=(const char* pString) const {
    return(strcmp(mChars, pString) ? true : false);
}

GString& GString::operator=(const GString& pString) {
    int tLength = strlen(pString.mChars) + 1;
    if (tLength > mLength)  {
	mLength = tLength + tLength % 16;
	char* tChars = mChars;
	mChars = new char[mLength];
	strcpy(mChars, tChars);
	delete tChars;
    }
    strcpy(mChars, pString.mChars);
    assert(mLength > getLength());
    return(self);
}

GString GString::substr(int from, int to) {
    GString tStr(self);
    if (from > 0) tStr.del(0, from - 1);
    tStr.del(to + 1, tStr.getLength() - 1);
    return(tStr);
}

void GString::del(int from, int to) {
    if (from >= getLength()) return;
    if (to >= getLength()) to = getLength() - 1;
    strcpy(mChars + from, mChars + to + 1);
    assert(mLength > getLength());
}

char GString::operator[](int at) const {
    assert(at >= 0 && at <= getLength());
    return(mChars[at]);
}

GString GString::car(void) const {
    sscanf(mChars, "%s", sBuffer);
    // get the position of the first word
    int posi = getPos(sBuffer);
    // the length of the remainder of the string
    int lengthLeft = getLength() - posi - strlen(sBuffer);
    char *cp = mChars + posi + strlen(sBuffer);
    while (isspace(*cp++)) lengthLeft--;
    return(GString(mChars + strlen(mChars) - lengthLeft));
}

GString GString::cdr(void) const {
    sscanf(mChars, "%s", sBuffer);
    return(GString(sBuffer));
}
    
const char* GString::tolower(void) const {
    char* tFromPtr = mChars;
    char* tToPtr = sBuffer;
    while (*tToPtr++ = ::tolower(*tFromPtr++)) ;
    return(sBuffer);
}

int GString::getPos(const char* pStr) const {
    char *tStr = strstr(mChars, pStr);
    if (!tStr) return(-1);
    return(tStr - mChars);
}


GString operator+(const GString& pA, const char* pB) {
    strcpy(GString::sBuffer, (const char *)pA.mChars);
    strcat(GString::sBuffer, pB);
    
    return (GString(GString::sBuffer));
}

GString operator+(const char* pA, const GString& pB) {
    strcpy(GString::sBuffer, pA);
    strcat(GString::sBuffer, (const char *)pB.mChars);
    
    return (GString(GString::sBuffer));
}

GString operator+(const GString& pA, const GString& pB) {
    strcpy(GString::sBuffer, (const char *)pA.mChars);
    strcat(GString::sBuffer, (const char *)pB.mChars);
    
    return (GString(GString::sBuffer));
}

bool GString::operator< (const GString& pS) const {
    return (strcmp(mChars, pS.mChars) < 0); 
}

bool GString::operator<=(const GString& pS) const {
    return (strcmp(mChars, pS.mChars) <= 0); 
}

bool GString::operator> (const GString& pS) const {
    return (strcmp(mChars, pS.mChars) > 0); 
}

bool GString::operator>=(const GString& pS) const {
    return (strcmp(mChars, pS.mChars) >= 0); 
}


char& GString::operator[](int at) {
    assert(at >= 0 && at <= getLength());
    return(mChars[at]);
}


unsigned int GString::hash(void) const {
    unsigned int h=0;
    unsigned int g;

    for (register int i=0; mChars[i] != '\0'; i++) {
	h = (h<<4) + mChars[i];
	if (g = h & 0xf0000000) {
	    h = h^(g>>24);
	    h = h^g;
	}
    }
    return (h);
}

// return the buffer from i-th -> j-th position of the old string
GString GString::operator()(int i, int j) const {
    if (j<i || i>=mLength || j>=mLength)
	gOutput->error("GString::operator() index out of range");

    char* buf = new (char[j-i+1]);
    strncpy(buf, mChars + i, j - i);
    buf[j-i] = '\0';

    return (GString(buf));
}

    


char GString::sBuffer[100];
