//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifdef GWITH_TCL

#ifdef GWINDOWS
#include <windows.h>
#endif

#include <malloc.h>
#include <ctype.h>
#include "goutput.h"
#include "util/gtcl.h"

GTclLoadProc GTclClass::sLoadProc;
GDLList<GTclClassDesc> GTclClass::sClasses(true);

class GTclFindClassIterator: public GIterator<GTclClassDesc> {
  public:
    GTclFindClassIterator(const char* pClass): mClass(pClass), mDesc(0) {}

    bool exec(GTclClassDesc& c) {
	GTclClassDesc& tDesc = c;
	if (!strcmp(mClass, tDesc.mConstr.mMethod)) {
	    mDesc = &tDesc;
	    return(true);
	}
	else return(false);
    }
    GTclClassDesc* get(void) const { return(mDesc); }
  private:
    const char* mClass;
    GTclClassDesc* mDesc;
};

class GTclFindMethodIterator: public GIterator<GTclMethodDesc> {
  public:
    GTclFindMethodIterator(const char* pMethod): mMethod(pMethod), mDesc(0) {}

    bool exec(GTclMethodDesc& m) {
	if (!strcmp(mMethod, m.mMethod)) {
	    mDesc = &m;
	    return(true);
	}
	else return(false);
    }
    GTclMethodDesc* get(void) const { return(mDesc); }
  private:
    const char* mMethod;
    GTclMethodDesc* mDesc;
};

//@ManMemo: Put all class names to the Tcl result.
class GTclGetClassesIterator: public GIterator<GTclClassDesc> {
  public:
    virtual bool exec(GTclClassDesc& GTclClassDesc) { 
	gTclAppendResult(GTclClassDesc.getClass()); 
	return(false);
    }
};

//@ManMemo: Put all class names to the Tcl result.
class GTclGetMethodsIterator: public GIterator<GTclMethodDesc> {
  public:
    virtual bool exec(GTclMethodDesc& m) { 
	gTclAppendResult(m.mMethod); 
	return(false);
    }
};

GTclClassDesc* GTclClass::getClassByName(const char* pName) {
    GTclFindClassIterator tIter(pName);
    sClasses.doWithElements(tIter);
    return(tIter.get());
}

bool GTclClass::objectCMD(char* argv[]) {
    bool r = false;
    int start, diff = 0, xdiff; 
    const char* pr;

    if ((start = gTclFindString(argv, "-setMembers")) >= 0) {
	r = true;
	while (gTclFromString((const char**)&argv[start + diff + 1], pr, xdiff)) {
	    diff += xdiff;
	    int xargc; char** xargv;
	    if (Tcl_SplitList(gTclTmpInterp, (char*)pr, &xargc, &xargv) == TCL_OK) {
		if (xargc == 2) {
		    mMembers.set(xargv[0], xargv[1]);
		    mMembers.setMembers(getName());
		}
		free((char*)xargv);
	    }
	}
    }
    gTclClearArgv(argv, start, diff);

    // help request:
    if (*argv && !strcmp(*argv, gTclHelp)) {
	mDesc->inquireInformation();
	// get infos of super classes:
	const GTclClassDesc* fdesc;
	const char* fclass = mDesc->getParentClass();

	while (1) {
	    GTclFindClassIterator tIter(fclass);
	    sClasses.doWithElements(tIter);
	    fdesc = tIter.get();
	    if (fdesc) {
		fclass = fdesc->getParentClass();
		fdesc->inquireInformation();
	    }
	    else break;
	}
	return(true); 
    }

    if (!*argv) return(r);
    
    // evaluate arguments:
    if (callMethod(mDesc, argv)) return(true); 

    // call super class methods:
    const GTclClassDesc* fdesc;
    const char* fclass = mDesc->getParentClass();

    while (1) {
	GTclFindClassIterator tIter(fclass);
	sClasses.doWithElements(tIter);
	fdesc = tIter.get();
	if (fdesc) {
	    fclass = fdesc->getParentClass();
	    if (callMethod(fdesc, argv)) return(true); 
	}
	else break;
    }
    return(r);
}

GTclClass::GTclClass(const char* pName, GTclClassDesc* pDesc): GTclInterface(pName), mName(pName), mDesc(pDesc) {
    addMembers(mDesc->mMembers);
    
    // add recursively super class members:
    const GTclClassDesc* fdesc;
    const char* fclass = mDesc->getParentClass();

    while (1) {
	GTclFindClassIterator tIter(fclass);
	sClasses.doWithElements(tIter);
	fdesc = tIter.get();
	if (fdesc) {
	    addMembers(fdesc->mMembers);
	    fclass = fdesc->getParentClass();
	}
	else break;
    }
}

bool GTclClass::callMethod(const GTclClassDesc* pDesc, char** argv, bool co) {
    static depth = 0;
    const GTclMethodDesc* mdesc;
    if (co) mdesc = &(pDesc->mConstr);
    else { 
	// find the method: 
	GTclFindMethodIterator tIter(*argv++);
	pDesc->mMethods.doWithElements(tIter);
	if (!(mdesc = tIter.get())) return(false);
    }
    
    // still remaining arguments:
    int argc = 0; char** tmp = argv; 
    while (*tmp) { argc++; tmp++; }
    
    int xargc; char** xargv;
    if (Tcl_SplitList(gTclTmpInterp, mdesc->mParams, &xargc, &xargv) == TCL_OK) {
	// error message for a method
	// for constructor this already has been done (in classCMD)
	if (argc < xargc) {
	    gOutput->errorVar(getName(), ": Need following arguments for method ",
			      (char*)mdesc->mMethod, ": ", (char*)mdesc->mParams, "!", 0);
	    free((char*)xargv);
	    return(false);
	}

	GString code(100);
	{	
	    char tmp[10];
	    sprintf(tmp, "%i", ++depth);
	    code = "proc _gen";
	    code += tmp;
	    code += "_ {} {\n";
	}
	
	// set "this":
	code += "set this "; 
	code += getName(); 
	code += "\n";

	// set temporary values as arguments:
	char** tmp = xargv;
	while (*tmp) {
	    code += "set "; 
	    code += *tmp++; 
	    code += " {"; 
	    code += *argv++; 
	    code += "}\n";
	}
	free((char*)xargv);

	// put the parameter which are not bound to variables in a parameter list 
	if (xargc < argc) {
	    code += "set argv {";
	    while (*argv) { 
		code += *argv++; 
		code += " "; 
	    }
	    code += "} \n";
	    char count[20];
	    sprintf(count, "set argc %d\n", argc - xargc);
	    code += count;
	} else code += "set argv {}\nset argc 0\n";
	
	// set the member variables:
	mMembers.globalize(getName(), code);
	
	if (co) { 
	    // merge recursively super classes code into this code:
	    const GTclClassDesc* fdesc;
	    const char* fclass = mDesc->getParentClass();
	    GString tmp1(100), tmp2(100);
	    while (1) {
		GTclFindClassIterator tIter(fclass);
		sClasses.doWithElements(tIter);
		fdesc = tIter.get();
		if (fdesc) {
		    fclass = fdesc->getParentClass();
		    tmp1 = fdesc->mConstr.mBody;
		    tmp1 += tmp2;
		    tmp1 += "\n";
		    tmp2 = tmp1;
		}
		else {
		    code += tmp1;
		    break;
		}
	    }
	}
	
	// write the code:
	code += mdesc->mBody;

	code += "\n}";

	if (gTclGlobalEval(code) != TCL_OK) gOutput->errorVar(getName(), ": Error in function body: ", gTclGetResult(), 0);
	else {
	    char tmp[10];
	    sprintf(tmp, "_gen%i_", depth);
	    if (gTclGlobalEval(tmp) != TCL_OK) 
		gOutput->errorVar(getName(), ".", co ? "constructor" : (char*)mdesc->mMethod, ": Error when evaluating! ", gTclGetResult(), 0);
	    else {
		GTclInterface::setResult(gTclGetResult());
		depth--;
		return(true);
	    }
	    depth--;
	}
    }
    return(false);
}

void GTclClass::addMembers(const char* mbs) {
    int xargc; char** xargv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)mbs, &xargc, &xargv) == TCL_OK) {
	for (int i = 0; i < xargc; i++) mMembers.add(xargv[i]);
	free((char*)xargv);
    }
}

int GTclClass::getClassesCMD(ClientData, Tcl_Interp*, int, char* []) {
    GTclGetClassesIterator tIter;
    sClasses.doWithElements(tIter);
    return(TCL_OK);
}

int GTclClass::getMethodsCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 2) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <class>\"", 0);
	return(TCL_ERROR);
    }
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    const GTclClassDesc* tDesc = tIter.get();
    if (!tDesc) {
	Tcl_AppendResult(ip, argv[0], ": There is no class ", argv[1], 0);
	return(TCL_ERROR);
    }
    Tcl_AppendResult(ip, (char*)tDesc->mConstr.mMethod, " ", 0);
    GTclGetMethodsIterator tIter2;
    tDesc->mMethods.doWithElements(tIter2);
    return(TCL_OK);
}

int GTclClass::getParentClassCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 2) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <class>\"", 0);
	return(TCL_ERROR);
    }
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    const GTclClassDesc* desc = tIter.get();
    if (!desc) {
	Tcl_AppendResult(ip, argv[0], ": there is no class ", argv[1], 0);
	return(TCL_ERROR);
    }
    Tcl_AppendResult(ip, desc->getParentClass(), 0);
    return(TCL_OK);
}

int GTclClass::getParameterCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    GTclMethodDesc* rec;
    if (rec = find(ip, argc, argv)) {
	Tcl_AppendResult(ip, (char*)rec->mParams, 0);
	return(TCL_OK);   
    }
    return(TCL_ERROR);
}

int GTclClass::getDescriptionCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    GTclMethodDesc* rec;
    if (rec = find(ip, argc, argv)) {
	Tcl_AppendResult(ip, (char*)rec->mDesc, 0);
	return(TCL_OK);   
    }
    return(TCL_ERROR);
}

int GTclClass::getBodyCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    GTclMethodDesc* rec;
    if (rec = find(ip, argc, argv)) {
	Tcl_AppendResult(ip, (char*)rec->mBody, 0);
	return(TCL_OK);   
    }
    return(TCL_ERROR);
}

int GTclClass::classCMD(ClientData cd, Tcl_Interp*, int argc, char* argv[]) {
    if (!cd ) return(TCL_OK);
    GTclClassDesc* pDesc = (GTclClassDesc*)cd;

    if (argc == 3) {
	if (!strcmp(argv[1], "-description"))  {
	    pDesc->setDescription(argv[2]);
	    return(TCL_OK);
	}
	if (!strcmp(argv[1], "-keywords")) {
	    pDesc->setKeywords(argv[2]);
	    return(TCL_OK);
	}
    }
    return(TCL_OK);
}

int GTclClass::addClassCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 8) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <class> <parent class> <members> <parameters> <parameter types> <description> <body>\"", 0);
	return(TCL_ERROR);
    }
    { 
	// check the father class:
	if (strcmp(argv[2], "")) {
	    GTclFindClassIterator tIter(argv[2]);
	    sClasses.doWithElements(tIter);
	    if (!tIter.get() && loadClass(argv[2]) != TCL_OK) {
		Tcl_AppendResult(ip, argv[0], ": No parent class ", argv[2], " found.", 0);
		return(TCL_ERROR);
	    }
	}
    }
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    GTclClassDesc* desc = tIter.get();

    if (!desc) {
	// no such class found - must be a new class
	desc = new GTclClassDesc(argv[2], argv[3]);
	desc->mConstr.mMethod = argv[1];
	desc->mClass = desc->mConstr.mMethod;
	sClasses.append(desc);
    } 
    GTclMethodDesc& tmp = desc->mConstr;
    
    // put the new values into the record:
    tmp.mParams = argv[4];
    tmp.mPTypes = argv[5];
    tmp.mDesc = argv[6];
    tmp.mBody = argv[7];

    return(TCL_OK);
}

int GTclClass::addMethodCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 7) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <classname> <methodname> <parameters> <parametertypes> <description> <body>\"", 0);
	return(TCL_ERROR);
    }
    // find the class:
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    GTclClassDesc* cdesc = tIter.get();
    if (!cdesc) {
	Tcl_AppendResult(ip, argv[0], ": There is no class ", argv[1], " ", 0);
	return(TCL_ERROR);
    }
    
    // find the method:
    GTclFindMethodIterator tIter2(argv[2]);
    cdesc->mMethods.doWithElements(tIter2);
    GTclMethodDesc* mdesc = tIter2.get();
    
    if (!mdesc) {
	mdesc = new GTclMethodDesc;
	mdesc->mMethod = argv[2];
	cdesc->mMethods.append(mdesc);
    } 
    GTclMethodDesc* tmp = mdesc;
    
    // put the new values into the record:
    mdesc->mClass = cdesc->mClass;
    tmp->mParams = argv[3];
    tmp->mPTypes = argv[4];
    tmp->mDesc = argv[5];
    tmp->mBody = argv[6]; 
    return(TCL_OK);
}

int GTclClass::deleteClassCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 2) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <classname>\"", 0);
	return(TCL_ERROR);
    }
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    GTclClassDesc* desc = tIter.get();
    if (desc) sClasses.remove(desc);
    else {
	Tcl_AppendResult(ip, argv[0], ": there is no class ", argv[1], " ", 0);
	return(TCL_ERROR);
    }
    return(TCL_OK);
}

int GTclClass::deleteAllClassesCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 1) {
	Tcl_AppendResult(ip, argv[0], ": No args supported.", 0);
	return(TCL_ERROR);
    }
    sClasses.reInitialize();
    return(TCL_OK);
}

int GTclClass::deleteMethodCMD(ClientData, Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 3) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0],
			 " <classname> <method>\"", 0);
	return(TCL_ERROR);
    }
    // first find the class:
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    GTclClassDesc* cdesc = tIter.get();

    if (!cdesc) {
	Tcl_AppendResult(ip, argv[0], ": there is no class ", argv[1], " ", 0);
	return(TCL_ERROR);
    }
    
    // then find the method:
    if (!strcmp(cdesc->mConstr.mMethod, argv[2])) {
	Tcl_AppendResult(ip, argv[0], ": You cannot delete the constructor separately.", 0);
	return(TCL_ERROR);
    }
    GTclFindMethodIterator tIter2(argv[2]);
    cdesc->mMethods.doWithElements(tIter2);
    GTclMethodDesc* mdesc = tIter2.get();
    if (!mdesc) {
	Tcl_AppendResult(ip, argv[0], ": There is no method ", argv[2], " in class ", argv[1], " ", 0);
	return(TCL_ERROR);
    }
    else cdesc->mMethods.remove(mdesc);
    return(TCL_OK);
}

GTclMethodDesc* GTclClass::find(Tcl_Interp* ip, int argc, char* argv[]) {
    if (argc != 3) {
	Tcl_AppendResult(ip, argv[0], ": Wrong # args: should be \"", argv[0], " <classname> <method>\"", 0);
	return(0);
    }
    // first find the class:
    GTclFindClassIterator tIter(argv[1]);
    sClasses.doWithElements(tIter);
    GTclClassDesc* cdesc = tIter.get();
    if (!cdesc) {
	Tcl_AppendResult(ip, argv[0], ": There is no class ", argv[1], " ", 0);
	return(0);
    }
    
    // then find the method:
    if (cdesc->mConstr.mMethod == argv[2]) return(&cdesc->mConstr);
    GTclFindMethodIterator tIter2(argv[2]);
    cdesc->mMethods.doWithElements(tIter2);
    GTclMethodDesc* mdesc = tIter2.get();
    return(mdesc);
}

void gTclInitClassCommands(Tcl_Interp* ip) {
    GTclClass::setLoadProc(GTclClass::localLoadClass);

    // interface to class browser:
    Tcl_CreateCommand(ip, "TclClass.getClasses", GTclClass::getClassesCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.deleteAllClasses", GTclClass::deleteAllClassesCMD, 0, 0);

    // access to a class:
    Tcl_CreateCommand(ip, "Tcl_Method", GTclClass::addMethodCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.getMethods", GTclClass::getMethodsCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.getParentClass", GTclClass::getParentClassCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.deleteClass", GTclClass::deleteClassCMD, 0, 0);

    // access to a method:
    Tcl_CreateCommand(ip, "TclClass.getParameter", GTclClass::getParameterCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.getDescription", GTclClass::getDescriptionCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.getBody", GTclClass::getBodyCMD, 0, 0);
    Tcl_CreateCommand(ip, "TclClass.deleteMethod", GTclClass::deleteMethodCMD, 0, 0);
}

int GTclClass::localLoadClass(const char* pClass) {
    GString tStr(pClass);

    GString tClassFile(100);
    tClassFile += tStr.tolower();
    tClassFile += ".cls";

    FILE* tFile;
    tFile = fopen(tClassFile, "r");
    if (tFile) {
	fclose(tFile);
	int tRes = Tcl_EvalFile(gGlobalInterp, tClassFile);
	if (tRes == TCL_OK) Tcl_SetResult(gGlobalInterp, "1", TCL_VOLATILE);
	return(tRes);
    }
    else {
	Tcl_SetResult(gGlobalInterp, "0", TCL_VOLATILE);
	return(TCL_OK);
    }
}

bool GTclClass::isA(const char* pClass) const {
    if (mDesc->mClass == pClass) return(true);

    const GTclClassDesc* fdesc;
    const char* fclass = mDesc->getParentClass();
    
    while (1) {
	GTclFindClassIterator tIter(fclass);
	sClasses.doWithElements(tIter);
	fdesc = tIter.get();
	if (fdesc) {
	    fclass = fdesc->getParentClass();
	    if (fdesc->mClass == pClass) return(true);
    }
	else break;
    }
    return(false);
}

GTclClass::~GTclClass(void) {
    GString code(100);
    static depth = 0;
    char tmps[25];
    sprintf(tmps, "proc _gen%i_ {} {\n", ++depth);
    code = tmps;
	
    // set this:
    code += "set this ";
    code += getName();
    code += "\n";

    // set the member variables:
    mMembers.globalize(getName(), code);
    
    // merge recursively super class destructors into this code:
    const GTclClassDesc* fdesc;
    const char* fclass = getClass();
    while (1) {
	GTclFindClassIterator tIter(fclass);
	sClasses.doWithElements(tIter);
	fdesc = tIter.get();
	if (fdesc) {
	    GTclMethodDesc* mdesc;

	    GString tstr(20);
	    tstr = "~"; 
	    tstr += fclass;

	    fclass = fdesc->getParentClass();

	    GTclFindMethodIterator tIter2(tstr);
	    fdesc->mMethods.doWithElements(tIter2);
	    if (mdesc = tIter2.get()) { 
		code += mdesc->mBody; 
		code += '\n'; 
	    }
	}
	else break;
    }

    code += "}";

    if (gTclGlobalEval(code) != TCL_OK) 
	gOutput->errorVar(getName(), ": Error in function body: ", gTclGetResult(), 0);
    else {
	char tmp[10];
	sprintf(tmp, "_gen%i_", depth);
	if (gTclGlobalEval(tmp) != TCL_OK) 
	    gOutput->errorVar(getName(), ".destructor: Error when evaluating! ", gTclGetResult(), 0);
    }
    depth--;

}

bool GTclInterface::exists(const char* pName) {
    Tcl_CmdInfo tInfo;
    if (pName) return((bool)Tcl_GetCommandInfo(gGlobalInterp, (char*)pName, &tInfo));
    else return(true);
}

GTclInterface::GTclInterface(const char* pName): mName(pName ? pName : "") {
    if (pName) Tcl_CreateCommand(gGlobalInterp, (char*)pName, GTclInterface::objectCMD, (ClientData)this, 0); 
}

GTclInterface::~GTclInterface(void) {
    if (mName != (const char*)"") Tcl_DeleteCommand(gGlobalInterp, (char*)mName);
}

int GTclInterface::classCMD(ClientData, Tcl_Interp*, int argc, char* argv[]) {
    if (argc < 2) return(TCL_ERROR);
    if (!strcmp(argv[1], gTclHelp)) return(TCL_HELP);
    return(TCL_OK);
}

int GTclInterface::objectCMD(ClientData cd, Tcl_Interp* ip, int, char* argv[]) {
    static const char* gNoResult = "";
    setResult(gNoResult);
    bool tHelp = false;
    if (argv[1] && !strcmp(argv[1], gTclHelp)) tHelp = true;  
    if (((GTclInterface*)cd)->objectCMD(&argv[1])) {
	if (strcmp(getResult(), gNoResult))
	    Tcl_VarEval(ip, "set {} ", (char*)getResult(), 0);
	else Tcl_SetResult(ip, (char*)gNoResult, 0);
	return(TCL_OK);
    }
    else {
	if (!tHelp) {
	    Tcl_AppendResult(ip, argv[0], ": Couldn't resolve a method.", 0);
	    return(TCL_ERROR);
	}
	else return(TCL_OK);
    }
}

GString GTclInterface::sResult;

#endif   
