//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifdef GWITH_TCL

#ifdef GWINDOWS
#include <windows.h>
#endif

#include <math.h>
#include <stdlib.h> 
// Sun needs that for rand() prototype.

#include "util/gtclpars.h"
#include "util/gtclmath.h"

double exp10(GFloat x) {
    return(exp(x*log(10)));
}

#if defined(GWINDOWS) || defined(GHPUX)

double acosh(double val) {
    return(log(val + sqrt(val * val - 1)));
}

double asinh(double val) {
    return(log(val + sqrt(val * val + 1)));
}

double atanh(double val) {
    return(0.5 * log((1 + val) / (1 - val)));
}

#endif

typedef double (*gMath1Func)(double);

static int gTclMath1CMD(ClientData cd, Tcl_Interp *ip, int argc, char *argv[]) {
    GFloat ret, val;
    if (argc != 2) {
        Tcl_AppendResult(ip,argv[0],": wrong # args: should be \"",
                         argv[0]," <float>!\"", 0);
        return(TCL_ERROR);
    }
    if (!(gTclFromString(argv[1], val))) {
        Tcl_AppendResult(ip, argv[0], ": wrong arg #1: should be a float!", 0);
        return(TCL_ERROR); 
    } 
    ret = ((gMath1Func)cd)(val);
    
    Tcl_AppendResult(ip, gTclToString(ret), 0);
    return(TCL_OK);
}

typedef int (*gMath2Func)(void);

static int gTclMath2CMD(ClientData cd, Tcl_Interp *ip, int argc, char *argv[]) {
    if (argc != 1) {
        Tcl_AppendResult(ip, argv[0], ": no arguments needed.", 0);
        return(TCL_ERROR);
    }
    Tcl_AppendResult(ip, gTclToString(((gMath2Func)cd)()), 0);
    return(TCL_OK);
}

void gTclMathInit(Tcl_Interp* ip) {
    Tcl_CreateCommand(ip, "acos", gTclMath1CMD, (ClientData)acos, 0);
    Tcl_CreateCommand(ip, "acosh", gTclMath1CMD, (ClientData)acosh, 0);
    Tcl_CreateCommand(ip, "asin", gTclMath1CMD, (ClientData)asin, 0);
    Tcl_CreateCommand(ip, "asinh", gTclMath1CMD, (ClientData)asinh, 0);
    Tcl_CreateCommand(ip, "atan", gTclMath1CMD, (ClientData)atan, 0);
    Tcl_CreateCommand(ip, "atanh", gTclMath1CMD, (ClientData)atanh, 0);
    Tcl_CreateCommand(ip, "cos", gTclMath1CMD, (ClientData)cos, 0);
    Tcl_CreateCommand(ip, "cosh", gTclMath1CMD, (ClientData)cosh, 0);
    Tcl_CreateCommand(ip, "exp", gTclMath1CMD, (ClientData)exp, 0);
    Tcl_CreateCommand(ip, "exp10", gTclMath1CMD, (ClientData)exp10, 0);
    Tcl_CreateCommand(ip, "log", gTclMath1CMD, (ClientData)log, 0);
    Tcl_CreateCommand(ip, "log10", gTclMath1CMD, (ClientData)log10, 0);
    Tcl_CreateCommand(ip, "sin", gTclMath1CMD, (ClientData)sin, 0);
    Tcl_CreateCommand(ip, "sinh", gTclMath1CMD, (ClientData)sinh, 0);
    Tcl_CreateCommand(ip, "sqrt", gTclMath1CMD, (ClientData)sqrt, 0);
    Tcl_CreateCommand(ip, "tan", gTclMath1CMD, (ClientData)tan, 0);
    Tcl_CreateCommand(ip, "tanh", gTclMath1CMD, (ClientData)tanh, 0);

    Tcl_CreateCommand(ip, "rand", gTclMath2CMD, (ClientData)rand, 0);
}

#endif
