//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifdef GWITH_TCL

#include <malloc.h>
#include "util/gtclpars.h"

const int gTclNextParseEnum = 11;
const char* gTclHelp = "?";
Tcl_Interp* gTclTmpInterp = Tcl_CreateInterp();
GString gTclTmpString(1000);
Tcl_Interp* gGlobalInterp = 0;

int gTclFindString(char** argv, const char *str) {
    int i = 0;
    while (*argv) {
	if (!strcmp(*argv, str)) return(i);
	argv++; 
	i++;
    }
    return(-1); 
}

void gTclClearArgv(char **argv, int start, int diff) {
    if (start >= 0) 
	for (int i = start; argv[i]; i++ ) 
	    argv[i] = argv[i + diff + 1];
}

void gTclParseTable(char** argv, GTclParseEntry* entry) {
    if (argv[0] && !strcmp(argv[0], gTclHelp)) {
	while (entry->mType != gTclEnd) {
	    if (entry->mIsSet) *entry->mIsSet = 0; 
	    // reset the entry flags
	    if (entry->mName[0] != '_') {
		// Do not print out hidden methods.
		sprintf(gTclTmpString, "{%s {%s} {%s}}\n", entry->mName, entry->mTclTypes, entry->mInfo ); 
		gTclAppendResult(gTclTmpString);
	    }
	    entry++;
	}
	return;
    }
    while (entry->mType != gTclEnd) {
	if (entry->mIsSet) *entry->mIsSet = 0;
	int start, diff;
	if ((entry->mType != gTclSpecial) && (start = gTclFindString(argv, entry->mName)) >= 0) {
	    diff = 0;
	    switch (entry->mType) {
	      case gTclNone:
		*entry->mIsSet = true;
		break;
	      case gTclString: 
		const char* fnd;
		if (gTclFromString((const char**)&argv[start + 1], fnd, diff)) {
		    *(char**)entry->mParam = (char*)fnd; 
		    *entry->mIsSet = true;
		}
		break;
	      case gTclBool:
		bool b;
		if (gTclFromString((const char**)&argv[start + 1], b, diff)) {
		    *(bool*)entry->mParam = b; 
		    *entry->mIsSet = true;
		}
		break;
	      case gTclInt:
		int i;
		if (gTclFromString((const char**)&argv[start + 1], i, diff)) {
		    *(int*)entry->mParam = i; 
		    *entry->mIsSet = true;
		}
		break;
	      case gTclFloat:
		GFloat f;
		if (gTclFromString((const char**)&argv[start + 1], f, diff)) {
		    *(GFloat*)entry->mParam = f; 
		    *entry->mIsSet = true;
		}
		break;
	      case gTclVector2: {
		  GVector2 vc2;
		  if (gTclFromString((const char**)&argv[start + 1], vc2, diff) ) {
		      *(GVector2*)entry->mParam = vc2; 
		      *entry->mIsSet = true;
		  }
	      }
		break;
	      case gTclVector3: {
		  GVector3 vc3;
		  if (gTclFromString((const char**)&argv[start + 1], vc3, diff) ) {
		      *(GVector3*)entry->mParam = vc3; 
		      *entry->mIsSet = true;
		  }
	      }
		break;
	      case gTclMatrix4: {
		  GMatrix4 ma;
		  if (gTclFromString((const char**)&argv[start + 1], ma, diff) ) {
		      *(GMatrix4*)entry->mParam = ma; 
		      *entry->mIsSet = true;
		  }
	      }
		break;
	      case gTclSurface: {
		  GSurface surf;
		  if (gTclFromString((const char**)&argv[start + 1], surf, diff) ) {
		      *(GSurface*)entry->mParam = surf;
		      *entry->mIsSet = true;
		  }
	      }

	      // just to make the compiler happy:
	      case gTclSpecial: break;
	      case gTclEnd: break;
	    }
	    gTclClearArgv(argv, start, diff);
	}
	entry++;
    }
}

// The utility functions return true if successful:

#ifndef GPRE_ANSI
bool gTclFromString(const char** argv, bool& bn, int& nr) {
    nr = 0;
    if (*argv) {
	bool b;
	if (gTclFromString(*argv, b)) {
	    bn = b;
	    nr = 1;
	    return(true);
	}
	return(false);
    }
    return(false);
}
#endif

bool gTclFromString(const char** argv, int& in, int& nr) {
    nr = 0;
    if (*argv) {
	int i;
	if (gTclFromString(*argv, i)) {
	    in = i;
	    nr = 1;
	    return(true);
	}
	return(false);
    }
    return(false);
}

bool gTclFromString(const char** argv, GFloat& fl, int& nr) {
    nr = 0;
    if (*argv) {
	GFloat f;
	if (gTclFromString(*argv, f)) {
	    fl = f;
	    nr = 1;
	    return(true);
	}
	return(false);
    }
    return(false);
}

bool gTclFromString(const char** argv, const char*& s, int& nr) {
    nr = 0;
    if (*argv) {
	s = argv[0];
	nr = 1;
	return(true);
    }
    return(false);
}

bool gTclFromString(const char** s, GVector2& v, int& nr) {
    nr = 0;
    if (!*s) return(false);
    if (gTclFromString(*s, v)) {
	nr = 1;
	return(true);
    }
    else return(false);
}

bool gTclFromString(const char** s, GVector3& v, int& nr) {
    nr = 0;
    if (!*s) return(false);
    if (gTclFromString(*s, v)) {
	nr = 1;
	return(true);
    }
    else return(false);
}

bool gTclFromString(const char** s, GMatrix4& m, int& nr) {
    nr = 0;
    if (!*s) return(false);
    if (gTclFromString(*s, m)) {
	nr = 1;
	return(true);
    }
    else return(false);
}

bool gTclFromString(const char** s, GColorRGB& co, int& nr) {
    nr = 0;
    if (!*s) return(false);
    if (gTclFromString(*s, co)) {
	nr = 1;
	return(true);
    }
    else return(false);
}

bool gTclFromString(const char** s, GSurface& su, int& nr) {
    nr = 0;
    if (!*s) return(false);
    if (gTclFromString(*s, su)) {
	nr = 1;
	return(true);
    }
    else return(false);
}

#ifndef GPRE_ANSI
bool gTclFromString(const char* s, bool& br) {
    int i;
    if (sscanf(s, "%i", &i)) {
	br = i ? true : false;
	return(true);
    }
    return(false);  
}
#endif

bool gTclFromString(const char* s, int& ir) {
    int i;
    if (sscanf(s, "%i", &i)) {
	ir = i;
	return(true);
    }
    return(false);  
}

bool gTclFromString(const char* s, GFloat& fr) {
    GFloat f;
    if (sscanf(s, 
#ifdef GSINGLE_PRECISION
	       "%f",
#else
	       "%lf",
#endif
	       &f)) {
	fr = f;
	return(true);
    }
    return(false);
}

bool gTclFromString(const char* s, GVector2& v) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
	bool ok = false;
	GFloat x, y;
	if (argc == 2) {
	    if (gTclFromString(argv[0], x))
		if (gTclFromString(argv[1], y)) {
			ok = true;
			v.setX(x);
			v.setY(y);
		    }
	}
	free((char*)argv);
	return(ok);
    }
    else return(false);
}

bool gTclFromString(const char* s, GVector3& v) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
	bool ok = false;
	GFloat x, y, z;
	if (argc == 3) {
	    if (gTclFromString(argv[0], x))
		if (gTclFromString(argv[1], y))
		    if (gTclFromString(argv[2], z)) {
			ok = true;
			v.setX(x);
			v.setY(y);
			v.setZ(z);
		    }
	}
	free((char*)argv);
	return(ok);
    }
    else return(false);
}

bool gTclFromString(const char*s, GMatrix4& m) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
	bool ok = false;
	if (argc == 16) {
	    for (int j = 0; j < 4; j++) 
		for (int i = 0; i < 4; i++) {
		    GFloat f;
		    ok = gTclFromString(argv[j * 4 + i], f);
		    m.set(j, i, f);
		    if (!ok) break;
		}
	}
	free((char*)argv);
	return(ok);
    }
    else return(false);
}

bool gTclFromString(const char*s, GColorRGB& c) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
        bool ok = false;
        if(argc == 3) {
            GFloat rgb[3];
            ok = gTclFromString(argv[0], rgb[0]);
            ok = ok && gTclFromString(argv[1], rgb[1]);
            ok = ok && gTclFromString(argv[2], rgb[2]);
            
            c.setRGB(rgb);
            c.correct();
        }
        free((char*)argv);
        return(ok);
    }
    else return(false);
}

bool gTclFromString(const char*s, GSurface& surf) {
    int argc; char **argv;
    if (Tcl_SplitList(gTclTmpInterp, (char*)s, &argc, &argv) == TCL_OK) {
        bool ok = false;
        if(argc == 6) {
            GColorRGB amb, diff, spec;
            GFloat shin, refr, trans;
            ok = gTclFromString(argv[0], amb);
            ok = ok && gTclFromString(argv[1], diff);
            ok = ok && gTclFromString(argv[2], spec);

            ok = ok && gTclFromString(argv[3], shin);
            ok = ok && gTclFromString(argv[4], refr);
            ok = ok && gTclFromString(argv[5], trans);

            surf.setAmbient(amb);
            surf.setDiffuse(diff);
            surf.setSpecular(spec);

            surf.setShininess(shin);
            surf.setRefraction(refr);
            surf.setTransparency(trans);
        }
        free((char*)argv);
        return(ok);
    }
    else return(false);
}

// global Tcl functions:

void gTclAppendResult(const char* pRes) {
    assert(gGlobalInterp);
    Tcl_AppendResult(gGlobalInterp, (char*)pRes, 0);
}

int gTclGlobalEval(const char* pExpr) {
    assert(gGlobalInterp);
    return(Tcl_GlobalEval(gGlobalInterp, (char*)pExpr)); 
}

const char* gTclGetResult(void) {
    assert(gGlobalInterp);
    return(gGlobalInterp->result);
}

#ifndef GPRE_ANSI
const char* gTclToString(bool b) {
    return(b ? "1" : "0");
}
#endif

const char* gTclToString(int i) {
    static char sString[10];
    sprintf(sString, "%i", i);
    return(sString);
}

const char* gTclToString(GFloat f) {
    static char sString[10];
    sprintf(sString, 
#ifdef GSINGLE_PRECISION
	    "%lf",
#else
	    "%f",
#endif
	    f);
    return(sString);
}

const char* gTclToString(const GVector2& v) {
    static GString sString(10);
    sString = "{";
    sString += gTclToString(v.getX());
    sString += " ";
    sString += gTclToString(v.getY());
    sString += "}";
    return(sString);
}

const char* gTclToString(const GVector3& v) {
    static GString sString(30);
    sString = "{";
    sString += gTclToString(v.getX());
    sString += " ";
    sString += gTclToString(v.getY());
    sString += " ";
    sString += gTclToString(v.getZ());
    sString += "}";
    return(sString);
}

#endif
