//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#ifdef GWITH_TCL

#include "util/gtclutil.h"

const GString& GTclMethodDesc::getInformation(void) const {
    static GString sString(1000);
    sString = "{";
    sString += mMethod; 
    sString += " {";
    sString += mPTypes; 
    sString += "} {";
    sString += mDesc;
    sString += "}}\n";
    return(sString);
}

template<class M>
class GTclGetInformationIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that collects
    // the information about the methods of a class
    // in the Tcl result.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(M& m) { 
	gTclAppendResult(m.getInformation()); 
	return(false);
    }
};

void GTclClassDesc::inquireInformation(void) const {
    GTclGetInformationIterator<GTclMethodDesc> tIter; 
    mMethods.doWithElements(tIter);
};

template<class M>
class GTclGlobalizeIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that makes
    // object data variables global.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTclGlobalizeIterator(const char* pName, GString& pString): mFirst(true), mName(pName), mString(pString) {}
    virtual bool exec(M& m) { 
	if (mFirst) { 
	    mFirst = false;
	    mString += "global "; 
	}
	mString += mName; 
	mString += "->"; 
	mString += m.getName(); 
	mString += " ";
	return(false);
    }
    void newline(void) const {
	mString += '\n';
    }
  private:
    bool mFirst;
    const char* mName;
    GString& mString;
};

void GTclMemberList::globalize(const char* pName, GString& s) {
    GTclGlobalizeIterator<GTclMember> tIter(pName, s);
    mMembers.doWithElements(tIter);
    tIter.newline();
}

template<class M>
class GTclSetMemberIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that sets
    // the member as Tcl variables.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTclSetMemberIterator(const char* pName): mName(pName) {}
    virtual bool exec(M& m) {
	gTclTmpString = mName; 
	gTclTmpString += "->"; 
	gTclTmpString += m.getName();
	Tcl_SetVar(gGlobalInterp, (char*)gTclTmpString, (char*)m.getValue(), TCL_GLOBAL_ONLY);
	return(false);
    }
  private:
    const char* mName;
};

void GTclMemberList::setMembers(const char* pName) {
    GTclSetMemberIterator<GTclMember> tIter(pName);
    mMembers.doWithElements(tIter);
}

template<class M>
class GTclGetMemberIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that gets
    // the values of the Tcl variables back in the member data.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTclGetMemberIterator(const char* pName): mName(pName) {}
    virtual bool exec(M& m) {
	gTclTmpString = mName; 
	gTclTmpString += "->"; 
	gTclTmpString += m.getName();
	m.setValue(Tcl_GetVar(gGlobalInterp, (char*)gTclTmpString, TCL_GLOBAL_ONLY));
	return(false);
    }
 private:
    const char* mName;
};

void GTclMemberList::getMembers(const char *pName) const {
    GTclGetMemberIterator<GTclMember> tIter(pName);
    mMembers.doWithElements(tIter);
}

template<class M>
class GTclSetValueIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that sets
    // a value for a specified member.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTclSetValueIterator(const char* pName, const char* pValue): mName(pName), mValue(pValue) {}
    virtual bool exec(M& m) {
	if (!strcmp(m.getName(), mName)) {
	    m.setValue(mValue);
	    return(true);
	}
	else return(false);
    }
  private:
    const char* mName; 
    const char* mValue; 
};

void GTclMemberList::set(const char* pName, const char* pValue) {
    GTclSetValueIterator<GTclMember> tIter(pName, pValue);
    mMembers.doWithElements(tIter);
}

template<class M>
class GTclAddMemberIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: Add a data member to the 
    // member list.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    GTclAddMemberIterator(const char* pName): mFound(false), mName(pName) {}
    virtual bool exec(M& m) {
	if (!strcmp(m.getName(), mName)) {
	    mFound = true;
	    return(true);
	}
	else return(false);
    }
    bool found(void) const { return(mFound); }
  private:
    bool mFound;
    const char* mName; 
};

void GTclMemberList::add(const char* pName) {
    GTclAddMemberIterator<GTclMember> tIter(pName);
    mMembers.doWithElements(tIter);
    if (!tIter.found()) mMembers.append(new GTclMember(pName, ""));
}

template<class M>
class GTclGetMemberValuesIterator: public GIterator<M> {
    // CONCEPT: Generic graphics kernel
    //
    // ABSTRACT PURPOSE: An iterator that gets
    // the values of the Tcl variables into a string.
    // INVARIANT:
    //
    // IMPLEMENTATION ISSUES:
    //
    // CONSTANTS: 
  public:
    virtual bool exec(M& m) {
	gTclTmpString += "{"; 
	gTclTmpString += m.getName();
	gTclTmpString += " {"; 
	gTclTmpString += m.getValue();
	gTclTmpString += "}} "; 
	return(false);
    }
};

const GString& GTclMemberList::getMemberValues(void) const {
    gTclTmpString = "";
    if (mMembers.getNumber()) {
	GTclGetMemberValuesIterator<GTclMember> tIter;
	mMembers.doWithElements(tIter);
    }
    return(gTclTmpString);
}

#endif
