//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Mon Dec 11 12:39:43 1995 by paul(@prakinf.tu-ilmenau.de)
//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel. Copying, distribution and     
//  legal info is in the file COPYRIGHT which should be distributed with this 
//  file. If COPYRGHT is not available or for more info please contact: 
//		ekki@prakinf.tu-ilmenau.de
/////////// (C) '95 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gtopolgy.h"
#include "goutput.h"

GNormal gCalcNormal(int n, GPoint3* points) {
    GFloat xu, yu, zu, ru, xv, yv, zv, rv, xn, yn, zn, rn;
    GFloat deltax, deltay, deltaz;
    int arete, aretex, aremix, aremiy, aremiz;
    GFloat xmin, xmax, ymin, ymax, zmin, zmax;

    xmin = ymin = zmin = HUGE;
    xmax = ymax = zmax = -HUGE;
    for (int i = 0; i < n; i++) {
	if (points[i].getX() < xmin) {
	    aremix = i;
	    xmin = points[i].getX();
	}
	if (points[i].getY() < ymin) {
	    aremiy = i;
	    ymin = points[i].getY();
	}
	if (points[i].getZ() < zmin) {
	    aremiz = i;
	    zmin = points[i].getZ();
	}
	xmax = gmax(points[i].getX(), xmax);
	ymax = gmax(points[i].getY(), ymax);
	zmax = gmax(points[i].getZ(), zmax);
    }

    deltax = xmax - xmin;
    deltay = ymax - ymin;
    deltaz = zmax - zmin;
    if (deltax >= deltay && deltay >= deltaz) aretex = aremix;
    else if (deltax >= deltaz && deltaz >= deltay) aretex = aremix;
    else if (deltay >= deltax && deltax >= deltaz) aretex = aremiy;
    else if (deltay >= deltaz && deltaz >= deltax) aretex = aremiy;
    else if (deltaz >= deltax && deltax >= deltay) aretex = aremiz;
    else if (deltaz >= deltay && deltay >= deltax) aretex = aremiz;
    GVector3 pta = points[aretex];
    arete = aretex + 1;
    if (arete >= n) arete = 0;
    GVector3 ptb = points[arete];

    xu = ptb.getX() - pta.getX(); 
    yu = ptb.getY() - pta.getY(); 
    zu = ptb.getZ() - pta.getZ();

    ru = sqrt(xu * xu + yu * yu + zu * zu);
    if (fabs(ru) > gEpsilon) {
	xu /= ru; 
	yu /= ru; 
	zu /= ru; 
    }
    do {
	arete++; 
	if (arete >= n) arete = 0;
	GVector3 ptc = points[arete];
	xv = ptc.getX() - ptb.getX(); 
	yv = ptc.getY() - ptb.getY();
	zv = ptc.getZ() - ptb.getZ();
	rv = sqrt(xv * xv + yv * yv + zv * zv);
	if (fabs(rv) > gEpsilon) {
	    xv /= rv;
	    yv /= rv;
	    zv/= rv; 
	}
	xn = yu * zv - zu * yv;
	yn = zu * xv - xu * zv;
	zn = xu * yv - yu * xv;
	rn = sqrt(xn * xn + yn * yn + zn * zn);
	if (fabs(rn) > gEpsilon) {
	    xn /= rn; 
	    yn /= rn; 
	    zn /= rn;
	}
    }
    while (rn < 0.01 && (xu * xv + yu * yv + zu * zv) > 0.);
    
    // clockwise:
    GNormal norm(-xn, -yn, -zn);
    if (norm.abs() < gEpsilon) norm = GNormal(0.0, 0.0, 1.0);
    else norm.unitize();
    return(norm);
}

GTopology1D::GTopology1D(const GVertexAggregate& pVertices): mList(true) {
    GAppendAggregateIterator<GVertex> tIter(mList, true);
    pVertices.doWithElements(tIter);
}

void GTopology1D::setVertex(int pIndex, const GVertex& pVertex) {
    setVertex(pIndex, new GVertex(pVertex));
    geometryChanged();
}

void GTopology1D::setVertex(int pIndex, GVertex* pVertex) {
    mList.setElement(pIndex, *pVertex);
}

const GVertex& GTopology1D::getVertex(int pIndex) const {
    return(mList.getElement(pIndex));
}

void GTopology1D::deleteVertex(int pIndex) {
    mList.deleteElement(pIndex);
    geometryChanged();
}


GTopology2D::GTopology2D(void): mX(0), mY(0), mVertices(0) {}

GTopology2D::~GTopology2D(void) {
    int p = mX * mY; 
    assert(p);

    while(p--) delete mVertices[p];
    delete mVertices;
    
    mVertices = 0;
}

GTopology2D::GTopology2D(int pX, int pY) {
    assert(pX > 0);
    assert(pY > 0);
    mX = pX; mY = pY;

    mVertices = new GVertex*[mX*mY];

    for(int i = 0; i < mX; i++)
	for(int j = 0; j < mY; j++)
	    mVertices[i + mX * j] = 0;
    // set all pointers to nill
}

GTopology2D::GTopology2D(int pX, int pY, const GVertexAggregate& pVertices) {
    assert(pX > 0);
    assert(pY > 0);
    assert(pVertices.getNumber() == pX*pY);

    mX = pX; mY = pY;

    mVertices = new GVertex*[mX*mY];
    for(int i = 0; i < mX; i++)
	for(int j = 0; j < mY; j++)
	    mVertices[i + mX * j] = new GVertex(pVertices.getElement(i + mX * j));
}

bool GTopology2D::checkVertex(int i, int j) {
    if(!mVertices) return false;
    if(!mVertices[i + mX * j])
	return false;

    return true;
}

bool GTopology2D::checkVertices() {
    if(!mVertices) return false;

    for(int i = 0; i < getX()*getY(); i++)
	if(!mVertices[i]) 
	    return false;
	else
	    continue;

    return true;
}

void GTopology2D::setVertex(int pX, int pY, GVertex* pVertex) {
    if (checkIndices(pX, pY)) {
	if(mVertices[pX + mX * pY]) {
	    delete mVertices[pX + mX * pY];
	}
	// delete old element
	mVertices[pX + mX * pY] = pVertex;
    }
    else gOutput->error("Indices out of area. Will not change vertex point."); 
}

void GTopology2D::setVertex(int pX, int pY, const GVertex& pVertex) {
    if (checkIndices(pX, pY)) {
	setVertex(pX, pY, new GVertex(pVertex));
	geometryChanged();
    }
    else gOutput->error("Indices out of area. Will not change vertex point.");
}
	    
const GVertex& GTopology2D::getVertex(int pX, int pY) const {
    if (checkIndices(pX, pY)) {
	assert(mVertices[pX + mX * pY]);
	return(*mVertices[pX + mX * pY]);
    }
    else {
	gOutput->error("Indices out of area. Will return a wrong vertex value.");
	assert(mVertices[0]);
	return(*mVertices[0]);
    }
}

bool GTopology2D::checkIndices(int pX, int pY) const {
    if (0 > pX || pX >= mX) return(false);
    if (0 > pY || pY >= mY) return(false);
    return(true);
}

void GTopology1D::computeNormals(void) {
//    gOutput->error("Sorry, not implemented: normal computing for 1D topology");
}

void GTopology2D::computeNormals(void) {
//    gOutput->error("Sorry, not implemented: normal computing for quadmesh");
}

void GTopologyPh::computeNormals(void) {
    gOutput->error("Sorry, not implemented: normal computing for polyhedron");
}

GTopologyPh::GTopologyPh(const GVertexAggregate& pVList, const GFacetAggregate& pFList): mVList(true), mFList(true) {
    GAppendAggregateIterator<GVertex> tIter1(mVList, true);
    pVList.doWithElements(tIter1);
    GAppendAggregateIterator<GFacet> tIter2(mFList, true);
    pFList.doWithElements(tIter2);
}

void GTopologyPh::appendVertices(const GVertexAggregate& pVList) {
    GAppendAggregateIterator<GVertex> tIter(mVList, true);
    pVList.doWithElements(tIter);
    geometryChanged();
}

void GTopologyPh::appendFacets(const GFacetAggregate& pFList) {
    GAppendAggregateIterator<GFacet> tIter(mFList, true);
    pFList.doWithElements(tIter);
    geometryChanged();
}

const GVertex& GTopologyPh::getVertex(int pIndex) const {
    return(mVList[pIndex]);
} 

const GFacet& GTopologyPh::getFacet(int pIndex) const {
    return(mFList[pIndex]);
} 

const GVertex& GTopologyPh::getVertex(int pFIndex, int pVIndex) const {
    int tIndex = mFList[pFIndex].getPointerToVertex(pVIndex);
    return(mVList[tIndex]);
}
