//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include "util/gvariabl.h"

GDLList<GDbVariable>* gDbAllVariables;

GDbVariable::GDbVariable(char* pName, bool c): mName(pName) {
    con = c;
    determinedBy = 0;
    mark = 0;
    stay = true;
    if (c) walkStrength = G_REQUIRED; 
    // it's a constant
    else walkStrength = G_WEAKEST;  
    // it's a variable
    gDbAllVariables->append(this);
}

GDbVariable::~GDbVariable(void) {
    GDbConstraint* c;
    c = constraints.removeFirst();
    while (c) {
	delete c;
	c = constraints.removeFirst();
    }
    gDbAllVariables->remove(this);
}

template<class P>
class GDbAddUnsatisfiedToListIterator: public GIterator<P> {
  public:
    GDbAddUnsatisfiedToListIterator(GDbConstraintList& pList): mList(pList) {}
    virtual bool exec(P& p) { 
	if (!p.satisfied()) mList.add(&p);
	return(false);
    }  
  private:
    GDbConstraintList& mList;
};

void GDbVariable::removePropagateFrom(GDbConstraintList* pList) {
    GDbVariable* v = this;
    delete gDbTodo2;
    gDbTodo2 = new GDbConstraintList;
    determinedBy = 0;
    walkStrength = G_WEAKEST;
    stay = true;

    GDbAddUnsatisfiedToListIterator<GDbConstraint> tIter(*pList);

    while (true) {
	v->constraints.doWithElements(tIter);
	GDbConstraint* nextC = v->nextDownstreamConstraint(gDbTodo2);
	if (!nextC) break; 
	else {
	    nextC->recalculate();
	    v = gDbOutVar(nextC);
	}
    }
}

void GDbVariable::free(void) {
    GDbConstraint* c = constraints.removeFirst();
    while (c) {
	for (int i = c->varCount; i > 0; i--) 
	    c->variables[i-1]->constraints.remove(c);
	delete c;
	c = constraints.removeFirst();
    }
    delete this;
}

template<class P>
class GDbDownStreamIterator: public GIterator<P> {
  public:
    GDbDownStreamIterator(GDbConstraintList* pTodo, GDbConstraint* pDeter): 
    mTodo(pTodo), mDeter(pDeter), mFirst(0) {}

    virtual bool exec(P& p) { 
	if (&p != mDeter && p.satisfied()) {
	    if (!mFirst) mFirst = &p;
	    else mTodo->add(&p);
	}
	return(false);
    }  
    GDbConstraint* getFirst(void) const {
	return(mFirst);
    }
  private:
    GDbConstraintList* mTodo;
    GDbConstraint* mDeter;
    GDbConstraint* mFirst;
};

GDbConstraint* GDbVariable::nextDownstreamConstraint(GDbConstraintList* todo) {
    GDbDownStreamIterator<GDbConstraint> tIter(todo, determinedBy);
    constraints.doWithElements(tIter);
    GDbConstraint* tFirst = tIter.getFirst();
    if (!tFirst) tFirst = todo->removeFirst();
    return(tFirst);
}

