//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Tue Jan 30 11:00:17 1996 - juhu@rz.tu-ilmenau.de
#include "util/gvector.h"
#include "util/gmatrix.h"


const GPoint3 gOrigin3(0.0, 0.0, 0.0);
long GRay3::mID = 0L;
int GRay3::mPath = 0;

bool GVector2::operator==(const GVector2& v) const {
    if (mX != v.mX) return(false);
    if (mY != v.mY) return(false);
    return(true);
}

ostream& operator<<(ostream& pStream, const GVector2& pVector) {
    pStream << '{' << pVector.mX << ' ' << pVector.mY << '}';
    return(pStream);
}

bool GVector3::operator==(const GVector3& v) const {
    if (mX != v.mX) return(false);
    if (mY != v.mY) return(false);
    if (mZ != v.mZ) return(false);
    return(true);
}

ostream& operator<<(ostream& pStream, const GVector3& pVector) {
    pStream << '{' << pVector.mX << ' ' << pVector.mY << ' ' << pVector.mZ << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GRay3& pRay) {
    pStream << '{' << pRay.mPoint << " + " << pRay.mLength << " * " << pRay.mVector << '}';
    return(pStream);
}



inline GFloat  qCos(GFloat x) {
    return ( 1 - x*x/2.4684);
}

inline GFloat qACos(GFloat x) {
    return(sqrt(2.4684*(1-x)));
}

// generate a random ray within a solid angle along this ray
GRay3 GRay3::distribute(GFloat pAngle, GFloat pR1, GFloat pR2) const{
    GFloat tPhi = pR1 * pAngle;
    GFloat tPsi = pR2 * pAngle;

    // construct a distrib ray in modelling space, toward to +Z
    GVector3 tDir(tan(tPhi), tan(tPsi), 1.0);
    
    GVector3 tUp(0, 0, 1.0);
    if(gabs(mVector.getX()) >= gabs(mVector.getY())) {
	tUp.setX(1.0); tUp.setY(0.0); tUp.setZ(0.0); 
    }
    
    GVector3 tU = (mVector.cross(tUp)).unitize();
    GVector3 tV = (mVector.cross(tU)).unitize();
    
    // construct a coord. XForm w/ U, V, W
    GMatrix4 tM(tU, tV, mVector);
    
    // transform the distrib ray from modelling coord. to world coord.
    tDir = tM * tDir;
    
    

//     GVector3 tDir(mVector);
//     if( tDir.getX() > .5 ) {
// 	tDir.setX(qCos(qACos(tDir.getX()) + tPhi));
// 	tDir.setY(qCos(qACos(tDir.getY()) + tPsi));
//     } 
//     else {
// 	if( tDir.getY() < .5 ) {
// 	    tDir.setY(qCos(qACos(tDir.getY()) + tPhi));
// 	    tDir.setZ(qCos(qACos(tDir.getZ()) + tPsi));
// 	} else {
// 	    tDir.setZ(qCos(qACos(tDir.getZ()) + tPhi));
// 	    tDir.setX(qCos(qACos(tDir.getX()) + tPsi));
// 	}
//     }
    
    

    return (GRay3(mPoint, tDir.unitize(), mFlags));
}



















































