//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
//  LAST EDIT: Mon Dec 11 12:32:45 1995 by paul(@prakinf.tu-ilmenau.de)
#include "util/gvertex.h"

ostream& operator<<(ostream& pStream, const GVertex& pVertex) {
    pStream << '{' << pVertex.getPoint() << "} {" ;
    if (pVertex.getNormal()) pStream << *pVertex.getNormal();
    pStream << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GVertexWithColorRGB& pVertex) {
    pStream << (GVertex&)pVertex;
    pStream << " {" ;
    if (pVertex.getColor()) pStream << *pVertex.getColor();
    pStream << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GVertexWithSurface& pVertex) {
    pStream << (GVertex&)pVertex;
    pStream << " {" ;
    if (pVertex.getSurface()) pStream << *pVertex.getSurface();
    pStream << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GCurveVertex& pVertex) {
    pStream << (GVertex&)pVertex;
    pStream << " {" ;
    if (pVertex.getTangent()) pStream << *pVertex.getTangent();
    if (pVertex.getGradient()) pStream << *pVertex.getGradient();
    pStream << '}';
    return(pStream);
}

ostream& operator<<(ostream& pStream, const GSurfaceVertex& pVertex) {
    pStream << (GVertex&)pVertex;
    pStream << " {" ;
    if (pVertex.getTangentU()) pStream << *pVertex.getTangentU();
    if (pVertex.getTangentW()) pStream << *pVertex.getTangentW();
    pStream << '}';
    return(pStream);
}

bool GCurveVertex::operator==(const GCurveVertex& pVertex) const {
    bool equal;
    
    if(getPoint() == pVertex.getPoint())
	equal = true;
    else
	return false;
    
    const GVector3* tNormal1 = getNormal();
    const GVector3* tNormal2 = pVertex.getNormal();
    if(tNormal1 || tNormal2) {
	if(tNormal1 && tNormal2)
	    if(*tNormal1 == *tNormal2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    const GVector3* tTangent1 = getTangent();
    const GVector3* tTangent2 = pVertex.getTangent();
    if(tTangent1 || tTangent2) {
	if(tTangent1 && tTangent2)
	    if(*tTangent1 == *tTangent2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    const GVector3* tGradient1 = getGradient();
    const GVector3* tGradient2 = pVertex.getGradient();
    if(tGradient1 || tGradient2) {
	if(tGradient1 && tGradient2)
	    if(*tGradient1 == *tGradient2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    return equal;
}

bool GSurfaceVertex::operator==(const GSurfaceVertex& pVertex) const {
    bool equal;
    
    if(getPoint() == pVertex.getPoint())
	equal = true;
    else
	return false;
    
    const GVector3* tNormal1 = getNormal();
    const GVector3* tNormal2 = pVertex.getNormal();
    if(tNormal1 || tNormal2) {
	if(tNormal1 && tNormal2)
	    if(*tNormal1 == *tNormal2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    const GVector3* tTangentU1 = getTangentU();
    const GVector3* tTangentU2 = pVertex.getTangentU();
    if(tTangentU1 || tTangentU2) {
	if(tTangentU1 && tTangentU2)
	    if(*tTangentU1 == *tTangentU2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    const GVector3* tTangentW1 = getTangentW();
    const GVector3* tTangentW2 = pVertex.getTangentW();
    if(tTangentW1 || tTangentW2) {
	if(tTangentW1 && tTangentW2)
	    if(*tTangentW1 == *tTangentW2)
		equal = true;
	    else
		return false;
	else
	    return false;
    }
    else
	equal = true;
    
    return equal;
}
