//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <util/gdeltabl.h>
#include <input/gtime.h>

static GTime tsTime;

void tsStart(void) { tsTime = GTime::getTime(); }

long tsFinish(void) { 
    GTime tTime = GTime::getTime() - tsTime;
    long t = tTime.getMicroSeconds();
    t += 1000000 * tTime.getSeconds();
    t /= 1000;
    return(t);
}

class TsVariable: public GDbVariable {
  public:
    TsVariable(char* c, long l, bool b = false):
    GDbVariable(c, b) {	value = l; }
    void assign(long);
    void change(long);
    void print(void) {
	printf("\t%s: %i\n", getName(), value);
    }
  public:
    long value;
};

void TsVariable::assign(long newValue) {
    GDbConstraint* editC = new GDbEditC(this, G_REQUIRED);
    if (editC->satisfied()) {
	value = newValue;
	GDbConstraintList* plan = editC->extractPlanFromConstraint();
	plan->executePlan();
	delete plan;
    }
    delete editC;
}

void TsVariable::change(long newValue) {
    printf("Changing %s...\n", getName());

    tsStart();
    GDbConstraint* editC = new GDbEditC(this, G_STRONG_DEFAULT);
    long msecs = tsFinish();
    printf("Adding Constraint: %ld msecs.\n", msecs);
   
    tsStart();
    GDbConstraintList* plan = editC->extractPlanFromConstraint();
    msecs = tsFinish();
    printf("Making Plan (length: %d): %ld msecs.\n", plan->getNumber(), msecs);
    
    tsStart();
    value = newValue;
    for (int i = 0; i < 100; i++) plan->executePlan();
    msecs = tsFinish();
    printf("Executing Plan: %.3f msecs.\n", msecs / 100.0);
    delete plan;

    tsStart();
    delete editC;
    msecs = tsFinish();
    printf("Removing Constraint: %ld msecs\n", msecs);
}

///// test constraints:

class TsEqualsC: public GDbConstraint {
    // a == b 
  public:
    TsEqualsC(TsVariable* a, TsVariable* b, GDbStrength strength): GDbConstraint(2, strength) {
	variables[0] = a;
	variables[1] = b;
	methodCount = 2;
	methodOuts[0] = 0;
	methodOuts[1] = 1;
	add();
    }

    virtual void execute(void);
};

void TsEqualsC::execute(void) {
    switch (whichMethod) {
      case 0:
	((TsVariable*)variables[0])->value
	    = ((TsVariable*)variables[1])->value;
	break;
      case 1:
	((TsVariable*)variables[1])->value
	    = ((TsVariable*)variables[0])->value;
	break;
    }
}

class TsAddC: public GDbConstraint {
    // a + b = sum 
  public:
    TsAddC(TsVariable* a, TsVariable* b, TsVariable* sum, GDbStrength strength): GDbConstraint(3, strength) {
	variables[0] = a;
	variables[1] = b;
	variables[2] = sum;
	methodCount = 3;
	methodOuts[0] = 2;
	methodOuts[1] = 1;
	methodOuts[2] = 0;
	add();
    }
    virtual void execute(void);
};

void TsAddC::execute(void) {
    switch (whichMethod) {
      case 0:
	((TsVariable*)variables[2])->value
	    = ((TsVariable*)variables[0])->value
		+ ((TsVariable*)variables[1])->value;
	break;
      case 1:
	((TsVariable*)variables[1])->value
	    = ((TsVariable*)variables[2])->value
		- ((TsVariable*)variables[0])->value;
	break;
      case 2:
	((TsVariable*)variables[0])->value
	    = ((TsVariable*)variables[2])->value
		- ((TsVariable*)variables[1])->value;
	break;
    }
}

class TsMultiplyC: public GDbConstraint {
    // a * b = prod 
  public:
    TsMultiplyC(TsVariable* a, TsVariable* b, TsVariable* prod, GDbStrength strength): GDbConstraint(3, strength) {
	variables[0] = a;
	variables[1] = b;
	variables[2] = prod;
	methodCount = 3;
	methodOuts[0] = 2;
	methodOuts[1] = 1;
	methodOuts[2] = 0;
	add();
    }
    virtual void execute(void);
};

void TsMultiplyC::execute(void) {
    switch (whichMethod) {
      case 0:
	((TsVariable*)variables[2])->value
	    = ((TsVariable*)variables[0])->value 
		* ((TsVariable*)variables[1])->value;
	break;
      case 1:
	((TsVariable*)variables[1])->value
	    = ((TsVariable*)variables[2])->value
		/ ((TsVariable*)variables[0])->value;
	break;
      case 2:
	((TsVariable*)variables[0])->value 
	    = ((TsVariable*)variables[2])->value 
		/ ((TsVariable*)variables[1])->value;
	break;
    }
}

class TsScaleOffsetC: public GDbConstraint {
    // (src * scale) + offset = dest 
  public:
    TsScaleOffsetC(TsVariable* src, TsVariable* scale, TsVariable* offset, TsVariable* dest, GDbStrength strength): GDbConstraint(4, strength) {
	variables[0] = src;
	variables[1] = scale;
	variables[2] = offset;
	variables[3] = dest;
	methodCount = 2;
	methodOuts[0] = 3;
	methodOuts[1] = 0;
	add();
    }
    void execute(void);
};

void TsScaleOffsetC::execute(void) {
    switch (whichMethod) {
      case 0:
	((TsVariable*)variables[3])->value 
	    = (((TsVariable*)variables[0])->value
	       * ((TsVariable*)variables[1])->value) 
		+ ((TsVariable*)variables[2])->value;
	break;
      case 1:
	((TsVariable*)variables[0])->value
	    = (((TsVariable*)variables[3])->value
	       - ((TsVariable*)variables[2])->value)
		/ ((TsVariable*)variables[1])->value;
	break;
    }
}

///// a Fahrenheit to Celcius temperature converter:

void tsTemperatureConverter(void) {
    gDbInit();
    TsVariable* celcius = new TsVariable("C", 0);
    TsVariable* fahrenheit = new TsVariable("F", 0);
    TsVariable* t1 = new TsVariable("t1", 1);
    TsVariable* t2 = new TsVariable("t2", 1);
    TsVariable* nine = new TsVariable("*const*", 9, true);
    TsVariable* five = new TsVariable("*const*", 5, true);
    TsVariable* thirtyTwo = new TsVariable("*const*", 32, true);

    new TsMultiplyC(celcius, nine, t1, G_REQUIRED);
    new TsMultiplyC(t2, five, t1, G_REQUIRED);
    new TsAddC(t2, thirtyTwo, fahrenheit, G_REQUIRED); 

    printf("***** Changing celcius to 0 *****\n");
    celcius->assign(0);
    celcius->print();
    fahrenheit->print();

    printf("***** Changing fahrenheit to 212 *****\n");
    fahrenheit->assign(212);
    celcius->print();
    fahrenheit->print();

    printf("***** Changing celcius to -40 *****\n");
    celcius->assign(-40);
    celcius->print();
    fahrenheit->print();

    printf("***** Changing fahrenheit to 70 *****\n");
    fahrenheit->assign( 70);
    celcius->print();
    fahrenheit->print();
    gDbExit();
}

///// This is the standard DeltaBlue benchmark. A long chain of equality
///// constraints is constructed with a stay constraint on one end. An edit
///// constraint is then added to the opposite end and the time is measured for
///// adding and removing this constraint, and extracting and executing a
///// constraint satisfaction plan. There are two cases. In case 1, the added
///// constraint is stronger than the stay constraint and values must propagate
///// down the entire length of the chain. In case 2, the added constraint is
///// weaker than the stay constraint so it cannot be accomodated. The cost in
///// this case is, of course, very low. Typical situations lie somewhere between
///// these two extremes.

void tsBenchmark(int n) {
    gDbInit();
    TsVariable* prev,* v,* first,* last;
    prev = first = last = NULL;

    tsStart();
    for (int i = 0; i < n; i++) {
	char name[20];
	sprintf(name, "v%ld", i);
	v = new TsVariable(name, 0);
	if (prev) new TsEqualsC(prev, v, G_REQUIRED);
	if (i == 0) first = v;
	if (i == (n-1)) last = v;
	prev = v;
    }
    long msecs = tsFinish();
    printf("\n%ld msecs to add %d constraints.\n", msecs, n);
    new GDbStayC(last, G_DEFAULT);

    tsStart();
    GDbConstraint* editC = new GDbEditC(first, G_STRONG_DEFAULT);
    msecs = tsFinish();
    printf("Add Constraint (case 1): %ld msecs.\n", msecs);

    tsStart();
    GDbConstraintList* plan = editC->extractPlanFromConstraint();
    msecs = tsFinish();
    printf( "Make Plan (case 1): %ld msecs (plan is length %d).\n",
	   msecs, plan->getNumber());

    tsStart();
    for (i = 0; i < 100; i++) plan->executePlan();
    msecs = tsFinish();
    printf("Execute Plan (case 1): %.3f msecs.\n", msecs / 100.0);
    delete plan;

    tsStart();
    delete editC;
    msecs = tsFinish();
    printf("Remove Constraint (case 1): %ld msecs\n", msecs);
    
    tsStart();
    editC = new GDbEditC(first, G_WEAK_DEFAULT);
    msecs = tsFinish();
    printf("Add Constraint (case 2): %ld msecs.\n", msecs);
    
    tsStart();
    plan = editC->extractPlanFromConstraint();
    msecs = tsFinish();
    printf( "Make Plan (case 2): %ld msecs (plan is length %d).\n",
	   msecs, plan->getNumber());

    tsStart();
    for (i = 0; i < 100; i++) plan->executePlan();
    msecs = tsFinish();
    printf("Execute Plan (case 2): %.3f msecs.\n", msecs / 100.0);
    delete plan;

    tsStart();
    delete editC;
    msecs = tsFinish();
    printf("Remove Constraint (case 2): %ld msecs\n", msecs);

    tsStart();
    editC = new GDbEditC(last, G_STRONG_DEFAULT);
    msecs = tsFinish();
    printf("Add Constraint (case 3): %ld msecs.\n", msecs);
    
    tsStart();
    plan = editC->extractPlanFromConstraint();
    msecs = tsFinish();
    printf( "Make Plan (case 3): %ld msecs (plan is length %d).\n",
	   msecs, plan->getNumber());

    tsStart();
    for (i = 0; i < 100; i++) plan->executePlan();
    msecs = tsFinish();
    printf("Execute Plan (case 3): %.3f msecs.\n", msecs / 100.0);
    delete plan;

    tsStart();
    delete editC;
    msecs = tsFinish();
    printf("Remove Constraint (case 3): %ld msecs\n\n", msecs);
}

///// This test constructs a two sets of variables related to each other by a
///// simple linear transformation (scale and offset). The time is measured to
///// change a variable on either side of the mapping and to change the scale or
///// offset factors. It has been tested for up to 2000 variable pairs.

void tsProjectionTest(int n) {
    TsVariable* src,* scale,* offset,* dest;
    long msecs;
    char name[20];
    gDbInit();
    tsStart();
    scale = new TsVariable("scale", 10);
    offset = new TsVariable("offset", 1000);
    
    for (int i = 1; i <= n; i++) {
	// make src and dest variables:
	sprintf(name, "src%ld", i);
	src = new TsVariable(name, i);
	sprintf(name, "dest%ld", i);
	dest = new TsVariable(name, i);
	
	// add stay on src:
	new GDbStayC(src, G_DEFAULT);
	
	// add scale/offset constraint:
	new TsScaleOffsetC(src, scale, offset, dest, G_REQUIRED);
    }
    msecs = tsFinish();
    printf("Setup time for %d points: %ld msecs.\n", n, msecs);
    
    src->change( 17);
    dest->change( 1050);
    scale->change( 5);
    offset->change( 2000);
    gDbExit();
}

///// This test constructs a full binary tree of add constraints of the given
///// depth and then measures the time to change the root variable. Log(depth)
///// constraints must be traversed. DeltaBlue chooses an arbitrary path to a
///// leaf of the tree.

// returns the root variable of an adder tree of the given depth 
TsVariable* tsMakeTree(int depth) {
    TsVariable *root, *left, *right;
    if (depth <= 0) {
	root = new TsVariable("leaf", 1);
	new GDbStayC(root, G_DEFAULT);
    } else {
	root = new TsVariable("nonleaf", 1);
	left = tsMakeTree(depth - 1);
	right = tsMakeTree(depth - 1);
	new TsAddC(left, right, root, G_REQUIRED);
    }
    return(root);
}

void tsTreeTest(int depth) {
    gDbInit();
    printf("Adder tree of depth %d\n", depth);
    tsStart();

    TsVariable *root = tsMakeTree(depth);
    
    long msecs = tsFinish();
    // Note: with stays, there is one constraint per variable 
    printf( "%i constraints added in %ld msecs.\n", gDbAllVariables->getNumber(), msecs);
    root->change(17);
    gDbExit();
}

int main(void) {
    tsTemperatureConverter();
    tsTreeTest(5);
    tsTreeTest(10);
    tsBenchmark(50);
    tsBenchmark(100);
    tsBenchmark(200);
    tsBenchmark(400);
    tsBenchmark(800);
    tsProjectionTest(250);
    tsProjectionTest(500);
    tsProjectionTest(1000);
    tsProjectionTest(2000);

    return(0);
}
