//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <render/gpixmap.h>
#include <render/gshade.h>
#include <input/gtime.h>
#include <util/gtempl.h>

class GForeground: public GShadable {
  public:
    virtual void shade(void) {
	setSurface(GSurface(GColorRGB(1.0, 0.0, 0.0)));
    }
};

int main(void) {
    const GColorRGB cPink(0.5, 0.0, 0.7);

    // initialize GENERIC:
    gInitialize();

    // create a pixmap display to be used for
    // shading output:
    GPixmapDisplay tPixmap("GTest", 600, 600);
    
    // create an aggregate representing a scene:
    GDLList<GShadable> tScene(false);
    tScene.append(new GForeground);

    GVertex tVertex;

    // create an object:
    GShadablePolygon* tPolygon = new GShadablePolygon;
    tPolygon->setNormal(GVector3(0.0, 0.0, 1.0));
    
    // setup 4 points in clockwise direction:
    tVertex.setPoint(GVector3(0.0, 0.0, 0.0));
    tPolygon->setVertex(0, tVertex);
    tVertex.setPoint(GVector3(0.0, 1.0, 0.0));
    tPolygon->setVertex(1, tVertex);
    tVertex.setPoint(GVector3(1.0, 1.0, 0.0));
    tPolygon->setVertex(2, tVertex);
    tVertex.setPoint(GVector3(1.0, 0.0, 0.0));
    tPolygon->setVertex(3, tVertex);

    tScene.append(tPolygon);

    // create another object:
    tPolygon = new GShadablePolygon;
    tPolygon->setNormal(GVector3(0.0, 0.0, 1.0));
    
    // setup 4 points in clockwise direction:
    tVertex.setPoint(GVector3(-1.0, -1.0, 0.0));
    tPolygon->setVertex(0, tVertex);
    tVertex.setPoint(GVector3(-1.0, -0.8, 0.0));
    tPolygon->setVertex(1, tVertex);
    tVertex.setPoint(GVector3(1.0, -0.8, 0.0));
    tPolygon->setVertex(2, tVertex);
    tVertex.setPoint(GVector3(1.0, -1.0, 0.0));
    tPolygon->setVertex(3, tVertex);

    tScene.append(tPolygon);

    // create a point light:
    GShadePointLight tLight;
    tLight.setPosition(GPoint3(3, 3, 3));
    tLight.setColor(GColorRGB(1, 1, 0));

    GDLList<GShadeLight>* tLights = new GDLList<GShadeLight>(false);
    // insert light:
    tLights->append(&tLight); 

    // initialize global shading light source aggregate:
    gShadeLights = tLights;

    // create the renderer:
    GLookatShader<GShadable, GPixmapDisplay> tRenderer;

    // create the output:
    tRenderer.initRealization(tPixmap);

    GTime tTime;
    for (int i = 0; i < 30; i++) {
	tPixmap.clear(cPink * i/30.0);
	tPixmap.activate();
	tRenderer.rendering(tScene);
	tPixmap.swapBuffers();
    }
    cout << "Rendering time (30 updates): " << GTime::getTime() - tTime << " secs" << endl;

    getchar();
    return(0);
}
