//////////////////////////////////////////////////////////////////////////////
//  This file belongs to the GENERIC 3D kernel, Version 1.5.
//  Copying, distribution and legal info is in the file COPYRGHT which should
//  be distributed with this file. If COPYRGHT is not available or for more
//  info please contact	generic@prakinf.tu-ilmenau.de (in English, please).
//  This project is sponsored by EasternGraphics GmbH, Germany.
/////////// (C) '96 Ekkehard 'Ekki' Beier - All rights reserved. /////////////
#include <render/gpixmap.h>
#include <render/gshade.h>
#include <input/gtime.h>
#include <input/gcallbck.h>
#include <input/gfile.h>
#include <input/gserver.h>
#include <input/gtimer.h>
#include <util/gtcl.h>
#include <format/ghfont.h>

// This is a really simple demo made on top of G.
// It's 2d, it's wireframe and, again, it's simple.
// But, I didn't want to spend more than one day 
// to implement it from scratch.
// Nevertheless, it covers rendering, modeling, interaction,
// and specific semantics :-), and takes full advantage of display lists.
// Have fun,
//       Ekki.
// TO DO: Add Tk user interface and display of remaining lives.

const GString gtFont("symbolic");
const GColorRGB gtBackground(0.0, 0.0, 0.0);
const int gtLevels = 7;
const int gtLifes = 3;
const int gtSteps = 50;
const GFloat gtDimension(10.0);
const GFloat gtInterval(0.05); 
const GFloat gtTranslation = 0.05;
int gtCounter = 0;
const GFloat gtVelocity = 0.2;
GFloat gtScoreFac = 1;

const GColorRGB gtG2RColors[gtLevels] = {
    {0.0, 1.0, 0.0},
    {0.0, 0.8, 0.0},
    {0.0, 0.6, 0.0},
    {0.0, 0.4, 0.0},
    {0.0, 0.2, 0.0},
    {0.6, 0.6, 0.0},
    {0.7, 0.0, 0.0}
};

const GColorRGB gtBlueColors[gtLevels] = {
    {0.0, 0.0, 0.9},
    {0.0, 0.0, 0.7},
    {0.0, 0.0, 0.6},
    {0.0, 0.0, 0.4},
    {0.0, 0.0, 0.3},
    {0.0, 0.0, 0.2},
    {0.0, 0.0, 0.1},
};

class GTPrimitive: public GShadable {
  protected:
    GTPrimitive(char);
  public:
    ~GTPrimitive(void) { sScene.remove(this); }
    static const GAggregate<GTPrimitive>& getPrimitives(void) { return(sScene); }
    virtual void shade(void);
  protected:
    GPoint2 mPos;
    GPoint2 mOff;
    GPoint2 mSize;
    int mLevel;
    const GColorRGB* mColors[gtLevels];

    class LFont: public GHersheyFont {
      public:
	LFont(void);
    };
    static LFont sFont;
    static GDLList<GTPrimitive> sScene;
};

class GTSymbol: public GTPrimitive {
  public:
    GTSymbol(char pC): GTPrimitive(pC) {
	mColors = &gtG2RColors;
	init(); 
    }
    virtual void shade(void);
  private:
    void init(void);
  private:
    int mSteps;
    int mStep;
    GFloat mVeloc;
};

class GTMarker: public GTPrimitive {
  public:
    GTMarker(void): GTPrimitive(0x57), mLive(0), mSteps(0), mScore(0) {
	mColors = &gtBlueColors;
	mPos = GPoint2(0, - 0.5 * gtDimension);
	mLevel = 0;
    }
    virtual void shade(void);
    void translate(GFloat);
    GFloat distance2(const GPoint2&);
    void incrLevel(void);
    void incrSteps(void);
    void decrLevel(void);
    void addScore(int pInc) {
	mScore += pInc;
	cout << mScore << endl;
	gtScoreFac = 1.0 + mScore * 0.001;
    }
  private:
    int mSteps;
    int mLive;
    int mScore;
}* gtMarker;

GTPrimitive::LFont GTPrimitive::sFont;
GDLList<GTPrimitive> GTPrimitive::sScene(false);

GTPrimitive::LFont::LFont(void): GHersheyFont(gtFont) {
    GString tPath = gtFont;
    FILE* tFile = fopen(tPath, "r");
    if (!tFile) gOutput->fatalVar("Could not open file ", (char*)tPath, ".", 0);
    load(tFile);
    fclose(tFile);
}

GTPrimitive::GTPrimitive(char pC) {
    // NOTE, currently multiple usage of identical display lists NOT supported.
    sScene.append(this);

    GPoint2 tMin;
    GPoint2 tMax;
    bool tFirst = false;
    openDisplayList();
    for (int i = 0; i < sFont.getLines(pC); i++) {
	int nrp = sFont.getNumberOfPoints(pC, i);
	GShadablePolyline tLine;
	for (int k = 0; k < nrp; k++) {
	    const GPoint3& tPoint = sFont.getPoint(pC, i, k);
	    tLine.setVertex(k, GVertex(tPoint));
	    if (tFirst) {
		tMin.setX(tPoint.getX());
		tMin.setY(tPoint.getY());
		tMax.setX(tPoint.getX());
		tMax.setY(tPoint.getY());
	    }
	    else {
		if (tPoint.getX() < tMin.getX()) tMin.setX(tPoint.getX());
		if (tPoint.getY() < tMin.getY()) tMin.setY(tPoint.getY());
		if (tPoint.getX() > tMax.getX()) tMax.setX(tPoint.getX());
		if (tPoint.getY() > tMax.getX()) tMax.setY(tPoint.getY());
	    }
	}
	tLine.shade();
    }
    closeDisplayList();
    
    mSize.setX(0.05 * (tMax.getX() - tMin.getX()));
    mSize.setY(0.05 * (tMax.getY() - tMin.getY()));

    mOff.setX(0.1 * tMin.getX() - mSize.getX()); 
    mOff.setY(0.1 * tMin.getY() + mSize.getY()); 
}

void GTPrimitive::shade(void) {
    static GSurface sSurface;
    sSurface.setAmbient((*mColors)[mLevel]);
    sSurface.setDiffuse((*mColors)[mLevel]);
    setSurface(sSurface);

    static GMatrix4 sMatrix;
    static bool sScale = true;
    if (sScale) {
	sScale = false;
	sMatrix.set(0, 0, 0.2);
	sMatrix.set(1, 1, 0.2);
    }
    sMatrix.set(0, 3, mPos.getX() + mOff.getX());
    sMatrix.set(1, 3, mPos.getY() + mOff.getY());
    setMatrix(sMatrix);

    callDisplayList();
    unsetMatrix();
}

void GTMarker::translate(GFloat pX) {
    mPos.setX(mPos.getX() + pX);
}

GFloat GTMarker::distance2(const GPoint2& pPoint) {
    GFloat tF = pPoint.getX() - mPos.getX();
    GFloat tRes = tF * tF;
    tF = pPoint.getY() - mPos.getY();
    return(tRes + tF * tF);
}

void GTMarker::decrLevel(void) {
    if (mLevel > 0) mLevel--;
}

void GTMarker::incrLevel(void) {
    mLevel++;
    if (mLevel == gtLevels) {
	mLevel = 0;
	if (++mLive > gtLifes) {
	    cout << "Final Score: " << mScore << " points." << endl;
	    exit(0);
	}
    }
}

void GTMarker::incrSteps(void) {
    mSteps++;
    if (mSteps == gtSteps) {
	mSteps = 0;
	incrLevel();
    }
}

void GTMarker::shade(void) {
    gtCounter++;
    incrSteps();
    GTPrimitive::shade();
}

void GTSymbol::shade(void) {
    if (mPos.getY() < (-gtDimension - 1)) init();
    else {
	mPos.setY(mPos.getY() - mVeloc);
	mStep++;
	if ((mStep == mSteps) && (mLevel < (gtLevels - 1))) {
	    mLevel++;
	    mStep = 0;
	}
    }
    
    if (gtMarker->distance2(mPos) < 0.5) {
	if (mLevel < (gtLevels - 2)) {
	    gtMarker->decrLevel();
	    gtMarker->addScore(10 * (gtLevels - mLevel - 1));
	}
	if (mLevel == (gtLevels - 1)) gtMarker->incrLevel();
	init();
    }    
    GTPrimitive::shade();
}

void GTSymbol::init(void) {
    GFloat tRand = (rand() & 0xff)/128.0 - 1;
    mLevel = 0;
    mStep = 0;
    mSteps = (int)((20.0 + 5 * tRand)/gtScoreFac);
    mPos.setX(gtDimension * tRand);
    mPos.setY(gtDimension + 1);
    mVeloc = gtVelocity * gtScoreFac * (1 + 0.3 * tRand);
}

class GTRenderer {
  public:
    GTRenderer(void);
    void rendering(void);
  private:
    GPixmapDisplay mPixmap;
    GLookatShader<GTPrimitive, GPixmapDisplay> mRend;

    class LTimerCB: public GCallback {
      public:
	LTimerCB(GTRenderer& pRend): mRend(pRend) {}
	virtual void action(GInputDevice&) { mRend.rendering();	}
      private:
	GTRenderer& mRend;
    } lTimerCB;
};

class GTInteractor: public GInputDevice {
  public:
    GTInteractor(void): GInputDevice((GInputDevice*)gNoInputDevice), mTrans(0) {}
    virtual bool event(GEvent*);
  private:
    int mTrans;
};

GTRenderer::GTRenderer(void): mPixmap(".p", 600, 600), lTimerCB(self) {
    GDLList<GShadeLight>* tLights = new GDLList<GShadeLight>(false);
    tLights->append(new GShadeAmbientLight);
    gShadeLights = tLights;
    mRend.orthographic();
    mRend.setZoom(85);
    mRend.setViewpoint(GPoint3(0.0, 0.0, 1.0));
    mRend.setRefpoint(gOrigin3);
    mRend.initRealization(mPixmap);
    (new GTInteractor)->setRealization(&mPixmap);
    (new GTimer(gtInterval))->addCB(&lTimerCB);
    rendering();
}

void GTRenderer::rendering(void) {
    mPixmap.clear(gtBackground);
    mPixmap.activate();
    mRend.rendering(GTPrimitive::getPrimitives());
    mPixmap.swapBuffers();
}

bool GTInteractor::event(GEvent* pEvent) {
    // Quick and dirty hack. No hierarchy. no callbacks.
    GRealization* tReal = getRealization();
    assert(tReal);
    pEvent = tReal->getEvent();
    if (pEvent && pEvent->isA(gMouseButtonEventID)) {
	GMouseButtonEvent* bev = (GMouseButtonEvent*)pEvent;
	mTrans = 0; // deselection (?)
	if (bev->left && !bev->middle && !bev->right) mTrans = -1; // left pick
	if (!bev->left && !bev->middle && bev->right) mTrans = 1; // right pick
    }

    if (mTrans) gtMarker->translate(mTrans * gtTranslation);
    if (pEvent) delete pEvent;
    return(pEvent ? true : false);
}

int main(void) {
    gInitialize();

    GTRenderer gtGame;

    gtMarker = new GTMarker;
    GTSymbol tStar(0x4c);
    GTSymbol tBell(0x61);
    GTSymbol tPalm(0x62);
    GTSymbol tTree(0x63);
    GTSymbol tClover(0x68);
    GTSymbol tHeart(0x6f);

    gInputLoop();
    return(0);
}
